/* $Id: omega_background.c,v 1.4 2012/01/31 13:38:50 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/31 13:38:50 $
 * $Revision: 1.4 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <stdlib.h>

#include "omega_background.h"
#include "omega_dfs.h"
#include "omega_utils.h"

/**
 * @defgroup omega_background  Background related functions
 *
 * This module provides functions to estimate the background by the usage
 * of Sextractor
 *
 * @par Synopsis:
 *   Functions to get the image background
 *
 * @code
 * #include "omega_background.h"
 * #include "omega_dfs.h"
 * #include "omega_utils.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/**
 * @brief    	Creates a background image using Sextractor
 *
 * @param    parlist     	the parameters list
 * @param    in_name   		the name of the input FITS image (science image)
 * @param    out_name   		the name of the output FITS image (background image)
 * @return   0 if everything is OK or -1 in case of error
 */

int omega_create_background(cpl_parameterlist *parlist, const char *in_name, const char *out_name)
{

	int i = 0;
	int npars = 0;
	double dthre = 0.0;
	int backsize=64;
	char *cmd = NULL;
	const char *path = NULL;
	const char *sex_conf = NULL;
	const char *sex_conv = NULL;
	const char *sex_nnw = NULL;
	const char *sex_par = NULL;
	const char *text = NULL;

	cpl_parameter 		*par;


	/*
	 *  Start the recipe
	 */

	if (!parlist) {
		cpl_msg_error (cpl_func,"Sextractor parameters list not found");
		return -1;
	}
	else {
		npars = cpl_parameterlist_get_size(parlist);
		if(npars < 1) {
			cpl_msg_error (cpl_func,"Sextractor parameters list is NULL");
			return -1;
		}
	}


	/*
	 *  Retrieve Sextractor configuration parameters
	 */

	par = cpl_parameterlist_get_first(parlist);

	for(i=0; i<npars; i++) {
		text = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
		if(strcmp("bin-path", text) == 0) {
			path = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-config", text) == 0) {
			sex_conf = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-conv", text) == 0) {
			sex_conv = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-param", text) == 0) {
			sex_par = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-nnw", text) == 0) {
			sex_nnw = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("backthre", text) == 0) {
			dthre = cpl_parameter_get_double(par) ;
		}
		else if(strcmp("backsize", text) == 0){
			backsize = cpl_parameter_get_int(par);
		}
		par = cpl_parameterlist_get_next(parlist);
	}

	/*
	 *  Create command line for Sextractor
	 */

	cmd = cpl_sprintf("%s/sex %s -c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
			"-FILTER N "
			"-DETECT_THRESH %g "
			"-BACK_SIZE %d "
			"-CHECKIMAGE_TYPE BACKGROUND "
			"-CHECKIMAGE_NAME %s",
			path,
			in_name,
			sex_conf,
			sex_par,
			sex_conv,
			sex_nnw,
			dthre,
			backsize,
			out_name);

	if (system(cmd) != 0) {
		cpl_free(cmd);
		cpl_msg_error(cpl_func,"Failed to run Sextractor for background creation");
		return -1;
	}

	cpl_free(cmd);

	return 0;
}


/**
 * @brief    	Creates a background image using Sextractor
 *
 * @param	 sci			Science image
 * @param    parlist     	the parameters list
 *
 * @return   back		Background image or NULL in case of error
 */

cpl_image *omega_background(const cpl_image *sci, const cpl_parameterlist *parlist)
{

	int i = 0;
	int npars = 0;
	double dthre = 0.0;
	char *cmd = NULL;
	const char *sciname = "omega_sci4back.fits";
	const char *out_name = "omega_back.fits";
	const char *path = NULL;
	const char *sex_conf = NULL;
	const char *sex_conv = NULL;
	const char *sex_nnw = NULL;
	const char *sex_par = NULL;
	const char *text = NULL;

	cpl_parameter 		*par;
	cpl_image *back;


	if((sci == NULL) || (parlist == NULL))
		return NULL;


	/*
	 *  Retrieve Sextractor configuration parameters
	 */

	npars = cpl_parameterlist_get_size(parlist);
	par = cpl_parameterlist_get_first(parlist);

	for(i=0; i<npars; i++) {
		text = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
		if(strcmp("bin-path", text) == 0) {
			path = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-config", text) == 0) {
			sex_conf = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-conv", text) == 0) {
			sex_conv = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-param", text) == 0) {
			sex_par = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-nnw", text) == 0) {
			sex_nnw = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("backthre", text) == 0) {
			dthre = cpl_parameter_get_double(par) ;
		}
		par = cpl_parameterlist_get_next(parlist);
	}

	/* Save image in disk to use in Sextractor */
	cpl_image_save(sci, sciname, BITPIX, NULL, CPL_IO_CREATE);

	/*
	 *  Create command line for Sextractor
	 */
	cmd = cpl_sprintf("%s/sex %s -c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
			"-FILTER N "
			"-DETECT_THRESH %g "
			"-CHECKIMAGE_TYPE BACKGROUND "
			"-CHECKIMAGE_NAME %s 2>/dev/null",
			path,
			sciname,
			sex_conf,
			sex_par,
			sex_conv,
			sex_nnw,
			dthre,
			out_name);

	if (system(cmd) != 0) {
		cpl_free(cmd);
		cpl_msg_error(cpl_func,"Failed to run Sextractor for background detection");
		return NULL;
	}

	cpl_free(cmd);

	back = cpl_image_load(out_name,CPL_TYPE_FLOAT,0,0);

	return back;
}

/**@}*/
