/* $Id: omega_bpm.c,v 1.6 2012/01/31 13:38:50 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/31 13:38:50 $
 * $Revision: 1.6 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "omega_dfs.h"
#include "omega_stats.h"
#include "omega_utils.h"
#include "omega_bpm.h"
#include "omega_trim.h"

/**
 * @defgroup omega_bpm  Bad Pixel Mask related functions
 *
 * This module provides functions to deal with bad pixel masks
 *
 * @par Synopsis:
 *   Functions to deal with bad pixel masks
 *
 * @code
 * #include "omega_dfs.h"
 * #include "omega_stats.h"
 * #include "omega_utils.h"
 * #include "omega_bpm.h"
 * #include "omega_trim.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Create an image of its dead pixels

    	  Replace dead pixels by the median of the correct neighbors in a
        [-boxsize/2, boxsize/2] neighborhood around each pixel. If no
        correct pixel can be found in the 8 neighbors, the pixel is set to
        zero.

  @param   dirty image		Image to clean
  @param    dead map		Dead pixel map
  @param	boxsize			Size of clean box
  @return   Newly allocated, clean image
  	Converted from OMEGA Eclipse image_clean_deadpix_median (dead_pixels.c)
  @deprecated Unstable API, may change or disappear. Do not use in new code!


 */
/*----------------------------------------------------------------------------*/
cpl_image * clean_deadpix_median(cpl_image *dirty, cpl_mask *dead, int boxsize)
{

    int i = 0;
    int j = 0;
    int k = 0;
    int l = 0;
    int nx = 0;
    int ny = 0;
    int lx = 0;
    int ly = 0;
    int pixelpos = 0;
    int numgood = 0;
    double replace = 0.0;
    float *goodpix;
    float *cleandata;
    float *dirtydata;

    const char *_id = "clean_deadpix_median";

    cpl_binary			*deaddata;
    cpl_image 			*cleaned;


    if(dirty == NULL){
        cpl_msg_error(_id,"Null input image");
        return NULL;
    }

    lx = cpl_image_get_size_x(dirty);
    ly = cpl_image_get_size_y(dirty);

    dirtydata = (float *)cpl_image_get_data(dirty);

    cleaned = cpl_image_new(lx,ly,CPL_TYPE_FLOAT);
    cleandata = (float *)cpl_image_get_data(cleaned);

    goodpix = calloc((boxsize+1)*(boxsize+1), sizeof(float));
    if(goodpix == NULL){
        cpl_msg_error(_id,"Error in allocating memory");
        cpl_image_delete(cleaned);
        return NULL;
    }

    boxsize = boxsize/2;

    deaddata = cpl_mask_get_data(dead);

/*Replace dead pixels by interpolated value*/
    for(j=0; j<ly; j++){
        for(i=0; i<lx; i++){
            pixelpos = i + j*lx;

            if(deaddata[pixelpos] == CPL_BINARY_1){
                cleandata[pixelpos] = dirtydata[pixelpos];
            }
            else{
                numgood = 0;
                replace = 0.0;
                for(l=-boxsize; l<=boxsize; l++){
                    ny = j+l;
                    for(k=-boxsize; k<=boxsize; k++){
                        nx = i+k;
                        /*Check neighbour id it is inside image
                    and pixel is valid in mask*/

                        if((nx >= 0) && (nx < lx) &&
                            (ny >= 0) && (ny < ly) &&
                            (deaddata[nx+ny*lx] == CPL_BINARY_1)){

                            goodpix[numgood] = (double)dirtydata[nx+ny*lx];
                            numgood++;
                        }
                    }
                }
                /*Take the median value over the good neighbours*/
                if(numgood <=1){
                    cleandata[pixelpos] = (float)0;
                }
                else{
/*		cleandata[pixelpos] = cpl_tools_get_median_float(goodpix, numgood);*/
                    cleandata[pixelpos] = get_median_float(goodpix, numgood);
                }
            }
        }
    }

    free(goodpix);

    return cleaned;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create a bad pixels map

    	  It takes a hot and/or a cold pixels map
        and creates a AND combination of them to
        make a Bad pixels Map.

  @param    hot		Hot pixels frame
  @param    cold		Cold pixels frame
  @param	cxn		Extension number of FITS file to load from
  @return   bpm		A cpl_mask
**/
/*----------------------------------------------------------------------------*/
cpl_mask *makebpm(const cpl_frame *hot, const cpl_frame *cold, int cxn)
{
    cpl_image		*hot_img;
    cpl_image		*cold_img;
    cpl_mask		*bpm;
    cpl_mask		*mask;

    if((hot == NULL) && (cold == NULL))
        return NULL;

    /* FIXME: Maybe I need to do a OR instead */
    if((hot != NULL) && (cold != NULL)){
        hot_img = cpl_image_load(cpl_frame_get_filename(hot), CPL_TYPE_INT, 0, cxn);
        cold_img = cpl_image_load(cpl_frame_get_filename(cold), CPL_TYPE_INT, 0, cxn);
        bpm = cpl_mask_threshold_image_create(hot_img, 0.5, 1.5) ;
        mask = cpl_mask_threshold_image_create(cold_img, 0.5, 1.5) ;
/*        cpl_mask_and(bpm, mask);*/
        cpl_mask_or(bpm, mask);

        cpl_image_delete(hot_img);
        cpl_image_delete(cold_img);
        cpl_mask_delete(mask);
    }
    else if((hot == NULL) && (cold != NULL)){
        cold_img = cpl_image_load(cpl_frame_get_filename(cold), CPL_TYPE_INT, 0, cxn);
        bpm = cpl_mask_threshold_image_create(cold_img, 0.5, 1.5) ;
        cpl_image_delete(cold_img);
    }
    else if((hot != NULL) && (cold == NULL)){
        hot_img = cpl_image_load(cpl_frame_get_filename(hot), CPL_TYPE_INT, 0, cxn);
        bpm = cpl_mask_threshold_image_create(hot_img, 0.5, 1.5) ;
        cpl_image_delete(hot_img);
    }
    else {
    	return NULL;
    }

    return bpm;

}

/**
 * @brief Create a map of the satured pixels of an image
 *
 * @param	frame	Frame where to create map from
 * @param	oc		Integer giving the overscan method to apply
 * @param	ext		Integer of FITS extension
 * @param	spars	Parameters list
 *
 * @return	mask	Returns a cpl_mask of the saturated pixels in
 * the image or NULL in case of error.
 * @deprecated Unstable API, may change or disappear. Do not use in new code!
 */
cpl_mask * create_saturated_map(const cpl_frame *frame, int oc, int ext, cpl_parameterlist *spars)
{

    int i = 0;
    int npars = 0;
    double low = 0.0;
    double high = 0.0;
    const char *alias = NULL;

    cpl_image *trim;
    cpl_mask *mask ;
    cpl_parameter *par;


    if((frame == NULL) || (spars == NULL)){
        cpl_msg_error(cpl_func,"Input image or parameter list is NULL");
        return NULL;
    }


    npars = cpl_parameterlist_get_size(spars);
    par = cpl_parameterlist_get_first(spars);

    for(i=0; i<npars; i++) {
        alias = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
        if(strcmp ("lt-satu", alias) == 0){
            low =  cpl_parameter_get_double(par);
        }
        else if(strcmp ("ht-satu", alias) == 0){
            high =  cpl_parameter_get_double(par);
        }
        par = cpl_parameterlist_get_next(spars);
    }

    trim = TrimOscanCorrect(frame, oc, ext);
    if (trim == NULL) {
        cpl_msg_error(cpl_func,"Cannot trim image");
        return NULL;
    }
    mask = cpl_mask_threshold_image_create(trim, low, high);
    cpl_mask_not(mask);

    freeimage(trim);


/*FIXME: how to count number of saturated pixels?*/

    return mask;
}

/**
 * @brief Create a map of the saturated pixels of an image
 *
 * @param	trim	Trimmed science image
 * @param	oc		Integer giving the overscan method to apply
 * @param	spars	Parameters list
 *
 * @return	mask	Returns a cpl_mask of the saturated pixels in
 * the image or NULL in case of error.
 */
cpl_mask * omega_saturated_map(cpl_image *trim, int oc, cpl_parameterlist *spars)
{

	int i = 0;
	int npars = 0;
	double low = 0.0;
	double high = 0.0;
	const char *alias = NULL;
	double imsize=0.0;
	double badpixel=0.0;
	cpl_mask *mask ;
	cpl_parameter *par;


	if((trim == NULL) || (spars == NULL))
		return NULL;


	npars = cpl_parameterlist_get_size(spars);
	par = cpl_parameterlist_get_first(spars);

	for(i=0; i<npars; i++) {
		alias = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
		if(strcmp ("lt-satu", alias) == 0){
			low =  cpl_parameter_get_double(par);
		}
		else if(strcmp ("ht-satu", alias) == 0){
			high =  cpl_parameter_get_double(par);
		}
		par = cpl_parameterlist_get_next(spars);
	}

	mask = cpl_mask_threshold_image_create(trim, low, high);
	cpl_mask_not(mask);

	imsize=cpl_image_get_size_x(trim)*cpl_image_get_size_y(trim);
	badpixel=cpl_mask_count(mask);

	cpl_msg_debug(cpl_func,"Ratio of saturated pixel: %g", badpixel/imsize);

	cpl_msg_debug(cpl_func,"Detected %g saturated pixels in image",badpixel);

	if(badpixel/imsize > 0.95) {
		cpl_msg_warning(cpl_func," %g percent of the pixels are saturated!",
				badpixel/imsize*100.);
	}

	return mask;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create a Weight Image Map for a science image

    		This function will perform logical AND in all maps
            created (BPM, Saturated map, cosmic map and satellite map)
            to create a final Weight Map

  @param	flat		flatfield image
  @param	bpmsatu	combined BPM and saturated mask
  @param	cosmic		cosmic rays mask
  @param	satellite	satellite tracks mask

  @return   weight		weight map image
 */
/*----------------------------------------------------------------------------*/
cpl_image * create_weightframe(cpl_image *flat, cpl_image *bpmsatu,
		cpl_mask *cosmic,cpl_mask *satellite)
{

	int count = 0;
	int nx = 0;
	int ny = 0;

	cpl_image 			*mask_img;
	cpl_image 			*weight;
	cpl_mask			*bpm;


	if((flat == NULL) || (bpmsatu == NULL)) {
		cpl_msg_debug(cpl_func,"NULL input frame");
		return NULL;
	}

	/*Create a mask from the combined BPM and Saturated images*/
	bpm = cpl_mask_threshold_image_create(bpmsatu, 0.5, 1.5);

	if(cosmic != NULL)
		cpl_mask_or(bpm, cosmic);

	if(satellite != NULL)
		cpl_mask_or(bpm, satellite);

	nx = cpl_mask_get_size_x(bpm);
	ny = cpl_mask_get_size_y(bpm);

	/* Count the bad pixels which have value CP_BINARY_1 */
	cpl_msg_debug(cpl_func,"Weight of % " CPL_SIZE_FORMAT "  bad pixels is set to 0", cpl_mask_count(bpm));

	/*Convert mask to INT*/
	/* The weight will be multiplied by the image, therefore
	 * the bad pixels need to be set to CPL_BINARY_0
	 */
	cpl_mask_not(bpm);
	mask_img = cpl_image_new_from_mask(bpm);

	/*Convert INT to FLOAT*/
	weight = cpl_image_cast(mask_img, CPL_TYPE_FLOAT);

	freeimage(mask_img);

	cpl_image_multiply(weight, flat);

	freemask(bpm);

	return weight;

}

/*----------------------------------------------------------------------------*/
/**
  @deprecated Unstable API, may change or disappear. Do not use in new code!

 */
/*----------------------------------------------------------------------------*/

cpl_mask *omega_bpm_create(cpl_image *img, cpl_parameterlist *pars)
{
	int i,n;
	double low,high;
	const char *name = "omega_temp.fits";
	const char *bname = "omega_back.fits";
	const char *text;
	cpl_parameter *par;
	cpl_image *norm, *back;
	cpl_mask *bpm;

	if(img == NULL)
		return NULL;

	/* Get parameters */
	par = cpl_parameterlist_get_first(pars);

	for(i=0; i<cpl_parameterlist_get_size(pars); i++) {
		text = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
		if(strcmp("low", text) == 0) {
			low = cpl_parameter_get_double(par) ;
			n++;
		}
		else if(strcmp("high", text) == 0){
			high = cpl_parameter_get_double(par);
			n++;
		}
		if (n == 2)
			break;

		par = cpl_parameterlist_get_next(pars);
	}


	/* Save image to disk */
	cpl_image_save(img, name, CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_CREATE);

    /* Call function to create background image*/
    if(omega_create_background(pars, name, bname) != 0){
        cpl_msg_warning(cpl_func,"Error in creating background image");
        return NULL;
    }

    if((back = cpl_image_load(bname, CPL_TYPE_FLOAT, 0, 0)) == NULL)
    	return NULL;

    norm = cpl_image_divide_create(img, back);
    freeimage(back);

    bpm = cpl_mask_threshold_image_create(norm, low, high);
    freeimage(norm);

//    cpl_mask_not(bpm);

    return bpm;

}

/**@}*/
