/* $Id: omega_catalog.c,v 1.9 2012/05/30 12:08:03 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/05/30 12:08:03 $
 * $Revision: 1.9 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "omega_catalog.h"
#include "omega_dfs.h"
#include "omega_pfits.h"
#include "omega_utils.h"

/**
 * @defgroup omega_catalog  Catalogues
 *
 * This module provides functions to manipulate USNOA2 and reference
 * standard star catalogues.
 *
 * @par Synopsis:
 *   Functions to manipulate catalogues
 *
 * @code
 *   #include "omega_catalog.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Static Functions
 -----------------------------------------------------------------------------*/

static float kselect(float *a, int n, int k) {
    while (n > 1) {
        int i = 0, j = n - 1;
        float x = a[j/2], w;

        do {
            while (a[i] < x) i++;
            while (a[j] > x) j--;
            if (i < j) {
                w = a[i]; a[i] = a[j]; a[j] = w;
            } else {
                if (i == j) i++;
                break;
            }
        } while (++i <= --j);

        if (k < i)
            n = i;
        else {
            a += i; n -= i; k -= i;
        }
    }

    return a[0];
}

static float omega_med(float *data, unsigned char *bpm, long npts) {
    int i,j,is_even,ilevel;
    float *buf,value;

    /* If there is not BPM, then just do a straight forward median */

    buf = cpl_malloc(npts*sizeof(*buf));
    if (bpm == NULL) {
        is_even = !(npts & 1);
        memmove((char *)buf,(char *)data,npts*sizeof(float));
        if (is_even) {
            ilevel = npts/2 - 1;
            value = kselect(buf,npts,ilevel);
            ilevel = npts/2;
            value = 0.5*(value + kselect(buf,npts,ilevel));
        } else {
            ilevel = npts/2;
            value = kselect(buf,npts,ilevel);
        }

    /* Otherwise get rid of the dodgy values and then do the median */

    } else {
        j = 0;
        for (i = 0; i < npts; i++) {
            if (bpm[i] == 0)
                buf[j++] = data[i];
        }
        if (j == 0) {
            cpl_free(buf);
            value = CX_MAXFLOAT;
            return(value);
        }
        is_even = !(j & 1);
        if (is_even) {
            ilevel = j/2 - 1;
            value = kselect(buf,j,ilevel);
            ilevel = j/2;
            value = 0.5*(value + kselect(buf,j,ilevel));
        } else {
            ilevel = j/2;
            value = kselect(buf,j,ilevel);
        }
    }
    cpl_free(buf);
    return(value);
}

static void omega_medmad(float *data, unsigned char *bpm, long np, float *med,
			  float *mad) {
    int i;
    float *work;

    /* First find the median value */

    *med = omega_med(data,bpm,np);

    /* Now work out the MAD. Start by getting a bit of workspace and filling
       it with absolute residuals */

    work = cpl_malloc(np*sizeof(*work));
    for (i = 0; i < np; i++)
	work[i] = (float)fabs((double)(data[i] - *med));

    /* Now get the median value of the absolute deviations */

    *mad = omega_med(work,bpm,np);

    /* Tidy and exit */

    cpl_free(work);
}

static int omega_fndmatch(float x, float y, float *xlist, float *ylist,
			   			int nlist, float err)
{
    int isp,ifp,index,i;
    float errsq,errmin,dx,dy,poserr;

    /* Find lower limit index */

    isp = 0;
    ifp = nlist - 1;
    errsq = err*err;
    index = (isp + ifp)/2;
    while (ifp-isp >= 2) {
        if (ylist[index] < y - err) {
            isp = index;
            index = (index+ifp)/2;
        } else if (ylist[index] > y - err) {
            ifp = index;
            index = (index+isp)/2;
        } else {
            isp = index;
            break;
        }
    }

    /* Now find nearest one within limit */

    index = -1;
    errmin = errsq;
    for (i = isp; i < nlist; i++) {
        if (ylist[i] > y+err)
            break;
        dx = x - xlist[i];
        dy = y - ylist[i];
        poserr = dx*dx + dy*dy;
        if (poserr < errsq) {
            if (poserr <= errmin) {
                index = i;
                errmin = poserr;
            }
        }
    }
    return(index);
}


/*-----------------------------------------------------------------------------*/
/**
 * @brief		Get the Ra and Dec coverage of an image
 *
 * @param	plist	Property list contaning NAXIS information
 * @param	fudge   Refining the search
 * @param	from    Matrix with pixel coordinates
 * @param	ra1     Output initial RA coordinate
 * @param	ra2     Output final RA coordinate
 * @param	dec1    Output initial DEC coordinate
 * @param	dec2    Output final DEC coordinate
 *
 * @return	0	if everything is OK or -1 in case of error.
 **/
/*------------------------------------------------------------------------------*/
int omega_get_coverage(cpl_propertylist *plist, int fudge, cpl_matrix *from, double *ra1,
			   double *ra2, double *dec1, double *dec2)
{

    cpl_wcs *wcs;
    double ra,dec,dra,ddec,boxfudge,min_4q,max_1q;
    int first_quad=0,fourth_quad=0, ns=0;
    long i=0,naxes[2];
    cpl_matrix *to=NULL;
    cpl_array *status = NULL;

    /* Initialise these in case of failure later*/

    *ra1 = 0.0;
    *ra2 = 0.0;
    *dec1 = 0.0;
    *dec2 = 0.0;

    /* Grab the WCS info from the property list */
    wcs = cpl_wcs_new_from_propertylist(plist);
    if(wcs == NULL){
        cpl_msg_error(cpl_func,"Cannot create wcs. %s",cpl_error_get_message());
        return -1;
    }

    /* Get the size of the data array */
    naxes[0] = (long)cpl_propertylist_get_int(plist, "NAXIS1");
    naxes[1] = (long)cpl_propertylist_get_int(plist, "NAXIS2");

    /* Find the RA and Dec limits of the image */

    *ra1 = 370.0;
    *ra2 = -370.0;
    *dec1 = 95.0;
    *dec2 = -95.0;
    first_quad = 0;
    fourth_quad = 0;
    min_4q = 370.0;
    max_1q = 0.0;

    cpl_wcs_convert(wcs, from, &to, &status, CPL_WCS_PHYS2WORLD);
    freearray(status);
    ns = cpl_matrix_get_nrow(to);

    for(i=0; i<ns; i++){
        ra = (double)cpl_matrix_get(to, i, 0);
        dec = (double)cpl_matrix_get(to, i, 1);
        if (ra >= 0.0 && ra <= 90.0) {
            first_quad = 1;
            max_1q = omega_max(ra,max_1q);
        }
        else if (ra >= 270.0 && ra <= 360.0) {
            fourth_quad = 1;
            min_4q = omega_min((ra-360.0),min_4q);
        }

        *ra1 = omega_min(*ra1,ra);
        *ra2 = omega_max(*ra2,ra);
        *dec1 = omega_min(*dec1,dec);
        *dec2 = omega_max(*dec2,dec);
    }

/*        for (j = 1; j < naxes[1]; j += 10) {
        j = min(j,naxes[1]);
        y = (double)j;
        for (i = 1; i < naxes[0]; i += 10) {
            i = min(i,naxes[0]);
            x = (double)i;
            vircam_xytoradec(wcs,x,y,&ra,&dec);
            if (ra >= 0.0 && ra <= 90.0) {
                first_quad = 1;
                max_1q = max(ra,max_1q);
            } else if (ra >= 270.0 && ra <= 360.0) {
                fourth_quad = 1;
                min_4q = min((ra-360.0),min_4q);
            }
            *ra1 = min(*ra1,ra);
            *ra2 = max(*ra2,ra);
            *dec1 = min(*dec1,dec);
            *dec2 = max(*dec2,dec);
        }
    }*/
    freewcs(wcs);

    /* Now have a look to see if you had RA values in both the first and
       fourth quadrants.  If you have, then make the minimum RA a negative
       value.  This will be the signal to the caller that you have the
       wraparound... */

    if (first_quad && fourth_quad) {
        *ra1 = min_4q;
        *ra2 = max_1q;
    }

    /* Pad out search a bit */

    if (fudge) {
        boxfudge = 0.01*(float)fudge;
        dra = 0.5*boxfudge*(*ra2 - *ra1);
        *ra1 -= dra;
        *ra2 += dra;
        ddec = 0.5*boxfudge*(*dec2 - *dec1);
        *dec1 -= ddec;
        *dec2 += ddec;
    }

    freematrix(to);

    /* Exit */

    return 0;
}

/*------------------------------------------------------------------------------*/
/**
 * @brief			Find matches between two catalogues
 * @param	objtbl		Objects table
 * @param   stdstab		Catalog table
 * @param	srad		radius for match in pixels
 * @param	outtab		output table with matches
 *
 * @return	the number of matches
 *
 *  It assumes that the two catalogues are sorted by the Y coordinate
 **/
/*--------------------------------------------------------------------------------*/
/*Note: adapted from vircam pipeline*/
int omega_matchstds(cpl_table *objtab, cpl_table *stdstab, float srad,
			    	cpl_table **outtab)
{

    char *colname;
    int nobj,nstd,ngrid,ngrid2,ibest,ig,jg,nmatch,k,*matches,jm,l,dont,null;
    float *xstd,*ystd,*xobj,*yobj,aveden,errlim,xoffbest,yoffbest,*xoffs;
    float *yoffs,x,y,x2,y2,r2,x1,y1,r1,xoffmed,sigx,yoffmed,sigy,xoff,yoff;
    cpl_propertylist *p;
    cpl_table *mstds;

     *outtab = NULL;

    /* Check the size of each of the tables */
    nobj = cpl_table_get_nrow(objtab);
    nstd = cpl_table_get_nrow(stdstab);
    if (nobj == 0) {
	    cpl_msg_error(cpl_func,"Object table has no rows");
        return -1;
    }
    else if (nstd == 0) {
	    cpl_msg_error(cpl_func,"Standards RA/DEC table has no rows");
        return -1;
    }

    /* Get the x,y coordinates for each table */
    xobj = cpl_table_get_data_float(objtab,"X_IMAGE");
    yobj = cpl_table_get_data_float(objtab,"Y_IMAGE");
    xstd = cpl_table_get_data_float(stdstab,"xpredict");
    ystd = cpl_table_get_data_float(stdstab,"ypredict");
    if (xstd == NULL || ystd == NULL || xobj == NULL || yobj == NULL){
		cpl_msg_error(cpl_func,"NULL input data");
        return -1;
    }

    /* Calculate the error limit and the number of grid points */
    aveden = (float)nstd/(float)(NX*NY);
    errlim = 1.0/sqrt(4.0*CPL_MATH_PI*aveden);
    errlim = omega_min(errlim,5.0);
    ngrid = (int)(srad/errlim);
    ngrid = (ngrid/2)*2 + 1;
    ngrid = omega_max(5,omega_min(NGRIDMAX,ngrid));
    ngrid2 = ngrid/2 + 1;

    /* Now search for the best solution */
    ibest = 0;
    xoffbest = 0.0;
    yoffbest = 0.0;
    for (ig = -ngrid2; ig <= ngrid2; ig++) {
	    xoff = (float)ig*errlim*CPL_MATH_SQRT2;
	    for (jg = -ngrid2; jg <= ngrid2; jg++) {
	        yoff = (float)jg*errlim*CPL_MATH_SQRT2;
	        nmatch = 0;
            for (k = 0; k < nobj; k++) {
		        x = xobj[k] + xoff;
		        y = yobj[k] + yoff;
		        if (omega_fndmatch(x,y,xstd,ystd,nstd,errlim) > -1)
		            nmatch++;
	        }
	        if (nmatch > ibest) {
		        ibest = nmatch;
		        xoffbest = xoff;
		        yoffbest = yoff;
	        }
	    }
    }

    /* Now allocate some workspace so that you can calculate a good median
       coordinate difference */
    xoffs = cpl_malloc(nstd*sizeof(*xoffs));
    yoffs = cpl_malloc(nstd*sizeof(*yoffs));
    matches = cpl_malloc(nstd*sizeof(*matches));
    for (k = 0; k < nstd; k++)
	    matches[k] = -1;

    /* Now get the best matches */
    nmatch = 0;
    for (k = 0; k < nstd; k++) {
	    x = xstd[k] - xoffbest;
	    y = ystd[k] - yoffbest;
	    jm = omega_fndmatch(x,y,xobj,yobj,nobj,errlim);
	    if (jm > -1) {
	        dont = 0;
	        x2 = xobj[jm] - x;
	        y2 = yobj[jm] - y;
	        r2 = sqrt(x2*x2 + y2*y2);
	        for (l = 0; l < nstd; l++) {
		        if (matches[l] == jm) {
 		            x1 = xobj[jm] - (xstd[l] - xoffbest);
		            y1 = yobj[jm] - (ystd[l] - yoffbest);
		            r1 = sqrt(x1*x1 + y1*y1);
		            if (r2 < r1)
			            matches[l] = -1;
		            else
			            dont = 1;
		            break;
		        }
	        }
	        if (dont == 0)
		        matches[k] = jm;
	    }
    }

    /* Now get the coordinate difference for the best matches */
    for (k = 0; k < nstd; k++) {
	jm = matches[k];
	if (jm != -1) {
	    xoffs[nmatch] = xobj[jm] - xstd[k];
	    yoffs[nmatch] = yobj[jm] - ystd[k];
	    nmatch++;
	}
    }
    if (nmatch == 0) {
	xoffmed = 0.0;
	sigx = 1.0;
	yoffmed = 0.0;
	sigy = 1.0;
    } else {
        omega_medmad(xoffs,NULL,nmatch,&xoffmed,&sigx);
        sigx *= 1.48;
        omega_medmad(yoffs,NULL,nmatch,&yoffmed,&sigy);
        sigy *= 1.48;
    }

    /* Now go through one final time with a reduced error box and get
       the final matches */
    errlim = 3.0*omega_max(sigx,sigy);
    for (k = 0; k < nstd; k++)
	    matches[k] = -1;
    for (k = 0; k < nstd; k++) {
	    x = xstd[k] + xoffmed;
	    y = ystd[k] + yoffmed;
	    jm = omega_fndmatch(x,y,xobj,yobj,nobj,errlim);
	    if (jm > -1) {
	        dont = 0;
	        x2 = xobj[jm] - x;
	        y2 = yobj[jm] - y;
	        r2 = sqrt(x2*x2 + y2*y2);
	        for (l = 0; l < nstd; l++) {
		        if (matches[l] == jm) {
 		            x1 = xobj[jm] - (xstd[l] + xoffmed);
		            y1 = yobj[jm] - (ystd[l] + yoffmed);
		            r1 = sqrt(x1*x1 + y1*y1);
		            if (r2 < r1)
			            matches[l] = -1;
		            else
			            dont = 1;
/* 		                break; */
		        }
	        }
	        if (dont == 0)
		        matches[k] = jm;
	    }
    }
    jm = matches[1];

    /* Make a copy of the standards table and add all the columns from the
       object catalogue to it. Ingore the RA and DEC columns in the catalogue
       as the standards table will already have these.*/

    mstds = cpl_table_duplicate(stdstab);
    colname = (char *)cpl_table_get_column_name(objtab);
    while (colname != NULL) {
        if (strcmp(colname,"X_WORLD") && strcmp(colname,"Y_WORLD"))
  	        cpl_table_new_column(mstds,colname,
  	        					cpl_table_get_column_type(objtab,colname));

	    colname = (char *)cpl_table_get_column_name(NULL);
    }
    cpl_table_unselect_all(mstds);

    /* Now go through and find the matches.  */

    for (k = 0; k < nstd; k++) {
	jm = matches[k];
	if (jm != -1) {
            colname = (char *)cpl_table_get_column_name(objtab);
	    while (colname != NULL) {
  	        if (!strcmp(colname,"X_WORLD") || !strcmp(colname,"Y_WORLD")) {
                    colname = (char *)cpl_table_get_column_name(NULL);
		    continue;
		}
		null = 0;
		switch (cpl_table_get_column_type(objtab,colname)) {
		case CPL_TYPE_INT:
		    cpl_table_set_int(mstds,colname,k,
				      cpl_table_get_int(objtab,colname,jm,
							&null));
		    break;
		case CPL_TYPE_FLOAT:
		    cpl_table_set_float(mstds,colname,k,
					cpl_table_get_float(objtab,colname,jm,
							    &null));
		    break;
		case CPL_TYPE_DOUBLE:
		    cpl_table_set_double(mstds,colname,k,
					 cpl_table_get_double(objtab,colname,
							      jm,&null));
		    break;
		default:
		    cpl_table_set_float(mstds,colname,k,
					cpl_table_get_float(objtab,colname,jm,
							    &null));
		    break;
		}
                colname = (char *)cpl_table_get_column_name(NULL);
	    }
	    cpl_table_select_row(mstds,k);
	}
    }

    /* Now extract the selected rows into the output table */

    *outtab = cpl_table_extract_selected(mstds);
    cpl_table_delete(mstds);

    /* Tidy up */

    cpl_free(matches);
    cpl_free(xoffs);
    cpl_free(yoffs);

    return 0;
}
/*-------------------------------------------------------------------------*/
/**	@brief			Read master usnoa2 fits table and return
 * 					a fits table that contains stars in a given
 * 					region.
 *
 * @param	catalog		cpl_frame of usnoa2 master table
 * @param	coords		vector with 4 fields, RA min,RA max, DEC min, DEC max
 * @params	nstds		NULL (number of sources to be populated by function)
 *
 * @return	all			table of all standards in the field
 * 		Input master table has the following structure:
 * 		USNOA2_CAT	RA_min	RA_max	Dec_min	Dec_max	N_sources
 *
 * 		Each usno FITS table contains sources in the whole range of RA
 * 		divided in slices of 1.5 degrees in DEC. The structure of these tables is:
 *      USNO_A2_number RA2000  Dec2000  MagB  MagR Plate
 **/
 /*--------------------------------------------------------------------------*/
cpl_table *omega_get_usnoa_table(const cpl_frame *catalog, cpl_vector *coords, int *nstds)
{
	int number = 0;
	int i = 0;
//	int irow = 0;
	double ramin = 0.0;
	double ramax = 0.0;
	double decmin = 0.0;
	double decmax = 0.0;
	char *pathname = NULL;
//	const char *name = NULL;
//	char *tname = NULL;
	cpl_table *table = NULL;
	cpl_table *extracted;
	cpl_table *all;

	/* Load master usnoa2 table */
	table = cpl_table_load(cpl_frame_get_filename(catalog), 1, 0);
	if( table == NULL){
		return NULL;
	}

	/* Get pathname */
	pathname = omega_get_pathname(cpl_frame_get_filename(catalog));

	/* Get coordinates of desired region */
	ramin = cpl_vector_get(coords, 0);
	ramax = cpl_vector_get(coords, 1);
	decmin = cpl_vector_get(coords, 2);
	decmax = cpl_vector_get(coords, 3);

	/* Select the fits tables that fall into the desired region */
	cpl_table_select_all(table);
	cpl_table_and_selected_double(table,"Dec_max",CPL_NOT_LESS_THAN,decmin);
	number = cpl_table_and_selected_double(table,"Dec_min",CPL_NOT_GREATER_THAN,decmax);
	if(number == 0){
		freetable(table);
		return NULL;
	}

	extracted = cpl_table_extract_selected(table);
	freetable(table);

	/* Merge the tables into one after selecting */
	for(i=0; i<number; i++){
		const char *name = cpl_table_get_string(extracted,"USNOA2_CAT",i);
		char *tname = cpl_sprintf("%s/%s",pathname,name);
		cpl_table *t = cpl_table_load(tname, 1, 0);

		/* Select stars falling in RA range */
		cpl_table_select_all(t);
		cpl_table_and_selected_double(t, "RA2000",CPL_NOT_LESS_THAN, ramin);
		cpl_table_and_selected_double(t, "RA2000", CPL_NOT_GREATER_THAN, ramax);

		/* Select stars falling in DEC range */
		cpl_table_and_selected_double(t, "Dec2000", CPL_NOT_LESS_THAN, decmin);
		int number = cpl_table_and_selected_double(t, "Dec2000", CPL_NOT_GREATER_THAN, decmax);
		if (number <= 0) {
			freetable(t);
			freespace(tname);
			freetable(extracted);
			return NULL;
		}

		/* Extract selected stars */
		cpl_table *ext = cpl_table_extract_selected(t);
		if(i==0){
			all = cpl_table_new(0);
			cpl_table_copy_structure(all, ext);
		}

		int irow = cpl_table_get_nrow(all);

		cpl_table_insert(all, ext, irow+1);
		freetable(ext);
		freetable(t);
		freespace(tname);
	}

	*nstds = cpl_table_get_nrow(all);
	//	printf("n=%d",nstds);

	freetable(extracted);

	return all;
}

/*-------------------------------------------------------------------------*/
/**	@brief			Read standard reference fits table and return
 * 					a fits table that contains stars in a given
 * 					region.
 *
 * @param	stdcat		pathname of std table
 * @param	coords		vector with 4 fields, RA min,RA max, DEC min, DEC max
 * @params	nstds		NULL (number of sources to be populated by function)
 *
 * @return	A table with all the standards in the field
 *
 * 		Input reference table has the following structure:
 * \code
 * 		SeqNr origin Name Ra Ra_err Dec Dec_err Epoch Flag JohnsonU JohnsonU_err etc. for
 * JohnsonB JohnsonV CousinsR,I SloanU,G,R,I,Z
 * \endcode
 *
 **/
 /*--------------------------------------------------------------------------*/
cpl_table *omega_get_stds_table(const char *stdcat, cpl_vector *coords, int *nstds)
{
	int number = 0;
	double ramin = 0.0;
	double ramax = 0.0;
	double decmin = 0.0;
	double decmax = 0.0;
	cpl_table *table = NULL;
	cpl_table *extracted;

	/* Load reference std table */
	table = cpl_table_load(stdcat, 1, 0);
	if( table == NULL){
		cpl_msg_warning(cpl_func,"Cannot load reference catalogue %s",stdcat);
		return NULL;
	}

	/* Get coordinates of desired region */
	ramin = cpl_vector_get(coords, 0);
	ramax = cpl_vector_get(coords, 1);
	decmin = cpl_vector_get(coords, 2);
	decmax = cpl_vector_get(coords, 3);

	/* Select the stars that fall into the desired region */
	cpl_table_select_all(table);

	cpl_table_and_selected_double(table,"Dec",CPL_NOT_LESS_THAN,decmin);

	cpl_table_and_selected_double(table,"Dec",CPL_NOT_GREATER_THAN,decmax);

	cpl_table_and_selected_double(table,"Ra",CPL_NOT_LESS_THAN,ramin);

	number = cpl_table_and_selected_double(table,"Ra",CPL_NOT_GREATER_THAN,ramax);

	if(number <= 0){
		cpl_msg_debug(cpl_func,"There are no stars in catalogue in given region");
		freetable(table);
		return NULL;
	}

	extracted = cpl_table_extract_selected(table);
	freetable(table);

	*nstds = cpl_table_get_nrow(extracted);


	return extracted;
}


/**
 * --------------------------------------------------------------------------
 * @brief			Filter a table with a column of type integer
 *
 * @param	src			table containing column @em col
 * @param   col         column name
 * @param   operator    CPL relational operator
 * @param   value       reference value
 * @param	nsrc		number of selected rows in table
 *
 * @return	out			table containing only rows fulfilling criterium
 *
 * 		It will select from unselected rows only the rows that fulfill
 *      the condition.
 */
cpl_table *omega_filter_table_int(cpl_table *src, const char *col, cpl_table_select_operator operator,
		                         int value, int *nsrc)
{
	cpl_table *out;

	if(cpl_table_has_column(src, col) != 1)
		return NULL;

	if(cpl_table_get_column_type(src, col) != CPL_TYPE_INT)
		return NULL;

	cpl_table_unselect_all(src);
	cpl_table_or_selected_int(src, col, operator, value);
	out = cpl_table_extract_selected(src);
	*nsrc = cpl_table_get_nrow(out);

	return out;

}

/*-------------------------------------------------------------------------*/
/**	@brief			Filter a table with a column of type double
 *
 * @param	src			table containing column col
 * @param   col         column name
 * @param   operator    CPL relational operator
 * @param   value       reference value
 * @param	nsrc		number of selected rows in table
 *
 * @return	out			table with only rows fulfilling criterium
 *
 * 		It will select from unselected rows only the rows that fulfill the condition
 **/

cpl_table *omega_filter_table_double(cpl_table *src, const char *col, cpl_table_select_operator operator,
		                     double value, int *nsrc)
{
	cpl_table *out;

	if(cpl_table_has_column(src, col) != 1)
		return NULL;

	if(cpl_table_get_column_type(src, col) != CPL_TYPE_DOUBLE)
		return NULL;

	cpl_table_unselect_all(src);
	cpl_table_or_selected_double(src, col, operator, value);
	out = cpl_table_extract_selected(src);
	*nsrc = cpl_table_get_nrow(out);

	return out;

}

/*-------------------------------------------------------------------------*/
/**	@brief			Filter a table with a column of type string
 *
 * @param	src			table containing column col
 * @param   col         column name
 * @param   operator    CPL relational operator
 * @param   value       reference value
 * @param	nsrc		number of selected rows in table
 *
 * @return	out			table with only rows fulfilling criterium
 *
 * 		It will select from unselected rows only the rows that fulfill the condition
 **/

cpl_table *omega_filter_table_string(cpl_table *src, const char *col, cpl_table_select_operator operator,
		                     const char *value, int *nsrc)
{
	cpl_table *out;

	if(cpl_table_has_column(src, col) != 1)
		return NULL;

	if(cpl_table_get_column_type(src, col) != CPL_TYPE_STRING)
		return NULL;

	cpl_table_unselect_all(src);
	cpl_table_or_selected_string(src, col, operator, value);
	out = cpl_table_extract_selected(src);
	*nsrc = cpl_table_get_nrow(out);

	return out;

}

/*-------------------------------------------------------------------------*/
/**	@brief			Filter a table with a column of type float
 *
 * @param	src			table containing column col
 * @param   col         column name
 * @param   operator    CPL relational operator
 * @param   value       reference value
 * @param	nsrc		number of selected rows in table
 *
 * @return	out			table with only rows fulfilling criterium
 *
 * 		It will select from unselected rows only the rows that fulfill the condition
 **/

cpl_table *omega_filter_table_float(cpl_table *src, const char *col, cpl_table_select_operator operator,
		                     float value, int *nsrc)
{
	cpl_table *out;

	if(cpl_table_has_column(src, col) != 1)
		return NULL;

	if(cpl_table_get_column_type(src, col) != CPL_TYPE_FLOAT)
		return NULL;

	cpl_table_unselect_all(src);
	cpl_table_or_selected_float(src, col, operator, value);
	out = cpl_table_extract_selected(src);
	*nsrc = cpl_table_get_nrow(out);

	return out;

}

/**
 * @brief		Create a catalogue from a science frame
 *
 * @param	sci		Name of science image
 * @param	conf	Name of weight map
 * @param	pars	Parameter list
 *
 * @return	A cpl_table with the catalogue or NULL in case of error.
 *
 */
cpl_table *omega_science_catalogue(const char *sci, const char *conf, double zeropoint_final,  cpl_parameterlist *pars)
{
	int i = 0;
	int nsources = 0;
	int npars = 0;
	double cat_thre = 1.0;
	char *cmd = NULL;
	const char *path = NULL;
	const char *sex_conf = NULL;
	const char *sex_conv = NULL;
	const char *sex_nnw = NULL;
	const char *sex_par = NULL;
	const char *text = NULL;
	const char *srcs = "omega_sci_cat.fits";

	cpl_table *table, *sources;
	cpl_parameter *par;


	/* Get parameters */
	npars = cpl_parameterlist_get_size(pars);
	par = cpl_parameterlist_get_first(pars);

	for(i=0; i<npars; i++) {
		text = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
		if(strcmp("bin-path", text) == 0) {
			path = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-config", text) == 0) {
			sex_conf = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-conv", text) == 0) {
			sex_conv = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-param", text) == 0) {
			sex_par = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-nnw", text) == 0) {
			sex_nnw = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("cat-thre", text) == 0){
			cat_thre = cpl_parameter_get_double(par);
		}
		par = cpl_parameterlist_get_next(pars);

	}

	cpl_msg_debug(cpl_func,"zeropoint_final: %f",zeropoint_final);


	/* Create catalog of sources from image */
	cmd = cpl_sprintf("%s/sex %s -c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
			"-FILTER Y -DETECT_THRESH %g -ANALYSIS_THRESH %g "
			"-WEIGHT_TYPE MAP_WEIGHT -WEIGHT_IMAGE %s -CATALOG_TYPE FITS_1.0 -CATALOG_NAME %s -MAG_ZEROPOINT %g "
			"-BACKPHOTO_TYPE LOCAL",
			path,
			sci,
			sex_conf,
			sex_par,
			sex_conv,
			sex_nnw,
			cat_thre,
			cat_thre,
			conf,
			srcs,
			zeropoint_final);

	cpl_msg_debug(cpl_func,"Sextractor call: %s",cmd);

	if (system(cmd) != 0) {
		cpl_msg_debug(cpl_func,"Failed to create catalogue using Sextractor");
		freespace(cmd);
		return NULL;
	}
	freespace(cmd);

	table = cpl_table_load(srcs, 1, 0);
	if((table == NULL) || (cpl_table_get_nrow(table) <= 0)){
		cpl_msg_debug(cpl_func,"Science catalogue table is empty or NULL");
		return NULL;
	}

	/* Filter for FLAGS = 0  */
	sources = omega_filter_table_int(table, "FLAGS", CPL_EQUAL_TO, 0, &nsources);

	freetable(table);


	return sources;
}

/**@}*/

/*
 * Alternative matching function for tests
 *
 */
cpl_table * wcscor_matchstds(cpl_table *objtab, cpl_table *stdstab, float srad)
{

    int i = 0;
    int j = 0;
    int k = 0;
    int nobj, nstd;
    float *xobj;
    float *yobj;
    float *xstd;
    float *ystd;
    const char *colname;
    cpl_table *matches, *temp;
    cpl_propertylist *p;

    nobj = cpl_table_get_nrow(objtab);
    nstd = cpl_table_get_nrow(stdstab);
    if (nobj == 0) {
        cpl_msg_warning(cpl_func,"Object table has no rows");
    } else if (nstd == 0) {
        cpl_msg_warning(cpl_func,"Reference standards table has no rows");
    }

/* First, sort the two tables by the Y coordinate */
/*    p = cpl_propertylist_new();
    cpl_propertylist_append_bool(p,"Y_IMAGE",0);
    if (cpl_table_sort(objtab,p) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func,"Cannot sort sources table");
            cpl_propertylist_delete(p);
            return NULL;
    }
    cpl_propertylist_erase(p,"Y_IMAGE");
    cpl_propertylist_append_bool(p,"ypredict",0);
    if (cpl_table_sort(stdstab,p) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func,"Cannot sort stds table");
            cpl_propertylist_delete(p);
            return NULL;
    }
    cpl_propertylist_delete(p);
*/
    cpl_table_unselect_all(objtab);
    cpl_table_unselect_all(stdstab);

    /* Get columns data */
    xobj = cpl_table_get_data_float(objtab,"X_IMAGE");
    xstd = cpl_table_get_data_float(stdstab, "xpredict");
    yobj = cpl_table_get_data_float(objtab,"Y_IMAGE");
    ystd = cpl_table_get_data_float(stdstab, "ypredict");
//    ra = cpl_table_get_data_double(stdstab, "RA2000");
//    dec = cpl_table_get_data_double(stdstab, "Dec2000");

    if (xstd == NULL || ystd == NULL || xobj == NULL || yobj == NULL){
        cpl_msg_error(cpl_func,"Fatal error in wcscor_matchstds");
        return NULL;
    }

    k = 0;

    for (i=0; i<nobj; i++){
        for(j=0; j<nstd; j++){
            if((xstd[j] <= xobj[i] +srad) && (xstd[j] >= xobj[i] -srad) &&
                (ystd[j] <= yobj[i] +srad) && (ystd[j] >= yobj[i] -srad)) {
                    cpl_table_select_row(stdstab, j);
                    cpl_table_select_row(objtab, i);
                k++;

                break;
           }
        }
    }

    matches = cpl_table_extract_selected(stdstab);
    temp = cpl_table_extract_selected(objtab);
    for (i=0; i<cpl_table_get_ncol(temp); i++){
        colname = cpl_table_get_column_name(temp);
        cpl_table_move_column(matches, colname, temp);
    }


    return matches;
}
