/* $Id: omega_cosmic.c,v 1.4 2012/01/31 13:38:50 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/31 13:38:50 $
 * $Revision: 1.4 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <stdlib.h>

#include "omega_background.h"
#include "omega_cosmic.h"
#include "omega_dfs.h"
#include "omega_utils.h"

/**
 * @defgroup omega_cosmics  Cosmic ray detection
 *
 * This module provides functions to detect cosmic rays
 *
 * @par Synopsis:
 *   Functions to detect cosmic rays
 *
 * @code
 * #include "omega_background.h"
 * #include "omega_cosmic.h"
 * #include "omega_dfs.h"
 * #include "omega_utils.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/


static cpl_mask *detcosmic(cpl_image *nscience, cpl_parameterlist *spars);

/*----------------------------------------------------------------------------*/
/**
 * @internal
 * @brief	nscience	Science image
 *
 * @param	spars		Parameters list
 *
 * @return	cosmic		Mask of detected cosmic rays
 */
/*----------------------------------------------------------------------------*/
static cpl_mask *detcosmic(cpl_image *nscience, cpl_parameterlist *spars)
{

	int count = 0;
	int i = 0;
	int npars = 0;

	double thre = 0.0;

	char *cmd = NULL;
	const char *cosmiccat = "omega_cosmic.cat";
	const char *cosmicname = "omega_cosmic.fits";
	const char *path = NULL;
	const char *sex_conf = NULL;
	const char *sex_conv = NULL;
	const char *sex_nnw = NULL;
	const char *sex_par = NULL;
	const char *text = NULL;
	const char *tmpname23 = "temp0023.fits";

	cpl_image 			*img;
	cpl_mask 			*cosmic_map;
	cpl_parameter 		*par;
	cpl_propertylist	*plist;


	if(nscience == NULL)
		return NULL;


	/* Retrieve Sextractor configuration parameters */
	par = cpl_parameterlist_get_first(spars);
	npars = cpl_parameterlist_get_size(spars);

	for(i=0; i<npars; i++) {
		text = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
		if(strcmp("bin-path", text) == 0) {
			path = cpl_parameter_get_string(par) ;
		}
		else if(strcmp("sex-cosmic", text) == 0) {
			sex_conf = cpl_parameter_get_string(par);
		}
		else if(strcmp("sex-cosfilt", text) == 0) {
			sex_conv = cpl_parameter_get_string(par);
		}
		else if(strcmp("sex-cosmic-param", text) == 0) {
			sex_par = cpl_parameter_get_string(par);
		}
		else if(strcmp("sex-nnw", text) == 0) {
			sex_nnw = cpl_parameter_get_string(par);
		}
		else if(strcmp("sex-cosmic-det", text) == 0) {
			thre = cpl_parameter_get_double(par);
		}
		par = cpl_parameterlist_get_next(spars);

	}


	/*Save temporary reduced science image to disk with no extension header*/
	cpl_image_save(nscience, tmpname23,BITPIX, NULL, CPL_IO_CREATE);


	/*
	 *  Create command line for Sextractor
	 */

	cmd = cpl_sprintf("%s/sex %s -c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
			"-FILTER Y "
			"-DETECT_THRESH %g "
			"-CATALOG_NAME %s "
			"-CATALOG_TYPE FITS_1.0 "
			"-CHECKIMAGE_NAME %s 2>/dev/null",
			path,
			tmpname23,
			sex_conf,
			sex_par,
			sex_conv,
			sex_nnw,
			thre,
			cosmiccat,
			cosmicname);


	if (system(cmd) != 0) {
		cpl_free(cmd);
		cpl_msg_debug(cpl_func,"Failed to run Sextractor");
		return NULL;
	}
	cpl_free(cmd);

	img = cpl_image_load(cosmicname, CPL_TYPE_INT, 0, 0);
	if(img == NULL){
		cpl_msg_debug(cpl_func,"Cannot load cosmic rays image");
		return NULL;
	}

	cosmic_map = cpl_mask_threshold_image_create(img, -0.5, 0.5);

	/* Count the number of CPL_BINARY_1 which in this image
	 * have pixels with high values (> 0.5)*/
	cpl_mask_not(cosmic_map);
	count = cpl_mask_count(cosmic_map);
	/*	cpl_msg_info("","Detected %d pixels with cosmic rays",count);*/

	/*Invert back*/
//	cpl_mask_not(cosmic_map);

	plist = cpl_propertylist_load(cosmiccat, 1);
	if(plist == NULL) {
		cpl_msg_debug(cpl_func,"Cannot load header of %s",cosmiccat);
		freeimage(img);
		freemask(cosmic_map);
		return NULL;
	}

	/*The length of the catalog tells how many cosmic rays were detected*/
	count = cpl_propertylist_get_int(plist,"NAXIS2");
	cpl_msg_debug(cpl_func,"Detected %d cosmic ray events",count);

	freeimage(img);
	freeplist(plist);

	return cosmic_map;

}

/**
 * @brief		Detect cosmic rays in an image
 *
 * @param	sci			Background subtracted science image
 * @param	flat		Flat field
 * @param	illum		Optional illumination map
 * @param	bpm			Bad pixels map combined or not with a saturated map
 * @param	pars		Parameters list
 *
 * @return	cosmic_map	Mask with detected cosmic rays
 */
cpl_mask *omega_cosmic_rays(const cpl_image *sci, const cpl_image *flat, const cpl_image *illum,
		const cpl_image *bpm, cpl_parameterlist *pars)
{

	cpl_mask *cosmic_map = NULL;
	cpl_mask *bpm_map    = NULL;
	cpl_image *weight, *fbpm, *image;

	if((sci == NULL) || (flat == NULL) || (bpm == NULL) || (pars == NULL))
		return NULL;

	/* Change pixel values in BPM image:
	 * 0->1  and 1-> 0 in order to use it more easily   */
	bpm_map=cpl_mask_threshold_image_create(bpm,-0.5,0.5);
	fbpm=cpl_image_new_from_mask(bpm_map);
	freemask(bpm_map);

	weight = cpl_image_multiply_create(flat,fbpm);
	freeimage(fbpm);

	if(illum != NULL)
		cpl_image_divide(weight,illum);

	cpl_image_power(weight, 0.5);

	image = cpl_image_multiply_create(sci, weight);

	freeimage(weight);

	cosmic_map = detcosmic(image, pars);

	freeimage(image);
	if(cosmic_map == NULL){
		cpl_msg_debug(cpl_func,"NULL cosmic rays map");
		return NULL;
	}

	return cosmic_map;
}

/**@}*/
