/* $Id: omega_fits.c,v 1.2 2012/01/31 13:38:50 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2005 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/31 13:38:50 $
 * $Revision: 1.2 $
 * $Name: omega-1_0_2 $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <math.h>
#include <string.h>


#include "omega_utils.h"
#include "omega_fits.h"

/**
    \defgroup omega_fits OMEGA Fits Structure

    \brief
    These are methods for manipulating the omega_fits object

    \code
    #include <stdio.h>
    #include <math.h>
    #include <string.h>
    #include "omega_utils.h"
    #include "omega_fits.h"
    \endcode
 */

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_load
    \par Purpose:
        Load an input image into a omega_fits object
    \par Description:
        An image from an input cpl_frame is loaded into a omega_fits object.
	Only the image for a given extension number is loaded. The rest
	of the object properties are initialised
    \par Language:
        C
    \param inframe
        The input cpl_frame object
    \param type
        The data type for the loaded image
    \param extnum
        The image extension that you want to load.
    \return
        The output omega_fits object.
    \author
        Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern omega_fits *omega_fits_load(const cpl_frame *inframe, cpl_type type,
		int extnum)
{
	omega_fits *p;
	cpl_image *im;
	cpl_frame *fr;
	int nf;

	/* Check for nonsense input */

	if (inframe == NULL)
		return(NULL);

	/* Duplicate the frame */
	fr = cpl_frame_duplicate(inframe);

	/* See if you can load the image */
	im = cpl_image_load(cpl_frame_get_filename(inframe),type,0,extnum);
	if (im == NULL) {
		cpl_msg_error(cpl_func,"Unable to load %s -- %s\n",
				cpl_frame_get_filename(inframe),cpl_error_get_message());
		return(NULL);
	}

	/* Get the omega_fits structure */

	p = cpl_malloc(sizeof(omega_fits));

	/* Load stuff in */

	p->frame = fr;
	p->image = im;
	p->extnum = extnum;
	p->phu = NULL;
	p->ehu = NULL;
	p->fname = cpl_strdup(cpl_frame_get_filename(inframe));
	p->status = 0;

	/* Get the extension header and the extension name */

	(void)omega_fits_get_ehu(p);
	if (p->ehu == NULL)
		return(NULL);
	if (cpl_propertylist_has(p->ehu,"EXTNAME")) {
		p->extname = cpl_strdup(cpl_propertylist_get_string(p->ehu,"EXTNAME"));
	}
	else {
		nf = 11 + log10(extnum);
		p->extname = cpl_malloc(nf);
		(void)snprintf(p->extname,nf,"DET1.CHIP%d",extnum);
	}
	nf = strlen(p->extname) + strlen(p->fname) + 3;
	p->fullname = cpl_malloc(nf);
	(void)snprintf(p->fullname,nf,"%s[%s]",p->fname,p->extname);

	/* Get out of here */

	return(p);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_duplicate
    \par Purpose:
        Copy a omega_fits structure into another one.
    \par Description:
        An input omega_fits structure is duplicated and returned
    \par Language:
        C
    \param in
        The input omega_fits object
    \return
        The output omega_fits object.
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern omega_fits *omega_fits_duplicate(omega_fits *in) {
	omega_fits *p;

	/* Check for nonsense input */

	if (in == NULL)
		return(NULL);

	/* Get the omega_fits structure */

	p = cpl_malloc(sizeof(omega_fits));

	/* Now copy everything over */

	p->frame = cpl_frame_duplicate(in->frame);
	p->image = cpl_image_duplicate(in->image);
	p->phu = cpl_propertylist_duplicate(in->phu);
	p->ehu = cpl_propertylist_duplicate(in->ehu);
	p->fname = cpl_strdup(in->fname);
	p->extname = cpl_strdup(in->extname);
	p->fullname = cpl_strdup(in->fullname);
	p->extnum = in->extnum;
	p->status = in->status;

	/* Get out of here */

	return(p);
}


/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_load_list
    \par Purpose:
        Load an input images into a omega_fits object array
    \par Description:
        Images from an input cpl_frameset are loaded into a list of
	omega_fits objects. Only the images for a given extension number are
	loaded. The rest of the object properties are initialised
    \par Language:
        C
    \param f
        The input cpl_frameset object
    \param type
        The data type for the loaded images
    \param exten
        The image extension that you want to load.
    \return
        The output omega_fits object list.
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern omega_fits **omega_fits_load_list(const cpl_frameset *f, cpl_type type,
		int exten) {
	int i;
	omega_fits **p;

	/* Check for nonsense input */

	if (f == NULL)
		return(NULL);

	/* Get some workspace */

	p = cpl_malloc(cpl_frameset_get_size(f)*sizeof(omega_fits *));

	/* Now load each of the frames... */

	for (i = 0; i < cpl_frameset_get_size(f); i++) {
		p[i] = omega_fits_load(cpl_frameset_get_frame_const(f,i),type,exten);
		if (p[i] == NULL) {
			omega_fits_delete_list(p,i-1);
			return(NULL);
		}
	}

	/* Now return the array */

	return(p);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_delete
    \par Purpose:
        Free all the workspace associated with a omega_fits object
    \par Description:
        Free all the workspace associated with a omega_fits object
    \par Language:
        C
    \param p
        The input omega_fits object
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void omega_fits_delete(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return;

	/* Free up workspace if it's been used */

	freeframe(p->frame);
	freeimage(p->image);
	freeplist(p->phu);
	freeplist(p->ehu);
	freespace(p->fname);
	freespace(p->extname);
	freespace(p->fullname);
	cpl_free(p);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_delete_list
    \par Purpose:
        Free all the workspace associated with a list of omega_fits objects
    \par Description:
        Free all the workspace associated with a list of omega_fits objects
    \par Language:
        C
    \param p
        The input list of omega_fits objects
    \param n
        The number of omega_fits objects in the above array
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void omega_fits_delete_list(omega_fits **p, int n) {
	int i;

	/* Check for nonsense input */

	if (p == NULL)
		return;

	/* Free up workspace if it's been used */

	for (i = 0; i < n; i++)
		omega_fits_delete(p[i]);

	freespace(p);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_frame
    \par Purpose:
        Get the CPL frame from the omega_fits object
    \par Description:
        Return the CPL frame from the input omega_fits object. This frame is
	suitable for use in all cpl_frame routines.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The cpl_frame object. NULL if there was an error.
    \author
        Sandra Castro
 */
/*---------------------------------------------------------------------------*/

extern cpl_frame *omega_fits_get_frame(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(NULL);

	/* Return it */

	return(p->frame);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_image
    \par Purpose:
        Get the CPL image from the omega_fits object
    \par Description:
        Return the CPL image from the input omega_fits object. This image is
	suitable for use in all cpl_image routines.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The cpl_image object. NULL if there was an error.
    \author
        Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern cpl_image *omega_fits_get_image(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(NULL);

	/* Return it */

	return(p->image);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_extnum
    \par Purpose:
        Get the FITS extension number for the current image in a omega_fits
	object
    \par Description:
        Get the FITS extension number for the current image in a omega_fits
	object
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The extension number (-1 in case of error)
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int omega_fits_get_extnum(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(-1);

	/* Return it */

	return(p->extnum);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_phu
    \par Purpose:
        Get the propertylist for the primary header for a given omega_fits image.
    \par Description:
        Get the propertylist for the primary header for a given omega_fits image.
	This should only need to be read once and then can be used to add
	things to the primary header.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The propertylist representing the primary header of the input image
	(NULL if there is an error).
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern cpl_propertylist *omega_fits_get_phu(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(NULL);

	/* If the propertylist hasn't already been loaded, then do it now */

	if (p->phu == NULL)
		p->phu = cpl_propertylist_load(p->fname,0);

	/* Return it */

	return(p->phu);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_ehu
    \par Purpose:
        Get the propertylist for the extension header for a given omega_fits
	image.
    \par Description:
        Get the propertylist for the extension header for a given omega_fits
	image. This is the extension that is relevant of the image.
	This should only need to be read once and then can be used to add
	things to the primary header.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The propertylist representing the extension header of the input image
	(NULL if there is an error).
    \author
          Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern cpl_propertylist *omega_fits_get_ehu(omega_fits *p)
{

	/* Check for nonsense input */

	if (p == NULL){
		return(NULL);
	}

	/* If the propertylist hasn't already been loaded, then do it now */

	if (p->ehu == NULL) {
		p->ehu = cpl_propertylist_load(p->fname,p->extnum);
	}

	/* Return it */

	return(p->ehu);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_extname
    \par Purpose:
        Get the extension name for a give omega_fits object
    \par Description:
        Get the extension name for a given omega_fits object.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The pointer to the extension name in the omega_fits structure -- not a
	copy! (NULL if there is an error).
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern char *omega_fits_get_extname(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(NULL);

	/* Return it */

	return(p->extname);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_filename
    \par Purpose:
        Get the filename from which the current omega_fits object originated
    \par Description:
        Get the filename from which the current omega_fits object originated. If
	this is null, then the image didn't originate in an FITS file.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The name of the file from which this image originated
    \author
        Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern char *omega_fits_get_filename(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(NULL);

	/* Return it */

	return(p->fname);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_fullname
    \par Purpose:
        Get the fullname of the FITS extension from which the current
	omega_fits object originated
    \par Description:
        Get the fullname of the FITS extension  from which the current
	omega_fits object originated. If this is null, then the image didn't
	originate in an FITS file.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The fullname name of the file from which this image originated
    \author
        Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern char *omega_fits_get_fullname(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(NULL);

	/* Return it */

	return(p->fullname);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_get_status
    \par Purpose:
        Get the error status of the current object.
    \par Description:
        Get the error status of the current object.
    \par Language:
        C
    \param p
        The input omega_fits object
    \return
        The error status
    \author
        Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int omega_fits_get_status(omega_fits *p) {

	/* Check for nonsense input */

	if (p == NULL)
		return(-1);

	/* Return it */

	return(p->status);
}


/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_set_error
    \par Purpose:
        Set the error status and message for an object
    \par Description:
        The input status is checked to see if there has been a problem
	with the current object. If there has been, then the status is
	stored away and any error message from the cpl_error system is
	copied down.
    \par Language:
        C
    \param p
        The input omega_fits object
    \param status
        The input error status
    \return
        A flag to say whether the input status was fatal.
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int omega_fits_set_error(omega_fits *p, int status) {

	/* Check for nonsense input */

	if (p == NULL)
		return(0);

	/* Get out of here if the status is OK */

	if (status == 0)
		return(0);

	/* Set the error status if there was an error */

	p->status = status;

	/* If there was a cpl error then spell that out now */

	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error("","%s",cpl_error_get_message());
		cpl_error_reset();
	}

	/* Get out of here */

	if (status == -1)
		return(-1);
	else
		return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_set_filename
    \par Purpose:
        Set the file name associated with a omega_fits structure
    \par Description:
        Set the file name associated with a omega_fits structure. This is
	useful if you have wrapped a cpl_image in a structure and you don't
	have a physical file to refer to. The name can be used to define
	an output file name if you ultimately decide to write this out.
	If a name already exists, it will be overwritten
    \par Language:
        C
    \param p
        The input omega_fits object
    \param fname
        The input file name.
    \return
        Nothing
    \author
        Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void omega_fits_set_filename(omega_fits *p, char *fname) {

	/* Check for nonsense input */

	if (p == NULL || fname == NULL)
		return;

	/* Set the name up and get out of here */

	freespace(p->fname);
	p->fname = cpl_strdup(fname);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_fits_wrap
    \par Purpose:
        Wrap an image in a omega_fits wrapper
    \par Description:
        The input image is inserted into a omega_fits wrapper. A model
	omega_fits object may be provided to give the new object
	headers. If the phu and ehu parameters are not null then they will
	be used as the propertylists for the new object. If not, then
	an attempt will be made to copy the propertylists from the model.
    \par Language:
        C
    \param im
        The input cpl_image
    \param model
        The input omega_fits model object
    \param phu
        The input propertylist for the primary header for the new object.
    \param ehu
        The input propertylist for the extension header for the new object.
    \return
        The new omega_fits structure.
    \author
         Sandra Castro - based on code written by Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern omega_fits *omega_fits_wrap(cpl_image *im, omega_fits *model,
		cpl_propertylist *phu,
		cpl_propertylist *ehu) {
	omega_fits *p;

	/* Check for nonsense input */

	if (im == NULL)
		return(NULL);

	/* Get the omega_fits structure */

	p = cpl_malloc(sizeof(omega_fits));

	/* Load stuff in */

	p->frame = NULL;
	p->image = im;
	p->extnum = -1;
	if (phu != NULL)
		p->phu = cpl_propertylist_duplicate(phu);
	else if (model != NULL)
		p->phu = cpl_propertylist_duplicate(omega_fits_get_phu(model));
	else
		p->phu = NULL;
	if (ehu != NULL)
		p->ehu = cpl_propertylist_duplicate(ehu);
	else if (model != NULL)
		p->ehu = cpl_propertylist_duplicate(omega_fits_get_ehu(model));
	else
		p->ehu = NULL;
	p->fname = NULL;
	p->status = 0;
	p->extname = NULL;
	p->fullname = NULL;

	/* Get out of here */

	return(p);
}

/**@}*/
