/* $Id: omega_pfits.c,v 1.11 2012/03/30 11:13:41 agabasch Exp $
 *
 * This file is part of the oc_ Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/03/30 11:13:41 $
 * $Revision: 1.11 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>

#include "omega_pfits.h"
#include "omega_stats.h"
#include "omega_utils.h"

static int omega_get_quadrant(const char * extname, int filtno);

/**
 * @defgroup  omega_pfits Header Manipulation
 *
 * \par Synopsis
 *  These functions manipulate the FITS headers.
 *
 *  \code
 *  #include omega_pfits.h
 *  #include omega_stats.h
 *  #include omega_utils.h
 *  \endcode
 *
 */
/**@{*/

/* Static Functions */
static void omega_property_dump(cpl_property *property)
{

    const char *name = cpl_property_get_name(property);
    const char *comment = cpl_property_get_comment(property);

    char c;

    long size = cpl_property_get_size(property);

    cpl_type type = cpl_property_get_type(property);


    cpl_msg_info(cpl_func, "Property at address %p", property);
    cpl_msg_info(cpl_func, "\tname   : %p '%s'", name, name);
    cpl_msg_info(cpl_func, "\tcomment: %p '%s'", comment, comment);
    cpl_msg_info(cpl_func, "\ttype   : %#09x", type);
    cpl_msg_info(cpl_func, "\tsize   : %ld", size);
    cpl_msg_info(cpl_func, "\tvalue  : ");


    switch (type) {
    case CPL_TYPE_CHAR:
        c = cpl_property_get_char(property);
        if (!c)
            fprintf(stderr, "''");
        else
            fprintf(stderr, "'%c'", c);
        break;

    case CPL_TYPE_BOOL:
        fprintf(stderr, "%d", cpl_property_get_bool(property));
        break;

    case CPL_TYPE_INT:
        fprintf(stderr, "%d", cpl_property_get_int(property));
        break;

    case CPL_TYPE_LONG:
        fprintf(stderr, "%ld", cpl_property_get_long(property));
        break;

    case CPL_TYPE_FLOAT:
        fprintf(stderr, "%.7g", cpl_property_get_float(property));
        break;

    case CPL_TYPE_DOUBLE:
        fprintf(stderr, "%.15g", cpl_property_get_double(property));
        break;

    case CPL_TYPE_STRING:
        fprintf(stderr, "'%s'", cpl_property_get_string(property));
        break;

    default:
        fprintf(stderr, "unknown.");
        break;

    }

    fprintf(stderr, "\n");

    return;

}

/*----------------------------------------------------------------------------
 *@brief    find out the date of observation
 *@param    plist       property list to read from
 *@return   pointer to statically allocated character string
 *
 *----------------------------------------------------------------------------*/

const char * omega_pfits_get_date_obs(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "DATE-OBS") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the ARCFILE keyword value from the header

  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * omega_pfits_get_arcfile(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ARCFILE");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the DIT keyword value from the header
  @param    plist       property list to read from
  @return   The DIT value
 */
/*----------------------------------------------------------------------------*/
double omega_pfits_get_dit(const cpl_propertylist * plist)
{
    if (cpl_propertylist_has(plist, "ESO DET DIT") == 1)
        return cpl_propertylist_get_double(plist, "ESO DET DIT");
    else
        return 0.0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the template ID
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * omega_pfits_get_templateid(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO TPL ID") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the MJD-OBS
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double omega_pfits_get_mjdobs(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "MJD-OBS") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the EXPTIME value
  @param    plist       property list to read from
  @return   the requested value or 0.0 otherwise
 */
/*----------------------------------------------------------------------------*/
double omega_pfits_get_exptime(const cpl_propertylist * plist)
{

	if(cpl_propertylist_has(plist, "EXPTIME")){
		return cpl_propertylist_get_double(plist, "EXPTIME")  ;
	}
	else{
		return 0.0;
	}
}


/*---------------------------------------------------------------------------*/
/**
    \brief    Get the value of DET_LIVE
    \param    plist       property list to read from
    \param    detlive     requested value of DET_LIVE
    \retval 0 if everything went right or -1 if there was an error
 */
/*---------------------------------------------------------------------------*/

extern int omega_pfits_get_detlive(const cpl_propertylist *plist,
				    int *detlive)
{
    int val;

    if(cpl_propertylist_has(plist, "ESO DET CHIP LIVE") == 0){
        cpl_error_reset();
        *detlive = 1;
       return 0;
    }

    val = cpl_propertylist_get_bool(plist,"ESO DET CHIP LIVE");
    if (cpl_error_get_code() == CPL_ERROR_NONE) {
	*detlive = val;
        return 0;
    } else {
	cpl_error_reset();
	*detlive = 1;
	return -1;
    }
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get chip orientation based on chip name
  @param    chip        Chip name
  @return   The chip orientation

  Takes the string with the chip name and returns its orientation.
 */
/*----------------------------------------------------------------------------*/
int omega_pfits_get_orientation(const char *chip)
{

	int orient = 0;

	/*
 For WFI with EXTNAME < 5
    if( (strcmp(chip, "ccd50")==0)||(strcmp(chip, "ccd51")==0)||
        (strcmp(chip, "ccd52")==0)||(strcmp(chip, "ccd53")==0) ) {

        orient = 2;
    }
    else { For all OmegaCAM EXTNAME
        orient = 1;
    }
	 */
	if(strcmp(chip, "ESO_CCD_#65" )==0 ||
			strcmp(chip, "ESO_CCD_#66" )==0 ||
			strcmp(chip, "ESO_CCD_#67" )==0 ||
			strcmp(chip, "ESO_CCD_#68" )==0 ||
			strcmp(chip, "ESO_CCD_#69" )==0 ||
			strcmp(chip, "ESO_CCD_#70" )==0 ||
			strcmp(chip, "ESO_CCD_#71" )==0 ||
			strcmp(chip, "ESO_CCD_#72" )==0 ||
			strcmp(chip, "ESO_CCD_#73" )==0 ||
			strcmp(chip, "ESO_CCD_#74" )==0 ||
			strcmp(chip, "ESO_CCD_#75" )==0 ||
			strcmp(chip, "ESO_CCD_#76" )==0 ||
			strcmp(chip, "ESO_CCD_#77" )==0 ||
			strcmp(chip, "ESO_CCD_#78" )==0 ||
			strcmp(chip, "ESO_CCD_#79" )==0 ||
			strcmp(chip, "ESO_CCD_#80" )==0){
		/*The readout port is located in the lower left of the ccd*/
		orient = 1;
	}
	else if (strcmp(chip, "ESO_CCD_#81" )==0 ||
			strcmp(chip, "ESO_CCD_#82" )==0 ||
			strcmp(chip, "ESO_CCD_#83" )==0 ||
			strcmp(chip, "ESO_CCD_#84" )==0 ||
			strcmp(chip, "ESO_CCD_#85" )==0 ||
			strcmp(chip, "ESO_CCD_#86" )==0 ||
			strcmp(chip, "ESO_CCD_#87" )==0 ||
			strcmp(chip, "ESO_CCD_#88" )==0 ||
			strcmp(chip, "ESO_CCD_#89" )==0 ||
			strcmp(chip, "ESO_CCD_#90" )==0 ||
			strcmp(chip, "ESO_CCD_#91" )==0 ||
			strcmp(chip, "ESO_CCD_#92" )==0 ||
			strcmp(chip, "ESO_CCD_#93" )==0 ||
			strcmp(chip, "ESO_CCD_#94" )==0 ||
			strcmp(chip, "ESO_CCD_#95" )==0 ||
			strcmp(chip, "ESO_CCD_#96" )==0){
		/*The readout port is located in the upper right of the ccd*/
		orient = 2;
	}
	else {
		orient = 0;
	}


	return orient;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get xy sizes of a chip
  @param    chip      Chip name
  @param    xs        Size of x-axis
  @param    ys        Size of y-axis
  @return   Nothing

  Based on the chip name, it reads the X and Y sizes of the image.
 */
/*----------------------------------------------------------------------------*/

void omega_pfits_get_xysize(const char *chip, int *xs, int *ys)
{

    int orientation = 0;
    int prscx = 48;	/*FIXME: these 4 values are default in Python code. Is this real??*/
    int prscy = 0;
    int ovscx = 48;
    int ovscy = 30;
    int naxis1 = 2142;
    int naxis2 = 4128;
    int x0 = 1;
    int x1 = 1;
    int y0 = 1;
    int y1 = 1;

    orientation = omega_pfits_get_orientation(chip);

/* Get trim coordinates based on orientation keyword */
    switch (orientation) {
        case 1:
            x0 = prscx + 1;
            y0 = prscy + 1;
            x1 = naxis1 - ovscx;
            y1 = naxis2 - ovscy;
            break;
        case 2:
            x0 = prscx + 1;
            y0 = ovscy + 1;
            x1 = naxis1 - ovscx;
            y1 = naxis2 - prscy;
            break;
        case 3:
            x0 = ovscx + 1;
            y0 = prscy+ 1;
            x1 = naxis1 - prscx;
            y1 = naxis2 - ovscy;
            break;
        case 4:
            x0 = ovscx + 1;
            y0 = ovscy + 1;
            x1 = naxis1 - prscx;
            y1 = naxis2 - prscy;
            break;
        default:
            x0 = prscx + 1;
            y0 = prscy + 1;
            x1 = naxis1 - ovscx;
            y1 = naxis2 - ovscy;
            break;
        }

    *xs = x1 - x0+1;
    *ys = y1 - y0+1;

}


/*
----------------------------------------------------------------------------
*
  @brief    Gets filter information from header
  @param    main      Main property list
  @param    ext       Extension property list
  @return   1 newly allocated property list or NULL in error case

	This function gets the following information from the main and extension headers.
	HIERARCH ESO INS FILT1 ID
	HIERARCH ESO INS FILT1 NAME
	HIERARCH ESO DET CHIP* ID

	and associate these values to other information, to return a new
	propertylist with the following format:

	const char *FILT_ID
	const char *FILT_NAME
	float        CWL
	int         HAS_FRINGES
	const char *CHIP_ID


----------------------------------------------------------------------------
cpl_propertylist * omega_pfits_get_filter(const cpl_propertylist *main, const cpl_propertylist *ext)
{

    int has_fringes = 0;
    int i = 0;
    int n = 4; change here for correct number of filters
    int check =0 ;
    int check1 = 0;

    double wave_central = 0.0f;

    char name[80] = "";
    const char *chipid = NULL;
    const char *filtid = NULL;
    const char *filtname = NULL;

    cpl_propertylist 		*list;

    if (main == NULL) {
        cpl_msg_error(cpl_func,"Main header is empty");
        return NULL;
    }
    if (ext == NULL) {
        cpl_msg_error(cpl_func,"Extension header is empty");
        return NULL;
    }

    list = cpl_propertylist_new();

    if (cpl_propertylist_has(main,"ESO INS FILT ID") == 1) {
        filtid = cpl_propertylist_get_string(main, "ESO INS FILT ID");
    }
    else if (cpl_propertylist_has(main,"ESO INS FILT1 ID") == 1) {
        filtid = cpl_propertylist_get_string(main, "ESO INS FILT1 ID");
    }
    else if (cpl_propertylist_has(main,"FILT_ID") == 1) {
        filtid = cpl_propertylist_get_string(main, "FILT_ID");
    }
    else {
        filtid = "";
        cpl_msg_warning(cpl_func,"Cannot find FILT_ID information on main header");
		return NULL;
    }

    cpl_propertylist_append_string(list,"FILT_ID", filtid);

Get filter name from header
    if (cpl_propertylist_has(main,"ESO INS FILT1 NAME") == 1){
        filtname = cpl_propertylist_get_string(main, "ESO INS FILT1 NAME");
    }


FIXME: Modify this once the names of filters are correct
    for(i=0; i<n; i++) {
        check = strcmp("842", filtid);
        check1 = strcmp("#842", filtid);
        if(check == 0 || check1 == 0) {
            filtname = "JohnsonB";
            wave_central = 4562.52;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("843", filtid);
        check1 = strcmp("#843", filtid);
        if(check == 0 || check1 == 0) {
            filtname = "JohnsonV";
            wave_central = 5395.62;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("844", filtid);
        check1 = strcmp("#844", filtid);
        if(check == 0 || check1 == 0) {
            filtname = "CousinsR";
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("845", filtid);
        check1 = strcmp("#845", filtid);
        if(check == 0 || check1 == 0) {
            filtname = "CousinsI";
		wave_central = 7838.45;
            wave_central = 7900.00;
            has_fringes = 1; has fringing
				cpl_msg_info("","%s This filter %s has fringes",_id,filtname);
            break;
        }
        check = strcmp("846", filtid);
        check1 = strcmp("#846", filtid);
        if(check == 0 || check1 == 0) {
				filtname = "CousinsI";
            wave_central = 9648.19;
            has_fringes = 1; has fringing
			cpl_msg_info("","%s This filter %s has fringes",_id,filtname);
            break;
        }
        check = strcmp("879", filtid);
        check1 = strcmp("#879", filtid);
        if(check == 0 || check1 == 0) {
            filtname = "CousinsI";
            wave_central = 8269.00;
            has_fringes = 1; has fringing
				cpl_msg_info("","%s This filter %s has fringes",_id,filtname);
            break;
        }
Filter names and IDs from Bernard Muschielok
FIXME: standard catalog columns must have filter names like these ones
FIXME: correct values of wave_central for each filter and the has_fringes too.
        check = strcmp("PS1", filtid);
        check1 = strcmp("u_SDSS", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS2", filtid);
        check1 = strcmp("g_SDSS", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS3", filtid);
        check1 = strcmp("r_SDSS", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS4", filtid);
        check1 = strcmp("i_SDSS", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS5", filtid);
        check1 = strcmp("z_SDSS", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS6", filtid);
        check1 = strcmp("B_JOHN", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS7", filtid);
        check1 = strcmp("V_JOHN", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS8", filtid);
        check1 = strcmp("V_STRM", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS9", filtid);
        check1 = strcmp("H_ALPHA", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS10", filtid);
        check1 = strcmp("H_ALPHA", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS11", filtid);
        check1 = strcmp("NB_852_861_869_878", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS12", filtid);
        check1 = strcmp("NB_453_494_535_575", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS13", filtid);
        check1 = strcmp("NB_615_710_755_815", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("P53", filtid);
        check1 = strcmp("SCAT_AUX_CCD", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("P54", filtid);
        check1 = strcmp("PINHOLE", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("P56", filtid);
        check1 = strcmp("frame", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS59", filtid);
        check1 = strcmp("test2", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("PS60", filtid);
        check1 = strcmp("test1", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("calib", filtid);
        check1 = strcmp("u_g_r_i_SDSS", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
        check = strcmp("opaque", filtid);
        check1 = strcmp("opaque", filtname);
        if(check == 0 && check1 == 0) {
            wave_central = 6517.25;
            has_fringes = 0; has no fringing
            break;
        }
    }

    cpl_propertylist_append_string(list,"FILT_NAME", filtname);
    cpl_propertylist_append_double(list,"CWL", wave_central);
    cpl_propertylist_append_int(list,"HAS_FRINGES", has_fringes);

 Get chip information and include in same plist
FIXME: 32 or 16 ccd? Depends on if 1 or 2 files in paranal workstation
    for(i=1; i<= 32; i++) {

        if (cpl_propertylist_has(ext, "ESO DET CHIP ID") == 1) {
            chipid = cpl_propertylist_get_string(ext, "ESO DET CHIP ID");
            break;
        }
        else if (cpl_propertylist_has(ext, "CHIP_ID") == 1) {
            chipid = cpl_propertylist_get_string(ext, "CHIP_ID");
            break;
        }
        else {
			char name[80] = "";
            sprintf(name,"ESO DET CHIP%d ID", i);

            if (cpl_propertylist_has(ext, name) == 1) {
                chipid = cpl_propertylist_get_string(ext, name);
                break;
            }
        }
    }

    cpl_propertylist_append_string(list,"CHIP_ID", chipid);

 It returns a plist with the following fields FILT_ID,FILT_NAME,CWL,HAS_FRINGES,CHIP_ID)
    return list;
}
*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Calculates the mean airmass from a FITS header
  @param    plist      A propertylist with the necessary information (ESO TEL AIRM START,END)
  @param    airm       Output pointer to the value of the mean airmass
  @return   Nothing

  It updates the pointer with the double value of the mean
  airmass which is calculated here.
 */
/*----------------------------------------------------------------------------*/

void omega_pfits_get_mean_airmass(const cpl_propertylist *plist, double *airm)
{

    double airm0 = 1.0;
    double airm1 = 1.0;


    if( (cpl_propertylist_has(plist, "ESO TEL AIRM START") == 1) &&
        (cpl_propertylist_has(plist, "ESO TEL AIRM END") == 1 ) ) {
        airm0 = cpl_propertylist_get_double(plist, "ESO TEL AIRM START");
        airm1 = cpl_propertylist_get_double(plist, "ESO TEL AIRM END");

        *airm = (airm0 + airm1)/2;
    }
    else if ( (cpl_propertylist_has(plist, "AIRMSTRT") == 1) &&
        (cpl_propertylist_has(plist, "AIRMEND")) == 1 ){
        airm0 = cpl_propertylist_get_double(plist, "AIRMSTRT");
        airm1 = cpl_propertylist_get_double(plist, "AIRMEND");

        *airm = (airm0 + airm1)/2;
    }
    else if (cpl_propertylist_has(plist, "AIRMASS") == 1){
        *airm = cpl_propertylist_get_double(plist, "AIRMASS");
    }
    else {
        *airm = 1.0;
    }

    if (*airm == 0.0){
    	*airm = 1.0;
    }

    return;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Reads the airmass keywords from the header
  @param    plist1      A propertylist with the necessary information (ESO TEL AIRM START,END)
  @param    start       A pointer to the keyword ESO TEL AIRM START
  @param    end         A pointer to the keyword ESO TEL AIRM END
  @return   Nothing

  It updates the already allocated pointers to the double values of the start
  and end airmasses
 */
/*----------------------------------------------------------------------------*/
void omega_pfits_get_airmass(const cpl_propertylist *plist1, double *start, double *end)
{


/*This needs to be the main header of the image*/

    if (plist1 == NULL) {
        cpl_msg_error(cpl_func,"Main header is empty");
    }

    if( (cpl_propertylist_has(plist1, "ESO TEL AIRM START") == 1) &&
        (cpl_propertylist_has(plist1, "ESO TEL AIRM END") == 1 ) ){
        *start = cpl_propertylist_get_double(plist1, "ESO TEL AIRM START");
        *end = cpl_propertylist_get_double(plist1, "ESO TEL AIRM END");
    }
    else if ( (cpl_propertylist_has(plist1, "AIRMSTRT") == 1) &&
        (cpl_propertylist_has(plist1, "AIRMEND")) == 1 ){
        *start = cpl_propertylist_get_double(plist1, "AIRMSTRT");
        *end = cpl_propertylist_get_double(plist1, "AIRMEND");
    }
    else if (cpl_propertylist_has(plist1, "AIRMASS") == 1) {
        *start = cpl_propertylist_get_double(plist1, "AIRMASS");
        *end = *start;
    }
    else {
        *start = 1.0;
        *end = 1.0;
    }


}

/*
 *  Check from which instrument is the frame (WFI or OMEGA)
 */
/**
 * @brief   Check if INSTRUME is WFI or OMEGA
 * @param fr  Frame
 * @return 0 if OMEGA, 1 if WFI or -1 in case of error.
 */
int omega_pfits_check_instrume(const cpl_frame *fr)
{

    const char *INSTRUMENT = NULL;
    cpl_propertylist		*plist;


    if (fr == NULL) {
        cpl_msg_error(cpl_func,"NULL frame");
        return -1;
    }

    plist = cpl_propertylist_load_regexp(cpl_frame_get_filename(fr), 0,
                                         "INSTRUME", 0);
    if (plist == NULL) {
        cpl_msg_error (cpl_func,"Cannot load main header for %s",cpl_frame_get_filename(fr));
        return -1;
    }

    INSTRUMENT = cpl_propertylist_get_string(plist, "INSTRUME");
    if(!INSTRUMENT) {
        cpl_msg_error(cpl_func,"Cannot find INSTRUME keyword in header");
        cpl_propertylist_delete(plist);
        return -1;
    }

    if(strcmp("OMEGA", INSTRUMENT)==0){
        cpl_propertylist_delete(plist);
        return 0;
    }
    else if(strcmp("WFI", INSTRUMENT)==0){
        cpl_propertylist_delete(plist);
        return 1;
    }
    else{
        cpl_propertylist_delete(plist);
        return -1;
    }

    return 0;
}

/**
 *  @brief    Get the INSTRUME keyword value
 *  @param plist	primary header
 *  @return 0 if OMEGA, 1 if WFI or -1 in case of error
 *
 *  @note plist must be primary header unit
 */
int omega_pfits_get_instrume(const cpl_propertylist *plist)
{

    const char *INSTRUMENT = NULL;


    if (plist == NULL) {
        cpl_msg_error(cpl_func,"NULL property list");
        return -1;
    }


    INSTRUMENT = cpl_propertylist_get_string(plist, "INSTRUME");
    if(!INSTRUMENT) {
        cpl_msg_warning(cpl_func,"Cannot find INSTRUME keyword in header");
        return 0;
    }

    if(strcmp("OMEGA", INSTRUMENT)==0){
        return 0;
    }
    else if(strcmp("WFI", INSTRUMENT)==0){
        return 1;
    }

    return 0;
}

/**
  @brief    Get the read noise value from a header

  @param    plist   Header with read noise keyword
  @param	rn		Output with the value of the readnoise
  @return	nothing
 */
/*--------------------------------------------------------------------------*/
void omega_pfits_get_readnoise(cpl_propertylist *plist, double *rn)
{

    int i = 0;

    const char *ronstring = "ESO DET OUT1 RON";


/* Start function algorithm */
/* This must be the extension header of the image*/
    if (plist == NULL) {
        cpl_msg_error(cpl_func,"Extension header is empty");
    }

    if (cpl_propertylist_has(plist, ronstring) == 1) {
        *rn = cpl_propertylist_get_double(plist, ronstring);
    }
    else if (cpl_propertylist_has(plist, "RON") == 1) {
        *rn = cpl_propertylist_get_double(plist, "RON");
    }
    else{

        for(i=1; i<= 32; i++) {
            char name[80] = "";
            sprintf(name,"ESO DET OUT%d RON", i);

            if (cpl_propertylist_has(plist, name) == 1) {
                *rn = cpl_propertylist_get_double(plist, name);
                break;
            }
        }
    }

}

/**
 * @brief  Get the DRS keyword of the overscan method used to reduce the image
 *
 * @param frame  Frame to get keyword from
 * @param xn     FITS extension to load
 * @return  The overscan method used
 */
int omega_pfits_get_overscan(const cpl_frame *frame, int xn)
{
    int value;
    cpl_propertylist *plist;

    if(frame == NULL){
        cpl_msg_warning(cpl_func,"Frame is NULL");
        return -1;
    }

    plist = cpl_propertylist_load_regexp(cpl_frame_get_filename(frame), xn,
    		"ESO DRS", 0);

    if(plist == NULL){
        cpl_msg_warning(cpl_func,"Cannot load property list");
        return -1;
    }

    if( cpl_propertylist_has(plist, "ESO DRS OVERSCAN METHOD") == 1 ){
        value = cpl_propertylist_get_int(plist, "ESO DRS OVERSCAN METHOD");
    }
    else {
        cpl_msg_warning(cpl_func,"Cannot find overscan method in header");
        value = -1;
    }

    freeplist(plist);

    return value;
}


/**
  @brief    Read the gain value from an image

  @param    plist       Header to read gain value from
  @param	gain		Output value of the gain
  @return	nothing
 */
/*--------------------------------------------------------------------------*/
void omega_pfits_get_gain(cpl_propertylist *plist, double *gain)
{

    int i = 0;
    char name[80] = "";

    *gain = 0.0;

    if(plist == NULL)
    	return;

/* Start function algorithm */
    if (plist == NULL) {
        cpl_msg_warning(cpl_func,"Header is NULL");
    }
    else if (cpl_propertylist_has(plist, "ESO DET OUT GAIN") == 1) {
        *gain = cpl_propertylist_get_double(plist, "ESO DET OUT GAIN");
    }
    else if (cpl_propertylist_has(plist, "GAIN") == 1) {
        *gain = cpl_propertylist_get_double(plist, "GAIN");
    }
    else {

        for(i=1; i<= 32; i++){
/*			char name[80] = "";			*/
            sprintf(name,"ESO DET OUT%d GAIN", i);

            if (cpl_propertylist_has(plist, name) == 1) {
                *gain = cpl_propertylist_get_double(plist, name);
                break;
            }
        }
    }

    return;
}

/**
 * @brief   Get the value of the keyword CONAD
 * @param plist  Property list containing the keyword
 * @param conad  Output with the CONAD value
 * return  nothing
 */
void omega_pfits_get_conad(cpl_propertylist *plist, double *conad)
{

    int i = 0;
    char name[80] = "";

    *conad = 0.0;

    if(plist == NULL)
    	return;

/* Start function algorithm */
    if (plist == NULL) {
        cpl_msg_error(cpl_func,"Header is empty");
    }
    else if (cpl_propertylist_has(plist, "ESO DET OUT CONAD") == 1) {
        *conad = cpl_propertylist_get_double(plist, "ESO DET OUT CONAD");
    }
    else if (cpl_propertylist_has(plist, "CONAD") == 1) {
        *conad = cpl_propertylist_get_double(plist, "CONAD");
    }
    else {

        for(i=1; i<= 32; i++){
/*			char name[80] = "";			*/
            sprintf(name,"ESO DET OUT%d CONAD", i);

            if (cpl_propertylist_has(plist, name) == 1) {
                *conad = cpl_propertylist_get_double(plist, name);
                break;
            }
        }
    }

    return;
}

/**
  @brief    Get CHIP ID from header

  @    header       Header to read keyword from
  @returns      string
 */
/*--------------------------------------------------------------------------*/
/**
 * @brief   Get the CHIP ID from the header
 * @param plist  Property list
 * @return ID of the chip as written in header
 */
const char *omega_pfits_get_chipid(const cpl_propertylist *plist)
{
    int i = 0;
    int nextensions = 32;   /*FIXME: check if OMEGA has 1 or 2 controllers*/
    const char *chipid = NULL;
    char name[80] = "";

    if (plist == NULL) {
        cpl_msg_error(cpl_func,"NULL input property list");
        return NULL;
    }

    /*Check if instrument is WFI*/
//    if (omega_pfits_get_instrume(plist) == 1)
//        nextensoins = 8;

    for(i=1; i<= nextensions; i++) {

        if (cpl_propertylist_has(plist, "ESO DET CHIP ID") == 1) {
            chipid = cpl_propertylist_get_string(plist, "ESO DET CHIP ID");
            break;
        }
        else {
            sprintf(name,"ESO DET CHIP%d ID", i);

            if (cpl_propertylist_has(plist, name) == 1) {
                chipid = cpl_propertylist_get_string(plist, name);
                break;
            }
        }
    }

    return chipid;
}

/**
  @brief    Get the EXTNAME keyword of a header

  @param    plist     Property list to read keyword from
  @return   EXTNAME value
 */
/*--------------------------------------------------------------------------*/

const char * omega_pfits_get_extname(const cpl_propertylist *plist)
{

	if (cpl_propertylist_has(plist, "EXTNAME") == 0){
		return NULL;
	}

	return cpl_propertylist_get_string(plist, "EXTNAME");

}

/**
  @brief    Compare the EXTNAME value of two headers

  @param  plist1     First header
  @param  plist2     Second header
  @return
     It returns 0 if the values of both keywords are the same or -1 if they are
     different.
 */
/*--------------------------------------------------------------------------*/

int omega_compare_extname(cpl_propertylist *plist1, cpl_propertylist *plist2)
{
	const char *extn1 = NULL;
	const char *extn2 = NULL;

	if(plist1 == NULL || plist2 == NULL){
		cpl_msg_error(cpl_func,"NULL input property list(s)");
		return -1;
	}

	if((cpl_propertylist_has(plist1, "EXTNAME") == 1)
			&& (cpl_propertylist_has(plist2, "EXTNAME") == 1)) {

		extn1 = cpl_propertylist_get_string(plist1, "EXTNAME");
		extn2 = cpl_propertylist_get_string(plist2, "EXTNAME");
		return strcmp(extn1, extn2);
	}
	else{
		return -1;
	}

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Gets filter information from header
  @param    main      Main propertylist
  @return   1 newly allocated propertylist or NULL in error case

	This function gets the following information from the main aheader
	HIERARCH ESO INS FILT1 ID
	HIERARCH ESO INS FILT1 NAME

	and associate these values to other information, to return a new
	propertylist with the following format:

	const char *FILT_ID
	const char *FILT_NAME
	const char *REF_FILT_NAME
	float        CWL
	int         HAS_FRINGES


 */
/*----------------------------------------------------------------------------*/
cpl_propertylist * omega_pfits_get_filter_info(const cpl_propertylist *main)
{

    int has_fringes = 0;
    int i = 0;
    int n = 4; /*change here for correct number of filters*/
    int check =0 ;
    int check1 = 0;
    int filtno = 0;

    double wave_central = 0.0f;

    const char *filtid = "none";
    const char *filtname = "none";
    const char *reffiltname = "none";
    const char *referr = "none";
    const char *extname = "none";


    cpl_propertylist 		*list;

    if (main == NULL) {
        cpl_msg_error(cpl_func,"Main header is empty");
        return NULL;
    }

    list = cpl_propertylist_new();

    if (cpl_propertylist_has(main,"ESO INS FILT ID") == 1) {
        filtid = cpl_propertylist_get_string(main, "ESO INS FILT ID");
        filtname = cpl_propertylist_get_string(main, "ESO INS FILT NAME");
    }
    else if (cpl_propertylist_has(main,"ESO INS FILT1 ID") == 1) {
        filtid = cpl_propertylist_get_string(main, "ESO INS FILT1 ID");
        filtname = cpl_propertylist_get_string(main, "ESO INS FILT1 NAME");
    }
    else if (cpl_propertylist_has(main,"FILT_ID") == 1) {
        filtid = cpl_propertylist_get_string(main, "FILT_ID");
        filtname = cpl_propertylist_get_string(main, "FILT_NAME");
      }
    else {
        filtid = "";
        filtname = "";
          cpl_msg_warning(cpl_func,"Cannot find filter information in header");
/*		return NULL;*/
    }

    if (cpl_propertylist_has(main,"EXTNAME") == 1) {
    	extname= cpl_propertylist_get_string(main, "EXTNAME");
    }
    else {
    	extname = "";
    	cpl_msg_debug(cpl_func,"Cannot find extension name in header");
    }

    if (cpl_propertylist_has(main,"ESO INS FILT1 NO") == 1) {
    	filtno= cpl_propertylist_get_int(main, "ESO INS FILT1 NO");
    }
    else {
    	filtno=0;
    }



    cpl_propertylist_append_string(list,"FILT_ID", filtid);



    for(i=0; i<n; i++) {
/*Filter names and IDs from Bernard Muschielok*/
/*FIXME: correct values of wave_central for each filter and the has_fringes too.*/
/* FIXME: check reffiltname for filters */

    	/* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS7", filtid);
        check1 = strcmp("u_SDSS", filtname);
        if(check == 0 || check1 == 0) {
        	reffiltname = "SloanU";
        	referr = "SloanU_err";
            wave_central = 3540.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS2", filtid);
        check1 = strcmp("g_SDSS", filtname);
        if(check == 0 || check1 == 0) {
        	reffiltname = "SloanG";
        	referr = "SloanG_err";
            wave_central = 4750.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS4", filtid);
        check1 = strcmp("r_SDSS", filtname);
        if(check == 0 || check1 == 0) {
        	reffiltname = "SloanR";
        	referr = "SloanR_err";
            wave_central = 6250.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS3", filtid);
        check1 = strcmp("i_SDSS", filtname);
        if(check == 0 || check1 == 0) {
        	reffiltname = "SloanI";
        	referr = "SloanI_err";
            wave_central = 7560.;
            has_fringes = 1; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS1", filtid);
        check1 = strcmp("z_SDSS", filtname);
        if(check == 0 || check1 == 0) {
        	reffiltname = "SloanZ";
        	referr = "SloanZ_err";
            wave_central = 8800.;
            has_fringes = 1; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS9", filtid);
        check1 = strcmp("B_JOHN", filtname);
        if(check == 0 || check1 == 0) {
        	reffiltname = "JohnsonB";
        	referr = "JohnsonB_err";
            wave_central = 4390;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS10", filtid);
        check1 = strcmp("V_JOHN", filtname);
        if(check == 0 || check1 == 0) {
        	reffiltname = "JohnsonV";
        	referr = "JohnsonV_err";
             wave_central = 5510;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS5", filtid);
        check1 = strcmp("v_STRM", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 4120.;
            has_fringes = 0; /*has no fringing*/
            break;
        }

/*
         From VST manual v 2.5 page 16 .
        check = strcmp("PS11", filtid);
        check1 = strcmp("H_ALPHA", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 0.;
            has_fringes = 0; has no fringing
            break;
        }
*/


        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS6", filtid);
        check1 = strcmp("NB_852_861_869_878", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 0.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS14", filtid);
        check1 = strcmp("NB_454_494_533_575", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 0.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS22", filtid);
        check1 = strcmp("NB_617_710_755_817", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 0.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS59", filtid);
        check1 = strcmp("Dummy_2", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 0.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS60", filtid);
        check1 = strcmp("Dummy_1", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 0.;
            has_fringes = 0; /*has no fringing*/
            break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("opaque", filtid);
        check1 = strcmp("opaque", filtname);
        if(check == 0 || check1 == 0) {
            wave_central = 0.;
            has_fringes = 0; /*has no fringing*/
            break;
        }

        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("calib", filtid);
        check1 = strcmp("u_g_r_i_SDSS", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 1) {
        	reffiltname = "SloanU";
        	referr = "SloanU_err";
        	wave_central = 3540.;
        	has_fringes = 0; /*has no fringing*/
        	break;
        }

        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("calib", filtid);
        check1 = strcmp("u_g_r_i_SDSS", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 4) {
        	reffiltname = "SloanG";
        	referr = "SloanG_err";
            wave_central = 4750.;
            has_fringes = 0; /*has no fringing*/
           break;
        }

        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("calib", filtid);
        check1 = strcmp("u_g_r_i_SDSS", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 3) {
        	reffiltname = "SloanR";
        	referr = "SloanR_err";
            wave_central = 6250.;
            has_fringes = 0; /*has no fringing*/
           break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("calib", filtid);
        check1 = strcmp("u_g_r_i_SDSS", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 2) {
        	reffiltname = "SloanI";
        	referr = "SloanI_err";
            wave_central = 7560.;
            has_fringes = 0; /*has no fringing*/
            break;
        }



        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS11", filtid);
        check1 = strcmp("H_ALPHA", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 1) {
        	reffiltname = "H_ALPHA";
        	referr = "H_ALPHA_err";
        	wave_central = 6791.;
        	has_fringes = 0; /*has no fringing*/
        	break;
        }

        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS11", filtid);
        check1 = strcmp("H_ALPHA", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 4) {
        	reffiltname = "H_ALPHA";
        	referr = "H_ALPHA_err";
            wave_central = 6590.;
            has_fringes = 0; /*has no fringing*/
           break;
        }

        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS11", filtid);
        check1 = strcmp("H_ALPHA", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 3) {
        	reffiltname = "H_ALPHA";
        	referr = "H_ALPHA_err";
            wave_central = 6660.;
            has_fringes = 0; /*has no fringing*/
           break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS11", filtid);
        check1 = strcmp("H_ALPHA", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 2) {
        	reffiltname = "H_ALPHA";
        	referr = "H_ALPHA_err";
            wave_central = 6726.;
            has_fringes = 0; /*has no fringing*/
            break;
        }


        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS30", filtid);
        check1 = strcmp("NB_659", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 1) {
        	reffiltname = "NB_659";
        	referr = "NB_659_err";
        	wave_central = 6586.;
        	has_fringes = 0; /*has no fringing*/
        	break;
        }

        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS30", filtid);
        check1 = strcmp("NB_659", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 4) {
        	reffiltname = "NB_659";
        	referr = "NB_659_err";
            wave_central = 6593.;
            has_fringes = 0; /*has no fringing*/
           break;
        }

        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS30", filtid);
        check1 = strcmp("NB_659", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 3) {
        	reffiltname = "NB_659";
        	referr = "NB_659_err";
            wave_central = 6586.;
            has_fringes = 0; /*has no fringing*/
           break;
        }
        /* From VST manual v 2.5 page 16 .*/
        check = strcmp("PS30", filtid);
        check1 = strcmp("NB_659", filtname);
        if((check == 0 || check1 == 0) && omega_get_quadrant(extname, filtno) == 2) {
        	reffiltname = "NB_659";
        	referr = "NB_659_err";
            wave_central = 6586.;
            has_fringes = 0; /*has no fringing*/
            break;
        }

    }

    cpl_propertylist_append_string(list,"FILT_NAME", filtname);
    cpl_propertylist_append_string(list,"REF_MAG_ID", reffiltname);
    cpl_propertylist_append_string(list,"REF_MAG_ID_ERR", referr);
    cpl_propertylist_append_double(list,"CWL", wave_central);
    cpl_propertylist_append_int(list,"HAS_FRINGES", has_fringes);


/* It returns a plist with the following fields FILT_ID,FILT_NAME,REF_FILT_NAME,CWL,HAS_FRINGES)*/
    return list;
}


/**
  @brief    Copy all properties from one plist to another

  @param   to    Property list to copy to
  @param   from  Property list to copy from
  @return
     It returns 0 on success or -1 in case of error.
 */
/*--------------------------------------------------------------------------*/
int omega_pfits_update_header(cpl_propertylist *to, const cpl_propertylist *from)
{

	int i = 0;
	int n = 0;

	if(from == NULL)
		return -1;

	if(to == NULL) {
		to = cpl_propertylist_new();
		cpl_propertylist_append(to, from);
		return 0;
	}

	n = cpl_propertylist_get_size(from);
	for (i=0; i < n; i++){
		const cpl_property *pro = cpl_propertylist_get_const(from, i);
		const char *name = cpl_property_get_name(pro);
		if(cpl_propertylist_has(to, name) == 1)
			cpl_propertylist_erase(to, name);

		cpl_propertylist_copy_property(to, from, name);
	}

//	cpl_propertylist_append(*to, from);

	return 0;
}


/**
  @brief   This function gets the prescan and overscan keywords from
            the header.

  @param    plist   Property list containing the keywords

  @return  cpl_vector with the keywords in the following order
           PRSCX, PRSCY, OVSCX, OVSCY

 */
/*--------------------------------------------------------------------------*/
cpl_vector * omega_pfits_get_preovscan(const cpl_propertylist *plist)
{
	int i = 0;
	int index = 0;
	int next = 32;	/*FIXME: check if this is necessary for OMEGA */
    int prscx = 48;
    int prscy = 0;
    int ovscx = 48;
    int ovscy = 100;
    char name[80] = "";
    const char *readkeys = "|ESO DET OUT[0-9]|ESO DET OUT|";
    cpl_vector    *sregion;

    if(plist == NULL)
    	return NULL;

    /*
     * Pre and over scan keywords are written in
     * HIERARCH ESO DET OUT# * keywords. The # value is
     * the Output index given in keywords
     * HIERARCH ESO DET OUT# INDEX. Get the index first
     * then, use it to get pre and over scan keywords
     */

    /* Get HIERARCH ESO DET OUT# INDEX */
	for(i=1; i<= next; i++) {
		sprintf(name,"ESO DET OUT%d INDEX", i);
		if(cpl_propertylist_has(plist, name) == 1) {
			index = cpl_propertylist_get_int (plist, name);
			break;
		}
	}

	if(index == 0){
		cpl_msg_warning(cpl_func,"Cannot find HIERARCH ESO DET OUT# INDEX");
		/* try anyway */
		index = 1;
	}

    /*Get pre and overscan regions from header*/
    sprintf(name, "ESO DET OUT%d PRSCX", index);
    if(cpl_propertylist_has(plist, name) == 1){
        prscx = cpl_propertylist_get_int (plist, name);
        sprintf(name,"ESO DET OUT%d PRSCY", index);
        prscy = cpl_propertylist_get_int (plist, name);
        sprintf(name,"ESO DET OUT%d OVSCX", index);
        ovscx = cpl_propertylist_get_int (plist, name);
        sprintf(name,"ESO DET OUT%d OVSCY", index);
        ovscy = cpl_propertylist_get_int (plist, name);
    }
    else {
    	return NULL;
    }

    sregion = cpl_vector_new(4);
    cpl_vector_set(sregion, 0, prscx);
    cpl_vector_set(sregion, 1, prscy);
    cpl_vector_set(sregion, 2, ovscx);
    cpl_vector_set(sregion, 3, ovscy);


    return sregion;

}

/**----------------------------------------------------------------*
 * @brief		Read offset keyword of JITTER header
 *
 * @param	plist	Property list
 *
 * @return	Value stored in keyword TEL.TARG.OFFSETSIZE or
 * 					0 in case of error.
 */
double omega_pfits_get_jitter_offset(cpl_propertylist *plist)
{

	double offset = 0.0;

	if(cpl_propertylist_has(plist,"ESO TEL TARG OFFSETSIZE") == 0)
		return 0;

	offset = cpl_propertylist_get_double(plist,"ESO TEL TARG OFFSETSIZE");

	return offset;
}

/**
 * @brief		Read offset X keyword of DITHER header
 *
 * @param	plist	Property list
 *
 * @return	Value stored in keyword TEL.TARG.OFFSETSIZEX or 0
 * 					in case of error.
 */
double omega_pfits_get_dither_offsetx(cpl_propertylist *plist)
{

	double offset = 0.0;

	if(cpl_propertylist_has(plist,"ESO TEL TARG OFFSETSIZEX") == 0)
		return 0;

	offset = cpl_propertylist_get_double(plist,"ESO TEL TARG OFFSETSIZEX");

	return offset;
}

/**
 * @brief		Read offset Y keyword of DITHER header
 *
 * @param	plist	Property list
 *
 * @return	Value stored in keyword TEL.TARG.OFFSETSIZEY or 0
 * 					in case of error.
 */
double omega_pfits_get_dither_offsety(cpl_propertylist *plist)
{

	double offset = 0.0;

	if(cpl_propertylist_has(plist,"ESO TEL TARG OFFSETSIZEY") == 0)
		return 0;

	offset = cpl_propertylist_get_double(plist,"ESO TEL TARG OFFSETSIZEY");

	return offset;
}

/**
 * @brief		Read offset ALPHA keyword of OFFSET header
 *
 * @param	plist	Property list
 *
 * @return	Value stored in keyword TEL.TARG.OFFSETALPHA or 0
 * 					in case of error.
 */
double omega_pfits_get_offset_alpha(cpl_propertylist *plist)
{

	double offset = 0.0;

	if(cpl_propertylist_has(plist,"ESO TEL TARG OFFSETALPHA") == 0)
		return 0;

	offset = cpl_propertylist_get_double(plist,"ESO TEL TARG OFFSETALPHA");

	return offset;
}

/**
 * @brief		Read offset DELTA keyword of OFFSET header
 *
 * @param	plist	Property list
 *
 * @return	Value stored in keyword TEL.TARG.OFFSETDELTA or 0
 * 					in case of error.
 */
double omega_pfits_get_offset_delta(cpl_propertylist *plist)
{

	double offset = 0.0;

	if(cpl_propertylist_has(plist,"ESO TEL TARG OFFSETDELTA") == 0)
		return 0;

	offset = cpl_propertylist_get_double(plist,"ESO TEL TARG OFFSETDELTA");

	return offset;
}

/*---------------------------------------------------------------------------*/
/**
    \brief    Get the value of cd1_1
    \param    plist       property list to read from
    \param    cd11        requested value of cd1_1
    \return The CD1_1 value or 0 in case of error.
 */
/*---------------------------------------------------------------------------*/

double omega_pfits_get_cd11(const cpl_propertylist *plist)
{

	if(cpl_propertylist_has(plist, "CD1_1") == 0)
		return 0;

    return(cpl_propertylist_get_double(plist,"CD1_1"));
}

/*---------------------------------------------------------------------------*/
/**
    \brief    Get the value of cd1_2
    \param    plist       property list to read from
    \param    cd12        requested value of cd1_2
    \return The CD1_2 value or 0 in case of error.
 */
/*---------------------------------------------------------------------------*/

double omega_pfits_get_cd12(const cpl_propertylist *plist)
{

	if(cpl_propertylist_has(plist, "CD1_2") == 0)
		return 0;

    return(cpl_propertylist_get_double(plist,"CD1_2"));

}

/*---------------------------------------------------------------------------*/
/**
    \brief    Get the value of cd2_1
    \param    plist       property list to read from
    \param    cd21        requested value of cd2_1
    \return the CD2_1 value or 0 in case of error
 */
/*---------------------------------------------------------------------------*/

double omega_pfits_get_cd21(const cpl_propertylist *plist)
{

	if(cpl_propertylist_has(plist, "CD2_1") == 0)
		return 0;

    return(cpl_propertylist_get_double(plist,"CD2_1"));

}

/*---------------------------------------------------------------------------*/
/**
    \brief    Get the value of cd2_2
    \param    plist       property list to read from
    \param    cd22        requested value of cd2_2
    \return The CD2_2 value or 0 in case of error.
 */
/*---------------------------------------------------------------------------*/

double omega_pfits_get_cd22(const cpl_propertylist *plist)
{

	if(cpl_propertylist_has(plist, "CD2_2") == 0)
		return 0;

    return(cpl_propertylist_get_double(plist,"CD2_2"));

}

/*---------------------------------------------------------------------------*/
/**
    \brief    Get the value of CDELT2
    \param    plist       property list to read from
    \param    cd22        requested value of cd2_2
    \return The CDELT2 value or 0 in case of error
 */
/*---------------------------------------------------------------------------*/

double omega_pfits_get_cdelt2(const cpl_propertylist *plist)
{

	if(cpl_propertylist_has(plist, "CDELT2") == 0)
		return 0;

    return(cpl_propertylist_get_double(plist,"CDELT2"));

}

/**
 * @brief		Get the size of a CCD excluding the pre/overscan regions
 *
 * @param plist		Property list containing keywords
 * @param x			Output X size
 * @param y			Output Y size
 *
 * @return		Nothing
 *
 */
void omega_pfits_get_chip_size(const cpl_propertylist *plist,int *xs, int *ys)
{
    int prscx = 48;
    int prscy = 0;
    int ovscx = 48;
    int ovscy = 100;
    int naxis1 = 2144;
    int naxis2 = 4200;
    int x0 = 1;
    int x1 = 1;
    int y0 = 1;
    int y1 = 1;
    cpl_vector *vec;

    if(plist == NULL){
    	/* Return default values of X,Y sizes */
    	cpl_msg_info(cpl_func,"Using default chip sizes");
        x0 = prscx + 1;
        y0 = prscy + 1;
        x1 = naxis1 - ovscx;
        y1 = naxis2 - ovscy;

        *xs = x1 - x0+1;
        *ys = y1 - y0+1;
        return;
    }

    /* NAXIS */
    naxis1 = cpl_propertylist_get_int(plist, "NAXIS1");
    naxis2 = cpl_propertylist_get_int(plist, "NAXIS2");

    /* PRSCX, PRSCY, OVSCX, OVSCY */
    vec = omega_pfits_get_preovscan(plist);
    prscx = cpl_vector_get(vec,0);
    prscy = cpl_vector_get(vec,1);
    ovscx = cpl_vector_get(vec,2);
    ovscy = cpl_vector_get(vec,3);

    x0 = prscx + 1;
    y0 = prscy + 1;
    x1 = naxis1 - ovscx;
    y1 = naxis2 - ovscy;

    *xs = x1 - x0+1;
    *ys = y1 - y0+1;

    freevector(vec);

    return;
}

/**
 * @brief		Returns the quadrant of the extension
 *
 * @param extname Name of the extension
 *
 * @return		Quadrant or 0 if quadrant can not be found
 *
 */

static int omega_get_quadrant(const char * extname, int filtno)
{

	/*
	Quadrant layout following Cartesian coordinate system
	  2  |  1
	 ---------
	  3  |  4
	 */
	/*

	if filtno 1-6 | filter is in Magazin A
	if filtno 7-12| filter is in Magazin B
	magazine B is rotated by 180 degree compared to magazine A
	 */
	if (extname == NULL){
		return 0;
	}
	else if((strcmp("ESO_CCD_#96", extname)== 0) ||
			(strcmp("ESO_CCD_#95", extname)== 0) ||
			(strcmp("ESO_CCD_#94", extname)== 0) ||
			(strcmp("ESO_CCD_#93", extname)== 0) ||
			(strcmp("ESO_CCD_#88", extname)== 0) ||
			(strcmp("ESO_CCD_#87", extname)== 0) ||
			(strcmp("ESO_CCD_#86", extname)== 0) ||
			(strcmp("ESO_CCD_#85", extname)== 0)) {
		if(filtno>=1 && filtno<=6){
			return 1;
		}
		else if (filtno>=7 && filtno<=12){
			return 3;
		}
	}
	else if((strcmp("ESO_CCD_#92", extname)== 0) ||
			(strcmp("ESO_CCD_#91", extname)== 0) ||
			(strcmp("ESO_CCD_#90", extname)== 0) ||
			(strcmp("ESO_CCD_#89", extname)== 0) ||
			(strcmp("ESO_CCD_#84", extname)== 0) ||
			(strcmp("ESO_CCD_#83", extname)== 0) ||
			(strcmp("ESO_CCD_#82", extname)== 0) ||
			(strcmp("ESO_CCD_#81", extname)== 0)) {
		if(filtno>=1 && filtno<=6){
			return 2;
		}
		else if (filtno>=7 && filtno<=12){
			return 4;
		}
	}
	else if((strcmp("ESO_CCD_#76", extname)== 0) ||
			(strcmp("ESO_CCD_#75", extname)== 0) ||
			(strcmp("ESO_CCD_#74", extname)== 0) ||
			(strcmp("ESO_CCD_#73", extname)== 0) ||
			(strcmp("ESO_CCD_#68", extname)== 0) ||
			(strcmp("ESO_CCD_#67", extname)== 0) ||
			(strcmp("ESO_CCD_#66", extname)== 0) ||
			(strcmp("ESO_CCD_#65", extname)== 0)) {
		if(filtno>=1 && filtno<=6){
			return 3;
		}
		else if (filtno>=7 && filtno<=12){
			return 1;
		}
	}
	else if((strcmp("ESO_CCD_#80", extname)== 0) ||
			(strcmp("ESO_CCD_#79", extname)== 0) ||
			(strcmp("ESO_CCD_#78", extname)== 0) ||
			(strcmp("ESO_CCD_#77", extname)== 0) ||
			(strcmp("ESO_CCD_#72", extname)== 0) ||
			(strcmp("ESO_CCD_#71", extname)== 0) ||
			(strcmp("ESO_CCD_#70", extname)== 0) ||
			(strcmp("ESO_CCD_#69", extname)== 0)){
		if(filtno>=1 && filtno<=6){
			return 4;
		}
		else if (filtno>=7 && filtno<=12){
			return 2;
		}
	}

	else {
		return 0;
	}


}

/**@}*/
