/* $Id: omega_photometry.c,v 1.21 2012/04/30 10:46:01 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/04/30 10:46:01 $
 * $Revision: 1.21 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "omega_photometry.h"
#include "omega_catalog.h"
#include "omega_dfs.h"
#include "omega_pfits.h"
#include "omega_utils.h"

/**
 * @defgroup  omega_photometry Photometric Functions
 *
 * \par Synopsis
 *  Functions to create a photometric table and for deriving the zeropoint of
 *  images.
 *
 *  \code
 *  #include omega_photometry.h
 *  #include omega_catalog.h
 *  #include omega_dfs.h
 *  #include omega_pfits.h
 *  #include omega_utils.h
 *  \endcode
 */
/**@{*/


/*----------------------------------------------------------------------------*/
/**	@brief	Match std stars with std reference catalogue and
 * 			calculate zero point for matched stars
 *
 * @param  star   name of reduced standard star in disk
 * @param  refcat frame of reference standard star catalogue
 * @param  pars   parameters list
 * @param  extra  table to write extra information
 *
 * @return table of matches between image and reference catalogue or NULL
 * 			in case of error.
 */
/*----------------------------------------------------------------------------*/

cpl_table *omega_photom_tbl(const char *star, double gain,
		const cpl_frame *refcat, cpl_parameterlist *pars, cpl_table **extra)
{
	int i, nsources, nstds;
	int npars = 0;
	int nmatches = 0;
	double exptime = 0.0;
    double omega_x1,omega_x2,omega_y1,omega_y2;
    double radius = 0.0;
    double photom_thre = 0.0;
    double airmass = 0.0;
    double pixscalex = 0.0;
    double pixscaley = 0.0;
    double seeing = 0.0;
    double ellipticity = 0.0;
    double aperture = 0.0;
 	char *cmd = NULL;
	const char *text = NULL;
	const char *path;
	const char *sex_conf = NULL;
	const char *sex_conv = NULL;
	const char *sex_par = NULL;
	const char *sex_nnw = NULL;
	const char *srcs = "omega_cat_photom_sources.fits";
	cpl_parameter *par;
	cpl_array *status;
	cpl_table *sources, *stds, *data;
	cpl_table *matchstds = NULL;
	cpl_matrix *from, *to;
	cpl_wcs *wcs;
	cpl_vector *coords;
	cpl_propertylist *xlist = NULL;
	cpl_propertylist *p, *filterlist;

	/* Get parameters */
	npars = cpl_parameterlist_get_size(pars);
	par = cpl_parameterlist_get_first(pars);

	for(i=0; i<npars; i++) {
		text = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
			if(strcmp("bin-path", text) == 0) {
    			path = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-config", text) == 0) {
    			sex_conf = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-conv", text) == 0) {
    			sex_conv = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-param", text) == 0) {
    			sex_par = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-nnw", text) == 0) {
    			sex_nnw = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("photom-thre", text) == 0){
				photom_thre = cpl_parameter_get_double(par);
			}
			else if(strcmp("radius", text) == 0){
				radius = cpl_parameter_get_double(par);
			}
			else if(strcmp("aperture", text) == 0){
				aperture = cpl_parameter_get_double(par);
			}
		par = cpl_parameterlist_get_next(pars);

	}

//	for(i = 0; i < 4; i++) {
	/* Create catalog of sources from image */
	cmd = cpl_sprintf("%s/sex %s -c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
			"-FILTER Y -DETECT_THRESH %g -ANALYSIS_THRESH %g -GAIN %g -PHOT_APERTURES %g -CATALOG_TYPE FITS_1.0 "
			"-CATALOG_NAME %s -CHECKIMAGE_NAME %s -CHECKIMAGE_TYPE OBJECTS -DETECT_MINAREA 5",
			path, star, sex_conf, sex_par, sex_conv, sex_nnw, photom_thre, 3.0, gain, aperture,
			srcs, "omega_ima_detected_photom_srcs.fits");

			cpl_msg_debug(cpl_func,"cmd: %s",cmd);
	if (system(cmd) != 0) {
		cpl_msg_error(cpl_func,"Failed to detect sources in image using Sextractor");
		freespace(cmd);
		return NULL;
	}
	freespace(cmd);

	data = cpl_table_load(srcs, 1, 0);
	if (data == NULL) {
		cpl_msg_error(cpl_func,"Sources table is NULL");
		return NULL;
	}

	/* Filter for FLAGS = 0  */
	sources = omega_filter_table_int(data, "FLAGS", CPL_EQUAL_TO, 0, &nsources);
	freetable(data);
	cpl_msg_debug(cpl_func,"Number of stars in reduced image: %d", nsources);

	/* Calculate the seeing */
	xlist = cpl_propertylist_load(star, 1);
	omega_get_pixelscale(xlist, &pixscalex, &pixscaley);

	/* Get extra information */

	p = cpl_propertylist_load(star, 0);

	/*Add extension name to the propertylist*/
	if (cpl_propertylist_has(xlist,"EXTNAME") == 1) {
		cpl_propertylist_update_string(p, "EXTNAME",
				cpl_propertylist_get_string(xlist, "EXTNAME"));

	}

	filterlist = omega_pfits_get_filter_info(p);
	omega_pfits_get_mean_airmass(p, &airmass);
	exptime = omega_pfits_get_exptime(p);
	freeplist(p);



    if(omega_basic_param(sources, pixscalex, &seeing, &ellipticity) == -1){
		cpl_msg_warning(cpl_func,"Cannot find stars with CLASS_STAR near 1 to calculate seeing");
    }

    /* Append extra info into a new extension table*/
 	*extra = cpl_table_new(1);
 	cpl_table_new_column(*extra, "FILT_ID", CPL_TYPE_STRING);
 	cpl_table_set_string(*extra, "FILT_ID",0, cpl_propertylist_get_string(filterlist, "FILT_ID"));
 	cpl_table_new_column(*extra, "FILT_NAME", CPL_TYPE_STRING);
 	cpl_table_set_string(*extra, "FILT_NAME",0, cpl_propertylist_get_string(filterlist, "FILT_NAME"));
 	cpl_table_new_column(*extra, "REF_MAG_ID", CPL_TYPE_STRING);
 	cpl_table_set_string(*extra, "REF_MAG_ID",0, cpl_propertylist_get_string(filterlist, "REF_MAG_ID"));
 	cpl_table_new_column(*extra, "REF_MAG_ID_ERR", CPL_TYPE_STRING);
 	cpl_table_set_string(*extra, "REF_MAG_ID_ERR",0, cpl_propertylist_get_string(filterlist, "REF_MAG_ID_ERR"));
 	cpl_table_new_column(*extra, "CWL", CPL_TYPE_DOUBLE);
 	cpl_table_set_double(*extra, "CWL",0, cpl_propertylist_get_double(filterlist, "CWL"));
 	cpl_table_new_column(*extra, "EXPTIME", CPL_TYPE_DOUBLE);
 	if(cpl_propertylist_has(xlist, "EXPTIME") == 0)
 		cpl_table_set_double(*extra, "EXPTIME",0, exptime);
 	else
 		cpl_table_set_double(*extra, "EXPTIME",0, cpl_propertylist_get_double(xlist, "EXPTIME"));

 	cpl_table_new_column(*extra, "AIRMASS", CPL_TYPE_DOUBLE);
 	cpl_table_set_double(*extra, "AIRMASS",0, airmass);
 	cpl_table_new_column(*extra, "CHIP_ID", CPL_TYPE_STRING);
 	cpl_table_set_string(*extra, "CHIP_ID", 0, omega_pfits_get_chipid(xlist));
 	cpl_table_new_column(*extra, "SEEING", CPL_TYPE_DOUBLE);
 	cpl_table_set_double(*extra, "SEEING", 0, seeing);
 	cpl_table_new_column(*extra, "ELLIPTICITY", CPL_TYPE_DOUBLE);
 	cpl_table_set_double(*extra, "ELLIPTICITY", 0, ellipticity);

	freeplist(filterlist);

    /* Sort sources table by Y_IMAGE */
    p = cpl_propertylist_new();
    cpl_propertylist_append_bool(p,"Y_IMAGE",0);
    if (cpl_table_sort(sources, p) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func,"Cannot sort sources table");
	    cpl_propertylist_delete(p);
	    cpl_propertylist_delete(xlist);
    	freetable(sources);
	    return NULL;
    }
    freeplist(p);

    /* Create matrix of X,Y coordinates */
    from = cpl_matrix_new(nsources, 2);
    for (i=0; i<nsources; i++){
        float xim = cpl_table_get_float(sources, "X_IMAGE", i, NULL);
        float yim = cpl_table_get_float(sources, "Y_IMAGE", i, NULL);
        cpl_matrix_set(from, i, 0, xim);
        cpl_matrix_set(from, i, 1, yim);
    }

    /* Get coverage of image */
    if(omega_get_coverage(xlist, 0, from, &omega_x1, &omega_x2, &omega_y1, &omega_y2) != 0){
    	cpl_msg_error(cpl_func,"Cannot get coverage of image");
    	freematrix(from);
    	freeplist(xlist);
    	freetable(sources);
    	return NULL;
    }
    freematrix(from);

    /* Get STD reference stars from catalogue */
    /* ramin = x1 ramax = x2 decmin = y1 decmax = y2;*/
    /* Write in 'nstds' the number of sources in the table */
    coords = cpl_vector_new(4);
    cpl_vector_set(coords, 0, omega_x1);
    cpl_vector_set(coords, 1, omega_x2);
    cpl_vector_set(coords, 2, omega_y1);
    cpl_vector_set(coords, 3, omega_y2);

    stds = omega_get_stds_table(cpl_frame_get_filename(refcat), coords, &nstds);
    if (stds == NULL) {
    	cpl_msg_warning(cpl_func,"Cannot find reference standard stars in this region: "
    			"{RA-RA:DEC-DEC}={%g-%g:%g=%g}", omega_x1,omega_x2,omega_y1,omega_y2);
    	freeplist(xlist);
    	freetable(sources);
    	freevector(coords);
    	return NULL;
    }
	freevector(coords);

    cpl_msg_debug(cpl_func,"%d reference standards fall in region: {RA-RA:DEC-DEC}={%g-%g:%g=%g}",
							nstds,omega_x1,omega_x2,omega_y1,omega_y2);

    /* Convert Ra, Dec to X,Y using the WCS information from image */
    from = cpl_matrix_new(nstds, 2);
    for (i=0; i<nstds; i++){
    	cpl_matrix_set(from, i, 0, cpl_table_get_double(stds, "Ra", i, NULL));
    	cpl_matrix_set(from, i, 1, cpl_table_get_double(stds, "Dec", i, NULL));
    }

    to = NULL;
    status = NULL;
    wcs = cpl_wcs_new_from_propertylist(xlist);
    if(cpl_wcs_convert(wcs, from, &to, &status, CPL_WCS_WORLD2PHYS) != CPL_ERROR_NONE){
    	freewcs(wcs);
    	cpl_array_delete(status);
    	freematrix(from);
    	freematrix(to);
    	freeplist(xlist);
    	freetable(sources);
    	freetable(stds);
    	cpl_msg_error(cpl_func,"Error in cpl_wcs conversion. %s",cpl_error_get_message());
    	return NULL;
    }
    freewcs(wcs);
    freematrix(from);
    freeplist(xlist);
	cpl_array_delete(status);

    /* Add the converted x,y coordinate columns to the reference stds table */
    cpl_table_new_column(stds,"xpredict",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(stds,"xpredict","pixels");
    cpl_table_new_column(stds,"ypredict",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(stds,"ypredict","pixels");

    for(i=0; i< nstds; i++){
    	float xp = (float)cpl_matrix_get(to, i, 0);
    	float yp = (float)cpl_matrix_get(to, i, 1);
    	cpl_table_set_float(stds,"xpredict", i, xp);
    	cpl_table_set_float(stds,"ypredict", i, yp);
    }
    freematrix(to);

    /* Sort stds table by ypredict */
    p = cpl_propertylist_new();
    cpl_propertylist_append_bool(p,"ypredict",0);
    if (cpl_table_sort(stds, p) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func,"Cannot sort stds table");
	    freeplist(p);
    	freetable(sources);
    	freetable(stds);
	    return NULL;
    }
    freeplist(p);

    /* Match sources and stds stars */
    if (omega_matchstds(sources, stds, radius, &matchstds) != 0){
    	cpl_msg_error(cpl_func,"Failed to match reference standards to image sources. %s",cpl_error_get_message());
    	freetable(stds);
    	freetable(sources);
    	return NULL;
    }

    nmatches = cpl_table_get_nrow(matchstds);
    if(nmatches < 2){
    	cpl_msg_info(cpl_func,"Not enough matches (n=%d). Trying again...",nmatches);
    	/* Try again with increased radius */
    	for(i=0; i<4; i++){
    		radius = radius + 10.0;
    		freetable(matchstds);
    		if (omega_matchstds(sources, stds, radius, &matchstds) != 0){
    			cpl_msg_error(cpl_func,"Failed to match reference standards to image sources. %s",
								cpl_error_get_message());
    			freetable(stds);
    			freetable(sources);
    			freetable(matchstds);
    			return NULL;
    		}
    		nmatches = cpl_table_get_nrow(matchstds);
    		if(nmatches < 2)
    			continue;

    	}
    }
    if(nmatches < 2){
    	cpl_msg_warning(cpl_func,"There are no standard stars in image");
    	freetable(stds);
    	freetable(sources);
    	freetable(matchstds);
    	return NULL;
    }

    cpl_msg_info(cpl_func,"Found %d standard stars in image", nmatches);

    freetable(stds);
    freetable(sources);

    return matchstds;
}


void omega_get_extinction(cpl_table *extra, cpl_table *extinct2,
		cpl_table *extinct1, double *extinction, cpl_table *monit2,
		double *shift_err, double *extinction_airmass,
		double *extinction_error_kwad)
{
	/* Calculate extinction */
	/*Get information from other tables */

	double upr = 0.;
	double slope = 0.;
	double shift = 0.;
	double temp_ext = 0.;
	double upr_ext = 0.;
	double lwr_ext = 0.;
	double cwl = 0.;
	double airmass = 0.;
	int i = 0;
	int numrows = 0;
	double wincr = 0;
	double lwr = 0;
	cwl = cpl_table_get_double(extra, "CWL", 0, NULL);
	airmass = cpl_table_get_double(extra, "AIRMASS", 0, NULL);
	wincr = cpl_table_get_float(extinct2, "wavel_incr", 0, NULL);
	lwr = cwl - fmod(cwl, wincr); /*takes the float remainder of operation*/
	numrows = cpl_table_get_nrow(extinct1);
	for(i=0; i< numrows; i++) {
		int number = cpl_table_get_float(extinct1,"wavelength", i, NULL);
		if(number == lwr) {
			lwr_ext = cpl_table_get_float(extinct1,"extinction",i, NULL);
			break;
		}
	}
	upr = cwl - fmod(cwl, wincr) + wincr;
	for(i=0; i< numrows; i++) {
		int number = cpl_table_get_float(extinct1,"wavelength",	i, NULL);
		if(number == upr) {
			upr_ext = cpl_table_get_float(extinct1,"extinction",i, NULL);
			break;
		}
	}
	slope = (upr_ext - lwr_ext) / wincr;
	temp_ext = fmod(cwl, wincr);
	*extinction = (temp_ext * slope) + lwr_ext;

	/* Get shift from monitoring report */

	shift = cpl_table_get_float(monit2, "shift", 0, NULL);
	*shift_err = cpl_table_get_float(monit2, "shift_err", 0, NULL);
	*extinction_airmass = *extinction + shift;
	*extinction = (*extinction_airmass) * (airmass);
	*extinction_error_kwad = ((*shift_err) * (airmass)) * ((*shift_err) * (airmass));

	return;
}

/*----------------------------------------------------------------------------*/
/**	@brief	Routine to calculate the zeropoint from standard stars.

  @param    matches       table calculated in omega_photom_tbl
  @param    extra    	  extra table calculated in omega_photom_tbl
  @param    monit         name of monitoring report table
  @param    extname       name of extinction curve table
  @param    zptbl		  output table where to write results

  @return   table with zeropoints or NULL in case of error.
 */
/*----------------------------------------------------------------------------*/
cpl_table *omega_zeropoints(cpl_table *matches, cpl_table *extra, const cpl_frame *monitfr,
							const cpl_frame *extfr, const cpl_frame *colfr)
{

    int i = 0;
    int j = 0;
    int nrows = 0;
    double extinction = 0.0;
    double shift_err = 0.0;
    double extinction_airmass = 0.0;
    double meanZP = 0.0;
    double ZPerror = 0.0;
    double extinction_error_kwad = 0.0;
    double weight_sum = 0.0;
    double sum_data = 0.0;
	double clip_median=0.;
	double clip_stdev=0.;
	double clip_min=0.;
	double clip_max=0.;

    const char *refmag;
    const char *refmagerr;
    const char *filt_id = NULL;
    const char *filt_name = NULL;

    cpl_table *raw_tbl, *monit1, *monit2;
    cpl_table *extinct1, *extinct2;
    cpl_table *coltbl = NULL;
    cpl_table *tbl, *zptbl;
	cpl_table * raw_tbl_tmp=NULL;


    /*Load monitoring report tables*/
    monit1 = cpl_table_load(cpl_frame_get_filename(monitfr), 1, 0);
    if (monit1 == NULL) {
    	cpl_msg_error(cpl_func, "Cannot load 1st extension of monitoring report table");
    	return NULL;
    }

    monit2 = cpl_table_load(cpl_frame_get_filename(monitfr), 2, 0);
    if (monit2 == NULL) {
    	cpl_msg_error(cpl_func, "Cannot load 2nd extension of monitoring report table");
    	freetable(monit1);
    	return NULL;
    }

    /*Load extinction curve tables*/
    extinct1 = cpl_table_load(cpl_frame_get_filename(extfr), 1, 0);
    if (extinct1 == NULL) {
    	cpl_msg_error(cpl_func, "Cannot load 1st extension of std extinction table");
    	freetable(monit1);
    	freetable(monit2);
    	return NULL;
    }

    extinct2 = cpl_table_load(cpl_frame_get_filename(extfr), 2, 0);
    if (extinct2 == NULL) {
    	cpl_msg_error(cpl_func, "Cannot load 2nd extension of std extinction table");
    	freetable(monit1);
    	freetable(monit2);
    	freetable(extinct1);
    	return NULL;
    }

    /* Load colour terms table if present*/
    if (colfr != NULL) {
    	coltbl = cpl_table_load(cpl_frame_get_filename(colfr), 1, 0);
    	if (coltbl == NULL){
    	cpl_msg_error(cpl_func, "Cannot load colour terms table");
    	freetable(monit1);
    	freetable(monit2);
    	freetable(extinct1);
    	freetable(extinct2);
    	return NULL;
    	}
    }

    filt_id = cpl_table_get_string(extra, "FILT_ID", 0);
    filt_name = cpl_table_get_string(extra, "FILT_NAME", 0);


    /* Calculate instrumental magnitudes and raw ZP */
    /*
     * inst_mag = -2.5 * log(FLUX_ISO/EXPTIME)
     * high = -2.5 * log( (FLUX_ISO + FLUX_ISO_ERR) / FLUX_ISO )
     * low = -2.5 * log( (FLUX_ISO - FLUX_ISO_ERR) / FLUX_ISO )
     * inst_mag_err = ( fabs(high) + fabs(low) ) / 2.0
     *
     * Using MAG_APER from Sextractor, not FLUX_ISO
     * raw_ZP = ref_mag - inst_mag - colour_term + extinction*airmass
     * raw_ZP_err = sqrt( ref_mag_err**2 + inst_mag_err**2 )
     *
     */
    nrows = cpl_table_get_nrow(matches);
    raw_tbl = cpl_table_new(nrows);
    refmag = cpl_table_get_string(extra, "REF_MAG_ID", 0);
    refmagerr = cpl_table_get_string(extra, "REF_MAG_ID_ERR", 0);

    cpl_table_duplicate_column(raw_tbl,"mag", matches, refmag);
    cpl_table_duplicate_column(raw_tbl,"mag_err", matches, refmagerr);
    cpl_table_duplicate_column(raw_tbl, "instmag", matches, "MAG_APER");
    cpl_table_duplicate_column(raw_tbl,"instmag_err", matches, "MAGERR_APER");

    /* Reject outliers from source list */
    tbl = omega_filter_table_float(raw_tbl, "instmag", CPL_LESS_THAN, 25.0, &nrows);
    freetable(raw_tbl);
    raw_tbl = omega_filter_table_float(tbl, "instmag", CPL_GREATER_THAN, -25.0, &nrows);
    freetable(tbl);
    tbl = omega_filter_table_float(raw_tbl, "mag", CPL_NOT_EQUAL_TO, 0.0, &nrows);
    freetable(raw_tbl);
    raw_tbl = cpl_table_duplicate(tbl);
    freetable(tbl);

    if(cpl_table_get_column_mean(raw_tbl, "mag") == 0.0) {
    	cpl_msg_warning(cpl_func, "Reference table magnitudes (%s) are NULL for the selected stds.",refmag);
    	freetable(raw_tbl);
    	freetable(monit1);
    	freetable(monit2);
    	freetable(extinct1);
    	freetable(extinct2);
    	if(coltbl) freetable(coltbl);
    	return NULL;
    }

   cpl_table_subtract_columns(raw_tbl,"mag","instmag");

    /* Use colour terms if present */
    if(coltbl != NULL){
    	double colour = 0.0;
    	for(j = 0; j < 8; j++){
    		if(strcmp(filt_name, cpl_table_get_string(coltbl, "FILT_NAME", j)) == 0){
    			colour = cpl_table_get_double(coltbl, "COLOUR_TERM",j, NULL);
    			break;
    		}
    		else if(strcmp(filt_id, cpl_table_get_string(coltbl, "FILT_ID", j)) == 0){
    			colour = cpl_table_get_double(coltbl, "COLOUR_TERM",j, NULL);
    			break;
    		}
    	}

        cpl_table_subtract_scalar(raw_tbl, "mag", colour);
        freetable(coltbl);
    }

    cpl_table_erase_column(raw_tbl,"instmag");
    cpl_table_name_column(raw_tbl,"mag","rawZP");


    /*Preselect the zeropoint table by applying a clipping algorithm
     * if there are enough entries */

    if(cpl_table_get_nrow(raw_tbl)>=3){

    	raw_tbl_tmp=cpl_table_duplicate(raw_tbl);
    	freetable(raw_tbl);

    	clip_median= cpl_table_get_column_median(raw_tbl_tmp, "rawZP");
    	clip_stdev = cpl_table_get_column_stdev(raw_tbl_tmp, "rawZP");
    	clip_min=clip_median - 2 * clip_stdev;
    	clip_max=clip_median + 2 * clip_stdev;

    	cpl_msg_info(cpl_func,"Using objects with magnitudes within [%g, %g] "
    			"to derive the weighted zeropoint", clip_min, clip_max);

    	/*select objects residing in the intervall [clip_min, clip_max]*/
    	cpl_table_and_selected_float(raw_tbl_tmp, "rawZP", CPL_NOT_GREATER_THAN,
    			clip_max);
    	cpl_table_and_selected_float(raw_tbl_tmp, "rawZP", CPL_NOT_LESS_THAN,
    			clip_min);

    	raw_tbl=cpl_table_extract_selected(raw_tbl_tmp);
    	freetable(raw_tbl_tmp);

    }

    /* Calculate instrumental magnitudes errors*/
    cpl_table_power_column(raw_tbl,"mag_err", 2.0);
    cpl_table_power_column(raw_tbl,"instmag_err", 2.0);
    cpl_table_add_columns(raw_tbl,"mag_err", "instmag_err");
    cpl_table_power_column(raw_tbl,"mag_err",0.5);
    cpl_table_erase_column(raw_tbl,"instmag_err");

    cpl_table_name_column(raw_tbl,"mag_err","rawZP_err");
    omega_get_extinction(extra, extinct2, extinct1, &extinction, monit2,
    		&shift_err, &extinction_airmass, &extinction_error_kwad);
    /* Calculate weighted average of ZP */
    cpl_table_add_scalar(raw_tbl, "rawZP", extinction);
    cpl_table_power_column(raw_tbl, "rawZP_err", 2.0);
    cpl_table_add_scalar(raw_tbl, "rawZP_err", extinction_error_kwad);
    cpl_table_power_column(raw_tbl, "rawZP_err", 0.5);
    cpl_table_name_column(raw_tbl, "rawZP", "zeropoint");
    cpl_table_name_column(raw_tbl, "rawZP_err", "zeropoint_err");

    cpl_table_duplicate_column(raw_tbl, "weights", raw_tbl, "zeropoint_err");

    cpl_table_power_column(raw_tbl, "weights", -2.);


    weight_sum = 0.0;
    for (i=0; i<nrows; i++) {
    	weight_sum += cpl_table_get_float(raw_tbl,"weights", i, NULL);
    }

    cpl_table_multiply_columns(raw_tbl, "zeropoint", "weights");

    sum_data = 0.0;
    for(i=0; i<nrows; i++) {
    	sum_data += cpl_table_get_float(raw_tbl, "zeropoint", i, NULL);
    }

    /* Mean ZP and ZP_err*/
    meanZP = sum_data/weight_sum;
    ZPerror = sqrt(1./weight_sum);

    cpl_msg_debug(cpl_func,"meanZP: %f ZPerror: %f",meanZP,ZPerror);


    cpl_msg_info("","Photometric parameters for filter %s",filt_name);
    cpl_msg_info("","Weighted Average Zeropoint (mag)  %0.5g", meanZP);
    cpl_msg_info("","Zeropoint error (mag)             %0.5g", ZPerror);
    cpl_msg_info("","Extinction (mag/airmass)          %0.5g", extinction_airmass);
    cpl_msg_info("","Extinction times airmass (mag)    %0.5g", extinction);
    cpl_msg_info("","Extinction error (mag/airmass)    %0.5g", shift_err);
    cpl_msg_info("","Seeing (arcsec)                   %0.5g", cpl_table_get_double(extra,"SEEING",0,NULL));
    cpl_msg_info("","Ellipticity                       %0.5g", cpl_table_get_double(extra,"ELLIPTICITY",0,NULL));



    /* Save a table  with the following columns */
    zptbl = cpl_table_new(1);
    cpl_table_duplicate_column(zptbl,"CHIP_ID", extra, "CHIP_ID");
    cpl_table_duplicate_column(zptbl,"FILT_ID", extra, "FILT_ID");
    cpl_table_duplicate_column(zptbl,"FILT_NAME", extra, "FILT_NAME");

    cpl_table_new_column(zptbl,"ZEROPOINT", CPL_TYPE_DOUBLE);
    cpl_table_set_double(zptbl,"ZEROPOINT", 0, meanZP);

    cpl_table_new_column(zptbl,"ZEROPOINT_ERR", CPL_TYPE_DOUBLE);
    cpl_table_set_double(zptbl,"ZEROPOINT_ERR", 0, ZPerror);

    cpl_table_new_column(zptbl,"EXTINCTION", CPL_TYPE_DOUBLE);
    cpl_table_set_double(zptbl,"EXTINCTION", 0, extinction_airmass);

    cpl_table_new_column(zptbl,"EXTINCTION_ERR", CPL_TYPE_DOUBLE);
    cpl_table_set_double(zptbl,"EXTINCTION_ERR", 0, shift_err);

    cpl_table_duplicate_column(zptbl,"SEEING", extra, "SEEING");
    cpl_table_duplicate_column(zptbl,"ELLIPTICITY", extra, "ELLIPTICITY");

    cpl_table_new_column(zptbl, "NSTARS", CPL_TYPE_INT);
    cpl_table_set_int(zptbl, "NSTARS", 0, nrows);

    /*Clean up */
    freetable(monit1);
    freetable(monit2);
    freetable(extinct1);
    freetable(extinct2);
    freetable(raw_tbl);

    return zptbl;
}


/*----------------------------------------------------------------------------*/
/**	@brief	Routine to calculate the seeing from stars.

  @param    sources       Sextractor table with stars
  @param    scale         scale in degrees (as in keyword CD1_1)
  @param    seeing		  double variable to write seeing

  @return   0 in case of success or -1 in case of error
 */
/*----------------------------------------------------------------------------*/
int omega_basic_param(cpl_table *sources, double scale, double *seeing,
		double * ellipticity)
{

	int n = 0;
	float stellarity = 0.9;
	double median = 0.0;
	cpl_table *tbl;

	*seeing = 0.0;
	cpl_table_select_all(sources);

	n = cpl_table_and_selected_float(sources, "CLASS_STAR", CPL_NOT_LESS_THAN, stellarity);
	if(n <= 0) {
		stellarity = 0.8;
		n = cpl_table_and_selected_float(sources, "CLASS_STAR", CPL_NOT_LESS_THAN, stellarity);
		if(n <= 0){
		    return -1;
		}
	}

	cpl_table_and_selected_float(sources, "FWHM_IMAGE", CPL_GREATER_THAN, 0.1);
	tbl = cpl_table_extract_selected(sources);
	median = cpl_table_get_column_median(tbl, "FWHM_IMAGE");
	cpl_msg_debug(cpl_func,"median1: %g",median);
	freetable(tbl);

	cpl_table_and_selected_float(sources, "FWHM_IMAGE", CPL_NOT_GREATER_THAN, median);
	tbl = cpl_table_extract_selected(sources);
	median = cpl_table_get_column_median(tbl, "FWHM_IMAGE");
	cpl_msg_debug(cpl_func,"median2: %g",median);

	*seeing = median * fabs(scale)*3600;

	if(cpl_table_has_column(tbl,"ELLIPTICITY")){
		*ellipticity = cpl_table_get_column_median(tbl, "ELLIPTICITY");
	}
	freetable(tbl);

	return 0;
}


/**@}*/
