/* $Id: omega_satellites.c,v 1.5 2012/08/22 08:51:45 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/08/22 08:51:45 $
 * $Revision: 1.5 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/
#include <string.h>
#include <math.h>

#include "omega_dfs.h"
#include "omega_background.h"
#include "omega_utils.h"
#include "omega_satellites.h"


/**
 * @defgroup omega_satellite  Satellite detections
 *
 * This module provides functions to detect satellite trails via Hough
 * transformation
 *
 * @par Synopsis:
 *   Functions to detect satellite trails
 *
 * @code
 * #include "omega_dfs.h"
 * #include "omega_background.h"
 * #include "omega_utils.h"
 * #include "omega_satellites.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/


/*----------------------------------------------------------------------------*/
/**
  @brief     A SatelliteMap is constructed from ScienceFrames using the Hough
    		Transform to detect long linear features, and using the inverse
    		transform to mask these features in a pixelmap

    		Note that for images without satellite tracks the inverse
    		transform produces a pxxelmap without bad pixels. It is
    		probably easier just to save these (highly compressed) images
    		than trying to distinguish in subsequent operations between
    		SatelliteMaps with and without satellite tracks
	  
  @param    
  @param    
  @return   
 */
/*----------------------------------------------------------------------------*/
/*cpl_mask * detsat(cpl_image *science, cpl_image *flat, cpl_image *weight0, cpl_parameterlist *hpars)*/
cpl_mask * detsat(cpl_image *nscience, cpl_parameterlist *hpars)
{

	int count = 0;
	int i = 0;
	int j = 0;
	int npars = 0;
	int nx = 0;
	int ny = 0;
	
	double dthre = 5.0;
	double hthre = 1000.0;
	double stdev = 1.0;
	
	float threshold = 0.0f;
	
	const char *_id = "detsat";
	const char *alias = NULL;
/*	const char *bckname = "OMEGA_back0062.fits";*/
	
	cpl_parameter 		*par;
	cpl_image 			*hough;
/*	cpl_image 			*nscience;*/
/*	cpl_image 			*temp;*/
/*	cpl_image 			*back;*/
	cpl_mask 			*hough_map;
/*	cpl_mask 			*mask;*/
	cpl_stats 			*stats;
	

		if(!hpars) {
			cpl_msg_error ("", "%s Parameters list not found", _id);
			return NULL;
		}		
	

/* Get parameters*/
	npars = cpl_parameterlist_get_size(hpars);
	par = cpl_parameterlist_get_first(hpars);
				
	for(i=0; i<npars; i++) {
		alias = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
			if(strcmp("det-sate", alias) == 0) {
    			dthre = cpl_parameter_get_double(par) ;
				j++;
			}
			else if(strcmp("hough-thre", alias) == 0) {
    			hthre = cpl_parameter_get_double(par) ;
				j++;
			}
		par = cpl_parameterlist_get_next(hpars);
			if(j == 2) break;
	}


	stats = cpl_stats_new_from_image (nscience, CPL_STATS_ALL);
	if(stats == NULL) {
		cpl_msg_warning("","%s Cannot calculate image statistics",_id);
		stdev = 1.0;
		cpl_error_reset();
	}
	else{
		stdev = cpl_stats_get_median_dev(stats);
		if(stdev == 0.0) stdev = 1.0;
		cpl_stats_delete(stats);
	}
	threshold = cpl_stats_get_median(stats) + (float)(dthre * stdev);
	//threshold = (float)(dthre * stdev);
		

	hough = hough_transform(nscience, threshold);
		if(hough == NULL) {
			cpl_msg_warning("","%s Error in Hough Transform", _id);
			cpl_image_delete(hough);
			return NULL;
		}
	threshold = (float)hthre;

	nx = cpl_image_get_size_x(nscience);
	ny = cpl_image_get_size_y(nscience);
			
	hough_map = hough_inverse_transform(hough, threshold, nx, ny);
		if(hough_map == NULL) {
			cpl_msg_warning("","%s Error in Inverse Hough Transform", _id);
			cpl_image_delete(hough);
			cpl_mask_delete(hough_map);
			return NULL;
		}

		
	count = cpl_mask_count(hough_map);
		if(count == -1) {
			cpl_msg_info("","No satellite tracks were detected in image");
		}
		else {
			cpl_msg_info("","Detected %d satellite tracks in image", count);
		}
		
/*FIXME: Don't know why but the satellites map need to have a mask_not*/
	//cpl_mask_not(hough_map);

	cpl_image_delete(hough);
	
	return hough_map; 
		
}

/* ************************************************************************
			Routines to compute Hough Transform
			Converted from Eclipse (in OMEGA pae2)
**************************************************************************/
/**
  @brief    Compute a Hough Transform on an image
  @param    img  		image
  @param    thre  		threshold for pixels to include in transform

  @return   hough	Hough image   
 
 */
/*--------------------------------------------------------------------------*/
cpl_image *hough_transform (cpl_image *img, float thre)
{
	
	int i = 0;
	int j = 0;
	int l = 0;
	int k = 0;
	int nx = 0;
	int ny = 0;
	int nrho = 0;
	int ntheta = 0;
	
	double drho = 0.0;
	double theta = 0.0;
	double dtheta = 0.0;
	double x = 0.0;
	double y = 0.0;

	float *sins;
	float *coss;
	float *hough_data;
	float *img_data;
	
	const char *_id = "hough_transform:";
			
	cpl_image 		*hough_img;



/* Get input image */	
		if (img == NULL) {
			cpl_msg_error("","<%s> No input image",_id);
			return NULL;
		}
	
/* Extract the image data */
	img_data = (float *)cpl_image_get_data(img);		

	drho = 1.0/sqrt(2.0);
	nx = cpl_image_get_size_x (img);
	ny = cpl_image_get_size_y (img);
	
	nrho = (int)(sqrt(nx*nx + ny*ny)/drho) + 1;
	ntheta = 3600;
	dtheta = 2*PI/(double)ntheta;
	
	hough_img = cpl_image_new(nrho,ntheta,CPL_TYPE_FLOAT);
		if(hough_img == NULL) {
			cpl_msg_error("","<%s>Cannot create Hough image",_id);
			return NULL;
		}
		
	hough_data = (float *)cpl_image_get_data(hough_img);

	
	for(i=0; i<nrho*ntheta; i++) {
		hough_data[i] = (float)0.0;
	}

	sins = calloc(1, ntheta*sizeof(float));
		if(sins == NULL) {
			cpl_msg_error("","%s Error in allocating memory", _id);
			cpl_image_delete(hough_img);
			return NULL;
		}
			
	coss = calloc(1, ntheta*sizeof(float));
		if(coss == NULL) {
			cpl_msg_error("","%s Error in allocating memory", _id);
			cpl_image_delete(hough_img);
			free(sins);
			return NULL;
		}
	
	for(i=0; i<ntheta; i++){
		theta = ((double)i + 0.5) * dtheta;
		sins[i] = sin(theta);
		coss[i] = cos(theta);
	}

	for(j=0; j<ny; j++){
		y = (double)j;
		for(i=0; i<nx; i++){
			if(img_data[j*nx+i] > thre){
				x = (double)i;
				for(k=0; k<ntheta; k++){
					l = (int)floor((x*coss[k] + y*sins[k]) / drho);
					hough_data[k*nrho+l] += (float)1.0;
				}
			}
		}
	}
	
	free(sins);
	free(coss);
	
	return hough_img;
}


cpl_mask *hough_inverse_transform(cpl_image *himg, float thre, int lx, int ly)
{

	int i = 0;
	int j = 0;
	int k = 0;
	int l = 0;
	int hnx = 0;
	int hny = 0;
	int nrho = 0;
	int ntheta = 0;
	
	double rho = 0.0;
	double drho = 0.0;
	double theta = 0.0;
	double dtheta = 0.0;
	double edge = 0.0;
	double a = 0.0;
	double b = 0.0;
	
	float *hdata;

	const char *_id = "hough_inverse_transform:";
	
	cpl_mask 			*pmap;
	cpl_binary 			*pdata;
	
	
		if(himg == NULL) {
			cpl_msg_error("","<%s> No input image",_id);
			return NULL;
		}

	drho = 1.0/sqrt(2.0);
	nrho = (int)(sqrt(lx*lx + ly*ly)/drho) + 1;
	ntheta = 3600;
	dtheta = 2*PI / (double)ntheta;

	hnx = cpl_image_get_size_x(himg);
	hny = cpl_image_get_size_y(himg);
	
		
		if( !((nrho == hnx) && (ntheta == hny)) ) {
			cpl_msg_error("","%s Dimensions of Hough image are incompatible",_id);
			return NULL;
		}
		
	pmap = cpl_mask_new(lx, ly);
		if(pmap == NULL) {
			cpl_msg_error("","<%s> Error creating pixelmap",_id);
			return NULL;
		}
	
	pdata = cpl_mask_get_data(pmap);
			
	edge = sqrt(2.0)/2.0;

	hdata = (float *)cpl_image_get_data(himg);
		
	for(j=0; j<ntheta; j++) {
		theta = ((double)j + 0.5) * dtheta;
		for(i=0; i<nrho; i++) {
			rho = (double)i * drho;
			if( hdata[j*nrho + i] > thre ) {
				
				if( fabs(sin(theta) > edge) ) {
					/*compute y = a*x+b */
					a = -cos(theta) / sin(theta);
					b = rho/sin(theta);
					for(l=0; l<lx; l++){
						k = floor( (double)l * a + b);
						if( (k>0) && (k<ly) ){
							pdata[k*lx+l] = CPL_BINARY_1;
						}
					}
				}
				else {
					/*compute x =a * y + b */
					a = -sin(theta) / cos(theta);
					b = rho/cos(theta);
					for(k=0; k<ly; k++){
						l = floor( (double)k * a + b);
						if( (l>=0) && (l<lx) ){
							pdata[k*lx+l] = CPL_BINARY_1;
						}
					}
				}
			}
		}
	}
	
	return pmap;
}


/**@}*/
