/* $Id: omega_stats.c,v 1.2 2012/01/31 13:26:37 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/31 13:26:37 $
 * $Revision: 1.2 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "omega_dfs.h"
#include "omega_stats.h"
#include "omega_utils.h"


/**
 * @defgroup omega_statistics  Statistics
 *
 * This module provides functions to calculate statistics
 *
 * @par Synopsis:
 *   Functions to derive statistics
 *
 * @code
 * #include "omega_dfs.h"
 * #include "omega_stats.h"
 * #include "omega_utils.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-------------------------------------------------------------------------

	VARIOUS ROUTINES TO CALCULATE STATISTICS


-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/**
  @brief	Compute statistics of an image iteratively
  @param	image_in	Input image.
  @param	threshold	The threshold rejecting outlying pixels
  @param	iter		Maximum number of iterations
  @return	1 newly allocated image.

  Compute various images statistics iteratively until reaching a
  convergency. Results are all stored into a returned cpl_stats
  object, that must be deallocated later.

 */
/*--------------------------------------------------------------------------*/
cpl_stats * omega_iter_stat(cpl_image *img, double threshold, int iter)
{

	int i = 0;
	double mean = 0.0;
	double median = 0.0;
	double stdev = 0.0;

	cpl_mask		*mask;
	cpl_image *image;
	cpl_stats		*stats = NULL;

	if(img == NULL)
		return NULL;

    image = cpl_image_duplicate(img);

	stats = cpl_stats_new_from_image(image, CPL_STATS_ALL);

	for(i=0; i<iter; i++){
		mean = cpl_stats_get_mean(stats);
		median = cpl_stats_get_median(stats);
		stdev = cpl_stats_get_stdev(stats);

		mask = cpl_mask_threshold_image_create(image,median-threshold*stdev,median+threshold*stdev);
		cpl_mask_not(mask);
		cpl_image_reject_from_mask(image, mask);
		freemask(mask);

		freestats(stats);

		stats = cpl_stats_new_from_image(image, CPL_STATS_ALL);
		if((abs(stdev/cpl_stats_get_stdev(stats)-1) <0.01)){
			break;
		}

	}

	freeimage(image);

	return stats;
}

/*-------------------------------------------------------------------------*/
/**
  @brief	Compute statistics of an image iteratively
  @param	image_in	Input image.
  @param	threshold	The threshold rejecting outlying pixels
  @param	iter		Maximum number of iterations
  @return	1 newly allocated image.

  Compute various images statistics iteratively until reaching a
  convergency. Results are all stored into a returned stats
  object, that must be deallocated.

 */
/*--------------------------------------------------------------------------*/
cpl_stats * omega_iter_stat_opts(cpl_image *img, cpl_vector *zone, double threshold, int iter)
{

    int x0 = 0;
	int zy0 = 0;
	int x1 = 0;
	int zy1 = 0;
	int i = 0;
	double mean = 0.0;
	double median = 0.0;
	double stdev = 0.0;

	cpl_mask		*mask;
	cpl_stats		*stats = NULL;
	cpl_image       *image;

	if(img == NULL)
		return NULL;

	    image = cpl_image_duplicate(img);


    if(zone != NULL){
	    x0 = (int)cpl_vector_get(zone,0);
	    zy0 = (int)cpl_vector_get(zone,1);
	    x1 = (int)cpl_vector_get(zone,2);
	    zy1 = (int)cpl_vector_get(zone,3);
		stats = cpl_stats_new_from_image_window(image, CPL_STATS_ALL,x0,zy0,x1,zy1);
    }
    else{
		stats = cpl_stats_new_from_image(image, CPL_STATS_ALL);
    }

	for(i=0; i<iter; i++){

		mean = cpl_stats_get_mean(stats);
		median = cpl_stats_get_median(stats);
		stdev = cpl_stats_get_stdev(stats);

		mask = cpl_mask_threshold_image_create(image,median-threshold*stdev,median+threshold*stdev);
		cpl_mask_not(mask);
		cpl_image_reject_from_mask(image, mask);
		freemask(mask);
		freestats(stats);

        if(zone != NULL){
		    stats = cpl_stats_new_from_image_window(image, CPL_STATS_ALL,x0,zy0,x1,zy1);
        }
		else{
		    stats = cpl_stats_new_from_image(image, CPL_STATS_ALL);
		}
		if( (abs(mean/cpl_stats_get_mean(stats)-1) < 0.01) &&
		    (abs(stdev/cpl_stats_get_stdev(stats)-1) < 0.01) ){
			break;
		}

	}

	freeimage(image);

	return stats;
}

/*-----------------------------------------------------------
		Hacked from cpl_tools
------------------------------------------------------------*/
float get_median_float(float *a, int n)
{
    return get_kth_float(a,n,(((n)&1)?((n)/2):(((n)/2)-1)));
}
/*----------------------------------------------------------------------------*/
/**
  @brief    Get the kth smallest value in a float array
  @param    a   the float array
  @param    n   the array size
  @param    k   the requested value position in the sorted array

  @return   the kth smallest value in the sorted array.
 */
/*----------------------------------------------------------------------------*/
float get_kth_float(float *a, int n, int k)
{
    const char *_id = "get_kth_float:";
	register float  x ;
    register int    i, j, l, m ;

/*    cpl_assure(a, CPL_ERROR_NULL_INPUT, "cpl_tools_get_kth_float", 0.00) ;*/

		if(a == NULL){
			cpl_msg_error("","%s Error in input for median of array",_id);
			return 0;
		}

    l=0 ; m=n-1 ;
    while (l<m) {
        x=a[k] ;
        i=l ;
        j=m ;
        do {
            while (a[i]<x) i++ ;
            while (x<a[j]) j-- ;
            if (i<=j) {
                FLOAT_SWAP(a[i],a[j]) ;
                i++ ; j-- ;
            }
        } while (i<=j) ;
        if (j<k) l=i ;
        if (k<i) m=j ;
    }

    return a[k] ;
}
/**@}*/
