/* $Id: omega_trim.c,v 1.3 2013/01/08 08:09:12 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2013/01/08 08:09:12 $
 * $Revision: 1.3 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "omega_trim.h"
#include "omega_fits.h"
#include "omega_pfits.h"
#include "omega_stats.h"
#include "omega_utils.h"

/**
 * @defgroup omega_trim  Image Bordering
 *
 * \par Synopsis
 * These are utility functions to extract an image from its pre/overscan regions, correct
 * for the overscan, get the FITS coordinates of the trim image.
 *
 * \code
 * #include omega_trim.h
 * #include omega_pfits.h
 * #include omega_stats.h
 * #include omega_fits.h
 * #include omega_utils.h
 * \endcode
 */
/**@{*/

/**
 * @brief   This method loads a raw image, trims it and (optionally)
 *          performs an overscan correction, using the statistics of the
 *          pre and overscan regions. This can be done with a single
 *          value for the whole image or on a per-row basis.
 *
 *          Methods 1-4 assume that DoStatistics() has been run.
 *
 * @param    frame  Frame to be trimmed and/or overscan corrected.
 * @param    oscan	overscan method
 * @param	extn	FITS extension number to load
 *
 * @return  a new image, extracted from the trim region
 *
 *           overscan_method -- which overscan correction to use:
 *              0 -- No overscan correction (default)
 *              1 -- Use median of prescan_x
 *              2 -- Use median of overscan_x
 *              3 -- Use median of prescan_y
 *              4 -- Use median of overscan_y
 *              5 -- Use per-row value of prescan_x
 *              6 -- Use per_rwo value of overscan_x
 *
 */
cpl_image * TrimOscanCorrect(const cpl_frame *frame, int oscan, int extn)
{

	int x0 = 1;
	int x1 = 1;
	int ry0 = 1;
	int ry1 = 1;

	const char *_id = "TrimOscanCorrect";

	cpl_image *image=NULL;
	cpl_image *trim_image=NULL;
	cpl_image *tmp_img=NULL;
	cpl_vector *region=NULL;
	cpl_stats *px = NULL; /*  *prescan_x_stat;*/
	cpl_stats *py = NULL; /*  *prescan_y_stat;*/
	cpl_stats *ox = NULL; /*  *ovrscan_x_stat;*/
	cpl_stats *oy = NULL; /*  *ovrscan_y_stat;*/

	if(frame == NULL) {
		cpl_msg_error(_id,"NULL input frame. %s",cpl_error_get_message());
		return NULL;
	}

	image = cpl_image_load(cpl_frame_get_filename(frame), CPL_TYPE_FLOAT, 0, extn);
	if(image == NULL) {
		cpl_msg_error(_id,"Cannot load input image. %s",cpl_error_get_message());
		return NULL;
	}

	if(oscan == 0){
		tmp_img = cpl_image_duplicate(image);
	}

	/* For overscan methods 1--4, do the following statistics */
	if (oscan > 0 && oscan < 5){
		/*
	      Here the global variables of pre and overscan
	      are populated
		 */

		DoStatistics(frame, image, extn, &px,&py,&ox,&oy);

		switch(oscan){
		case 1:
			if(px != NULL){
				tmp_img = cpl_image_subtract_scalar_create(image,cpl_stats_get_median(px));
			}
			else{
				tmp_img = cpl_image_duplicate(image);
				cpl_msg_warning(_id,"The overscan method %d may not suit these data",oscan);
			}
			break;

		case 2:
			if(ox != NULL){
				tmp_img = cpl_image_subtract_scalar_create(image,cpl_stats_get_median(ox));
			}
			else{
				tmp_img = cpl_image_duplicate(image);
				cpl_msg_warning(_id,"The overscan method %d may not suit these data",oscan);
			}
			break;

		case 3:
			if(py != NULL){
				tmp_img = cpl_image_subtract_scalar_create(image,cpl_stats_get_median(py));
			}
			else{
				tmp_img = cpl_image_duplicate(image);
				cpl_msg_warning(_id,"The overscan method %d may not suit these data",oscan);
			}
			break;

		case 4:
			if(oy != NULL){
				tmp_img = cpl_image_subtract_scalar_create(image,cpl_stats_get_median(oy));
			}
			else{
				tmp_img = cpl_image_duplicate(image);
				cpl_msg_warning(_id,"The overscan method %d may not suit these data",oscan);
			}
			break;
		}

	}
	else {
		cpl_propertylist *plist = cpl_propertylist_load(cpl_frame_get_filename(frame), extn);
		if(oscan == 5){
			region = omega_get_scan_coord(plist, 0);
			tmp_img = oscan_row(image, (int)cpl_vector_get(region,0),(int)cpl_vector_get(region,2));
			freevector(region);
		}
		if(oscan == 6){
			region = omega_get_scan_coord(plist, 2);
			tmp_img = oscan_row(image, (int)cpl_vector_get(region,0),(int)cpl_vector_get(region,2));
			freevector(region);
		}
		freeplist(plist);
	}

	region = get_trim_region(frame, extn);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	trim_image = cpl_image_extract(tmp_img, x0, ry0, x1, ry1);
	if (trim_image == NULL) {
		cpl_msg_error (_id,"Cannot extract trimmed image <%s>", cpl_error_get_message());
		freeimage(image);
		freeimage(tmp_img);
		freevector(region);
		freestats(px);
		freestats(py);
		freestats(ox);
		freestats(oy);
		return NULL;
	}

	freeimage(image);
	freeimage(tmp_img);
	freevector(region);
	freestats(px);
	freestats(py);
	freestats(ox);
	freestats(oy);

	return trim_image;
}

/**
 *
 * @brief   This method loads a raw image, trims it and (optionally)
 *          performs an overscan correction, using the statistics of the
 *          pre and overscan regions. This can be done with a single
 *          value for the whole image or on a per-row basis.
 *
 * @param    ofits  omega_fits with image to be trimmed and/or overscan corrected.
 * @param    oscan	overscan correction method
 *
 * @return  a new image, extracted from the trim region.
 *
 * Requires:
 * 		header keywords describing the pre/overscan layout
 *
 * 			overscan_method -- which overscan correction to use:
 *              0 -- No overscan correction (default)
 *              1 -- Use median of prescan_x
 *              2 -- Use median of overscan_x
 *              3 -- Use median of prescan_y
 *              4 -- Use median of overscan_y
 *              5 -- Use per-row value of prescan_x
 *              6 -- Use per_rwo value of overscan_x
 *
 * Definition of coordinates used in functions:
 *
 *      x0, ry0 -- lower lefthand corner
 *      x1, ry1 -- upper righthand corner
 *
 **/
cpl_image *omega_trim_oscan_correct(omega_fits *ofits, int oscan)
{

	int x0 = 1;
	int x1 = 1;
	int ry0 = 1;
	int ry1 = 1;

	cpl_image *trim_image;
	cpl_image *tmp_img;
	cpl_vector *region;
	cpl_propertylist *xlist;
	cpl_stats *px = NULL; /*  *prescan_x_stat;*/
	cpl_stats *py = NULL; /*  *prescan_y_stat;*/
	cpl_stats *ox = NULL; /*  *ovrscan_x_stat;*/
	cpl_stats *oy = NULL; /*  *ovrscan_y_stat;*/

	if(ofits == NULL) {
		cpl_msg_error(cpl_func,"NULL input file. %s",cpl_error_get_message());
		return NULL;
	}

	if(oscan == 0){
		tmp_img = cpl_image_duplicate(omega_fits_get_image(ofits));
	}

	/* For overscan methods 1--4, do the following statistics */
	if (oscan > 0 && oscan < 5){

		omega_scan_stats(ofits,&px ,&py, &ox, &oy);

		switch(oscan){
		case 1:
			if(px != NULL){
				tmp_img = cpl_image_subtract_scalar_create(omega_fits_get_image(ofits),cpl_stats_get_median(px));
			}
			else{
				tmp_img = cpl_image_duplicate(omega_fits_get_image(ofits));
				cpl_msg_warning(cpl_func,"The overscan method %d may not suit these data",oscan);
			}
			break;

		case 2:
			if(ox != NULL){
				tmp_img = cpl_image_subtract_scalar_create(omega_fits_get_image(ofits),cpl_stats_get_median(ox));
			}
			else{
				tmp_img = cpl_image_duplicate(omega_fits_get_image(ofits));
				cpl_msg_warning(cpl_func,"The overscan method %d may not suit these data",oscan);
			}
			break;

		case 3:
			if(py != NULL){
				tmp_img = cpl_image_subtract_scalar_create(omega_fits_get_image(ofits),cpl_stats_get_median(py));
			}
			else{
				tmp_img = cpl_image_duplicate(omega_fits_get_image(ofits));
				cpl_msg_warning(cpl_func,"The overscan method %d may not suit these data",oscan);
			}
			break;

		case 4:
			if(oy != NULL){
				tmp_img = cpl_image_subtract_scalar_create(omega_fits_get_image(ofits),cpl_stats_get_median(oy));
			}
			else{
				tmp_img = cpl_image_duplicate(omega_fits_get_image(ofits));
				cpl_msg_warning(cpl_func,"The overscan method %d may not suit these data",oscan);
			}
			break;
		}

	}
	else {
		xlist = omega_fits_get_ehu(ofits);
		if(oscan == 5){
			region = omega_get_scan_coord(xlist, 0);
			tmp_img = oscan_row(omega_fits_get_image(ofits), (int)cpl_vector_get(region,0),
					(int)cpl_vector_get(region,2));
		}
		if(oscan == 6){
			region = omega_get_scan_coord(xlist, 2);
			tmp_img = oscan_row(omega_fits_get_image(ofits), (int)cpl_vector_get(region,0),
					(int)cpl_vector_get(region,2));
		}
		xlist = NULL;
	}

	region = omega_get_trim_region(ofits);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	trim_image = cpl_image_extract(tmp_img, x0, ry0, x1, ry1);
	if (trim_image == NULL) {
		cpl_msg_error (cpl_func,"Cannot extract trimmed image <%s>", cpl_error_get_message());
		freeimage(tmp_img);
		freevector(region);
		freestats(px);
		freestats(py);
		freestats(ox);
		freestats(oy);
		return NULL;
	}

	freeimage(tmp_img);
	freevector(region);
	freestats(px);
	freestats(py);
	freestats(ox);
	freestats(oy);

	return trim_image;
}


/*-------------------------------------------------------------------------*/
/**
@brief    subtract median of overscan from each row
@param    img           Image.
@param    x1            first column of overscan region
@param    x2            last column of overscan region
@return
  New image, subtracted from the overscan
*/
/*--------------------------------------------------------------------------*/
cpl_image *oscan_row (cpl_image *img, int x1, int x2)
{

	int i = 0;
	int j = 0;
	int lx = 256;
	int ly = 256;
	int offset = 0;
	int oscanoffset = 0;
	int oscansize = 0;
	float med = 0.0f;
	/*	float * parr;*/
	float * pindata;
	float * poutdata;

	cpl_image 	*image_out;

	/*Start algorithm */
	lx = cpl_image_get_size_x (img);
	ly = cpl_image_get_size_y (img);

	image_out = cpl_image_new (lx, ly, CPL_TYPE_FLOAT);

	oscansize = x2-x1+1;
	/*	parr =  cpl_calloc(oscansize, sizeof(float));*/

	pindata = (float *)cpl_image_get_data (img);
	poutdata = (float *)cpl_image_get_data (image_out);

	for (i=0; i<ly; i++)
	{
		offset = i * lx;
		oscanoffset = offset+x1-1;
		med = 0.0;

		for (j=0; j<oscansize; j++)
		{
			med += pindata[oscanoffset+j];
		}

		/* med = median_pixelvalue(parr, oscansize); */
		med /= oscansize;

		for (j=0; j<lx; j++){
			poutdata[offset+j] = pindata[offset+j]-med;
		}
	}

	/*  cpl_free(parr);*/

	return image_out;
}


/**
@brief    	Get trim regions based on orientation keyword

@param    frame		cpl_frame with header keywords describing the pre/overscan layout
@param	extnum		FITS extension number to load
@return
  A cpl_vector with the four coordinates as defined below:

vector --> (x0, ry0, x1, ry1):
x0, ry0 -- lower lefthand corner
x1, ry1 -- upper righthand corner

*/
cpl_vector * get_trim_region(const cpl_frame *frame, int extnum)
{

	int orient = 0;
	int naxis1 = 0;
	int naxis2 = 0;
	int ovscx = 0;
	int ovscy = 0;
	int prscx = 0;
	int prscy = 0;
	int x0 = 1;
	int x1 = 1;
	int ry0 = 1;
	int ry1 = 1;

	const char *filename = NULL;

	cpl_propertylist 	*xplist;
	cpl_vector 		*region;
	cpl_vector 		*sregion;


	if(frame == NULL){
		cpl_msg_error(cpl_func,"Input frame is NULL");
		return NULL;
	}

	filename = cpl_frame_get_filename(frame);

	/*Get extension header*/
	xplist = cpl_propertylist_load(filename, extnum);
	if(xplist == NULL) {
		cpl_msg_error(cpl_func,"Cannot load header");
		return NULL;
	}

	naxis1 = cpl_propertylist_get_int(xplist, "NAXIS1");
	naxis2 = cpl_propertylist_get_int(xplist, "NAXIS2");
	if (naxis1 == 0 || naxis2 == 0){
		cpl_msg_error(cpl_func,"NAXIS1 or NAXIS2 is 0");
		freeplist(xplist);
		return NULL;
	}

	/*Get pre and overscan regions from header*/
	sregion = omega_pfits_get_preovscan(xplist);
	if(sregion == NULL){
		cpl_msg_warning(cpl_func,"Cannot get pre and overscan regions");
		freeplist(xplist);
		return NULL;
	}

	prscx = cpl_vector_get(sregion,0);
	prscy = cpl_vector_get(sregion,1);
	ovscx = cpl_vector_get(sregion,2);
	ovscy = cpl_vector_get(sregion,3);

	freeplist(xplist);

	/* Get trim coordinates */
	/* Switch by orientation of CCD layout */
	orient = omega_get_chip_orientation(extnum);

	switch(orient){
	case 1:
		x0 = prscx + 1;
		ry0 = prscy + 1;
		x1 = naxis1 - ovscx;
		ry1 = naxis2 - ovscy;
		break;
	case 2:
		x0 = ovscx + 1;
		ry0 = ovscy + 1;
		x1 = naxis1 - prscx;
		ry1 = naxis2 - prscy;
		break;
	default:
		x0 = prscx + 1;
		ry0 = prscy + 1;
		x1 = naxis1 - ovscx;
		ry1 = naxis2 - ovscy;
		cpl_msg_warning(cpl_func,"Using default chip layout");
		break;
	}


	/* Original: Get trim coordinates based on orientation keyword */
	/*    switch (chipornt) {
    case 1:
    	x0 = prscx + 1;
    	ry0 = prscy + 1;
    	x1 = naxis1 - ovscx;
    	ry1 = naxis2 - ovscy;
    	break;

    case 2:
    	x0 = prscx + 1;
    	ry0 = ovscy + 1;
    	x1 = naxis1 - ovscx;
    	ry1 = naxis2 - prscy;
    	break;

    case 3:
    	x0 = ovscx + 1;
    	ry0 = prscy+ 1;
    	x1 = naxis1 - prscx;
    	ry1 = naxis2 - ovscy;
    	break;

    case 4:
    	x0 = ovscx + 1;
    	ry0 = ovscy + 1;
    	x1 = naxis1 - prscx;
    	ry1 = naxis2 - prscy;
    	break;

    default:
    	x0 = prscx + 1;
    	ry0 = prscy + 1;
    	x1 = naxis1 - ovscx;
    	ry1 = naxis2 - ovscy;
    	break;
    }*/

	region = cpl_vector_new(4);
	cpl_vector_set(region, 0, x0);
	cpl_vector_set(region, 1, ry0);
	cpl_vector_set(region, 2, x1);
	cpl_vector_set(region, 3, ry1);

	freevector(sregion);

	return region;

}

/**
 * @brief       Get trim regions based on CCD layout
 *
 * @param ofits	 omega_fits with header keywords describing the
				 pre/overscan layout

 * @return  Vector with the four coordinates as defined below:
 *  		vector --> (x0, ry0, x1, ry1):
 *  		x0, ry0 -- lower left-hand corner
 *  		x1, ry1 -- upper right-hand corner
 *
 *  Note: Property list xplist must not be deallocated as it belongs to the ofits
 *  structure
 */
cpl_vector * omega_get_trim_region(omega_fits *ofits)
{

	int orient = 0;
	int naxis1 = 0;
	int naxis2 = 0;
	int ovscx = 0;
	int ovscy = 0;
	int prscx = 0;
	int prscy = 0;
	int x0 = 1;
	int x1 = 1;
	int ry0 = 1;
	int ry1 = 1;

	cpl_vector 		*region;
	cpl_vector 		*sregion;
	cpl_propertylist *xlist;


	if(ofits == NULL){
		cpl_msg_error(cpl_func,"Input frame is NULL");
		return NULL;
	}

	xlist = omega_fits_get_ehu(ofits);
	naxis1 = cpl_propertylist_get_int(xlist, "NAXIS1");
	naxis2 = cpl_propertylist_get_int(xlist, "NAXIS2");
	if (naxis1 == 0 || naxis2 == 0){
		cpl_msg_error(cpl_func,"NAXIS1 or NAXIS2 is 0");
		xlist = NULL;
		return NULL;
	}

	/*Get pre and overscan regions from header*/
	sregion = omega_pfits_get_preovscan(xlist);
	if(sregion == NULL){
		cpl_msg_warning(cpl_func,"Cannot get pre and overscan regions");
		return NULL;
	}

	prscx = cpl_vector_get(sregion,0);
	prscy = cpl_vector_get(sregion,1);
	ovscx = cpl_vector_get(sregion,2);
	ovscy = cpl_vector_get(sregion,3);

	/* Get trim coordinates */
	/* Switch by orientation of CCD layout */
	orient = omega_get_chip_orientation(omega_fits_get_extnum(ofits));
	switch(orient){
	case 1:
		x0 = prscx + 1;
		ry0 = prscy + 1;
		x1 = naxis1 - ovscx;
		ry1 = naxis2 - ovscy;
		break;
	case 2:
		x0 = ovscx + 1;
		ry0 = ovscy + 1;
		x1 = naxis1 - prscx;
		ry1 = naxis2 - prscy;
		break;
	default:
		x0 = prscx + 1;
		ry0 = prscy + 1;
		x1 = naxis1 - ovscx;
		ry1 = naxis2 - ovscy;
		cpl_msg_warning(cpl_func,"Using default chip layout");
		break;
	}

	region = cpl_vector_new(4);
	cpl_vector_set(region, 0, x0);
	cpl_vector_set(region, 1, ry0);
	cpl_vector_set(region, 2, x1);
	cpl_vector_set(region, 3, ry1);

	freevector(sregion);
	xlist = NULL;

	return region;

}

/**
* @brief   This function performs several statistics on the image and
*          populates the 4 cpl_stats with CPL_STATS_ALL from different
*		   pre and over scan regions.
*
* @param    frame      	Frame
* @param    image       Image related to frame.
* @param	extn	    FITS extension number to load
* @param    px          Pointer to a cpl_stats of prescan_x
* @param    py          Pointer to a cpl_stats of prescan_y
* @param    ox          Pointer to a cpl_stats of overscan_x
* @param    oy          Pointer to a cpl_stats of overscan_y
*
* @return  nothing.
**/
void DoStatistics (const cpl_frame *frame, cpl_image *image, int extnum,
		cpl_stats **px,cpl_stats **py,cpl_stats **ox,cpl_stats **oy)
{
	int x0 = 0;
	int ry0 = 0;
	int x1 = 0;
	int ry1 = 0;
	int iscan = 0;
	double threshold = 1.0;  /*FIXME: pass this as a parameter*/
	int iter = 5;           /*FIXME: pass this as a parameter*/
	cpl_vector    *region;
	cpl_propertylist *plist;


	if((plist = cpl_propertylist_load(cpl_frame_get_filename(frame), extnum)) == NULL)
		return;

	/*
	   Returns a vector of type x0,ry0,x1,ry1
	   which are the coordinates of the trim section
	 */
	/*
	 * Get pre/overscan regions based on iscan index:
	 * 0 = Prescan X region
	 * 1 = Prescan Y region
	 * 2 = Overscan X region
	 * 3 = Overscan Y region
	 */
	iscan = 0; /*PRSCX*/
	region = omega_get_scan_coord(plist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	/*    region = GetPrescanXRegion(frame, extnum);
    x0 = (int)cpl_vector_get(region,0);
    ry0 = (int)cpl_vector_get(region,1);
    x1 = (int)cpl_vector_get(region,2);
    ry1 = (int)cpl_vector_get(region,3);*/

	if (x1 - x0 +1 > 0){
		*px = omega_iter_stat_opts(image,region, threshold, iter);
	}
	freevector(region);

	iscan = 1; /*PRSCY*/
	region = omega_get_scan_coord(plist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	/*    region = GetPrescanYRegion(frame, extnum);
    x0 = (int)cpl_vector_get(region,0);
    ry0 = (int)cpl_vector_get(region,1);
    x1 = (int)cpl_vector_get(region,2);
    ry1 = (int)cpl_vector_get(region,3);*/

	if (ry1 - ry0 +1 > 0){
		*py = omega_iter_stat_opts(image,region,threshold, iter);
	}
	freevector(region);

	iscan = 2; /*OVSCX*/
	region = omega_get_scan_coord(plist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	/*    region = GetOverscanXRegion(frame, extnum);
    x0 = (int)cpl_vector_get(region,0);
    ry0 = (int)cpl_vector_get(region,1);
    x1 = (int)cpl_vector_get(region,2);
    ry1 = (int)cpl_vector_get(region,3);*/

	if (x1 - x0 +1 > 0){
		*ox = omega_iter_stat_opts(image,region,threshold, iter);
	}
	freevector(region);

	iscan = 3; /*OVSCY*/
	region = omega_get_scan_coord(plist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	/*    region = GetOverscanYRegion(frame, extnum);
    x0 = (int)cpl_vector_get(region,0);
    ry0 = (int)cpl_vector_get(region,1);
    x1 = (int)cpl_vector_get(region,2);
    ry1 = (int)cpl_vector_get(region,3);*/

	if (ry1 - ry0 +1 > 0){
		*oy = omega_iter_stat_opts(image,region,threshold, iter);
	}

	freevector(region);
	freeplist(plist);

	return;

}

/**
* @brief   This function performs several statistics on the image and
*          populates the 4 cpl_stats with CPL_STATS_ALL from different
*		   pre and over scan regions.
*
* @param    ofits       omega_fits structure
* @param    px          Pointer to a cpl_stats of prescan_x
* @param    py          Pointer to a cpl_stats of prescan_y
* @param    ox          Pointer to a cpl_stats of overscan_x
* @param    oy          Pointer to a cpl_stats of overscan_y
*
* @return  nothing.
**/
void omega_scan_stats (omega_fits *ofits,cpl_stats **px,cpl_stats **py,
		cpl_stats **ox,cpl_stats **oy)
{
	int x0 = 0;
	int ry0 = 0;
	int x1 = 0;
	int ry1 = 0;
	int iscan = 0;
	double threshold = 1.0;  /*FIXME: pass this as a parameter*/
	int iter = 5;           /*FIXME: pass this as a parameter*/
	cpl_vector    *region;
	cpl_propertylist *xlist;

	if(ofits == NULL)
		return;

	xlist = omega_fits_get_ehu(ofits);

	/*
	   Returns a vector of type x0,ry0,x1,ry1
	   which are the coordinates of the trim section
	 */
	/*
	 * Get pre/overscan regions based on iscan index:
	 * 0 = Prescan X region
	 * 1 = Prescan Y region
	 * 2 = Overscan X region
	 * 3 = Overscan Y region
	 */
	iscan = 0; /*PRSCX*/
	region = omega_get_scan_coord(xlist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);


	if (x1 - x0 +1 > 0){
		*px = omega_iter_stat_opts(omega_fits_get_image(ofits),region, threshold, iter);
	}
	freevector(region);

	iscan = 1; /*PRSCY*/
	region = omega_get_scan_coord(xlist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	if (ry1 - ry0 +1 > 0){
		*py = omega_iter_stat_opts(omega_fits_get_image(ofits),region,threshold, iter);
	}
	freevector(region);

	iscan = 2; /*OVSCX*/
	region = omega_get_scan_coord(xlist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	if (x1 - x0 +1 > 0){
		*ox = omega_iter_stat_opts(omega_fits_get_image(ofits),region,threshold, iter);
	}
	freevector(region);

	iscan = 3; /*OVSCY*/
	region = omega_get_scan_coord(xlist, iscan);
	x0 = (int)cpl_vector_get(region,0);
	ry0 = (int)cpl_vector_get(region,1);
	x1 = (int)cpl_vector_get(region,2);
	ry1 = (int)cpl_vector_get(region,3);

	if (ry1 - ry0 +1 > 0){
		*oy = omega_iter_stat_opts(omega_fits_get_image(ofits),region,threshold, iter);
	}

	freevector(region);
	xlist = NULL;

	return;

}

/**
@brief   This method calculates the FITS coordinates of the pre/overscan
region in the iscan direction.

@param    plist     Property list to read from
@param	iscan     Index translating which keywords to read from header

@return  cpl_vector with the FITS coordinates in the following order
x0 ry0, x1, ry1. The iscan index is: 0 = Prescan X region;
1 = Prescan Y region; 2 = Overscan X region; 3 = Overscan Y region.
*/

/**
 * @brief
 *    This method calculates the FITS coordinates of the pre/overscan region in the iscan direction.
 *
 * @param plist   Property list to read from
 * @param iscan   Index translating which keywords to read from header
 * @return
 *   A cpl_vector with the FITS coordinates in the following order:
 *
 *   x0 ry0, x1, ry1. The iscan index is: 0 = Prescan X region;
 *   1 = Prescan Y region; 2 = Overscan X region; 3 = Overscan Y region.
 */
cpl_vector * omega_get_scan_coord(const cpl_propertylist *plist, int iscan)
{

	int naxis1 = 0;
	int naxis2 = 0;
	int ovscx = 0;
	int ovscy = 0;
	int prscx = 0;
	int prscy = 0;
	int x0 = 0;
	int x1 = 0;
	int ry0 = 0;
	int ry1 = 0;
	const char * chipid;
	int chipornt=0;

	cpl_vector *region;
	cpl_vector *sregion;


	if(plist == NULL)
		return NULL;

	naxis1 = cpl_propertylist_get_int(plist, "NAXIS1");
	naxis2 = cpl_propertylist_get_int(plist, "NAXIS2");
	if (naxis1 == 0 || naxis2 == 0){
		cpl_msg_error(cpl_func,"NAXIS1 or NAXIS2 is 0");
		return NULL;
	}

	chipid = omega_pfits_get_chipid(plist);
	chipornt = omega_pfits_get_orientation(chipid);

	if(chipornt == 0) chipornt = 1;

	/*Get pre and overscan regions from header*/
	sregion = omega_pfits_get_preovscan(plist);

	prscx = (int)cpl_vector_get(sregion,0);
	prscy = (int)cpl_vector_get(sregion,1);
	ovscx = (int)cpl_vector_get(sregion,2);
	ovscy = (int)cpl_vector_get(sregion,3);



	if(chipornt==1)
	{
		switch(iscan){
		case 0:
			/* FITS coordinates of prescan in X direction */
			x0 = 1;
			ry0 = prscy+1;
			x1 = prscx;
			ry1 = naxis2 - ovscy;
			cpl_msg_debug(cpl_func,"px:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		case 1:
			/* FITS coordinates of prescan in Y direction */
			x0 = prscx+1;
			ry0 = 1;
			x1 = naxis1-ovscx;
			ry1 = prscy;
			cpl_msg_debug(cpl_func,"py:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		case 2:
			/* FITS coordinates of overscan in X direction */
			x0 = naxis1-ovscx+1;
			ry0 = prscy+1;
			x1 = naxis1;
			ry1 = naxis2 - ovscy;
			cpl_msg_debug(cpl_func,"ox:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		case 3:
			/* FITS coordinates of overscan in Y direction */
			x0 = prscx+1;
			ry0 = naxis2-ovscy+1;
			x1 = naxis1-ovscx;
			ry1 = naxis2;
			cpl_msg_debug(cpl_func,"oy:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		default:
			cpl_msg_error(cpl_func,"Unable to get pre/overscan coordinates");
			freevector(sregion);
			return NULL;

		}
	}
	if(chipornt==2)
	{
		switch(iscan){
		case 0:
			/* FITS coordinates of prescan in X direction */
			x0 = naxis1 - prscx + 1;
			ry0 = ovscy+1;
			x1 = naxis1;
			ry1 = naxis2-prscy;
			cpl_msg_debug(cpl_func,"px:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		case 1:
			/* FITS coordinates of prescan in Y direction */
			x0 = ovscx+1;
			ry0 = naxis2-prscy;
			x1 = naxis1-prscx;
			ry1 = naxis2;
			cpl_msg_debug(cpl_func,"py:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		case 2:
			/* FITS coordinates of overscan in X direction */
			x0 = 1;
			ry0 = ovscy+1;
			x1 = ovscx;
			ry1 = naxis2 - prscy;
			cpl_msg_debug(cpl_func,"ox:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		case 3:
			/* FITS coordinates of overscan in Y direction */
			x0 = ovscx+1;
			ry0 = 1;
			x1 = naxis1-prscx;
			ry1 = ovscy;
			cpl_msg_debug(cpl_func,"oy:  x0:%5d ry0:%5d x1:%5d ry1:%5d", x0,ry0,x1,ry1 );
			break;
		default:
			cpl_msg_error(cpl_func,"Unable to get pre/overscan coordinates");
			freevector(sregion);
			return NULL;

		}
	}

	region = cpl_vector_new(4);
	cpl_vector_set(region, 0, x0);
	cpl_vector_set(region, 1, ry0);
	cpl_vector_set(region, 2, x1);
	cpl_vector_set(region, 3, ry1);


	freevector(sregion);

	return region;

}

/**@}*/
