/* $Id: omega_utils.c,v 1.5 2012/01/12 12:05:09 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/12 12:05:09 $
 * $Revision: 1.5 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include "omega_utils.h"
#include "omega_dfs.h"
#include "omega_pfits.h"

#define MAXNAMESZ       4096

/**
 * @defgroup utils  Utility Functions
 *
 * \par Synopsis
 *  Several utility functions to manipulate images or headers.
 *
 *  \code
 *  #include omega_utils.h
 *  #include omega_dfs.h
 *  \endcode
 */

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief	Get the pipeline copyright and license
  @return   The copyright and license string

  The function returns a pointer to the statically allocated license string.
  This string should not be modified using the returned pointer.
 */
/*----------------------------------------------------------------------------*/
const char * omega_get_license(void)
{
    const char  *   omega_license =
        "This file is part of the OMEGA Instrument Pipeline\n"
        "Copyright (C) 2002,2003 European Southern Observatory\n"
        "\n"
        "This program is free software; you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation; either version 2 of the License, or\n"
        "(at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program; if not, write to the Free Software\n"
        "Foundation, Inc., 59 Temple Place, Suite 330, Boston, \n"
        "MA  02111-1307  USA" ;
    return omega_license ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Open a new PAF file, output a default header.
  @param    filename    Name of the file to create.
  @param    paf_id      PAF identificator.
  @param    paf_desc    PAF description.
  @return   Opened file pointer.

  This function creates a new PAF file with the requested file name.
  If another file already exists with the same name, it will be overwritten
  (if the file access rights allow it).

  This function returns an opened file pointer, ready to receive more data
  through fprintf's. The caller is responsible for fclose()ing the file.
 */
/*----------------------------------------------------------------------------*/
FILE * oc_paf_print_header(
        char    *   filename,
        char    *   paf_id,
        char    *   paf_desc)
{
    FILE * paf ;

    if ((paf=fopen(filename, "w"))==NULL) return NULL ;
    fprintf(paf, "PAF.HDR.START                  ;# start of header\n");
    fprintf(paf, "PAF.TYPE                      \"pipeline product\" ;\n");
    fprintf(paf, "PAF.ID                        \"%s\"\n", paf_id);
    fprintf(paf, "PAF.NAME                      \"%s\"\n", filename);
    fprintf(paf, "PAF.DESC                      \"%s\"\n", paf_desc);
    fprintf(paf, "PAF.CHCK.CHECKSUM             \"\"\n");
    fprintf(paf, "PAF.HDR.END                    ;# end of header\n");
    fprintf(paf, "\n");
    return paf ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the root part of a basename (name without extension).
  @param    filename    File name to scan.
  @return   Pointer to statically allocated string.

  Find out the root part of a file name, i.e. the file name without extension.
  Since in Unix a file name can have several dots, only a number of extensions
  are supported. This includes:

  - .fits and .FITS
  - .tfits and .TFITS
  - .paf and .PAF
  - .ascii and .ASCII
  - .dat and .DAT
  - .txt and .TXT

  This function does not check for the existence of the path or the file.

  Examples:
  @verbatim
    omega_get_root_name("/cdrom/filename.fits") returns "/cdrom/filename"
    omega_get_root_name("filename.paf") returns "filename"
    omega_get_root_name("filename") returns "filename"
    omega_get_root_name("filename.ext") returns "filename.ext"
  @endverbatim

  Since the returned string is statically allocated in this module, do not try
  to free it or modify its contents.

  Note: hacked from qfits.
 */
/*----------------------------------------------------------------------------*/
char * omega_get_root_name(const char * filename)
{
    static char path[MAXNAMESZ+1];
    char * lastdot ;

    if (strlen(filename)>MAXNAMESZ) return NULL ;
    memset(path, MAXNAMESZ, 0);
    strcpy(path, filename);
    lastdot = strrchr(path, '.');
    if (lastdot == NULL) return path ;
    if ((!strcmp(lastdot, ".fits")) || (!strcmp(lastdot, ".FITS")) ||
        (!strcmp(lastdot, ".paf")) || (!strcmp(lastdot, ".PAF")) ||
        (!strcmp(lastdot, ".dat")) || (!strcmp(lastdot, ".DAT")) ||
        (!strcmp(lastdot, ".txt")) || (!strcmp(lastdot, ".TXT")) ||
        (!strcmp(lastdot, ".tfits")) || (!strcmp(lastdot, ".TFITS")) ||
        (!strcmp(lastdot, ".ascii")) || (!strcmp(lastdot, ".ASCII")))
    {
        lastdot[0] = (char)0;
    }
    return path ;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_compare_tags
    \par Purpose:
        Comparison function to identify different input frames
    \par Description:
        The tags for two frames are compared to see whether they are
	the same. An error occurs if either frame is missing its tag.
    \par Language:
        C
    \param frame1
        The first frame
    \param frame2
        The second frame
    \retval 0
        if frame1 != frame2
    \retval 1
        if frame1 == frame2
    \retval -1
        if either frame is missing its tag
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int omega_compare_tags(const cpl_frame *frame1, const cpl_frame *frame2) {
    char *v1,*v2;

    /* Test entries */

    if (frame1 == NULL || frame2 == NULL)
	return(-1);

    /* Get the tags */

    if ((v1 = (char *)cpl_frame_get_tag(frame1)) == NULL)
	return(-1);
    if ((v2 = (char *)cpl_frame_get_tag(frame2)) == NULL)
	return(-1);

    /* Compare the tags */

    if (strcmp(v1,v2))
	return(0);
    else
	return(1);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_frameset_subgroup
    \par Purpose:
        Extract a frameset of a given label from a master frameset
    \par Description:
        For each label in an input frameset a frameset with that label
        is extracted. The tag for the first frame in the extracted set is
        compared to an input tag. If they match, then the extracted frameset
        is returned. If none of the frames match the input tag then a NULL
        frameset is returned
    \par Language:
        C
    \param frameset
        The input frameset
    \param labels
        The labels for the input frameset
    \param nlab
        The number of labels for the input frameset
    \param tag
        The tag for the frames in the output frameset
    \return The set of all frames matching input tag or NULL if none match
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern cpl_frameset *omega_frameset_subgroup(cpl_frameset *frameset,
					      cpl_size *labels, cpl_size nlab,
					      const char *tag) {
    cpl_size i;
    cpl_frameset *cur_set,*ret_set;
    cpl_frame *cur_frame;
    char *cur_tag;

    ret_set = NULL;
    for (i = 0; i < nlab; i++) {
	cur_set = cpl_frameset_extract(frameset,labels,i);
	if (cur_set == NULL)
	    break;
	cur_frame = cpl_frameset_get_frame(cur_set,0);
	cur_tag = (char *)cpl_frame_get_tag(cur_frame);
        if (!strcmp(cur_tag,tag)) {
	    ret_set = cur_set;
	    break;
	}
	cpl_frameset_delete(cur_set);
    }
    return(ret_set);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_exten_range
    \par Purpose:
        Define the range of image extensions to use
    \par Description:
        If the input extension specified is -1, then all the extensions
        in a OmegaCAM MEF should be done. If not, then just the extension
        specified should be done. 0 means that there are no extensions.
    \par Language:
        C
    \param inexten
        Input image extension from the recipe invocation
    \param out1
        Lower part of extension range
    \param out2
        Upper part of extension range
    \return Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void omega_exten_range(int inexten, int *out1, int *out2) {

	if(inexten == -1){
		*out1 = 1;
		*out2 = 32;
	}
	else if (inexten == 0) {
		*out1 = 0;
		*out2 = 0;
	} else {
		*out1 = inexten;
		*out2 = inexten;
	}

    return;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_extensions
    \par Purpose:
        Define the range of image extensions to use
    \par Description:
        If the input extension specified is -1, then all the extensions
        in a OmegaCAM MEF should be done. If not, then just the extension
        specified should be done. 0 means that there are no extensions.
    \par Language:
        C
    \param frame
        Frame
    \param inexten
        Input image extension from the recipe invocation
    \param out1
        Lower part of extension range
    \param out2
        Upper part of extension range
    \return Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void omega_extensions(const cpl_frame *frame, int inexten, int *out1, int *out2)
{

	if(inexten == -1){
		/* All extensions */
		*out1 = 1;
//		*out2 = 32;
		*out2 = cpl_fits_count_extensions(cpl_frame_get_filename(frame));
	}
	else if (inexten == 0) {
		/* Unsupported */
		*out1 = 0;
		*out2 = 0;
	} else {
		/* One specific extension */
		*out1 = inexten;
		*out2 = inexten;
	}

    return;
}


/*---------------------------------------------------------------------------*/
/**
    \par Name:
        omega_reference_compare
    \par Purpose:
        Compare master frame with reference frame
    \par Description:
         Takes the difference between a master frame and a
         reference frame and calculates statistics of it.
    \par Language:
        C
    \param master
        Input master image
    \param refframe
        Reference frame
    \param diffstats
        statistics of the difference
    \return 0 if OK, -1 in error case.
    \author
        Sandra Castro
 */
/*---------------------------------------------------------------------------*/
int omega_compare_reference(const cpl_image *master, const cpl_frame *refframe,
                           int ext, cpl_stats **diffstats)
{

    cpl_image *refimage = NULL;
    cpl_image *diff = NULL;

    if ((master == NULL) || (refframe == NULL)){
        cpl_msg_warning(cpl_func,"NULL input");
        return -1;
    }

    refimage = cpl_image_load(cpl_frame_get_filename(refframe),CPL_TYPE_FLOAT,0, ext);
    if (refimage == NULL){
        cpl_msg_warning(cpl_func,"Cannot load reference image");
        return -1;
    }

    diff = cpl_image_subtract_create(master, refimage);
    *diffstats = cpl_stats_new_from_image(diff,CPL_STATS_ALL);
    if (*diffstats == NULL){
        freeimage(diff);
        freeimage(refimage);
        cpl_msg_warning(cpl_func,"Cannot calculate statistics of difference image");
        return -1;
    }


    return 0;
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Find the directory name in the given string.
  @param    filename    Full path name to scan.
  @return   Pointer to statically allocated string.

  Provide a full path name and you get in return a pointer to a statically
  allocated string containing the name of the directory only, without trailing
  slash. If the input string does not contain a slash (i.e. it is not a full
  path), the returned string is '.', corresponding to the current working
  directory. Since the returned string is statically allocated, do not try to
  free it or modify it.

  This function does not check for the existence of the path or the file.

  Examples:
  @verbatim
    omega_get_pathname("/cdrom/data/image.fits") returns "/cdrom/data"
    omega_get_pathname("filename.fits") returns "."
  @endverbatim
 */

char * omega_get_pathname(const char * filename)
{

    static char path[MAXNAMESZ];
    char *last_slash;

    if (strlen(filename)>MAXNAMESZ) return NULL ;
    strcpy(path, filename);
    /* Find last '/'.  */
    last_slash = path != NULL ? strrchr (path, '/') : NULL;

    if (last_slash == path)
    /* The last slash is the first character in the string.  We have to
    return "/".  */
        ++last_slash;
    else if (last_slash != NULL && last_slash[1] == '\0')
        /* The '/' is the last character, we have to look further.  */
        last_slash = memchr (path, last_slash - path, '/');

    if (last_slash != NULL)
        /* Terminate the path.  */
        last_slash[0] = '\0';
    else
        /* This assignment is ill-designed but the XPG specs require to
        return a string containing "." in any case no directory part is
        found and so a static and constant string is required.  */
        strcpy(path, ".");

    return path;


}

/*---------------------------------------------------------------------------*/
/**
  @brief    Setup a frame to save a product
  @param    filename    Name of product
  @params	tag		PRO CATG tag of product
  @params	type		Type of product (image or table)
  @return   Pointer to product frame.
 */
/*---------------------------------------------------------------------------*/
cpl_frame *omega_product_frame(const char *filename, const char *tag, cpl_frame_type type)
{
	cpl_frame *frame;

	if ((filename == NULL) || (tag == NULL)){
		return NULL;
	}

	frame = cpl_frame_new();
    cpl_frame_set_filename(frame, filename);
    cpl_frame_set_tag(frame, tag);
    cpl_frame_set_type(frame, type);
    cpl_frame_set_group(frame, CPL_FRAME_GROUP_PRODUCT);
    cpl_frame_set_level(frame, CPL_FRAME_LEVEL_FINAL);

    return frame;

}

/*---------------------------------------------------------------------------*/
/**
  @brief    Save a dummy image or table in an extension
  @param	set		frameset
  @param	parlist	Parameter list
  @param    type    Type of product
  @param	bitpix	BITPIX
  @param	outfile	name of output file
  @param	plist	extension property list
  @param    recipe  recipe name
  @param	product_frame	product frame

  @return   0 in success of -1 in case of error.
 */
/*---------------------------------------------------------------------------*/
int omega_save_dummy(cpl_frameset *set, cpl_parameterlist *parlist, cpl_frame_type type,
                           cpl_type_bpp bitpix, const char *outfile, const cpl_propertylist *plist,
                           const char *recipe, cpl_frame *product_frame)
{
    int naxis1 = 2046;
    int naxis2 = 4098;
    const char delkeys[] = "^(ORIGIN|TELESCOPE|INSTRUME|OBJECT|RA|DEC|EPOCH|EQUINOX|RADECSYS|DATE-OBS|" \
                           "MJD-OBS|UTC|LST|PI-COI|OBSERVER|PIPEFILE)$";
    cpl_image *image = NULL;
    cpl_table *table = NULL;
    cpl_propertylist *list = NULL;


    /* Setup the extension */
    /* Add DataFlow keywords in extension header */
    list = cpl_propertylist_duplicate(plist);

    if (cpl_dfs_setup_product_header(list, product_frame, set, parlist,
                                         recipe,PIPEID,DICID,NULL) != CPL_ERROR_NONE) {

        cpl_msg_warning(cpl_func, "Problem in the extension header of product DFS-compliance") ;
    }

    /*Remove undesired keywords from extension header*/
    cpl_propertylist_erase_regexp(list, delkeys, 0);

     /* Save the product */
    if(type == CPL_FRAME_TYPE_IMAGE){
    	image = cpl_image_new(naxis1, naxis2, CPL_TYPE_FLOAT);
        if (cpl_image_save(image,outfile,bitpix,list,
                           CPL_IO_EXTEND) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func,"Cannot save product %s", cpl_error_get_message());
            freeimage(image);
            freeplist(list);
            return -1;
        }
        freeimage(image);
    }
    else {
    	table = cpl_table_new(1);
        if (cpl_table_save(table, NULL, list,outfile, CPL_IO_EXTEND) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot save the extension table");
            freetable(table);
            freeplist(list);
            return -1;
        }
        freetable(table);
   }

    freeplist(list);

    return 0;
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Save the primary header of a product
  @param	set		      frameset
  @param	parlist	      Parameter list
  @param	outfile	      name of output file
  @param    recipe        recipe name
  @param	plist	      primary header or NULL
  @param	product_frame product frame
  @param    inherit       Frame to inherit header from or NULL

  @return   0 in success of -1 in case of error.
 */
/*---------------------------------------------------------------------------*/
int omega_save_primary(cpl_frameset *set, const cpl_parameterlist *parlist,
                     const char *outfile, const char *recipe, const cpl_propertylist *plist,
                     cpl_frame *product_frame, const cpl_frame *inherit)
{

	cpl_propertylist *pplist;

	/* Create an empty primary header */
	if(plist == NULL)
		pplist = cpl_propertylist_new();
	else
		pplist = cpl_propertylist_duplicate(plist);

	/* Add DataFlow keywords in primary header */
	if (cpl_dfs_setup_product_header(pplist, product_frame, set, parlist,
			recipe,PIPEID,DICID,inherit) != CPL_ERROR_NONE) {

		cpl_msg_debug(cpl_func, "Problem in the main header of product DFS-compliance") ;
	}

	/*Remove undesired keywords from main header*/
    /* Once CPL recalculates these values, the following
     * line should be removed
     */
    cpl_propertylist_erase_regexp(pplist,REM_PRIM_KEYS,0);

	/* Save the empty primary unit */
	if (cpl_propertylist_save(pplist,outfile,CPL_IO_DEFAULT) != CPL_ERROR_NONE){
		freeplist(pplist);
		return -1;
	}

	freeplist(pplist);
	cpl_frameset_insert(set,product_frame);

	return 0;
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Smooth an image using a median filter
  @param	in		image to be smoothed
  @param	size	size of kernel matrix to use in filtering

  @return   out 	smoothed image
 */
/*---------------------------------------------------------------------------*/
cpl_image *omega_smooth_image(cpl_image *in, int size)
{

	int i,nx, ny, type;
	cpl_matrix *kernel;
	cpl_mask *mask;
	cpl_image *out;

	if (in == NULL)
		return NULL;

//	if(size < 3)
//		return NULL;

	nx = cpl_image_get_size_x(in);
	ny = cpl_image_get_size_y(in);
	type = cpl_image_get_type(in);

	kernel = cpl_matrix_new(size, size);
//    cpl_matrix_fill(kernel, 1.0);
//    out = cpl_image_filter_median(in, kernel);
    cpl_matrix_delete(kernel);

    /* new code */

    mask = cpl_mask_new(size, size);
    for (i = 0; i < size; ++i) {
    	int j = 0;
    	for (j = 0; j < size; j++)
    	{
    		cpl_mask_set(mask, i + 1, j + 1, CPL_BINARY_1);
    	}

    }

    out = cpl_image_new(nx, ny, type);
    if(cpl_image_filter_mask(out, in, mask, CPL_FILTER_AVERAGE, CPL_BORDER_FILTER)
    		!= CPL_ERROR_NONE){
    	cpl_msg_debug(cpl_func,"Cannot filter image. %s", cpl_error_get_message());
    	freeimage(out);
    	freemask(mask);
    	return NULL;
    }


    freemask(mask);
    return (out);

}

/*---------------------------------------------------------------------------*/
/**
  @brief    Create a string with a name
  @param	instrument		string with instrument name
  @param	name			middle part of the output name
  @param	suffix			suffix name or NULL in which case fits is used

  @return   outname			or NULL in case of error

  It composes a string in the following way:
	  instrument_name.suffix
 */
/*---------------------------------------------------------------------------*/
char *omega_create_name(const char *instrument, const char *name, const char *suffix)
{
    char *outname;
    const char *end = "fits";

    if(instrument == NULL || name == NULL){
    	return NULL;
    }
    if(suffix == NULL)
    	suffix = end;

    sprintf(outname, "%s_%s.%s", instrument, name, suffix);

    return outname;
}

/**
 * @brief
 *   Subtract bias from an image
 *
 * @param	image	Image to subtract bias brom
 * @param	bias	Bias image
 *
 * @return
 *   0 on success or -1 in case of error
 */
int omega_biascor(cpl_image *image, const cpl_image *bias)
{

	if((image == NULL) || (bias == NULL))
		return -1;


	if(cpl_image_subtract(image, bias) != CPL_ERROR_NONE){
		return -1;
	}


	return 0;
}

/**
 * @brief
 *   Get the value of CHIP ID for a given extension
 *
 * @param ext	FITS extension for the particular CHIP
 *
 * @return
 *   The name stored in the keyword HIERARCH ESO DET CHIP# ID of OmegaCAM
 * 	 images or NULL in case of error.
 */
const char * omega_get_extension_chipid(int ext)
{

	const char *chipid;

	switch (ext){
	case 1:
		chipid = "ESO_CCD_#65";
		break;
	case 2:
		chipid = "ESO_CCD_#66";
		break;
	case 3:
		chipid = "ESO_CCD_#67";
		break;
	case 4:
		chipid = "ESO_CCD_#68";
		break;
	case 5:
		chipid = "ESO_CCD_#73";
		break;
	case 6:
		chipid = "ESO_CCD_#74";
		break;
	case 7:
		chipid = "ESO_CCD_#75";
		break;
	case 8:
		chipid = "ESO_CCD_#76";
		break;
	case 9:
		chipid = "ESO_CCD_#81";
		break;
	case 10:
		chipid = "ESO_CCD_#82";
		break;
	case 11:
		chipid = "ESO_CCD_#83";
		break;
	case 12:
		chipid = "ESO_CCD_#84";
		break;
	case 13:
		chipid = "ESO_CCD_#89";
		break;
	case 14:
		chipid = "ESO_CCD_#90";
		break;
	case 15:
		chipid = "ESO_CCD_#91";
		break;
	case 16:
		chipid = "ESO_CCD_#92";
		break;
	case 17:
		chipid = "ESO_CCD_#69";
		break;
	case 18:
		chipid = "ESO_CCD_#70";
		break;
	case 19:
		chipid = "ESO_CCD_#71";
		break;
	case 20:
		chipid = "ESO_CCD_#72";
		break;
	case 21:
		chipid = "ESO_CCD_#77";
		break;
	case 22:
		chipid = "ESO_CCD_#78";
		break;
	case 23:
		chipid = "ESO_CCD_#79";
		break;
	case 24:
		chipid = "ESO_CCD_#80";
		break;
	case 25:
		chipid = "ESO_CCD_#85";
		break;
	case 26:
		chipid = "ESO_CCD_#86";
		break;
	case 27:
		chipid = "ESO_CCD_#87";
		break;
	case 28:
		chipid = "ESO_CCD_#88";
		break;
	case 29:
		chipid = "ESO_CCD_#93";
		break;
	case 30:
		chipid = "ESO_CCD_#94";
		break;
	case 31:
		chipid = "ESO_CCD_#95";
		break;
	case 32:
		chipid = "ESO_CCD_#96";
		break;
	default:
		cpl_msg_warning(cpl_func,"Unable to find CHIP ID for extension %d",ext);
		chipid = NULL;
		break;
	}

	return chipid;
}

/**
 * @brief
 *   Get the readout orientation of a chip based on the FITS extension number.
 *
 * @param ext   FITS extension
 *
 * @return
 *   The orientation number, 1 or 2.
 */
int omega_get_chip_orientation(int ext)
{

	int orient = 1;

	if(((ext >= 1) && (ext <= 8)) || ((ext >=17) && (ext <= 24)))
		orient = 1;
	else if(((ext >= 9) && (ext <=16)) || ((ext >= 25) && (ext <= 32)))
		orient = 2;
	else
		cpl_msg_warning(cpl_func,"Unable to get chip orientation. Using default 1");


	return orient;
}

/**
 * @brief
 *   Save an image as a DFS compliant product
 * @param img     Image to save
 * @param set     Frame set
 * @param pars    Parameters list
 * @param alist   Optional property list
 * @param qclist  Optional QC list
 * @param bpp     CPL_TYPE_BPP of product
 * @param name    Name of product
 * @param recipe  Name of recipe
 * @param frame   Frame describing product
 * @param inherit Frame to inherit header from or NULL
 * @param isfirst Tell whether to create a primary header or not.
 *
 * @return
 *   0 on success or -1 in case of error.
 */
int omega_save_image(const cpl_image *img,cpl_frameset *set,const cpl_parameterlist *pars,
		cpl_propertylist *alist, const cpl_propertylist *qclist,cpl_type_bpp bpp,
		const char *name, const char *recipe, cpl_frame *frame,
		const cpl_frame *inherit, int isfirst)
{

	cpl_propertylist *plist;

	if((img == NULL) || (set == NULL) || (pars == NULL) ||
			(name == NULL) || (frame == NULL))
		return -1;

	if(isfirst){
		/* Save primary header */
		if(omega_save_primary(set,pars,name,recipe,NULL,frame,inherit) == -1){
			cpl_msg_error(cpl_func,"Cannot save primary header of product");
			return -1;
		}
	}

	/* Save extension */
	plist = cpl_propertylist_new();

	if(alist != NULL)
		omega_pfits_update_header(plist,alist);

	/* Add DataFlow keywords in extension header */
	if (cpl_dfs_setup_product_header(plist, frame, set, pars,
							recipe,PIPEID,DICID,inherit) != CPL_ERROR_NONE) {
		cpl_msg_warning(cpl_func, "Unable to setup extension header of %s. %s",name,
							cpl_error_get_message()) ;
	}

	/*Remove undesired keywords from extension header*/
	cpl_propertylist_erase_regexp(plist, REM_EXT_KEYS, 0);

	/* Append QC list to the extension header */
	if(qclist != NULL)
		cpl_propertylist_append(plist, qclist);

	/* Save the product */
	if (cpl_image_save(img,name,bpp,plist,CPL_IO_EXTEND) != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func,"Cannot save product. %s", cpl_error_get_message());
		freeplist(plist);
		return -1;
	}

	freeplist(plist);

	return 0;

}

/**
 * @brief
 *   Save a table as a DFS compliant product
 *
 * @param tbl     Table to save
 * @param set     Set of frames
 * @param pars    Parameters list
 * @param alist   Optional property list
 * @param qclist  Optional QC list
 * @param name    Name of product
 * @param recipe  Name of recipe
 * @param frame   Frame describing the product
 * @param inherit Frame to inherit header from or NULL
 * @param isfirst Tell whether to create a primary header or not.
 *
 * @return
 *   0 on success or -1 in case of error.
 */
int omega_save_table(const cpl_table *tbl,cpl_frameset *set,const cpl_parameterlist *pars,
		cpl_propertylist *alist, const cpl_propertylist *qclist,const char *name,
		const char *recipe, cpl_frame *frame,const cpl_frame *inherit, int isfirst)
{

	cpl_propertylist *plist;

	if((tbl == NULL) || (set == NULL) || (pars == NULL) ||
			(name == NULL) || (frame == NULL))
		return -1;

	if(isfirst){
		/* Save primary header */
		if(omega_save_primary(set,pars,name,recipe,NULL,frame,inherit) == -1){
			cpl_msg_error(cpl_func,"Cannot save primary header of product");
			return -1;
		}
	}

	/* Save the extension */
	plist = cpl_propertylist_new();

	if(alist != NULL)
		omega_pfits_update_header(plist,alist);

	/* Add DataFlow keywords in extension header */
	if (cpl_dfs_setup_product_header(plist, frame, set, pars,
			recipe,PIPEID,DICID,inherit) != CPL_ERROR_NONE) {

		cpl_msg_warning(cpl_func, "Unable to setup extension header of %s. %s",name,
							cpl_error_get_message()) ;
	}

	/*Remove undesired keywords from extension header*/
	cpl_propertylist_erase_regexp(plist, REM_EXT_KEYS, 0);

	/* Append QC list to the extension header */
	if(qclist != NULL)
		cpl_propertylist_append(plist, qclist);

	/* Save the product */
	if (cpl_table_save(tbl,NULL,plist,name,CPL_IO_EXTEND) != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func,"Cannot save product. %s", cpl_error_get_message());
		freeplist(plist);
		return -1;
	}

	freeplist(plist);

	return 0;
}

/**
 * @brief	Save an image which is taken from a omega_fits strucutre
 *
 * @param	scifits			omega_fits structure containing the image to save
 * @param	set				original set of frames
 * @param	pars			original parameters list
 * @param	qclist			QC property list or NULL
 * @param   bpp             CPL_TYPE_BPP to save product
 * @param	name   		    output name of product
 * @param   recipe          recipe name
 * @param	product_frame   output frame of product
 * @param	inherit			optional frame to inherit header from
 * @param	isfirst			Tells whether to create primary header or not
 *
 * @return	status		0 on success or -1 in case of error
 *
 */
int omega_save_fits(omega_fits *scifits,cpl_frameset *set, const cpl_parameterlist *pars,
							const cpl_propertylist *qclist, cpl_type_bpp bpp,const char *name,
							const char *recipe,cpl_frame *product_frame, const cpl_frame *inherit,
							const int isfirst)
{

	cpl_propertylist *plist;

	if((scifits == NULL) || (set == NULL) || (pars == NULL) ||
			(name == NULL) || (product_frame == NULL))
			return -1;


	if(isfirst){

		plist = cpl_propertylist_duplicate(omega_fits_get_phu(scifits));

		if(omega_save_primary(set,pars,name,recipe,plist,product_frame,inherit) == -1){
			cpl_msg_error(cpl_func,"Cannot save primary header of product");
			freeplist(plist);
			return -1;
		}

		freeplist(plist);
	}

	/* Save the extension */
	plist = cpl_propertylist_duplicate(omega_fits_get_ehu(scifits));

	/* Add DataFlow keywords in extension header */
	if (cpl_dfs_setup_product_header(plist, product_frame, set, pars,
			recipe,PIPEID,DICID,inherit) != CPL_ERROR_NONE) {

		cpl_msg_warning(cpl_func, "Unable to setup extension header of %s. %s",name,
							cpl_error_get_message()) ;
	}

	/*Remove undesired keywords from extension header*/
	cpl_propertylist_erase_regexp(plist, REM_EXT_KEYS, 0);

	/* Append QC list to the extension header */
	if(qclist != NULL)
		cpl_propertylist_append(plist, qclist);

	/* Save the product */
	if (cpl_image_save(omega_fits_get_image(scifits),name,bpp,
						plist,CPL_IO_EXTEND) != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func,"Cannot save product %s. %s", name, cpl_error_get_message());
		freeplist(plist);
		return -1;
	}

	freeplist(plist);

	return 0;
}

cpl_error_code omega_get_pixelscale(cpl_propertylist *plist,
		double * pixscale_x, double * pixscale_y){

	double cd11= 0.;
	double cd12= 0.;
	double cd21= 0.;
	double cd22= 0.;

	if(cpl_propertylist_has(plist,"CD1_1") &&
			cpl_propertylist_has(plist,"CD1_2") &&
			cpl_propertylist_has(plist,"CD2_1") &&
			cpl_propertylist_has(plist,"CD2_2")){
		cd11= cpl_propertylist_get_double(plist, "CD1_1");
		cd12= cpl_propertylist_get_double(plist, "CD1_2");
		cd21= cpl_propertylist_get_double(plist, "CD2_1");
		cd22= cpl_propertylist_get_double(plist, "CD2_2");
	}
	else{
		cpl_error_set_message(cpl_func,CPL_ERROR_DATA_NOT_FOUND, "CD matrix not"
				" complete. Pixel scale can not be computed");
		return cpl_error_get_code();
	}

		*pixscale_x = sqrt (cd11*cd11 + cd21*cd21);
		*pixscale_y = sqrt (cd12*cd12 + cd22*cd22);

		cpl_msg_debug(cpl_func,"Pixelscale  "
				"x: %g y: %g",*pixscale_x, *pixscale_y);

	return cpl_error_get_code();

}

/**
 * @brief
 *   Compute the wcs shift in x and y direction.
 *
 * @param wcslist1  Propertylist containg the wcs informations.
 * @param wcslist2  Propertylist containg the wcs informations.
 * @param xshift    Pointer where the shift in x direction is written
 * @param yshift    Pointer where the shift in y direction is written.
 * @return An appropriate error code if failing.
 *
 */
cpl_error_code omega_get_wcsshift(cpl_propertylist * wcslist1,
		cpl_propertylist * wcslist2, double * xshift, double * yshift){

	cpl_wcs * wcs1=NULL;
	cpl_wcs	* wcs2=NULL;

	wcs1=cpl_wcs_new_from_propertylist(wcslist1);
	wcs2=cpl_wcs_new_from_propertylist(wcslist2);

	if(wcs1 == NULL || wcs2 == NULL )
	{
		cpl_msg_warning(cpl_func, "WCS informations missing - assuming no shift "
				"of the x and y coordinate");
		*xshift=0;
		*yshift=0;
		cpl_msg_debug(cpl_func, "Coordinateshift: xshif=%g, yshift=%g",
				*xshift,*yshift);
		cpl_wcs_delete(wcs1);
		cpl_wcs_delete(wcs2);
		return cpl_error_get_code();
	}
	else{
		*xshift=cpl_array_get(cpl_wcs_get_crpix(wcs1),0,NULL) -
				cpl_array_get(cpl_wcs_get_crpix(wcs2),0,NULL);
		*yshift=cpl_array_get(cpl_wcs_get_crpix(wcs1),1,NULL) -
				cpl_array_get(cpl_wcs_get_crpix(wcs2),1,NULL);
	}

	cpl_wcs_delete(wcs1);
	cpl_wcs_delete(wcs2);

	cpl_msg_debug(cpl_func, "Coordinateshift: xshif=%g, yshift=%g",
			*xshift,*yshift);

	return cpl_error_get_code();
}

/**@}*/
