/* $Id: omega_wcscor.c,v 1.38 2011/12/21 14:53:55 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2011/12/21 14:53:55 $
 * $Revision: 1.38 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cpl.h>

#include "omega_wcscor.h"
#include "omega_catalog.h"
#include "omega_dfs.h"
#include "omega_pfits.h"
#include "omega_trim.h"
#include "omega_utils.h"

/*-----------------------------------------------------------------------------
                            Function prototypes
 -----------------------------------------------------------------------------*/
static char *my_strchr(char *, char);

/* Search for a char in a string and return its position*/
static char * my_strchr(char *string, char find)
{
    while(*string != find)
    {
        if(*string == '\0')
            return NULL;

        ++string;
    }

    return (string);
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup astrom  Astrometry functions
 *
 * @par Synopsis:
 * This module provides functions to calculate the deal with astrometric correction
 *
 * @code
 *   #include omega_wcscor.h
 *   #include omega_catalog.h
 *   #include omega_dfs.h
 *   #include omega_pfits.h
 *   #include omega_trim.h
 *   #include omega_utils.h
 * @endcode
 */
/**@{*/

/**
 * @brief
 * 		Shift reference pixel of a trimmed image based on the trimming regions
 *
 * @param	frame		Frame of raw image (not trimmed)
 * @param	xn			FITS extension to load
 * @param	plist		Property list of original frame which will be updated
 *
 * @return
 *   0 on success or -1 in case of error.
 **/
int omega_shift_refpix(const cpl_frame *frame, int xn, cpl_propertylist *xlist)
{
	int xoffset = 0;
	int yoffset = 0;
	double crpix1 = 0.0;
	double crpix2 = 0.0;
	cpl_vector *region_vec;
	cpl_propertylist *plist;


	if(omega_pfits_check_instrume(frame) == 1){
		/* It's a WFI image */
		plist = cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
        omega_wfi_correct_wcs(plist, xlist);
        freeplist(plist);
	}

	region_vec = get_trim_region(frame, xn);

	if(region_vec != NULL){
		/*
		 * Shift reference pixels of trimmed image
		 */
		xoffset = (int)cpl_vector_get(region_vec, 0);
		xoffset = xoffset -1;
		yoffset = (int)cpl_vector_get(region_vec, 1);
		yoffset = yoffset -1;

		crpix1 = cpl_propertylist_get_double(xlist, "CRPIX1");
		crpix2 = cpl_propertylist_get_double(xlist, "CRPIX2");
		crpix1 = crpix1 - xoffset;
		crpix2 = crpix2 - yoffset;

		cpl_propertylist_update_double(xlist, "CRPIX1", crpix1);
		cpl_propertylist_update_double(xlist, "CRPIX2", crpix2);

		cpl_vector_delete(region_vec);
	}

	return 0;
}


/**
 * @brief
 *   Calculate WCS correction for a reduced image using pattern matching
 * @param	reduced		Name of reduced image (STD or science)
 * @param	catalog		Frame of MASTER_USNOA@_CATALOGUE
 * @param	pars		Parameters list
 * @param   ext         FITS extension to load
 * @param	nmatches	Output number of matches
 * @param	mratio		Output ratio USNOA2 matches/detected stars
 *
 * @return
 *  A property list with the WCS coordinates or NULL in case of error.
 *
 * This function will create a catalogue of sources from the image and convert
 * the X_IMAGE and Y_IMAGE coordinates of each source to Ra and Dec. It will then
 * get the coverage of the image and extract all USNOA stars that lie in this
 * region. The RA2000 and Dec2000 coordinates of these extracted stars will
 * be converted to X_predict Y_predict and will be added to the same table.
 * the two catalogues will be sorted in ascending magnitude (MAG_AUTO and MagB).
 * Two matrices will be created from these lists and used in the pattern matching
 * routine. If there is a minimum of matches, then a plate solution will
 * be derived from this list.
 **/
cpl_propertylist *omega_match_points(const char *reduced,  const char *reduced_weight, const cpl_frame *catalog, cpl_parameterlist *pars,
									int ext, int *nmatches, double *mratio)
{

	int i, nsources=0, nsources_flag0=0, nstds, nstars, npattern;
	int mode = 0;
	int niter = 1;
	int npars = 0;
	int SextractorFlagLimit=0;
	int imsizex=0;
	int imsizey=0;
    //int shiftvalue=0;
    int nsources_firstpass=0;
	double matchlimit=0.;
	double zp = 0.0;
	double gain = 1.0;
    double x1,x2,y1,y2;
    double pradius = 0.0;
    double maxpradius = 0.0;
    double tol = 0.0;
    double astrom_thre = 0.0;
    double scale = 0.0;
    double angle = 0.0;
    double SextractorClassStar=0.0;
    double centralwl=0.0;
 	char *cmd = NULL;
	const char *text = NULL;
	const char *path;
	const char *sex_conf = NULL;
	const char *sex_conv = NULL;
	const char *sex_par = NULL;
	const char *sex_nnw = NULL;
	const char *srcs = "omega_cat_astrom_sources.fits";
	cpl_parameter *par;
	cpl_array *status, *matches;
	cpl_table *sources=NULL, *sources_flag0=NULL, *stds, *extracted, *data;
	cpl_table *dummy_table1=NULL;
	cpl_table * dummy_table=NULL;
	//cpl_table * tshift1=NULL;
	//cpl_table * tshift2=NULL;
	cpl_matrix *from, *to, *xymat, *rdmat, *stars, *pattern;
	cpl_wcs *wcs;
	cpl_vector *coords;
	cpl_wcs_platesol_fitmode fitmode = CPL_WCS_PLATESOL_4;
	cpl_propertylist *xlist = NULL;
	cpl_propertylist *plist = NULL;
	cpl_propertylist *filterlist = NULL;

	cpl_propertylist *aplist = NULL;
	cpl_propertylist *p;

	/* Get parameters */
	npars = cpl_parameterlist_get_size(pars);
	par = cpl_parameterlist_get_first(pars);

	for(i=0; i<npars; i++) {
		text = cpl_parameter_get_alias(par, CPL_PARAMETER_MODE_CLI);
			if(strcmp("bin-path", text) == 0) {
    			path = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-config", text) == 0) {
    			sex_conf = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-conv", text) == 0) {
    			sex_conv = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-param", text) == 0) {
    			sex_par = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("sex-nnw", text) == 0) {
    			sex_nnw = cpl_parameter_get_string(par) ;
			}
			else if(strcmp("astrom-thre", text) == 0){
				astrom_thre = cpl_parameter_get_double(par);
			}
			else if(strcmp("nstars", text) == 0){
				nstars = cpl_parameter_get_int(par);
			}
			else if(strcmp("npattern", text) == 0){
				npattern = cpl_parameter_get_int(par);
			}
			else if(strcmp("pradius", text) == 0){
				pradius = cpl_parameter_get_double(par);
			}
			else if(strcmp("maxpradius", text) == 0){
				maxpradius = cpl_parameter_get_double(par);
			}
			else if(strcmp("tol", text) == 0){
				tol = cpl_parameter_get_double(par);
			}
			else if(strcmp("platesol", text) == 0){
				mode = cpl_parameter_get_int(par);
				if(mode == 6) fitmode = CPL_WCS_PLATESOL_6;
			}
			else if(strcmp("niter", text) == 0){
				niter = cpl_parameter_get_int(par);
			}
			else if(strcmp("SextractorFlagLimit", text) == 0){
				SextractorFlagLimit = cpl_parameter_get_int(par);
			}
			else if(strcmp("SextractorClassStar", text) == 0){
				SextractorClassStar= cpl_parameter_get_double(par);
			}

		par = cpl_parameterlist_get_next(pars);

	}


	/*
	 * Get Zeropoint for science frames. For STD,
	 * this will be calculated later.
	 */
    xlist = cpl_propertylist_load(reduced, 1);
	if(cpl_propertylist_has(xlist,"ESO QC SCI ZEROPNT"))
		zp = cpl_propertylist_get_double(xlist,"ESO QC SCI ZEROPNT");

	if(cpl_propertylist_has(xlist,"NAXIS1"))
		imsizex = cpl_propertylist_get_int(xlist,"NAXIS1");
	if(cpl_propertylist_has(xlist,"NAXIS2"))
		imsizey = cpl_propertylist_get_int(xlist,"NAXIS2");

	/*Get filter info - used to sort the right filter in the USNO table*/
	plist = cpl_propertylist_load(reduced, 0);
	if(cpl_propertylist_has(xlist,"EXTNAME") &&
			!cpl_propertylist_has(plist,"EXTNAME"))
		cpl_propertylist_copy_property(plist,xlist,"EXTNAME");

	filterlist=omega_pfits_get_filter_info(plist);

	if(cpl_propertylist_has(filterlist,"CWL")){
		centralwl=cpl_propertylist_get_double(filterlist, "CWL");
	}
	else{
		centralwl=0.;
	}
	freeplist(plist);
	freeplist(filterlist);

	/*
	 * Get the gain for this extension.
	 */

	omega_pfits_get_conad(xlist, &gain);
	cpl_msg_debug(cpl_func,"gain: %f",gain);


	sources = NULL;

	for(i = 0; i < 15; i++) {
		/* Create catalog of sources from image */

		if(reduced_weight != NULL){
			cmd = cpl_sprintf("%s/sex %s -WEIGHT_TYPE MAP_WEIGHT -WEIGHT_IMAGE %s "
					"-c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
					"-FILTER Y -DETECT_THRESH %g -ANALYSIS_THRESH %g -GAIN %g -MAG_ZEROPOINT %f "
					"-CATALOG_TYPE FITS_1.0 -CATALOG_NAME %s -CHECKIMAGE_NAME %s "
					"-CHECKIMAGE_TYPE OBJECTS -DEBLEND_NTHRESH  1 2>/dev/null",
					path, reduced, reduced_weight, sex_conf, sex_par, sex_conv, sex_nnw, astrom_thre, astrom_thre,
					gain, zp, srcs, "omega_ima_detected_astrom_srcs.fits");
		}
		else
		{
			cmd = cpl_sprintf("%s/sex %s -c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
					"-FILTER Y -DETECT_THRESH %g -ANALYSIS_THRESH %g -GAIN %g -MAG_ZEROPOINT %f "
					"-CATALOG_TYPE FITS_1.0 -CATALOG_NAME %s -CHECKIMAGE_NAME %s "
					"-CHECKIMAGE_TYPE OBJECTS -DEBLEND_NTHRESH  1 2>/dev/null",
					path, reduced, sex_conf, sex_par, sex_conv, sex_nnw, astrom_thre, astrom_thre,
					gain, zp, srcs, "omega_ima_detected_astrom_srcs.fits");
		}

		cpl_msg_debug(cpl_func,"cmd: %s", cmd);
		if (system(cmd) != 0) {
			cpl_msg_error(cpl_func,"Failed to detect sources in image using Sextractor");
			freespace(cmd);
			freeplist(xlist);
			return NULL;
		}
		freespace(cmd);

		data = cpl_table_load(srcs, 1, 0);
		if (data == NULL) {
			cpl_msg_error(cpl_func,"Sources table is NULL");
			freeplist(xlist);
			return NULL;
		}

		freetable(sources);
		sources_flag0 = omega_filter_table_int(data,            "FLAGS",
				CPL_NOT_GREATER_THAN, SextractorFlagLimit, &nsources_flag0);
		sources       = omega_filter_table_float(sources_flag0, "CLASS_STAR",
				CPL_NOT_LESS_THAN, SextractorClassStar, &nsources);
		freetable(data);
		freetable(sources_flag0);

		if (i==0){
			nsources_firstpass=nsources;
		}

		if(nsources > 1000) {
			/* Detect again with greater threshold */
			astrom_thre = astrom_thre + 30;
		}
		else
			break;
	}


	/* Sort sources table by magnitude */
    p = cpl_propertylist_new();
    cpl_propertylist_append_bool(p,"MAG_AUTO",0);
    if (cpl_table_sort(sources, p) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func,"Cannot sort sources table");
	    freeplist(p);
	    freeplist(xlist);
    	freetable(sources);
	    return NULL;
    }
    freeplist(p);


	/*If there are too many sources in the image, pattern matching will very
	 *  likely fail. In this case for the pattern-matching-first-guess only
	 *  objects in a smaller spatial region are used*/

	if(nsources_firstpass>500)
    {
		/*cpl_msg_info(cpl_func,"There are too many sources for pattern matching,"
				" shrinking the field size ... ");*/
    	dummy_table=cpl_table_duplicate(sources);
    	freetable(sources);
    	cpl_table_and_selected_float(dummy_table,"X_IMAGE",CPL_NOT_LESS_THAN,1.+500.);
    	cpl_table_and_selected_float(dummy_table,"Y_IMAGE",CPL_NOT_LESS_THAN,1.+500.);
    	cpl_table_and_selected_float(dummy_table,"X_IMAGE",CPL_NOT_GREATER_THAN,(float)imsizex-500.);
    	cpl_table_and_selected_float(dummy_table,"Y_IMAGE",CPL_NOT_GREATER_THAN,(float)imsizey-500.);

    	sources= cpl_table_extract_selected(dummy_table);

    	cpl_table_not_selected(dummy_table);
    	dummy_table1=cpl_table_extract_selected(dummy_table);

    	cpl_table_insert(sources, dummy_table1, cpl_table_get_nrow(sources));
    	freetable(dummy_table);
    	freetable(dummy_table1);
    	nsources=cpl_table_get_nrow(sources);
    }

    cpl_msg_info(cpl_func,"Number of sources in image: %d", nsources);
    if(nsources < nstars){
    	nstars = nsources -1;
    }


    if(pradius > maxpradius)
    	pradius = maxpradius;


    /* Create matrix of X,Y coordinates */
    from = cpl_matrix_new(nsources, 2);
    for (i=0; i<nsources; i++){
        float xim = cpl_table_get_float(sources, "X_IMAGE", i, NULL);
        float yim = cpl_table_get_float(sources, "Y_IMAGE", i, NULL);
        cpl_matrix_set(from, i, 0, xim);
        cpl_matrix_set(from, i, 1, yim);
    }

    /* Create matrix to be used in PPM */
    stars = cpl_matrix_transpose_create(from);

    /* Get coverage of image */
    if(omega_get_coverage(xlist, 0, from, &x1, &x2, &y1, &y2) != 0){
    	cpl_msg_error(cpl_func,"Cannot get coverage of image");
    	freematrix(from);
    	freematrix(stars);
    	freeplist(xlist);
    	freetable(sources);
    	return NULL;
    }
    freematrix(from);

    /* Get USNOA stars that fall in region of image */
    /* ramin = x1 ramax = x2 decmin = y1 decmax = y2;*/
    /* Write in 'nstds' the number of sources in the table */
    coords = cpl_vector_new(4);
    cpl_vector_set(coords, 0, x1);
    cpl_vector_set(coords, 1, x2);
    cpl_vector_set(coords, 2, y1);
    cpl_vector_set(coords, 3, y2);

    stds = omega_get_usnoa_table(catalog, coords, &nstds);
    if (stds == NULL) {
    	cpl_msg_warning(cpl_func,"Cannot find USNOA2 stars in this region: {RA-RA:DEC-DEC}={%g-%g:%g=%g}",
							x1,x2,y1,y2);
    	freeplist(xlist);
    	freematrix(stars);
    	freetable(sources);
    	freevector(coords);
    	return NULL;
    }
	freevector(coords);


    cpl_msg_info(cpl_func,"Number of USNOA2 stars in this region: %d", nstds);
    cpl_msg_debug(cpl_func,"%d USNOA2 stars fall in region: {RA-RA:DEC-DEC}={%g--%g:%g--%g}", nstds,x1,x2,y1,y2);

    /* Convert Ra, Dec to X,Y using the WCS information from image */
    from = cpl_matrix_new(nstds, 2);
    for (i=0; i<nstds; i++){
    	cpl_matrix_set(from, i, 0, cpl_table_get_double(stds, "RA2000", i, NULL));
    	cpl_matrix_set(from, i, 1, cpl_table_get_double(stds, "Dec2000", i, NULL));
    }

    to = NULL;
    status = NULL;
    wcs = cpl_wcs_new_from_propertylist(xlist);
    if(cpl_wcs_convert(wcs, from, &to, &status, CPL_WCS_WORLD2PHYS) != CPL_ERROR_NONE){
    	freewcs(wcs);
    	freearray(status);
    	freematrix(from);
    	freematrix(stars);
    	freematrix(to);
    	freeplist(xlist);
    	freetable(sources);
    	freetable(stds);
    	cpl_msg_error(cpl_func,"Error in cpl_wcs conversion. %s",cpl_error_get_message());
    	return NULL;
    }
    freewcs(wcs);
    freematrix(from);
	freearray(status);

    /* Add the predicted x,y coordinate columns to the  usnoa table */
    cpl_table_new_column(stds,"xpredict",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(stds,"xpredict","pixels");
    cpl_table_new_column(stds,"ypredict",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(stds,"ypredict","pixels");



    for(i=0; i< nstds; i++){
    	float xp = (float)cpl_matrix_get(to, i, 0);
    	float yp = (float)cpl_matrix_get(to, i, 1);
    	cpl_table_set_float(stds,"xpredict", i, xp);
    	cpl_table_set_float(stds,"ypredict", i, yp);
    }
    freematrix(to);

 	//cpl_table_dump(stds,0,1000,NULL);

     /*Unselect objects falling outside the image borders +- fudge*/



    /* Sort stds table by magnitude */
    p = cpl_propertylist_new();

    if (centralwl>6000){
    	cpl_propertylist_append_bool(p,"MagR",0);
    	cpl_msg_info(cpl_func,"Sorting standard star catalogue by R-band mag");
    }
    else{
    	cpl_propertylist_append_bool(p,"MagB",0);
    	cpl_msg_info(cpl_func,"Sorting standard star catalogue by B-band mag");
    }

    if (cpl_table_sort(stds, p) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func,"Cannot sort stds table");
	    freeplist(p);
    	freetable(sources);
        freematrix(stars);
    	freetable(stds);
    	freeplist(xlist);
	    return NULL;
    }
    freeplist(p);


    /*  If there are too many objects in the catalog, pattern matching will very
	 *  likely fail. In this case for the pattern-matching-first-guess only
	 *  objects in a smaller spatial region are used */

	dummy_table=cpl_table_duplicate(stds);
	freetable(stds);

	if(nsources_firstpass>500){
		cpl_table_and_selected_float(dummy_table,"xpredict",CPL_NOT_LESS_THAN,1.+600.);
		cpl_table_and_selected_float(dummy_table,"ypredict",CPL_NOT_LESS_THAN,1.+600.);
		cpl_table_and_selected_float(dummy_table,"xpredict",CPL_NOT_GREATER_THAN,(float)imsizex-600.);
		cpl_table_and_selected_float(dummy_table,"ypredict",CPL_NOT_GREATER_THAN,(float)imsizey-600.);
	}
	else {
		cpl_table_and_selected_float(dummy_table,"xpredict",CPL_NOT_LESS_THAN,1.+100.);
		cpl_table_and_selected_float(dummy_table,"ypredict",CPL_NOT_LESS_THAN,1.+100.);
		cpl_table_and_selected_float(dummy_table,"xpredict",CPL_NOT_GREATER_THAN,(float)imsizex-100.);
		cpl_table_and_selected_float(dummy_table,"ypredict",CPL_NOT_GREATER_THAN,(float)imsizey-100.);
	}

	stds= cpl_table_extract_selected(dummy_table);
	cpl_table_not_selected(dummy_table);
	dummy_table1=cpl_table_extract_selected(dummy_table);

	cpl_table_insert(stds, dummy_table1, cpl_table_get_nrow(stds));
	freetable(dummy_table);
	freetable(dummy_table1);

	if (stds == NULL) {
    	cpl_msg_warning(cpl_func,"Cannot find USNOA2 stars within the image "
    			"borders");
    	freewcs(wcs);
    	freearray(status);
    	freematrix(from);
    	freematrix(stars);
    	freematrix(to);
    	freeplist(xlist);
    	freetable(sources);
    	return NULL;
    }






	nstds=cpl_table_get_nrow(stds);
    cpl_msg_info(cpl_func,"Number of USNOA2 stars falling on the image: %d", nstds);


    /* Prepare the matrices for the pattern matching */
    pattern = cpl_matrix_new(2, nstds);
    for(i=0; i<nstds; i++){
    	float x = cpl_table_get_float(stds, "xpredict", i, NULL);
    	float y = cpl_table_get_float(stds, "ypredict", i, NULL);
    	cpl_matrix_set(pattern, 0, i, x);
    	cpl_matrix_set(pattern, 1, i, y);
    }


    if(nstds < npattern){
    	npattern = nstds -1;
    }

    cpl_msg_debug(cpl_func,"stars: % " CPL_SIZE_FORMAT " , nstars: %d pattern: % " CPL_SIZE_FORMAT "  npattern: %d",
    		cpl_matrix_get_ncol(stars), nstars, cpl_matrix_get_ncol(pattern),
    		npattern);


/*


    Define the number of (bright)objects that should not be used for the
      * first guess in the pattern matching as they might be saturated, ...



    shift the matrix

    if(nsources_firstpass>1000){
    	shiftvalue = nstds < nsources ? (int)nstds*0.0 : (int)nsources*0.0;
    }
    else{
    	shiftvalue = nstds < nsources ? (int)nstds*0.0 : (int)nsources*0.0;
    }


     cpl_matrix_shift(stars,  0,0 - shiftvalue);
     cpl_matrix_shift(pattern,0,0 - shiftvalue);
     cpl_msg_debug(cpl_func,"shift_value: %d",shiftvalue);

     Shift the tables to keep them in sync with the matrix
     tshift1 = cpl_table_extract(sources,0,shiftvalue);
     tshift2 = cpl_table_extract(sources,shiftvalue,cpl_table_get_nrow(sources)-shiftvalue);
     freetable(sources);
     sources=cpl_table_duplicate(tshift2);
     cpl_table_insert(sources,tshift1,cpl_table_get_nrow(sources));
     freetable(tshift1);
     freetable(tshift2);
     //cpl_table_dump_structure(sources,NULL);

     tshift1 = cpl_table_extract(stds,0,shiftvalue);
     tshift2 = cpl_table_extract(stds,shiftvalue,cpl_table_get_nrow(stds)-shiftvalue);
     freetable(stds);
     stds=cpl_table_duplicate(tshift2);
     cpl_table_insert(stds,tshift1,cpl_table_get_nrow(stds));
     freetable(tshift1);
     freetable(tshift2);
     //cpl_table_dump_structure(sources,NULL);

*/

    /* Call the PPM routine
     * If no matches, less than 25%, or less then 5 are found, change nstar and
     * try again
     */
    matchlimit = nstds < nsources ? (double)nstds*0.25 : (double)nsources*0.25;
    if(matchlimit < 5) matchlimit=5;
    cpl_msg_debug(cpl_func,"Match limit: %f", matchlimit);
    for (i = 0; i < 4; i++) {

    	*nmatches = 0;
    	matches = NULL;
    	cpl_msg_debug(cpl_func,"nstars: %d  - npattern: %d ", nstars, npattern);
    	matches = cpl_ppm_match_points(stars, nstars, 0.5, pattern, npattern,
    			15, tol, pradius, NULL, NULL, &scale, &angle);
    	if(matches != NULL){
    		*nmatches = cpl_array_get_size(matches) - cpl_array_count_invalid(matches);
    	}
    	if((matches == NULL) || (*nmatches < matchlimit )) {
    		cpl_msg_debug(cpl_func,"nstars: %d npattern: %d  i: %d",nstars,npattern,i );

    		if (i==1){
    			npattern=20;
    			nstars=35;
    		}

    		if (i==2){
    			npattern=25;
    			nstars=50;
    		}

    		if (i==3){
    			freearray(matches);
    			break;
    		}

    		cpl_msg_debug(cpl_func,"Increasing number of stars in PPM; scale=%g angle=%g "
    				"nstars=%d npattern=%d nmatches=%d",scale,angle,nstars,npattern,*nmatches);

    		freearray(matches);
    		continue;
    	}
    	else
    		break;

    }
    *mratio = (double )(*nmatches / (double)nsources);

	if((matches == NULL) || (*nmatches < matchlimit)){
		cpl_msg_error(cpl_func,"Error in PPM module: Not enough matches found");
		freematrix(stars);
		freematrix(pattern);
		freetable(sources);
		freetable(stds);
		freeplist(xlist);
		freearray(matches);
		return NULL;
	}
	freematrix(stars);

	cpl_msg_info(cpl_func, "There are %d matches between USNOA2 and image", *nmatches);

    /* Get the RA and DEC from the pattern identifications */
    cpl_table_new_column(sources, "RA2000", CPL_TYPE_DOUBLE);
    cpl_table_new_column(sources, "Dec2000", CPL_TYPE_DOUBLE);
    cpl_table_unselect_all(stds);
    cpl_table_unselect_all(sources);
    for(i=0; i<cpl_matrix_get_ncol(pattern); i++){
    	int null;
    	int index = cpl_array_get_int(matches,i, &null);
    	if(!null){
    		double ra = cpl_table_get_double(stds, "RA2000", i, NULL);
    		double dec = cpl_table_get_double(stds, "Dec2000", i, NULL);
    		cpl_table_set(sources, "RA2000", index, ra);
    		cpl_table_set(sources, "Dec2000", index, dec);
    		cpl_table_select_row(sources, index);
    	}
    }

	freearray(matches);
	freematrix(pattern);

    /* Create two matrices for plate solution fitting */
    extracted = cpl_table_extract_selected(sources);
    rdmat = cpl_matrix_new(cpl_table_get_nrow(extracted), 2);
    xymat = cpl_matrix_new(cpl_table_get_nrow(extracted), 2);
    for (i=0; i<cpl_table_get_nrow(extracted); i++){
    	double ra = cpl_table_get_double(extracted, "RA2000", i, NULL);
    	double dec = cpl_table_get_double(extracted, "Dec2000", i, NULL);
    	cpl_matrix_set(rdmat, i, 0, ra);
    	cpl_matrix_set(rdmat, i, 1, dec);
    	float x = cpl_table_get_float(extracted, "X_IMAGE", i, NULL);
    	float y = cpl_table_get_float(extracted, "Y_IMAGE", i, NULL);
    	cpl_matrix_set(xymat, i, 0, x);
    	cpl_matrix_set(xymat, i, 1, y);
    }
    freetable(extracted);
   	freetable(stds);
 	freetable(sources);

    if(cpl_wcs_platesol(xlist, rdmat, xymat, niter, 3.0, fitmode,
    		CPL_WCS_MV_CRPIX, &aplist) != CPL_ERROR_NONE) {
    	cpl_msg_error(cpl_func,"Cannot calculate plate solution. %s", cpl_error_get_message());
    	freematrix(xymat);
    	freematrix(rdmat);
    	freeplist(xlist);
    	return NULL;
    }
    freematrix(xymat);
    freematrix(rdmat);
    freeplist(xlist);

    /* Return updated WCS in header */

	return aplist;

}

/*----------------------------------------------------------------

    This function will read the WCS information from a WFI header
    and correct it using CASU initwcsimage functions

------------------------------------------------------------------*/

void omega_wfi_correct_wcs(cpl_propertylist *mplist, cpl_propertylist *xplist)
{
    int i = 0;
    double xoff = 0.;
    double yoff = 0.;
    double tpa = 0.;
    double tpd = 0;
    double pixsca = 0.238;
    double rotskypa = 0.0;
    double angle = 0.0;
    double xapoff = 0.0;
    double yapoff = 0.0;
    double secd = 0.0;
    double aa = 0.0;
    double aap = 0.0;
    double bb = 0.0;
    double bbp = 0.0;
    double dd = 0.0;
    double ddp = 0.0;
    double eep = 0.0;
    double ee = 0.0;
    double cc = 0.0;
    double ff = 0.0;
    char key[80]= "";
    char ccdstr[80] = "";
    const char *ccd = NULL;

/*Get the CCD index*/
    if(cpl_propertylist_has(xplist, "ESO DET CHIP ID") == 1){
        ccd = cpl_propertylist_get_string(xplist, "ESO DET CHIP ID");
    }
    else {
    	for(i=1; i<=8; i++){
    		sprintf(key,"ESO DET CHIP%d ID", i);
    		if(cpl_propertylist_has(xplist,key)==1){
    			break;
    		}
    	}
    	sprintf(ccdstr,"ESO DET CHIP%d ID", i);
    	ccd = cpl_propertylist_get_string(xplist, ccdstr);
    }

    /*Identify wich CCD is in use and set the correct X Y offset */
    if(strcmp(ccd, "ccd50")==0){
        xoff = 4127.;
        yoff = -240.;
    }
    else if(strcmp(ccd, "ccd51")==0){
        xoff = 1990.;
        yoff = -233.;
    }
    else if(strcmp(ccd, "ccd52")==0){
        xoff = -154.;
        yoff = -233.;
    }
    else if(strcmp(ccd, "ccd53")==0){
        xoff = -2295.;
        yoff = -233.;
    }
    else if(strcmp(ccd, "ccd54")==0){
        xoff = -2294.;
        yoff = 3917.;
    }
    else if(strcmp(ccd, "ccd55")==0){
        xoff = -151.;
        yoff = 3919.;
    }
    else if(strcmp(ccd, "ccd56")==0){
        xoff = 1988.;
        yoff = 3920.;
    }
    else if(strcmp(ccd, "ccd57")==0){
        xoff = 4131.;
        yoff = 3912.;
    }

/* Get RA and DEC from main header*/
    tpa = cpl_propertylist_get_double(mplist, "RA");
    tpd = cpl_propertylist_get_double(mplist, "DEC");

    rotskypa = 180.0;
    angle = (rotskypa-180.0)*PI/180.0;
    xapoff = 0.0;
    yapoff = 0.0;

    /* tangent points of optical axis/rotator centre*/
    secd = 1.0/cos(tpd*PI/180.0);
    tpa = tpa - yapoff * secd * cos(angle) + xapoff *secd * sin(angle);
    tpd = tpd + yapoff * sin(angle) + xapoff * cos(angle);
    /*    printf (" %10.5f %10.5f %8.4f %6.3f %6.3f \n",tpa,tpd,angle,xapoff,yapoff);*/

    /* transformation matrix*/
    aap = -pixsca/3600.0;
    bbp = 0.0;
    ddp = 0.0;
    eep = pixsca/3600.0;

    /* apply rotation matrix to all constants to allow for different sky pa
    since n = A x ; x = R x'; o = R o'
    where A = aa bb; R = c -s;
              cc dd      s  c */

    aa = aap * cos(angle) + bbp * sin(angle);
    bb =-aap * sin(angle) + bbp * cos(angle);
    dd = ddp * cos(angle) + eep * sin(angle);
    ee =-ddp * sin(angle) + eep * cos(angle);

    cc = xoff;
    ff = yoff;
    /*    printf (" %10.5f %10.5f %10.5f %10.5f %10.5f %10.5f \n",aa,bb,cc,dd,ee,ff);*/


    /* Update extension header */
    cpl_propertylist_update_double(xplist,"CRPIX1",cc);
    cpl_propertylist_update_double(xplist,"CRPIX2",ff);
    cpl_propertylist_update_double(xplist,"CRVAL1",tpa);
    cpl_propertylist_update_double(xplist,"CRVAL2",tpd);
    cpl_propertylist_erase(xplist,"CDELT1");
    cpl_propertylist_erase(xplist,"CDELT2");
    cpl_propertylist_update_double(xplist,"CD1_1",aa);
    cpl_propertylist_update_double(xplist,"CD1_2",bb);
    cpl_propertylist_update_double(xplist,"CD2_1",dd);
    cpl_propertylist_update_double(xplist,"CD2_2",ee);
    /*    cpl_propertylist_update_double(xplist,"PROJP1",1.0);
    cpl_propertylist_update_double(xplist,"PROJP3",0.0);*/
    cpl_propertylist_update_string(xplist,"CTYPE1", "RA---TAN");
    cpl_propertylist_update_string(xplist,"CTYPE2","DEC--TAN");

}

/**@}*/
