/* $Id: omega_darkcurrent.c,v 1.5 2012/08/30 07:46:52 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/08/30 07:46:52 $
 * $Revision: 1.5 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "omega_recipe.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup omega_dark  Dark Current
 *
 * @par Synopsis:
 * Measure the dark current and the particle event rate. The raw dark frames are trimmed, overscan-corrected and \n
 * debiased. The dark current is calculated by median averaging three reduced dark frames and iteratively \n
 * rejecting outliers from the result, and computing the mean of the remaining pixels. The recipes expects \n
 * the following tags in the set of frames:
 *
 * @par Input frames:
 * DARK_CURRENT\n
 * MASTER_BIAS
 *
 * @par Output frames:
 * MEAN_DARK
 *
 * @code
 *      #include omega_recipe.h
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_darkcurrent_create(cpl_plugin *) ;
static int omega_darkcurrent_exec(cpl_plugin *) ;
static int omega_darkcurrent_destroy(cpl_plugin *) ;
static int omega_darkcurrent(cpl_frameset *set,cpl_parameterlist *pars) ;

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/
int omega_darkcurrent_combine(cpl_parameterlist *pars, int xn);
int omega_darkcurrent_save(cpl_image *image, cpl_frameset *set, cpl_parameterlist *parlist,
                           cpl_propertylist *qclist, cpl_type_bpp bitpix,
                           const char *outfile, cpl_frame *product_frame);
static void omega_darkcurrent_init(void);
static void omega_darkcurrent_tidy(void);

/*-----------------------------------------------------------------------------
                            Static structures
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs. Parameters */
/*	double			rej_sig;*/
    int         extnum;
    int         oc;
    int         paf;

    /* Outputs. QC parameters */
    double	dark_current;
    double	rate;

}omega_darkcurrent_config;

/* Input and Output */
static struct {
    cpl_size *labels;
    const cpl_frame *mbframe;
    cpl_frameset *darklist;
    omega_fits *firstdark;
    cpl_propertylist *eh;

    /* Products */
    cpl_image *mdark;

} ps;


/* Static variables */
#define RECIPE "omega_darkcurrent"

static int isfirst;
static int dummy;


/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    OMEGA_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "omega_darkcurrent",
                    "OMEGA - Measure the dark current and the particle event rate.(Calfile 531).",
                    "The raw dark frames are trimmed, overscan-corrected and"
                    "debiased.  The dark current is calculated by median averaging"
                    "three reduced dark frames and iteratively rejecting outliers"
                    "from the result, and computing the mean of the remaining pixels.",
                    "Sandra Castro",
                    "scastro@eso.org",
                    omega_get_license(),
                    omega_darkcurrent_create,
                    omega_darkcurrent_exec,
                    omega_darkcurrent_destroy) ;

    cpl_pluginlist_append(list, plugin) ;

    return 0;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief    Setup the recipe options
 * @param    plugin  the plugin
 * @return   0 if everything is ok
 *
 * Create the recipe instance and make it available to the application using the
 * interface. The recipe parameters are defined here.
 */
/*----------------------------------------------------------------------------*/
static int omega_darkcurrent_create(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    cpl_parameter * p ;
    char *path = NULL;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new() ;
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
   p = cpl_parameter_new_value("omega.omega_darkcurrent.ExtensionNumber",
                                CPL_TYPE_INT,
                                "FITS extension number to load (1 to 32). (-1 = all)",
                                "omega_darkcurrent",
                                -1) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    p = cpl_parameter_new_range("omega.omega_darkcurrent.OverscanMethod",
                                CPL_TYPE_INT,
                                "Overscan Correction Method",
                                "omega_darkcurrent",
                                6, 0, 6);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_darkcurrent.PAF",
                                CPL_TYPE_BOOL,
                                "Boolean value to create PAF files. 1(Yes), 0(No)",
                                "omega_darkcurrent",
                                1) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "paf") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_darkcurrent.RejThre",
                                CPL_TYPE_DOUBLE,
                                "The threshold rejecting outlying pixels",
                                "omega_darkcurrent",
                                5.0) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"rej-thre") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_darkcurrent.DetThre",
                                CPL_TYPE_DOUBLE,
                                "The detection threshold for cosmic ray event ",
                                "omega_darkcurrent",
                                6.0) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"det-thre") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    p = cpl_parameter_new_value("omega.omega_darkcurrent.MaxIter",
                                CPL_TYPE_INT,
                                "The maximum number of iterations",
                                "omega_darkcurrent",
                                3) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"niter") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_darkcurrent.MaxCurrent",
                                CPL_TYPE_DOUBLE,
                                "The maximum dark current in ADU/pixel/hour",
                                "omega_darkcurrent",
                                5.0) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "maxcur") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    p = cpl_parameter_new_value("omega.omega_darkcurrent.MaxDiffCurrent",
                                CPL_TYPE_DOUBLE,
                                "The maximum difference of dark current between previous ??",
                                "omega_darkcurrent",
                                0.2) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "maxdifcur") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

/* ------------------ Sextractor parameters ------------------------- */
    path = cpl_sprintf("%s", OMEGA_BIN_PATH);
    p = cpl_parameter_new_value("omega.omega_darkcurrent.BinPath",
                                CPL_TYPE_STRING,
                                "Path to any external executable program.",
                                "omega.BinPath",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "bin-path");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    path = cpl_sprintf("%s/omega.cosmic.sex", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_darkcurrent.SexCosmic",
                                CPL_TYPE_STRING,
                                "Path to Sextractor cosmic detection mode config file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-cosmic");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    path = cpl_sprintf("%s/omega.conv", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_darkcurrent.SexConv",
                                CPL_TYPE_STRING,
                                "Path to Sextractor convolution mask file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-conv");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    path = cpl_sprintf("%s/omega.param", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_darkcurrent.SexParam",
                                CPL_TYPE_STRING,
                                "Path to Sextractor parameters file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"sex-param");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    path = cpl_sprintf("%s/omega.nnw", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_darkcurrent.SexNnw",
                                CPL_TYPE_STRING,
                                "Path to Sextractor neural network config file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-nnw");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);


    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_darkcurrent_exec(cpl_plugin * plugin)
{

    cpl_recipe * recipe;
    int recipe_status;

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

     /* Invoke the recipe */
    recipe_status = omega_darkcurrent(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }


   return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_darkcurrent_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0 ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here
  @param    set         the frames list
  @param    pars        the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_darkcurrent(cpl_frameset *set, cpl_parameterlist *pars)
{

    int j,jst,jfn;
    cpl_size nlab;
    int ndark = 0;
    int oscan1 = 0;
    int status = 0;
    float bias = BIAS;
    char *outmdark = NULL;
    const char *_id = "omega_darkcurrent";

    cpl_parameter    *par = NULL;
    cpl_frame        *firstframe = NULL;
    cpl_frame        *prframe_mdark = NULL;
    cpl_propertylist *qclist = NULL;
    cpl_propertylist *alist = NULL;



 /*Start the recipe */

	if (pars == NULL) {
		cpl_msg_error (_id, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (_id, "Frameset not found");
		return -1;
	}

/* Retrieve input parameters */
    par = cpl_parameterlist_find(pars, "omega.omega_darkcurrent.ExtensionNumber") ;
    omega_darkcurrent_config.extnum = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_darkcurrent.OverscanMethod") ;
    omega_darkcurrent_config.oc = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_darkcurrent.PAF");
    omega_darkcurrent_config.paf = cpl_parameter_get_bool(par);



/* Identify the RAW and CALIB frames in the input frameset */
    if (oc_dfs_set_groups(set)) {
	cpl_msg_error(_id, "Cannot identify RAW and CALIB frames") ;
	return -1 ;
    }

    /*Initialized things*/
    omega_darkcurrent_init();

/* Verify the frameset contents */
    if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
					   &nlab)) == NULL) {
        cpl_msg_error(_id,"Cannot labelise the input frameset");
        omega_darkcurrent_tidy();
        return -1;
    }
    if ((ps.darklist = omega_frameset_subgroup(set,ps.labels,nlab,
						DARK_RAW)) == NULL) {
        cpl_msg_error(_id,"Cannot find bias frames in input frameset");
        omega_darkcurrent_tidy();
        return -1;
    }

    ndark = cpl_frameset_count_tags (set, DARK_RAW);
    if (ndark < 3) {
	cpl_msg_error (_id, "Need at least 3 (%s) frames to run this recipe", DARK_RAW);
	omega_darkcurrent_tidy();
	return -1;
    }

    cpl_msg_info (_id,"There are %d %s frames in frame set",ndark, DARK_RAW);

    /* Get first frame from frame set */
    firstframe = cpl_frameset_get_first(ps.darklist);


    /* Check for calibration frames */
    /* Master Bias */
    ps.mbframe = cpl_frameset_find_const(set, OMEGA_CALIB_BIAS);
    if (ps.mbframe == NULL) {
        cpl_msg_info(_id,"A Master Bias is not present in frame set. Using default value %f", bias);
    }
    else{
        cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(ps.mbframe));
    }

    /* Loop for each of the image extensions */
    omega_exten_range(omega_darkcurrent_config.extnum,&jst,&jfn);
    if(omega_darkcurrent_config.extnum == 0){
    	cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_darkcurrent_config.extnum);
    	omega_darkcurrent_tidy();
    	return -1;
    }

    /* Create output names for the products */
    for (j = jst; j <= jfn; j++) {

        isfirst = (j == jst);
        cpl_msg_info(_id,"Beginning work on extension %d",j);
        omega_darkcurrent_config.dark_current = 0.0;
        omega_darkcurrent_config.rate = 0;

        /* Check overscan correction method consistency if oc mode == 0*/

        if(omega_darkcurrent_config.oc==0){
        	if(ps.mbframe != NULL){
        		oscan1 = omega_pfits_get_overscan(ps.mbframe, j);
        		if(oscan1 != omega_darkcurrent_config.oc) {
        			cpl_msg_warning (_id, "Overscan correction mode for Master Bias (oc = %d) differs from "
        					"the one used here (oc = %d)", oscan1, omega_darkcurrent_config.oc);
        		}
        	}
        }

        ps.firstdark = omega_fits_load(firstframe,CPL_TYPE_FLOAT,j);
        ps.eh = omega_fits_get_ehu(ps.firstdark);

        /* Call the combining routine */
        status = omega_darkcurrent_combine(pars, j);
        if(status == 1){
            cpl_msg_warning(_id, "Image detector is not live");
            /* Save dummy product */
            freefits(ps.firstdark);
            ps.eh = NULL;
            continue;
        }
        else if(status == -1){
            cpl_msg_error(_id,"Cannot combine images");
            freespace(outmdark);
            omega_darkcurrent_tidy();
            return -1;
        }

        status = 0;

        /* Save the products */
        /* Save the MEAN DARK product */
        qclist = cpl_propertylist_new();
        cpl_propertylist_append_double(qclist, "ESO QC DARK CURRENT",
        		omega_darkcurrent_config.dark_current) ;
        cpl_propertylist_set_comment(qclist, "ESO QC DARK CURRENT",
        		"Dark current") ;
        cpl_propertylist_append_double(qclist, "ESO QC PARTICLE RATE",
        		omega_darkcurrent_config.rate) ;
        cpl_propertylist_set_comment(qclist, "ESO QC PARTICLE RATE",
					"Value of particle rate" ) ;

		if(isfirst){
			outmdark = cpl_sprintf("%s_%s.fits", INSTRUME,DARK_PROCATG);
			prframe_mdark = omega_product_frame(outmdark, DARK_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

	       alist = cpl_propertylist_new();
	        cpl_propertylist_append_string(alist, "EXTNAME",
						cpl_propertylist_get_string(ps.eh, "EXTNAME"));
				cpl_propertylist_set_comment(alist,"EXTNAME", "Extension name");

				cpl_propertylist_copy_property_regexp(alist, ps.eh, WCS_KEYS, 0);

		if(omega_save_image(ps.mdark,set,pars,alist,qclist,CPL_BPP_IEEE_FLOAT,outmdark,
					RECIPE,prframe_mdark,NULL,isfirst) == -1){
            cpl_msg_error(_id,"Cannot save product %s", DARK_PROCATG);
            freeplist(qclist);
            freeplist(alist);
            freespace(outmdark);
            omega_darkcurrent_tidy();
            return -1;;
		}

        freeplist(qclist);
        freeplist(alist);
        freeimage(ps.mdark);
        freefits(ps.firstdark);
        ps.firstdark = NULL;
        ps.eh = NULL;

    }  /* work on next extension */


    freespace(outmdark);
    omega_darkcurrent_tidy();


	return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Trim and combine input images

  @param    pars     the input list of parameters
  @param    xn       extension number

  @return   0 if everything is ok, 1 if this detector is dead and -1 otherwise
 */
/*----------------------------------------------------------------------------*/
int omega_darkcurrent_combine(cpl_parameterlist *pars, int xn)
{

    int count = 0;
    int i = 0;
    int maxiter = 0;
    int ndark = 0;
    int live = 0;
    float bias = BIAS;
    double dethres = 3.0;
    double exptime = 0.0;
    double rate = 0.0;
    double rejt = 5.0;
    double detector_size_x=1.;
    double detector_size_y=1.;
    double detector_size=1.;

    const char *cmd = NULL;
    const char *cosmiccat = "OMEGA_cosmicrays.fits";
    const char *cosmic_conf = NULL;
    const char *_id = "";
    const char *path = NULL;
    const char *sex_conv = NULL;
    const char *sex_nnw = NULL;
    const char *sex_par = NULL;
    const char *temp1739 = "OMEGA_temp001739.fits";

    const cpl_frame *darkfr = NULL;
    cpl_image *mbias = NULL, *trim_raw = NULL;
    cpl_imagelist *ilist = NULL;
    cpl_stats *stats = NULL;
    cpl_parameter *par = NULL;
    cpl_propertylist *mplist = NULL, *xplist = NULL;


    par = cpl_parameterlist_find(pars, "omega.omega_darkcurrent.MaxIter") ;
    maxiter = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_darkcurrent.RejThre") ;
    rejt = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_darkcurrent.DetThre") ;
    dethres = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_darkcurrent.BinPath");
    path = cpl_parameter_get_string(par) ;

    par = cpl_parameterlist_find(pars,"omega.omega_darkcurrent.SexCosmic" );
    cosmic_conf = cpl_parameter_get_string(par) ;

    par = cpl_parameterlist_find(pars,"omega.omega_darkcurrent.SexConv" );
    sex_conv = cpl_parameter_get_string(par) ;

    par = cpl_parameterlist_find(pars,"omega.omega_darkcurrent.SexParam" );
    sex_par = cpl_parameter_get_string(par) ;

    par = cpl_parameterlist_find(pars,"omega.omega_darkcurrent.SexNnw" );
    sex_nnw = cpl_parameter_get_string(par) ;

    ndark = cpl_frameset_get_size(ps.darklist);
    darkfr = cpl_frameset_get_first_const(ps.darklist);

    /* Check that this detector is live*/
    omega_pfits_get_detlive(ps.eh,&live);
    if (! live) {
        return 1;
    }

    /* Load master bias image */
    if(ps.mbframe != NULL){
        mbias = cpl_image_load(cpl_frame_get_filename(ps.mbframe), CPL_TYPE_FLOAT,0,xn);
        if (mbias == NULL) {
            cpl_msg_warning(_id,"Cannot load image %s", OMEGA_CALIB_BIAS);
        }
    }

    ilist = cpl_imagelist_new();
    cpl_msg_info (_id,"Doing trim and overscan correction on images");

    /* Loop through all images in frame list */
    /* Loop through raw dark list and trim each image*/

    for (i=0; i< ndark; i++)
    {

    	trim_raw = TrimOscanCorrect(darkfr, omega_darkcurrent_config.oc, xn);
    	if(trim_raw == NULL){
    		cpl_msg_error(_id,"Cannot trim image");
    		freeilist(ilist);
    		return -1;
    	}

    	if(omega_darkcurrent_config.oc==0){
    		if (mbias != NULL){
    			cpl_msg_info(cpl_func,"Subtracting bias frame");
    			cpl_image_subtract(trim_raw, mbias);
    		}
    		else
    			cpl_image_subtract_scalar(trim_raw, bias);
    	}

        cpl_imagelist_set(ilist, trim_raw, i);
        darkfr = cpl_frameset_get_next_const(ps.darklist);
        if (darkfr == NULL)
            break;

    }

    freeimage(mbias);
    if (ilist == NULL) {
        cpl_msg_error(_id,"Error in image list <%s>",cpl_error_get_message());
        freeimage(trim_raw);
        return -1;
    }

    cpl_msg_info (_id,"Computing the median of all images...");
    ps.mdark = cpl_imagelist_collapse_median_create(ilist);

    if (ps.mdark == NULL) {
        cpl_msg_error (_id,"Cannot take median of list <%s>",cpl_error_get_message());
        freeilist(ilist);
        return -1;
    }

    cpl_msg_info(_id,"Estimating the RMS using %d maximum iterations",maxiter);
    cpl_msg_info(_id,"Rejection threshold is %g (sigma)",rejt);


    /*Calculate statistics iteratively*/
    stats = omega_iter_stat(ps.mdark,rejt,maxiter);
    if(stats == NULL){
	cpl_msg_warning(_id,"Cannot calculate statistics iteratively");
    }

    darkfr = cpl_frameset_get_first(ps.darklist);
    mplist = cpl_propertylist_load_regexp(cpl_frame_get_filename(darkfr), 0, "EXPTIME", 0);
    xplist = cpl_propertylist_load_regexp(cpl_frame_get_filename(darkfr), xn, "EXPTIME", 0);
    exptime = omega_pfits_get_exptime(xplist);
    if (exptime == 0.0){
        exptime = omega_pfits_get_exptime(mplist);
    }

    omega_darkcurrent_config.dark_current = cpl_stats_get_mean(stats) / exptime*3600.0;
    freeplist(mplist);
    freeplist(xplist);

    freestats(stats);

    /*Calculate the particle event rate*/
    rate = 0.0;
    for(i=0; i < ndark; i++){

	    cpl_image_save(cpl_imagelist_get(ilist, i),temp1739,BITPIX,NULL,CPL_IO_DEFAULT);

        /*
         *  Run Sextractor on a cosmic detection mode
         */

        cmd = cpl_sprintf("%s/sex %s -c %s -PARAMETERS_NAME %s -FILTER_NAME %s -STARNNW_NAME %s "
                          "-DETECT_THRESH %g "
                          "-CHECKIMAGE_NAME chkimage.fits "
                          "-CATALOG_TYPE FITS_1.0 "
                          "-CATALOG_NAME %s",
                          path,
                          temp1739,
                          cosmic_conf,
                          sex_par,
                          sex_conv,
                          sex_nnw,
                          dethres,
                          cosmiccat);

        if (system(cmd) != 0) {
            cpl_free((char *)cmd);
	        freeilist(ilist);
            cpl_msg_error("","%s Failed to run Sextractor", _id);
            return -1;
        }

        cpl_free((char *)cmd);

        mplist = cpl_propertylist_load_regexp(cosmiccat, 1, "NAXIS2", 0);
        if(mplist == NULL){
            cpl_msg_error(_id,"Cannot load header for %s",cosmiccat);
	        freeilist(ilist);
	        return -1;
	    }

	    count = cpl_propertylist_get_int(mplist,"NAXIS2");
 /*       cpl_msg_info(_id,"Detected %d cosmic ray events",count);*/

	    rate += count;
	    freeplist(mplist);
    }
    /*Detector size in cm^2 assuming 15mue/pixel */

    detector_size_x=cpl_image_get_size_x(cpl_imagelist_get(ilist,0));
    detector_size_y=cpl_image_get_size_y(cpl_imagelist_get(ilist,0));
    detector_size=detector_size_x * detector_size_y * 0.0015 * 0.0015;

    rate  = rate/ndark;
    rate  = rate/exptime*3600.0;
    if(detector_size>0){
    	rate  = rate/detector_size;
    }
    omega_darkcurrent_config.rate = rate;

    freeilist(ilist);


    return 0;
}

/**
  @brief    Save the omega_darkcurrent recipe products on disk

  @param    image       reduced image to be saved
  @param    set         the input frame set
  @param    parlist     the input list of parameters
  @param    qclist      QC property list
  @param    bitpix      cpl_type_bpp
  @param    procatg     PRO CATG of product
  @param    outfile     output name (without .suffix)

  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
/**
 * @brief
 *    Save the omega_darkcurrent products
 * @param image          Reduced image to be saved
 * @param set            The input frame set
 * @param parlist        The input list of parameters
 * @param qclist         QC property list
 * @param bitpix         CPL_BPP to save product
 * @param outfile        Name of output
 * @param product_frame  Frame describing the product
 *
 * @return
 *   0 on success or -1 in case of error.
 */
int omega_darkcurrent_save(cpl_image *image, cpl_frameset *set, cpl_parameterlist *parlist,
                           cpl_propertylist *qclist, cpl_type_bpp bitpix,
                           const char *outfile, cpl_frame *product_frame)

{

    const char *fctid = "omega_darkcurrent_save";
    cpl_propertylist *plist = NULL;

    /* If it is the first time, setup initial frame */
    if (isfirst) {

        /* Create a new product frame object and define some tags */
        plist = cpl_propertylist_new();

        /* Add DataFlow keywords in primary header */
        if (cpl_dfs_setup_product_header(plist, product_frame, set, parlist,
                                        RECIPE,PIPEID,DICID,NULL) != CPL_ERROR_NONE) {

            cpl_msg_warning(fctid, "Problem in the main header of product DFS-compliance") ;
        }

        /* Once CPL recalculates these values, the following
         * line should be removed
         */
        cpl_propertylist_erase_regexp(plist,REM_PRIM_KEYS,0);

        /* Save the empty primary unit */
        if (cpl_propertylist_save(plist,outfile,CPL_IO_DEFAULT) != CPL_ERROR_NONE){
            cpl_msg_error(fctid,"Cannot save product PHU");
            freeplist(plist);
            return 1;
        }

        freeplist(plist);
        cpl_frameset_insert(set,product_frame);
    }

    /* Setup the extension */
    plist = cpl_propertylist_duplicate(ps.eh);


    /* Add DRS keywords */
    cpl_propertylist_append_int(plist, "ESO DRS OVERSCAN METHOD", omega_darkcurrent_config.oc);
    cpl_propertylist_set_comment(plist, "ESO DRS OVERSCAN METHOD", "overscan correction method");

    /* Add DataFlow keywords in extension header */
    if (cpl_dfs_setup_product_header(plist, product_frame, set, parlist,
                                         RECIPE,PIPEID,DICID,NULL) != CPL_ERROR_NONE) {

        cpl_msg_warning(fctid, "Problem in the extension header of product DFS-compliance") ;
    }

    /*Remove undesired keywords from extension header*/
    cpl_propertylist_erase_regexp(plist, REM_EXT_KEYS, 0);


    /* Append QC list to the extension header */
    cpl_propertylist_append(plist, qclist);

     /* Save the product */
     if (cpl_image_save(image,outfile,bitpix,plist,
                           CPL_IO_EXTEND) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product. %s", cpl_error_get_message());
        freeplist(plist);
        return -1;
    }


    /* Save the paf file */
    if(omega_darkcurrent_config.paf){
        int xtnum = omega_fits_get_extnum(ps.firstdark);
        char *prefix = omega_get_root_name(outfile);
        char *outpaf = cpl_sprintf("%s_%d.paf",prefix,xtnum);
        const char pafcopy[] = "^(DATE-OBS|ARCFILE|ESO TPL ID|ESO DET WIN1 DIT1|MJD-OBS|EXTNAME|ESO PRO CATG)$";
        cpl_frame *fframe = cpl_frameset_get_first(set);
        cpl_propertylist *mlist = cpl_propertylist_load_regexp(cpl_frame_get_filename(fframe),0,pafcopy, 0);
        cpl_propertylist_copy_property_regexp(qclist, mlist, pafcopy, 0);
        if (cpl_propertylist_copy_property_regexp(qclist, plist,
                                              pafcopy, 0) != CPL_ERROR_NONE){

           cpl_msg_warning(fctid,"Some mandatory keywords are missing in PAF file");
        }

        cpl_dfs_save_paf(INSTRUME, RECIPE, qclist, outpaf);
        cpl_free(outpaf);
        freeplist(mlist);
    }

    freeplist(plist);

    return 0;
}

/* Initialize the pointers */
static void omega_darkcurrent_init(void) {
    ps.labels = NULL;
    ps.darklist = NULL;
    ps.mdark = NULL;
    ps.mbframe = NULL;
    ps.eh = NULL;
    ps.firstdark = NULL;
}

/* Free any allocated memory */
static void omega_darkcurrent_tidy(void) {
    freespace(ps.labels);
    freeframeset(ps.darklist);
    freeimage(ps.mdark);
    freefits(ps.firstdark);
}

/**@}*/
