/* $Id: omega_fringes_flat.c,v 1.2 2012/01/12 12:05:09 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/12 12:05:09 $
 * $Revision: 1.2 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "omega_recipe.h"
#include "omega_wcscor.h"

/**
 * @defgroup omega_fringes_flat  Master Fringes Flat
 *
 * @par Synopsis:
 * This recipe is used to create a Fringes Flat if the filter requires it. The recipe always takes as input \n
 * a list of at least 3 science frames and reduced them, by trimming, correcting for the bias and the flat. \n
 * The recipes expects the following tags in the Set of Frames:
 *
 * @par Input frames:\n
 * FRINGES\n
 * MASTER_BIAS\n
 * MASTER_FLAT\n
 * BAD_PIXELS_MAP\n
 *
 * @par Output frames:\n
 * FRINGES_FLAT
 *
 *
 * @code
 *      #include omega_recipe.h
 *      #include omega_wcscor.h
 * @endcode
 */
/**@{*/
/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_fringes_flat_create(cpl_plugin *) ;
static int omega_fringes_flat_exec(cpl_plugin *) ;
static int omega_fringes_flat_destroy(cpl_plugin *) ;
static int omega_fringes_flat(cpl_frameset *, cpl_parameterlist *) ;
cpl_image *omega_fringes_flat_process(omega_fits *scifits, cpl_image *bias, cpl_image *flat,
		cpl_image *bpm, cpl_parameterlist *pars, int ext);

static int omega_fringes_load_calib(int ext);
static void omega_fringes_init(void);
static void omega_fringes_tidy(int level);


/* Input and Output structures */
static struct {
	/* Inputs. Parameters */
	int         extnum;
	int         oc;

}omega_fringes_config;

static struct {

	/* Calib frames */
	cpl_size         *labels;
	const cpl_frame  *mbiasfr;
	const cpl_frame  *mflatfr;
	const cpl_frame  *bpmfr;
	cpl_frameset     *sciset;
	omega_fits        *scifits;

	/* Products */
	cpl_image        *bpm;
	cpl_image        *mbias;
	cpl_image        *mflat;
	cpl_image        *fringes;

}ps;

#define RECIPE "omega_fringes_flat"

/*----------------------------------------------------------------------------*/
/**
@internal
@brief    Build the list of available plugins, for this module.
@param    list    the plugin list
@return   0 if everything is ok

This function is exported.
*/
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
	cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
	cpl_plugin  *   plugin = &recipe->interface ;

	cpl_plugin_init(plugin,
			CPL_PLUGIN_API,
			OMEGA_BINARY_VERSION,
			CPL_PLUGIN_TYPE_RECIPE,
			"omega_fringes_flat",
			"OMEGA - Create Fringes Flat.",
			"This recipe is used to create a Fringes Flat if the filter \n"
			"requires it. The recipe always takes as input a list of at least 3 \n"
			"science frames, a master bias and a master flat. In addition to these, \n"
			"an optional Bad Pixels Map may be provided.",
			"Sandra Castro",
			"scastro@eso.org",
			omega_get_license(),
			omega_fringes_flat_create,
			omega_fringes_flat_exec,
			omega_fringes_flat_destroy) ;

	cpl_pluginlist_append(list, plugin) ;

	return 0;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Setup the recipe options
@param    plugin  the plugin
@return   0 if everything is ok

Create the recipe instance and make it available to the application using the
interface.
*/
/*----------------------------------------------------------------------------*/
static int omega_fringes_flat_create(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	cpl_parameter * p ;

	/* Do not create the recipe if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Create the parameters list in the cpl_recipe object */
	recipe->parameters = cpl_parameterlist_new() ;
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Parameter list allocation failed");
		cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
	}

	/* Fill the parameters list */
	p = cpl_parameter_new_value("omega.omega_fringes_flat.ExtensionNumber",
			CPL_TYPE_INT,
			"FITS extension number to load (1 to 32). (-1 == all)",
			"omega_fringes_flat",
			-1) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_range("omega.omega_fringes_flat.OverscanMethod",
			CPL_TYPE_INT,
			"Overscan Correction Method",
			"omega_fringes_flat",
			0, 0, 6);
	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "oc-meth") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	/* Return */
	return 0;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Execute the plugin instance given by the interface
@param    plugin  the plugin
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_fringes_flat_exec(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	int recipe_status;

	/* Return immediately if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Verify parameter and frame lists */
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}
	if (recipe->frames == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Invoke the recipe */
	recipe_status = omega_fringes_flat(recipe->frames, recipe->parameters);

	/* Ensure DFS-compliance of the products */
	if (cpl_dfs_update_product_header(recipe->frames)) {
		if (!recipe_status) recipe_status = (int)cpl_error_get_code();
	}

	return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Destroy what has been created by the 'create' function
@param    plugin  the plugin
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_fringes_flat_destroy(cpl_plugin * plugin)
{
	cpl_recipe  *   recipe;

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	cpl_parameterlist_delete(recipe->parameters) ;

	return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
@brief    The recipe data reduction part is implemented here
@param    parlist     the parameters list
@param    framelist   the frames list
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_fringes_flat(cpl_frameset *set, cpl_parameterlist *pars)
{

	int i,j,jst,jfn,isfirst;
	cpl_size nlab;
	int nraw = 0;
	char *outfile = NULL;

	cpl_frame *sciframe, *product_frame;
	cpl_parameter *par;
	cpl_image *image;
	cpl_imagelist *ilist;
	cpl_propertylist *qclist;
	cpl_stats *stats;


	/*Start the recipe */
	if (!pars) {
		cpl_msg_error (cpl_func, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (cpl_func, "Frameset not found");
		return -1;
	}

	/* Retrieve input parameters */
	par = cpl_parameterlist_find(pars, "omega.omega_fringes_flat.ExtensionNumber") ;
	omega_fringes_config.extnum = cpl_parameter_get_int(par) ;

	/* Identify the RAW and CALIB frames in the input frameset */
	if (oc_dfs_set_groups(set)) {
		cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames") ;
		return -1 ;
	}

	/*Initialized things*/
	omega_fringes_init();

	/* Verify the frameset contents. */
	if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
			&nlab)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot labelise the input frameset");
		omega_fringes_tidy(0);
		return -1;
	}
	if ((ps.sciset = omega_frameset_subgroup(set,ps.labels,nlab,
			FRINGES_RAW)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot find fringes frames in input frameset");
		omega_fringes_tidy(0);
		return -1;
	}


	/* Verify the frameset contents. */
	nraw = cpl_frameset_count_tags(ps.sciset, FRINGES_RAW);
	if (nraw < 3) {
		cpl_msg_error (cpl_func, "Need at least 3 (%s) frames to run "
				"this recipe", FRINGES_RAW);
		omega_fringes_tidy(0);
		return -1;
	}

	cpl_msg_info (cpl_func,"There are %d %s frames in frame set",nraw, FRINGES_RAW);

	/* Check for calibration frames */
	/* Master Bias */
	ps.mbiasfr = cpl_frameset_find_const(set, OMEGA_CALIB_BIAS);
	if (ps.mbiasfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set", OMEGA_CALIB_BIAS);
		omega_fringes_tidy(0);
	}
	cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(ps.mbiasfr));

	/* Master Flat */
	ps.mflatfr = cpl_frameset_find_const(set, OMEGA_CALIB_FLAT);
	if(ps.mflatfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set", OMEGA_CALIB_FLAT);
		omega_fringes_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s", OMEGA_CALIB_FLAT, cpl_frame_get_filename(ps.mflatfr));

	/* Bad Pixels Map */
	ps.bpmfr = cpl_frameset_find (set,OMEGA_CALIB_BPM);
	if(ps.bpmfr != NULL)
		cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BPM,cpl_frame_get_filename(ps.bpmfr));

	/* Get first frame from frame set */
	sciframe = cpl_frameset_get_first(ps.sciset);

	/* Loop for each of the image extensions */
	omega_extensions(sciframe, omega_fringes_config.extnum,&jst,&jfn);
	if(omega_fringes_config.extnum == 0){
		cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_fringes_config.extnum);
		omega_fringes_tidy(0);
		return -1;
	}

	for (j = jst; j <= jfn; j++) {
		isfirst = (j == jst);
		cpl_msg_indent_more();
		cpl_msg_info(cpl_func,".....Working on extension %d.....",j);
		cpl_msg_indent_less();

		/* Load calibration frames */
		if(omega_fringes_load_calib(j) != 0){
			cpl_msg_error(cpl_func,"Cannot load calibration frame(s)");
			freespace(outfile);
			omega_fringes_tidy(0);
			return -1;
		}

		ilist = cpl_imagelist_new();

		sciframe = cpl_frameset_get_first(ps.sciset);
		for(i = 0; i < nraw; i++) {
			/* Load omega_fits structure for this extension */
			ps.scifits = omega_fits_load(sciframe, CPL_TYPE_FLOAT, j);

			/* Call processing routine */
			image = omega_fringes_flat_process(ps.scifits,ps.mbias,ps.mflat,ps.bpm, pars, j);
			if(image == NULL){
				cpl_msg_error(cpl_func,"Cannot reduce image %d of %d",i,nraw);
				freespace(outfile);
				freeilist(ilist);
				omega_fringes_tidy(0);
				return -1;
			}

			cpl_imagelist_set(ilist, image,i);
			sciframe = cpl_frameset_get_next(ps.sciset);
			freefits(ps.scifits);
			ps.scifits = NULL;

		}
		ps.fringes = omega_fringes_create(ilist,ps.bpm,nraw);
		freeilist(ilist);

		if(isfirst){
			outfile = cpl_sprintf("%s_%s.fits", INSTRUME,FRINGES_PROCATG);
			product_frame = omega_product_frame(outfile, FRINGES_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

		stats = cpl_stats_new_from_image(ps.fringes, CPL_STATS_ALL);
		qclist = cpl_propertylist_new();
		cpl_propertylist_append_double(qclist, "HIERARCH ESO QC FRINGESMIN", cpl_stats_get_min(stats));
		cpl_propertylist_append_double(qclist, "HIERARCH ESO QC FRINGESMAX", cpl_stats_get_max(stats));
		cpl_propertylist_append_double(qclist, "HIERARCH ESO QC FRINGESMEAN", cpl_stats_get_mean(stats));
		cpl_propertylist_append_double(qclist, "HIERARCH ESO QC FRINGESDEV", cpl_stats_get_stdev(stats));
		cpl_propertylist_append_double(qclist, "HIERARCH ESO QC FRINGESMED", cpl_stats_get_median(stats));
		freestats(stats);

		cpl_msg_info(cpl_func,"Saving Fringes Flat");
		/* Save Fringes Flat */
		if(omega_save_image(ps.fringes,set,pars,NULL,qclist,CPL_BPP_IEEE_FLOAT,outfile,
					RECIPE,product_frame,NULL,isfirst) == -1){
			cpl_msg_error(cpl_func,"Cannot save product %s",FRINGES_PROCATG);
			freeplist(qclist);
			freespace(outfile);
			omega_fringes_tidy(0);
			return -1;
		}

		freeplist(qclist);
		omega_fringes_tidy(1);
	}

	freespace(outfile);
	omega_fringes_tidy(0);

	return 0;
}

/**
* @brief			Reduce a science image
*
* @param scifits	omega_fits structure of science image
* @param mbias		master bias
* @param mflat		master flat
* @param bpm		bad pixels image
* @param pars		parameters list
* @param ext		FITS extension to laod
*
* @return fringes		Reduced science image
*/
cpl_image *omega_fringes_flat_process(omega_fits *scifits, cpl_image *mbias, cpl_image *mflat,
		cpl_image *bpm, cpl_parameterlist *pars, int ext)
{
	int oc = 0;

	cpl_image *fringes;
	cpl_parameter *p;
	cpl_propertylist *xlist;

	if((scifits == NULL) || (mbias == NULL) || (mflat == NULL) ||
			(bpm == NULL))
		return NULL;

	/* Get a few parameters */
	p = cpl_parameterlist_find(pars, "omega.omega_fringes_flat.OverscanMethod") ;
	omega_fringes_config.oc = cpl_parameter_get_int(p);

	/* Check overscan correction method consistency */
	oc = omega_pfits_get_overscan(ps.mbiasfr, ext);
	if(oc != omega_fringes_config.oc) {
		cpl_msg_warning (cpl_func, "Overscan correction mode for Master Bias (oc = %d) differs from "
				"the one used here (oc = %d)", oc, omega_fringes_config.oc);
	}

	/* Trim and overscan correct all images */
	if((fringes = omega_trim_oscan_correct(scifits, omega_fringes_config.oc)) == NULL){
		cpl_msg_error(cpl_func,"Unable to trim image");
		return NULL;
	}

	/* Correct coordinates after trimming */
	xlist = omega_fits_get_ehu(scifits);
	omega_shift_refpix(omega_fits_get_frame(scifits), ext, xlist);

	/* Subtract the bias */
	omega_biascor(fringes, mbias);

	/* Divide by the flat */
	if(omega_flatcor(fringes, mflat, NULL) != 0){
		cpl_msg_error(cpl_func,"Error in flat correction");
		freeimage(fringes);
		xlist = NULL;
		return NULL;
	}

	xlist = NULL;

	return fringes;
}

static int omega_fringes_load_calib(int ext)
{

	/* Load mandatory calibration frames */
	/* master bias */
	ps.mbias = cpl_image_load(cpl_frame_get_filename(ps.mbiasfr), CPL_TYPE_FLOAT, 0, ext);
	if(ps.mbias == NULL){
		cpl_msg_error(cpl_func,"Cannot load Master Bias. %s", cpl_error_get_message());
		return -1;
	}
	/* master flat */
	ps.mflat = cpl_image_load(cpl_frame_get_filename(ps.mflatfr),CPL_TYPE_FLOAT,0,ext);
	if (ps.mflat == NULL){
		cpl_msg_error(cpl_func,"Cannot load Master Flat. %s",cpl_error_get_message());
		return -1;
	}

	/* Optional */
	/* bad pixels map */
	if(ps.bpmfr != NULL){
		ps.bpm = cpl_image_load(cpl_frame_get_filename(ps.bpmfr), CPL_TYPE_INT, 0, ext);
		if(ps.bpm == NULL)
			cpl_msg_warning(cpl_func, "Cannot load BPM. %s", cpl_error_get_message());

	}

	return 0;
}

/* Initialize the pointers */
static void omega_fringes_init(void)
{
	ps.sciset = NULL;
	ps.bpmfr = NULL;
	ps.mbiasfr = NULL;
	ps.mflatfr = NULL;
	ps.scifits = NULL;
	ps.labels = NULL;
	ps.mbias = NULL;
	ps.mflat = NULL;
	ps.bpm = NULL;
	ps.fringes = NULL;
}

/*
 * Free any allocated memory
 */
static void omega_fringes_tidy(int level)
{
	freeimage(ps.mbias);
	freeimage(ps.mflat);
	freeimage(ps.bpm);
	freeimage(ps.fringes);
	if(level == 1)
		return;

	freeframeset(ps.sciset);
	freespace(ps.labels);
	freefits(ps.scifits);
}

/**@}*/


