/* $Id: omega_illumination.c,v 1.3 2011/11/21 10:20:01 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2011/11/21 10:20:01 $
 * $Revision: 1.3 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "omega_recipe.h"
#include "omega_utils.h"

/**
 * @defgroup omega_illumination  Illumination
 *
 * @par Synopsis:
 * This recipe is used to derive an illumination correction frame for one particular chip and filter. The \n
 * recipe always takes as input a fit of the illumination correction over the full focal plane for a given \n
 * filter. The recipes expects the following tags in the Set of Frames:
 *
 * @par Input frames:\n
 * ILLUM_FIT_PAR
 *
 * @par Output frames:\n
 * ILLUM_CORR
 *
 * @code
 *      #include omega_recipe.h
 * @endcode
 */
/**@{*/
/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_illumination_create(cpl_plugin *) ;
static int omega_illumination_exec(cpl_plugin *) ;
static int omega_illumination_destroy(cpl_plugin *) ;
static int omega_illumination(cpl_frameset *, cpl_parameterlist *) ;
static cpl_image *omega_illum_process(cpl_frame *frame, cpl_parameterlist *pars, int ext);
static void omega_illum_init(void);
static void omega_illum_tidy(void);


/* Input and Output structures */
static struct {
	/* Inputs. Parameters */
	int         extnum;

}omega_illum_config;

static struct {

	/* Calib frames */

	/* Products */
	cpl_image        *illum;

}ps;

#define RECIPE "omega_illumination"

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    OMEGA_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "omega_illumination",
                    "OMEGA - Generates a valid illumination correction frame (Calfile 548).",
           			"This recipe is used to derive an illumination correction frame for one \n"
					"particular chip and filter. The recipe always takes as input a fit of \n"
					"the illumination correction over the full focal plane for a given filter. \n"
					"A chip name should also be specified.",
                    "Sandra Castro",
                    "scastro@eso.org",
                    omega_get_license(),
                    omega_illumination_create,
                    omega_illumination_exec,
                    omega_illumination_destroy) ;

    cpl_pluginlist_append(list, plugin) ;

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
static int omega_illumination_create(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	cpl_parameter * p ;

	/* Do not create the recipe if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Create the parameters list in the cpl_recipe object */
	recipe->parameters = cpl_parameterlist_new() ;
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Parameter list allocation failed");
		cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
	}

	/* Fill the parameters list */
	p = cpl_parameter_new_value("omega.omega_illumination.ExtensionNumber",
			CPL_TYPE_INT,
			"FITS extension number to load (1 to 32). (-1 == all)",
			"omega_nightsky_flat",
			-1) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_illumination_exec(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	int recipe_status;

	/* Return immediately if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Verify parameter and frame lists */
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}
	if (recipe->frames == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Invoke the recipe */
	recipe_status = omega_illumination(recipe->frames, recipe->parameters);

	/* Ensure DFS-compliance of the products */
	if (cpl_dfs_update_product_header(recipe->frames)) {
		if (!recipe_status) recipe_status = (int)cpl_error_get_code();
	}

	return recipe_status;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_illumination_destroy(cpl_plugin * plugin)
{
	cpl_recipe  *   recipe;

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	cpl_parameterlist_delete(recipe->parameters) ;

    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_illumination(cpl_frameset *set, cpl_parameterlist *pars)
{
	int j,jst,jfn,isfirst;
	char *outfile = NULL;

	cpl_frame *frame, *product_frame;
	cpl_parameter *par;
	cpl_propertylist *qclist;

	/*Start the recipe */
	if (!pars) {
		cpl_msg_error (cpl_func, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (cpl_func, "Frameset not found");
		return -1;
	}

	/* Retrieve input parameters */
	par = cpl_parameterlist_find(pars, "omega.omega_illumination.ExtensionNumber") ;
	omega_illum_config.extnum = cpl_parameter_get_int(par) ;

	/* Identify the RAW and CALIB frames in the input frameset */
	if (oc_dfs_set_groups(set)) {
		cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames") ;
		return -1 ;
	}

	/*Initialized things*/
	omega_illum_init();

	/* Verify the frameset contents. */
	frame = cpl_frameset_find(set,OMEGA_CALIB_ILLFIT);
	if(frame == NULL){
		cpl_msg_error(cpl_func,"Cannot find coefficients table in input frameset");
		omega_illum_tidy();
		return -1;
	}

	/* Loop for each of the image extensions */
	omega_exten_range(omega_illum_config.extnum,&jst,&jfn);
	if(omega_illum_config.extnum == 0){
		cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_illum_config.extnum);
		omega_illum_tidy();
		return -1;
	}

	for (j = jst; j <= jfn; j++) {
		isfirst = (j == jst);
		cpl_msg_indent_more();
		cpl_msg_info(cpl_func,".....Working on extension %d.....",j);
		cpl_msg_indent_less();

		/* Call processing routine */
		ps.illum = omega_illum_process(frame, pars, j);
		if(ps.illum == NULL){
			cpl_msg_error(cpl_func,"Cannot create illumination correction frame");
			freespace(outfile);
			omega_illum_tidy();
			return -1;
		}

		/* Save illumination image */
		if(isfirst){
			outfile = cpl_sprintf("%s_%s.fits", INSTRUME,ILLUM_PROCATG);
			product_frame = omega_product_frame(outfile, ILLUM_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

		cpl_msg_info(cpl_func,"Saving Illumination frame");

		/* Save Illumination */
		if(omega_save_image(ps.illum,set,pars,NULL,NULL,CPL_BPP_IEEE_FLOAT,outfile,
					RECIPE,product_frame,NULL,isfirst) == -1){
			cpl_msg_error(cpl_func,"Cannot save product %s",ILLUM_PROCATG);
			freespace(outfile);
			omega_illum_tidy();
			return -1;
		}

/*
		if(omega_illum_save(ps.illum, set, pars, NULL, NULL, outfile, product_frame, isfirst) == -1){
			cpl_msg_error(cpl_func,"Cannot save product %s",ILLUM_PROCATG);
			freespace(outfile);
			omega_illum_tidy();
			return -1;
		}
*/
		freeimage(ps.illum);
		ps.illum = NULL;

	}  /* go to next extension */

	freespace(outfile);
	omega_illum_tidy();

	return 0;
}

/**
 * @brief		Create a illumination correction frame
 *
 * @param frame		Table with coefficients of the fit
 * @param pars		Parameters list
 * @param ext		FITS extension of chip to work on
 *
 * @return		Image of the illumination on the detector
 *
 */
static cpl_image *omega_illum_process(cpl_frame *frame, cpl_parameterlist *pars, int ext)
{

	int i = 0;
	int status = 0;
	int xsize, ysize;
	cpl_size pows[2];
	int dim = 2;
	float cfs[6];
	const char *chipid;
//	cpl_propertylist *plist;
	cpl_table *objects;
	cpl_image *illum;
	cpl_polynomial *poly;

	/* Assuming that the table has the HIERARCH ESO DET CHIP keywords */
//	plist = cpl_propertylist_load_regexp(cpl_frame_get_filename(frame),2,"CHIP", 0);

/*
	if((chipid = omega_pfits_get_chipid(plist)) == NULL){
		cpl_msg_error(cpl_func,"Chip ID not found");
		freeplist(plist);
		return NULL;
	}
*/

	/* Another possibility is to get the CHIP ID of a FITS extension. TBD */
	if((chipid = omega_get_extension_chipid(ext)) == NULL){
		cpl_msg_error(cpl_func,"CHIP ID not found");
//		freeplist(plist);
		return NULL;
	}
	cpl_msg_info(cpl_func,"CHIP ID for this extension is %s",chipid);

	if((objects = cpl_table_load(cpl_frame_get_filename(frame), 1, 0)) == NULL){
		cpl_msg_error(cpl_func,"NULL input OBJECTS table. %s",cpl_error_get_message());
//		freeplist(plist);
		return NULL;
	}

	for(i = 0; i < cpl_table_get_nrow(objects); i++){
		const char *name = cpl_table_get_string(objects,"chip_name",i);
		status = strcmp(chipid,name);
		if(status == 0){
			cfs[0] = cpl_table_get_float(objects,"C_o", i, NULL);
			cfs[1] = cpl_table_get_float(objects,"C_x", i, NULL);
			cfs[2] = cpl_table_get_float(objects,"C_y", i, NULL);
			cfs[3] = cpl_table_get_float(objects,"C_xx", i, NULL);
			cfs[4] = cpl_table_get_float(objects,"C_xy", i, NULL);
			cfs[5] = cpl_table_get_float(objects,"C_yy", i, NULL);
			break;
		}
	}

	if(status != 0){
		cpl_msg_error(cpl_func,"Cannot find CHIP ID in table");
		freetable(objects);
		return NULL;
	}
	freetable(objects);

	omega_pfits_get_chip_size(NULL, &xsize, &ysize);
//	freeplist(plist);

	poly = cpl_polynomial_new(dim);
	pows[0] = 0;
	pows[1] = 0;
	cpl_polynomial_set_coeff(poly, pows, cfs[0]);
	pows[0] = 1;
	pows[1] = 0;
	cpl_polynomial_set_coeff(poly, pows, cfs[1]);
	pows[0] = 0;
	pows[1] = 1;
	cpl_polynomial_set_coeff(poly, pows, cfs[2]);
	pows[0] = 2;
	pows[1] = 0;
	cpl_polynomial_set_coeff(poly, pows, cfs[3]);
	pows[0] = 1;
	pows[1] = 1;
	cpl_polynomial_set_coeff(poly, pows, cfs[4]);
	pows[0] = 0;
	pows[1] = 2;
	cpl_polynomial_set_coeff(poly, pows, cfs[5]);

	illum = cpl_image_new(xsize, ysize, CPL_TYPE_FLOAT);
	if(cpl_image_fill_polynomial(illum, poly,1,1,1,1) != CPL_ERROR_NONE){
		cpl_msg_error(cpl_func,"Cannot fill polynomial. %s",cpl_error_get_message());
		cpl_polynomial_delete(poly);
	}

	cpl_polynomial_delete(poly);

	cpl_image_multiply_scalar(illum, -0.4);
	cpl_image_exponential(illum, 10);

	return illum;
}


/* Initialize the pointers */
static void omega_illum_init(void)
{
	ps.illum = NULL;
}

/*
 * Free any allocated memory
 */
static void omega_illum_tidy(void)
{
	freeimage(ps.illum);
}

/**@}*/








