/* $Id: omega_mflat.c,v 1.6 2012/08/30 07:46:52 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/08/30 07:46:52 $
 * $Revision: 1.6 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "omega_recipe.h"

#include "omega_background.h"
#include "omega_bpm.h"
#include "omega_flats.h"
#include <cpl_errorstate.h>

/**
 * @defgroup omega_mflat  Master Flat
 *
 * @par Synopsis:
 * This recipe is used to derive a master twilight flat frame and a master flat for one particular chip and \n
 * filter. The recipe always takes as input a list of raw twilight flat frames, a master bias and an optional \n
 * master dome flat to be used in creating a final Master Flat Field. In addition, a hot pixel map and a cold\n
 * pixel map can be provided, which will be used to create a bad pixels map. Optionally, an overscan correction\n
 * mode can be set. The default is to apply no overscan correction. The raw twilight flats are trimmed and \n
 * overscan corrected. The data are then normalized, averaged, and the result is normalized again. Image \n
 * statistics are computed for the resulting frames. The recipe expects the following tags in the Set of Frames:
 *
 * @par Input frames:\n
 * TWILIGHT_FLAT\n
 * MASTER_BIAS\n
 * MASTER_FLAT_DOME (optional)\n
 * HOTE_PIXELS_MAP (optional)\n
 * COLD_PIXELS_MAP (optional)\n
 * REFERENCE_MFLAT (optional)
 *
 * @par Output frames:\n
 * MASTER_FLAT\n
 * BAD_PIXELS_MAP
 *
 * @code
 *      #include omega_recipe.h
 * @endcode
 */
/**@{*/
/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_mflat_create(cpl_plugin *) ;
static int omega_mflat_exec(cpl_plugin *) ;
static int omega_mflat_destroy(cpl_plugin *) ;
static int omega_mflat(cpl_frameset *,cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/
int omega_mflat_combine(cpl_parameterlist *pars, int xn);
static void omega_mflat_init(void);
static void omega_mflat_tidy(void);

/*-----------------------------------------------------------------------------
                            Static structures
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs. Parameters */
    int         extnum;
    int         oc;
    int         paf;
    double      sigma;

    /* Outputs. QC parameters */
    int         bpm;
    double	Mean;
    double	Median;
    double 	Stdev;
    double	RawMin;
    double	RawMax;
    double	RawMean;
    double	RawMedian;
    double	RawStdev;

}omega_mflat_config;

/* Input and Output */
static struct {

    /* Calib frames */
    const cpl_frame	*cframe;
    const cpl_frame 	*hframe;
    const cpl_frame     *mbframe;
    const cpl_frame     *mdomefr;
    const cpl_frame     *nskyfr;
    const cpl_frame     *bpframe;
    cpl_size            *labels;
    cpl_frameset     *flatlist;
    cpl_propertylist *eh;
    cpl_stats        *stats;
    omega_fits        *firstflat;

    /* Products */
    cpl_image        *bpixels;
    cpl_image        *mflat;
    cpl_image        *mtwilight;

}ps;

/*-----------------------------------------------------------------------------
                            Static global variables
 -----------------------------------------------------------------------------*/

/* Static variables */
#define RECIPE "omega_mflat"

static int isfirst;
static int dummy;

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    OMEGA_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "omega_mflat",
                    "OMEGA - Create Master Twilight and Master Flat for each chip",
                    "This recipe is used to derive a valid twilight flat frame (Calfile 543, 546) \n"
                    "for one particular chip and filter. The recipe always takes as input \n"
                    "a list of raw twilight flat frames, a master bias and an optional master \n"
                    "dome flat to be used in creating a final Master Flat Field. In addition, a hot  \n"
                    "pixel map (and a cold pixel map can be provided, which will be used to create a \n"
                    "bad pixels map. Optionally, an overscan correction mode \n"
                    "can be set. The default is to apply no overscan correction.\n\n"
                    "The raw twilight flats are trimmed and overscan corrected. The data are then \n"
                    "normalized, averaged, and the result is normalized again. Image statistics \n"
                    "are computed for the resulting frames.",
                    "Sandra Castro",
                    "scastro@eso.org",
                    omega_get_license(),
                    omega_mflat_create,
                    omega_mflat_exec,
                    omega_mflat_destroy) ;

    cpl_pluginlist_append(list, plugin) ;

    return 0;
}

/*
 * Create the plugin and the parameters list
 */
static int omega_mflat_create(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    cpl_parameter * p ;
    char *path = NULL;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new() ;
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }


    /* Fill the parameters list */

/* --------------- General parameters ---------------------------*/
   p = cpl_parameter_new_value("omega.omega_mflat.ExtensionNumber",
                                CPL_TYPE_INT,
                               "FITS extension number to load (1 to 32). (-1 = all)",
                                "omega_mflat",
                                -1) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    p = cpl_parameter_new_range("omega.omega_mflat.OverscanMethod",
                                    CPL_TYPE_INT,
                                    "Overscan Correction Method",
                                    "omega_mflat",
                                    0, 0, 6);

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mflat.PAF",
                                CPL_TYPE_BOOL,
                                "Boolean value to create PAF files. 1(Yes), 0(No)",
                                "omega_mflat",
                                1) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "paf") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mflat.SigmaClip",
                                CPL_TYPE_DOUBLE,
                                "Sigma Clipping Threshold",
                                "omega_mflat",
                                3.0) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sig-clip") ;
    cpl_parameterlist_append(recipe->parameters, p) ;



/*------------------ Parameters to create a Master Flat ------------------ */
    path = cpl_sprintf("%s", OMEGA_BIN_PATH);
    p = cpl_parameter_new_value("omega.omega_mflat.BinPath",
                                CPL_TYPE_STRING,
                                "Path to any external executable program.",
                                "omega.BinPath",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "bin-path");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    p = cpl_parameter_new_value("omega.omega_mflat.GaussianFiltSize",
                                CPL_TYPE_DOUBLE,
                                "Standard deviation of Gaussian Fourier filter size",
                                "omega_mflat",
                                9.0) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "gau_filt_sz") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mflat.MirrorXpix",
                                CPL_TYPE_INT,
                                "Width of X region for mirroring edges (FFT continuity)",
                                "omega_mflat",
                                75) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "mr_xpix") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mflat.MirrorYpix",
                                CPL_TYPE_INT,
                                "Width of Y region for mirroring edges (FFT continuity)",
                                "omega_mflat",
                                150) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "mr_ypix") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    /* Parameters to create bad pixels map */
    p = cpl_parameter_new_range("omega.omega_mflat.LowThre",
                                CPL_TYPE_DOUBLE,
                                "Low flagging threshold for cold pixels map",
                                "omega_mflat",
                                0.96, 0.90, 1.00) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "low") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_range("omega.omega_mflat.HighThre",
                                CPL_TYPE_DOUBLE,
                                "High flagging threshold for cold pixels map",
                                "omega_flat",
                                1.04, 1.00, 1.10) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"high") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    return 0;
}

/*
 * Execute the plugin
 */
static int omega_mflat_exec(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    int recipe_status;

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = omega_mflat(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    return recipe_status;
}

/*
 * Destroy the plugin
 */
static int omega_mflat_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters) ;

    return 0 ;
}

/*
 * Implement the recipe
 */
static int omega_mflat(cpl_frameset *set, cpl_parameterlist *pars)
{
    int j,jst,jfn;
    cpl_size nlab;
    int nflats = 0;
    int oscan1 = 0;
    int status = 1;

    float bias = BIAS;

    const char *_id = "omega_mflat";
    char *outbpm = NULL;
    char *outtwilight = NULL;
    char *outmflat = NULL;
    const cpl_frame 	*firstframe;
    const cpl_frame     *refframe;
    cpl_frame *prframe_twil = NULL;
    cpl_frame *prframe_mflat = NULL;
    cpl_frame *prframe_bpm = NULL;
    cpl_parameter 	*par;
    cpl_propertylist    *qclist;
    cpl_propertylist    *alist;
    cpl_stats *diffstats = NULL;
    cpl_errorstate prestate = cpl_errorstate_get();


 /*Start the recipe */

    if (!pars) {
        cpl_msg_error (_id, "Parameters list not found");
        return -1;
    }

    if (cpl_frameset_is_empty(set) == 1) {
        cpl_msg_error (_id, "Frameset not found");
        return -1;
    }

/* Retrieve input parameters */
    par = cpl_parameterlist_find(pars, "omega.omega_mflat.ExtensionNumber") ;
    omega_mflat_config.extnum = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mflat.OverscanMethod") ;
    omega_mflat_config.oc = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mflat.SigmaClip") ;
    omega_mflat_config.sigma = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mflat.PAF") ;
    omega_mflat_config.paf = cpl_parameter_get_bool(par) ;

/* Identify the RAW and CALIB frames in the input frameset */
    if (oc_dfs_set_groups(set)) {
        cpl_msg_error(_id, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /*Initialized things*/
    omega_mflat_init();

/* Verify the frameset contents. */
    if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
					   &nlab)) == NULL) {
        cpl_msg_error(_id,"Cannot labelise the input frameset");
        omega_mflat_tidy();
        return -1;
    }
    if ((ps.flatlist = omega_frameset_subgroup(set,ps.labels,nlab,
						MTWIL_RAW)) == NULL) {
	    cpl_msg_error(_id,"Cannot find twilight flats in input frameset");
        omega_mflat_tidy();
        return -1;
    }
    nflats = cpl_frameset_count_tags (ps.flatlist, MTWIL_RAW);
    if (nflats < 2) {
        cpl_msg_error (_id, "Need at least 2 (%s) frames to run "
                            "this recipe", MTWIL_RAW);
        omega_mflat_tidy();
        return -1;
    }

    cpl_msg_info (_id,"There are %d %s frames in frame set",nflats, MTWIL_RAW);

    /* Check for calibration frames */
    /* Master Bias */
    ps.mbframe = cpl_frameset_find_const(set, OMEGA_CALIB_BIAS);
    if (ps.mbframe == NULL) {
        cpl_msg_error(cpl_func,"Cannot find OMEGA_CALIB_BIAS input in frame set");
        omega_mflat_tidy();
        return -1;
    }
    cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(ps.mbframe));

    /* Cold Pixels Map */
    ps.cframe = cpl_frameset_find_const(set, OMEGA_CALIB_CPM);
    if(ps.cframe == NULL){
    	cpl_msg_error(cpl_func,"Cannot find %s input in frame set",OMEGA_CALIB_CPM);
    	omega_mflat_tidy();
    	return -1;
    }
    cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_CPM, cpl_frame_get_filename(ps.cframe));

    /* Hot Pixels Map */
    ps.hframe = cpl_frameset_find_const(set, OMEGA_CALIB_HPM);
    if(ps.hframe == NULL){
    	cpl_msg_error(cpl_func,"Cannot find %s input in frame set",OMEGA_CALIB_HPM);
    	omega_mflat_tidy();
    	return -1;
    }
    cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_HPM, cpl_frame_get_filename(ps.hframe));

    /* Optional Master Dome Flat */
    ps.mdomefr = cpl_frameset_find_const(set, OMEGA_CALIB_DOME);
    if(ps.mdomefr != NULL) {
        cpl_msg_info(_id,"Using %s %s", OMEGA_CALIB_DOME, cpl_frame_get_filename(ps.mdomefr));
    }

    /* Optional Master Night Sky Flat */
    ps.nskyfr = cpl_frameset_find_const(set, OMEGA_CALIB_NSKY);
    if(ps.nskyfr != NULL) {
            cpl_msg_info(_id,"Using %s %s", OMEGA_CALIB_NSKY, cpl_frame_get_filename(ps.nskyfr));
    }

    /* Optional Reference master flat frame */
    refframe = cpl_frameset_find_const(set, REFMFLAT);
    if (refframe != NULL)
        cpl_msg_info(cpl_func,"Using %s for comparison",cpl_frame_get_filename(refframe));


    /* Get first frame from frame set */
    firstframe = cpl_frameset_get_first(ps.flatlist);

    /* Loop for each of the image extensions */
    omega_extensions(firstframe,omega_mflat_config.extnum,&jst,&jfn);
    if(omega_mflat_config.extnum == 0){
    	cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_mflat_config.extnum);
    	omega_mflat_tidy();
    	return -1;
    }

    for (j = jst; j <= jfn; j++) {

        isfirst = (j == jst);
        cpl_msg_info(_id,"Beginning work on extension %d",j);

        omega_mflat_config.bpm = 0;
        omega_mflat_config.Mean = 0.0;
        omega_mflat_config.Median = 0.0;
        omega_mflat_config.Stdev = 0.0;
        omega_mflat_config.RawMin = 0.0;
        omega_mflat_config.RawMax = 0.0;
        omega_mflat_config.RawMean = 0.0;
        omega_mflat_config.RawMedian = 0.0;
        omega_mflat_config.RawStdev = 0.0;

        /* Check overscan correction method consistency */
        if(ps.mbframe != NULL){
            oscan1 = omega_pfits_get_overscan(ps.mbframe, j);
            if(oscan1 != omega_mflat_config.oc) {
                    cpl_msg_warning (_id, "Overscan correction mode for Master Bias (oc = %d) differs from "
                                        "the one used here (oc = %d)", oscan1, omega_mflat_config.oc);
            }
        }

        /* Load first twilight flat */
        ps.firstflat = omega_fits_load(firstframe,CPL_TYPE_FLOAT,j);
        ps.eh = omega_fits_get_ehu(ps.firstflat);


        /* Call the combining routine */
        status = omega_mflat_combine(pars, j);
        if(status == 1){
            cpl_msg_warning(_id, "Image detector is not live");
            /* Save dummy product */
            freefits(ps.firstflat);
            ps.eh = NULL;
            continue;
        }
        else if(status == -1){
            cpl_msg_error(_id,"Cannot combine images");
            freespace(outbpm);
            freespace(outtwilight);
            freespace(outmflat);
            omega_mflat_tidy();
            return -1;
        }

        status = 0;

       qclist = cpl_propertylist_new();


        /* Compare with reference frame */
        if(refframe != NULL){
            if(omega_compare_reference(ps.mflat,refframe,j,&diffstats) == -1){
                cpl_msg_warning(cpl_func,"Cannot compare with reference frame");
            }
            else{
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFMFLAT MEAN",
                                               cpl_stats_get_mean(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFMFLAT MEAN",
                                             "Mean of difference with reference");
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFMFLAT MEDIAN",
                                               cpl_stats_get_median(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFMFLAT MEDIAN",
                                             "Median of difference with reference");
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFMFLAT STDEV",
                                               cpl_stats_get_stdev(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFMFLAT STDEV",
                                             "Stdev of difference with reference");

                freestats(diffstats);
            }
        }

        /* Save the MASTER FLAT product */
        /* Compute statistics on master flat */
         ps.stats = cpl_stats_new_from_image(ps.mflat, CPL_STATS_ALL);
        if(ps.stats != NULL) {
            omega_mflat_config.Mean = cpl_stats_get_mean(ps.stats);
            omega_mflat_config.Median = cpl_stats_get_median(ps.stats);
            omega_mflat_config.Stdev = cpl_stats_get_stdev(ps.stats);
            freestats(ps.stats);
        }

        /* Save QC parameters in list */
        cpl_propertylist_append_double(qclist, "ESO QC MASTER FLAT MEAN",
                                                    omega_mflat_config.Mean) ;
        cpl_propertylist_set_comment(qclist,"ESO QC MASTER FLAT MEAN",
                                            "Mean value of master flat");
        cpl_propertylist_append_double(qclist, "ESO QC MASTER FLAT MEDIAN",
                                                    omega_mflat_config.Median) ;
        cpl_propertylist_set_comment(qclist,"ESO QC MASTER FLAT MEDIAN",
                                            "Median value of master flat");
        cpl_propertylist_append_double(qclist, "ESO QC MASTER FLAT STDEV",
                                                    omega_mflat_config.Stdev) ;
        cpl_propertylist_set_comment(qclist,"ESO QC MASTER FLAT STDEV",
                                            "Standard deviation value of master flat");

        /* Create a new product frame object and define some tags */
		if(isfirst){
			outmflat = cpl_sprintf("%s_%s.fits", INSTRUME,MFLAT_PROCATG);
			prframe_mflat = omega_product_frame(outmflat, MFLAT_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}


		alist = cpl_propertylist_new();
		cpl_propertylist_append_string(alist, "EXTNAME",
				cpl_propertylist_get_string(ps.eh, "EXTNAME"));
		cpl_propertylist_set_comment(alist,"EXTNAME", "Extension name");

		cpl_propertylist_copy_property_regexp(alist, ps.eh, WCS_KEYS, 0);

		/* Add DRS keywords */
		cpl_propertylist_update_int(alist, "ESO DRS OVERSCAN METHOD", omega_mflat_config.oc);
		cpl_propertylist_set_comment(alist, "ESO DRS OVERSCAN METHOD", "overscan correction method");

		if(omega_save_image(ps.mflat,set,pars,alist,qclist,CPL_BPP_IEEE_FLOAT,outmflat,
				RECIPE,prframe_mflat,NULL,isfirst) == -1){
            cpl_msg_error(_id,"Cannot save product %s", MFLAT_PROCATG);
            freeplist(alist);
            freeplist(qclist);
            freespace(outmflat);
            freespace(outtwilight);
            freespace(outbpm);
            omega_mflat_tidy();
            return -1;;
		}


		freeplist(qclist);
        freeimage(ps.mflat);
        status = 0;

        /* Save the MASTER TWILIGHT FLAT product */
        /* Compute statistics on master twilight flat */
        ps.stats = cpl_stats_new_from_image(ps.mtwilight, CPL_STATS_ALL);
        if(ps.stats != NULL) {
            omega_mflat_config.Mean = cpl_stats_get_mean(ps.stats);
            omega_mflat_config.Median = cpl_stats_get_median(ps.stats);
            omega_mflat_config.Stdev = cpl_stats_get_stdev(ps.stats);
            freestats(ps.stats);
        }

        /* Save QC parameters in list */
        qclist = cpl_propertylist_new();
        cpl_propertylist_append_double(qclist, "ESO QC MASTER TWILIGHT MEAN",
                                                omega_mflat_config.Mean) ;
        cpl_propertylist_set_comment(qclist,"ESO QC MASTER TWILIGHT MEAN",
                                            "Mean value of master twilight flat");
        cpl_propertylist_append_double(qclist, "ESO QC MASTER TWILIGHT MEDIAN",
                                                omega_mflat_config.Median) ;
        cpl_propertylist_set_comment(qclist,"ESO QC MASTER TWILIGHT MEDIAN",
                                            "Median value of master twilight flat");
        cpl_propertylist_append_double(qclist, "ESO QC MASTER TWILIGHT STDEV",
                                                omega_mflat_config.Stdev) ;
        cpl_propertylist_set_comment(qclist,"ESO QC MASTER TWILIGHT STDEV",
                                            "Standard deviation value of master twilight flat");

        cpl_propertylist_append_double(qclist, "ESO QC RAW TWILIGHT MIN",
        		omega_mflat_config.RawMin);
        cpl_propertylist_append_double(qclist, "ESO QC RAW TWILIGHT MAX",
        		omega_mflat_config.RawMax);
        cpl_propertylist_append_double(qclist, "ESO QC RAW TWILIGHT MEAN",
        		omega_mflat_config.RawMean);
        cpl_propertylist_append_double(qclist, "ESO QC RAW TWILIGHT MEDIAN",
        		omega_mflat_config.RawMedian);
        cpl_propertylist_append_double(qclist, "ESO QC RAW TWILIGHT STDEV",
        		omega_mflat_config.RawStdev);

        cpl_propertylist_set_comment(qclist, "ESO QC RAW TWILIGHT MIN",
        		"median value of the raw flat flat having the lowest flux");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW TWILIGHT MAX",
        		"median value of the raw flat flat having the highest flux");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW TWILIGHT MEAN",
        		"mean value of all input raw flat flats (ADU)");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW TWILIGHT MEDIAN",
        		"median value of all input raw flat flats (ADU)");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW TWILIGHT STDEV",
        		"standard deviation of all input raw flat flats (ADU)");

        /* Create a new product frame object and define some tags */
		if(isfirst){
			outtwilight = cpl_sprintf("%s_%s.fits", INSTRUME,MTWIL_PROCATG);
			prframe_twil = omega_product_frame(outtwilight, MTWIL_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

		if(omega_save_image(ps.mtwilight,set,pars,alist,qclist,CPL_BPP_IEEE_FLOAT,outtwilight,
				RECIPE,prframe_twil,NULL,isfirst) == -1){
            cpl_msg_error(_id,"Cannot save product %s", MTWIL_PROCATG);
            freeplist(qclist);
    		freeplist(alist);
            freespace(outtwilight);
            freespace(outbpm);
            freespace(outmflat);
            omega_mflat_tidy();
            return -1;;
		}

        freeplist(qclist);
        freeimage(ps.mtwilight);
        status = 0;

        /* Save the BAD PIXELS MAP product */
        /* Save QC parameters in list */
        qclist = cpl_propertylist_new();
        cpl_propertylist_append_int(qclist, "ESO QC NUMBER BAD PIXELS",omega_mflat_config.bpm);
        cpl_propertylist_set_comment(qclist,"ESO QC NUMBER BAD PIXELS",
                                            "Number of bad pixels");

		if(isfirst){
			outbpm = cpl_sprintf("%s_%s.fits", INSTRUME,BPM_PROCATG);
			prframe_bpm = omega_product_frame(outbpm, BPM_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

		if(omega_save_image(ps.bpixels,set,pars,alist,qclist,CPL_BPP_16_SIGNED,outbpm,
				RECIPE,prframe_bpm,NULL,isfirst) == -1){
            cpl_msg_error(_id,"Cannot save product %s", BPM_PROCATG);
    		freeplist(alist);
            freeplist(qclist);
            freespace(outbpm);
            freespace(outtwilight);
            freespace(outmflat);
            omega_mflat_tidy();
            return -1;;
		}

        freeplist(qclist);
		freeplist(alist);
        freeimage(ps.bpixels);
        freefits(ps.firstflat);
        ps.firstflat = NULL;
        ps.eh = NULL;

    }    /* go and work on next extension */

    /*Clean up */
    freespace(outbpm);
    freespace(outtwilight);
    freespace(outmflat);
    omega_mflat_tidy();

    if(!cpl_errorstate_is_equal(prestate)){
    	cpl_errorstate_dump(prestate,CPL_FALSE,NULL);
    }

    return 0;

}


/**
 * @brief
 *   Process the twilight flats
 *
 * @param pars  Parameters list
 * @param xn    FITS extension to load
 *
 * @return
 *   0 on success or -1 in case of error.
 */
int omega_mflat_combine(cpl_parameterlist *pars, int xn)
{

    int i,nflats,live, naxis1, naxis2;
    float bias = BIAS;
    double gain = 0.0;
    double median = 1.0;
    double threshold = 0.0;
    double *data_scales;
    const char *_id = "omega_mflat_combine";

    cpl_error_code code;
    const cpl_frame *flatfr;
    cpl_vector *scales;
    cpl_vector *median_vector = NULL;
    cpl_image *trim_raw,*dev,*good_float,*good_int,*median_all,*new_image;
    cpl_image *sum_data,*sum_good;
    cpl_image *mbias;
    cpl_image *mdome;
    cpl_image *nsky;
    cpl_mask  *bpm_map;
    cpl_imagelist *ilist;
    cpl_mask  *good;

    nflats = cpl_frameset_get_size(ps.flatlist);
    flatfr = cpl_frameset_get_first_const(ps.flatlist);

    /* Check that this detector is live*/
    omega_pfits_get_detlive(ps.eh,&live);
    if (! live) {
        return 1;
     }

    /* Load master bias image */
    if(ps.mbframe != NULL){
        mbias = cpl_image_load(cpl_frame_get_filename(ps.mbframe), CPL_TYPE_FLOAT,0,xn);
        if (mbias == NULL) {
            cpl_msg_warning(_id,"Cannot load image %s", OMEGA_CALIB_BIAS);
        }
    }

    /*Create Bad Pixels Map */
    bpm_map = makebpm(ps.hframe, ps.cframe, xn);
    if(bpm_map != NULL){
        /*        cpl_mask_not(bpm_map);*/
        omega_mflat_config.bpm = cpl_mask_count(bpm_map);
    }

    scales = cpl_vector_new(nflats);
    data_scales = cpl_vector_get_data(scales);

    median_vector = cpl_vector_new (nflats);
    cpl_vector_fill(median_vector,0.0);

    ilist = cpl_imagelist_new();
    cpl_msg_info (_id,"Doing trim and overscan correction on images");

    /* Loop through all images in frame list */
    for (i=0; i< nflats; i++)
    {

        trim_raw = TrimOscanCorrect(flatfr, omega_mflat_config.oc, xn);
        if(trim_raw == NULL){
            freevector(scales);
            freevector(median_vector);
            freeilist(ilist);
            freemask(bpm_map);
            freeimage(mbias);
            return -1;
        }

        if (mbias != NULL)
            cpl_image_subtract(trim_raw, mbias);
        else
            cpl_image_subtract_scalar(trim_raw, bias);

        if (bpm_map == NULL) {
            ps.stats = cpl_stats_new_from_image(trim_raw, CPL_STATS_ALL);
            median = cpl_stats_get_median(ps.stats);
        }
        else {
            cpl_image_reject_from_mask(trim_raw, bpm_map);
            ps.stats = cpl_stats_new_from_image(trim_raw, CPL_STATS_ALL);
            median = cpl_stats_get_median(ps.stats);
        }

        cpl_vector_set(median_vector, i, median);

        data_scales[i] = (double)1.0/median;

        cpl_image_divide_scalar(trim_raw, median);

        cpl_imagelist_set(ilist, trim_raw, i);


        flatfr = cpl_frameset_get_next_const(ps.flatlist);
        if (flatfr == NULL)
            break;

        freestats(ps.stats);

    }

    omega_mflat_config.RawMin    = cpl_vector_get_min(median_vector);
    omega_mflat_config.RawMax    = cpl_vector_get_max(median_vector);
    omega_mflat_config.RawMean   = cpl_vector_get_mean(median_vector);
    omega_mflat_config.RawMedian = cpl_vector_get_median(median_vector);
    omega_mflat_config.RawStdev  = cpl_vector_get_stdev(median_vector);

    freevector(median_vector);

    freestats(ps.stats);
    freeimage(mbias);

    if (ilist == NULL) {
        cpl_msg_error(_id,"Error in image list <%s>",cpl_error_get_message());
        freevector(scales);
        freeimage(trim_raw);
        freemask(bpm_map);
        return -1;
    }

    /*  Read gain from the header */
    /* FIXME: or read gain from gain recipe product?? */
    /* The images have the CONAD as the correct value for
     * the gain, therefore the following lines should be
     * changed if the headers are fixed.
     */

//    omega_pfits_get_gain(ps.eh, &gain);
    omega_pfits_get_conad(ps.eh, &gain);

    /*FIXME: should we avoid having gain=0?*/
    if (gain < 0.1){
    	omega_pfits_get_gain(ps.eh, &gain);
    	//        omega_pfits_get_conad(ps.eh, &gain);
    	if(gain < 0.1) {
    		cpl_msg_info(_id,"Using default value for gain");
    		gain = 2.0;
    	}
    }
    else {
    	cpl_msg_info(_id,"Gain value from images is %g", gain);
    }

    cpl_msg_info (_id,"Computing the median of all images...");
    median_all = cpl_imagelist_collapse_median_create(ilist);

    if (median_all == NULL) {
        cpl_msg_error (_id,"Cannot take median of imset <%s>",cpl_error_get_message());
        freevector(scales);
        freeilist(ilist);
        freemask(bpm_map);
        return -1;
    }

    naxis1 = cpl_image_get_size_x(median_all);
    naxis2 = cpl_image_get_size_y(median_all);

    sum_data = cpl_image_new(naxis1,naxis2,CPL_TYPE_FLOAT);
    sum_good = cpl_image_new(naxis1,naxis2,CPL_TYPE_FLOAT);

    /* Clip the outliers from the images*/
    for (i=0; i< nflats; i++)
    {
        trim_raw = cpl_imagelist_get(ilist, i);
        new_image = cpl_image_duplicate(trim_raw);

        /* FIXME: setting negative values to 0 to avoid failure of SQRT in next step
         * when there are negative values in the image. These negative values happen when any
         * given pixel in the domeflat is smaller than the same pixel in the master_bias image.
         */
        cpl_image_threshold(trim_raw, 0, FLT_MAX, 0, 2e20);

        dev = cpl_image_subtract_create(median_all, trim_raw);

        code = cpl_image_power(trim_raw, 0.5);
        if(code != CPL_ERROR_NONE) {
            cpl_msg_error(_id,"Error in SQRT operation <%s>",cpl_error_get_message());
            freevector(scales);
            freeilist(ilist);
            freeimage(median_all);
            freeimage(sum_data);
            freeimage(sum_good);
            freeimage(new_image);
            freeimage(dev);
            freemask(bpm_map);
            return -1;
        }

        cpl_image_divide(dev, trim_raw);

        threshold = omega_mflat_config.sigma * (sqrt(gain * data_scales[i]));

        good = cpl_mask_threshold_image_create(dev, -threshold, threshold);

        freeimage(dev);
        good_int = cpl_image_new_from_mask(good) ;
        freemask(good) ;

        good_float = cpl_image_cast(good_int, CPL_TYPE_FLOAT);
        freeimage(good_int);

        cpl_image_multiply(new_image, good_float);
        cpl_image_add(sum_data, new_image);

        cpl_image_add(sum_good, good_float);

        freeimage(new_image);
        freeimage(good_float);

    }

    freeimage(median_all);
    freevector(scales);

    /* Create master twilight flat */
    ps.mtwilight = cpl_image_divide_create(sum_data, sum_good);
    if (ps.mtwilight == NULL) {
        cpl_msg_error(_id,"Error in division %s <%s>",MTWIL_PROCATG, cpl_error_get_message());
        freeilist(ilist);
        freeimage(sum_data);
        freeimage(sum_good);
        freemask(bpm_map);
        return -1;
    }

    freeimage(sum_data);
    freeimage(sum_good);
    freeilist(ilist);


    /* Create a MASTER FLAT */
    if(ps.mdomefr != NULL){
        /* Load Master Dome */
        mdome = cpl_image_load(cpl_frame_get_filename(ps.mdomefr), CPL_TYPE_FLOAT,0,xn);
        if (mdome == NULL){
            cpl_msg_warning(_id,"Cannot load image %s", OMEGA_CALIB_DOME);
            ps.mflat = cpl_image_duplicate(ps.mtwilight);
        }
        else{
            ps.mflat = omega_make_mflat(mdome,ps.mtwilight,bpm_map,pars);
            if(ps.mflat == NULL){
                cpl_msg_warning(_id,"Cannot calculate Master Flat Field");
                ps.mflat = cpl_image_duplicate(ps.mtwilight);
            }
            freeimage(mdome);
        }
    }
    else{
        cpl_msg_warning(_id,"There is no Master Dome Flat in frame set");
        ps.mflat = cpl_image_duplicate(ps.mtwilight);
    }

    /* Load Master Night Sky Flat */
    if (ps.nskyfr != NULL) {
        nsky = cpl_image_load(cpl_frame_get_filename(ps.nskyfr), CPL_TYPE_FLOAT,0,xn);
        if (nsky == NULL) {
            cpl_msg_warning(_id,"Cannot load image %s", OMEGA_CALIB_NSKY);
        }
        else{
            cpl_image_multiply (ps.mflat, nsky);
            freeimage(nsky);
        }
    }


    /* Create a BAD PIXELS MAP IMAGE to save later*/
    /*   cpl_mask_not(bpm_map);*/
    if(bpm_map != NULL)
    	ps.bpixels = cpl_image_new_from_mask(bpm_map);

    omega_mflat_config.bpm = cpl_mask_count(bpm_map);
	freemask(bpm_map);

    return 0;
}

/* Initialize the pointers */
static void omega_mflat_init(void) {
    ps.bpixels = NULL;
    ps.cframe = NULL;
    ps.eh = NULL;
    ps.firstflat = NULL;
    ps.flatlist = NULL;
    ps.hframe = NULL;
    ps.labels = NULL;
    ps.mbframe = NULL;
    ps.mdomefr = NULL;
    ps.mflat = NULL;
    ps.mtwilight = NULL;
    ps.nskyfr = NULL;
    ps.stats = NULL;
}

/* Free any allocated memory */
static void omega_mflat_tidy(void) {
    freefits(ps.firstflat);
    freeframeset(ps.flatlist);
    freeimage(ps.bpixels);
    freeimage(ps.mflat);
    freeimage(ps.mtwilight);
    freespace(ps.labels);
    freestats(ps.stats);
}

/**@}*/
