/* $Id: omega_qcheck.c,v 1.2 2012/01/12 12:05:09 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/01/12 12:05:09 $
 * $Revision: 1.2 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "omega_recipe.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup omega_qcheck  Quick Check
 *
 * @par Synopsis:
 * Generates a quick check on the detector responsivity for each chip. The recipe takes as input exactly \n
 * one raw dome flat frame, and a master bias. Optionally, an overscan correction mode can be set. The default \n
 * is to apply no overscan correction. The recipes expects the following tags in the set of frames:
 *
 * @par Input frames:
 * DOME_LIFETEST\n
 * MASTER_BIAS\n
 * REFERENCE_LIFETEST (optional)
 *
 * @par Output frames:
 * LIFETEST_CHECK
 *
 * @code
 *      #include omega_recipe.h
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_qcheck_create(cpl_plugin *) ;
static int omega_qcheck_exec(cpl_plugin *) ;
static int omega_qcheck_destroy(cpl_plugin *) ;
static int omega_qcheck_old(cpl_frameset *,cpl_parameterlist *) ;
static int omega_qcheck(cpl_frameset *,cpl_parameterlist *) ;

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/
static void omega_qcheck_init(void);
static void omega_qcheck_tidy(void);

/* Parameters and QC structure*/
static struct {
	/* Inputs. Parameters */
	float       rejt;
	int         extnum;
	int         niter;
	int         oc;
	int         paf;

	/* QC parameters */
	double	mean;
	double	median;
	double	stdev;

}omega_qcheck_config;

/* Input and Output */
static struct {
	cpl_size         *labels;
	cpl_stats        *stats;
	cpl_frameset     *domelist;
	omega_fits *domefits1;
	cpl_propertylist *eh;

	/*Products*/
	char *proname;
	cpl_table *result;
} ps;

/* Static variables */

/* Static variables */
#define RECIPE "omega_qcheck"

/*----------------------------------------------------------------------------*/
/**
@brief    Build the list of available plugins, for this module.
@param    list    the plugin list
@return   0 if everything is ok

This function is exported.
*/
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
	cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
	cpl_plugin  *   plugin = &recipe->interface ;

	cpl_plugin_init(plugin,
			CPL_PLUGIN_API,
			OMEGA_BINARY_VERSION,
			CPL_PLUGIN_TYPE_RECIPE,
			"omega_qcheck",
			"OMEGA - Generates a quick check on the detector responsivity for each chip.",
			"The recipe takes as input exactly one raw domeflat frame, and \n"
			"a master bias. Optionally, an overscan correction mode can be set. \n"
			"The default is to apply no overscan correction.",
			"Sandra Castro",
			"scastro@eso.org",
			omega_get_license(),
			omega_qcheck_create,
			omega_qcheck_exec,
			omega_qcheck_destroy) ;

	cpl_pluginlist_append(list, plugin) ;

	return 0;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Setup the recipe options
@param    plugin  the plugin
@return   0 if everything is ok

Create the recipe instance and make it available to the application using the
interface.
*/
/*----------------------------------------------------------------------------*/
static int omega_qcheck_create(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	cpl_parameter * p ;

	/* Do not create the recipe if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Create the parameters list in the cpl_recipe object */
	recipe->parameters = cpl_parameterlist_new() ;

	/* Fill the parameters list */
	p = cpl_parameter_new_value("omega.omega_qcheck.ExtensionNumber",
			CPL_TYPE_INT,
			"FITS extension number to load (1 to 32). (-1 == all)",
			"omega_qcheck",
			-1) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	p = cpl_parameter_new_range("omega.omega_qcheck.OverscanMethod",
			CPL_TYPE_INT,
			"Overscan Correction Method (0 to 6):\n"
			"0 = no overscan correction;\n"
			"1 = use median of prescan in X;\n"
			"2 = use median of overscan in X;\n"
			"3 = use median on prescan in Y;\n"
			"4 = use median of overscan in Y;\n"
			"5 = per-row subtraction of the median of row in prescan regions of X;\n"
			"6 = per-row subtraction of the median of row in overscan regions of X\n",
			"omega_qcheck",
			0, 0, 6);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_qcheck.PAF",
			CPL_TYPE_BOOL,
			"Boolean value to create PAF files. 1(Yes), 0(No)",
			"omega_qcheck",
			1) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "paf") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_qcheck.rej_threshold",
			CPL_TYPE_DOUBLE,
			"The rejection threshold for outlying pixels",
			"omega_qcheck",
			5.0) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"rej-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_qcheck.NumberIter",
			CPL_TYPE_INT,
			"The maximum number of iterations",
			"omega_qcheck",
			5) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"niter") ;
	cpl_parameterlist_append(recipe->parameters, p) ;



	return 0;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Execute the plugin instance given by the interface
@param    plugin  the plugin
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_qcheck_exec(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	int recipe_status;

	/* Return immediately if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Verify parameter and frame lists */
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}
	if (recipe->frames == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Invoke the recipe */
	recipe_status = omega_qcheck(recipe->frames, recipe->parameters);

	/* Ensure DFS-compliance of the products */
	if (cpl_dfs_update_product_header(recipe->frames)) {
		if (!recipe_status) recipe_status = (int)cpl_error_get_code();
	}


	return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Destroy what has been created by the 'create' function
@param    plugin  the plugin
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_qcheck_destroy(cpl_plugin * plugin)
{
	cpl_recipe  *   recipe = (cpl_recipe *)plugin ;

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	cpl_parameterlist_delete(recipe->parameters);

	return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
* @internal
@brief    The recipe data reduction part is implemented here
@param    parlist     the parameters list
@param    framelist   the frames list
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_qcheck_old(cpl_frameset *set, cpl_parameterlist *pars)
{

	int j,jst,jfn,live,oscan1,isfirst;
	cpl_size nlab;
	const char *_id = "omega_qcheck";
	char *outfile = NULL;

	cpl_frame 		*mbias_frame;
	const cpl_frame 	*frame1;
	const cpl_frame 	*frame2;
	cpl_frame *product_frame;
	cpl_image 		*mbias_image;
	cpl_image		*trim_raw1;
	cpl_image		*trim_raw2;
	cpl_parameter 	*par;
	cpl_propertylist	*plist,*qclist;


	/*Start the recipe */

	if (!pars) {
		cpl_msg_error (_id, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (_id, "Frameset not found");
		return -1;
	}

	/* Initialise a few things */
	omega_qcheck_init();

	/* Retrieve input parameters */
	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.ExtensionNumber") ;
	omega_qcheck_config.extnum = cpl_parameter_get_int(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.OverscanMethod") ;
	omega_qcheck_config.oc = cpl_parameter_get_int(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.NumberIter") ;
	omega_qcheck_config.niter = cpl_parameter_get_int(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.rej_threshold") ;
	omega_qcheck_config.rejt = cpl_parameter_get_double(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.PAF") ;
	omega_qcheck_config.paf = cpl_parameter_get_bool(par) ;

	/* Identify the RAW and CALIB frames in the input frameset */
	if (oc_dfs_set_groups(set)) {
		cpl_msg_error(_id, "Cannot identify RAW and CALIB frames") ;
		omega_qcheck_tidy();
		return -1 ;
	}


	/* Verify the frameset contents. */
	/*Make sure there are exactly 2 lifetest frames */
	if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
			&nlab)) == NULL) {
		cpl_msg_error(_id,"Cannot labelise the input frameset");
		omega_qcheck_tidy();
		return -1;
	}
	if ((ps.domelist = omega_frameset_subgroup(set,ps.labels,nlab,
			LIFETEST_RAW)) == NULL) {
		cpl_msg_error(_id,"Cannot find dome frames in input frameset");
		omega_qcheck_tidy();
		return -1;
	}
	if (cpl_frameset_get_size(ps.domelist) < 2) {
		cpl_msg_error(_id,"Need exactly 2 (%s) frames to run this recipe",LIFETEST_RAW);
		omega_qcheck_tidy();
		return -1;
	}

	/* Set up some convenience variables */
	frame1 = cpl_frameset_get_frame_const(ps.domelist,0);
	frame2 = cpl_frameset_get_frame_const(ps.domelist,1);
	cpl_msg_info (_id,"Using %s frames: %s and %s",LIFETEST_RAW,cpl_frame_get_filename(frame1),
			cpl_frame_get_filename(frame2));

	mbias_frame = cpl_frameset_find(set, OMEGA_CALIB_BIAS);
	if (mbias_frame == NULL) {
		cpl_msg_error (_id,"No Master Bias is present in frame set");
		omega_qcheck_tidy();
		return -1;
	}

	cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(mbias_frame));

	/* Loop for each of the image extensions */
	omega_exten_range(omega_qcheck_config.extnum,&jst,&jfn);
	if(omega_qcheck_config.extnum == 0){
		cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_qcheck_config.extnum);
		omega_qcheck_tidy();
		return -1;
	}

	/* Check which instrument */
	if(omega_pfits_check_instrume(cpl_frameset_get_first_const(set)) == 1 &&
			omega_qcheck_config.extnum == 0 && jfn == 32)
		jfn = 8;

	for (j = jst; j <= jfn; j++) {
		cpl_msg_info(_id,"Beginning work on extension %d",j);
		isfirst = (j == jst);
		omega_qcheck_config.mean = 0.0;
		omega_qcheck_config.median = 0.0;
		omega_qcheck_config.stdev = 0.0;


		/* Check that this detector is live*/
		plist = cpl_propertylist_load(cpl_frame_get_filename(frame1),j);
		omega_pfits_get_detlive(plist,&live);
		if (! live) {
			cpl_msg_warning(_id,"First dome image detector not live");
			/*save dummy extension*/
			freeplist(plist);
			continue;
		}
		freeplist(plist);

		plist = cpl_propertylist_load(cpl_frame_get_filename(frame2),j);
		omega_pfits_get_detlive(plist,&live);
		if (! live) {
			cpl_msg_warning(_id,"Second dome image detector not live");
			/*save dummy extension*/
			freeplist(plist);
			continue;
		}
		freeplist(plist);

		/* Check overscan correction method consistency */
		oscan1 = omega_pfits_get_overscan(mbias_frame, j);
		if(oscan1 != omega_qcheck_config.oc) {
			cpl_msg_warning (_id, "Overscan correction mode for Master Bias (oc = %d) differs from "
					"the one used here (oc = %d)", oscan1, omega_qcheck_config.oc);
		}

		/* Load master bias*/
		mbias_image = cpl_image_load(cpl_frame_get_filename(mbias_frame), CPL_TYPE_FLOAT, 0, j);
		if (mbias_image == NULL) {
			cpl_msg_error(_id,"Cannot load MASTER BIAS");
			omega_qcheck_tidy();
			return -1;
		}

		cpl_msg_info (_id,"Doing trim and overscan correction on images");

		trim_raw1 = TrimOscanCorrect(frame1, omega_qcheck_config.oc, j);
		if(trim_raw1 == NULL){
			cpl_msg_error(_id,"Cannot trim first image");
			freeimage(mbias_image);
			omega_qcheck_tidy();
			return -1;
		}

		cpl_image_subtract(trim_raw1, mbias_image);

		trim_raw2 = TrimOscanCorrect(frame2, omega_qcheck_config.oc, j);
		if(trim_raw2 == NULL){
			cpl_msg_error(_id,"Cannot trim second image");
			freeimage(mbias_image);
			omega_qcheck_tidy();
			return -1;
		}

		cpl_image_subtract(trim_raw2, mbias_image);
		freeimage(mbias_image);

		cpl_image_divide(trim_raw1, trim_raw2);
		if(trim_raw1 == NULL){
			cpl_msg_error(_id,"Error in image division");
			omega_qcheck_tidy();
			return -1;
		}

		freeimage(trim_raw2);

		/*Calculate statistics iteratively*/
		ps.stats = omega_iter_stat_opts(trim_raw1,NULL,omega_qcheck_config.rejt,omega_qcheck_config.niter);
		if(ps.stats != NULL){
			omega_qcheck_config.mean = cpl_stats_get_mean(ps.stats);
			omega_qcheck_config.median = cpl_stats_get_median(ps.stats);
			omega_qcheck_config.stdev = cpl_stats_get_stdev(ps.stats);
		}
		else{
			cpl_msg_warning(_id,"Cannot calculate statistics iteratively");
		}

		freestats(ps.stats);
		freeimage(trim_raw1);

		/* Save the product */
		ps.result = cpl_table_new(1);
		cpl_table_new_column(ps.result, "MEAN", CPL_TYPE_DOUBLE);
		cpl_table_new_column(ps.result, "MEDIAN", CPL_TYPE_DOUBLE);
		cpl_table_new_column(ps.result, "STDEV", CPL_TYPE_DOUBLE);
		cpl_table_set_double(ps.result, "MEAN", 0, omega_qcheck_config.mean);
		cpl_table_set_double(ps.result, "MEDIAN", 0, omega_qcheck_config.median);
		cpl_table_set_double(ps.result, "STDEV", 0, omega_qcheck_config.stdev);

/*
		ps.domefits1 = omega_fits_load(frame1,CPL_TYPE_FLOAT,j);
		ps.eh = omega_fits_get_ehu(ps.domefits1);
*/

		/* Create QC list */
		qclist = cpl_propertylist_new();
		cpl_propertylist_append_double(qclist, "ESO QC QUICK CHECK MEAN",
				omega_qcheck_config.mean) ;
		cpl_propertylist_set_comment(qclist,"ESO QC QUICK CHECK MEAN","Mean of difference");

		cpl_propertylist_append_double(qclist, "ESO QC QUICK CHECK MEDIAN",
				omega_qcheck_config.median) ;
		cpl_propertylist_set_comment(qclist,"ESO QC QUICK CHECK MEDIAN","Median of difference");

		cpl_propertylist_append_double(qclist, "ESO QC QUICK CHECK STDEV",
				omega_qcheck_config.stdev) ;
		cpl_propertylist_set_comment(qclist,"ESO QC QUICK CHECK STDEV","Standard deviation of difference");

		if(isfirst){
			ps.proname = cpl_sprintf("%s_%s.fits", INSTRUME,LTEST_PROCATG);
			product_frame = omega_product_frame(ps.proname, LTEST_PROCATG, CPL_FRAME_TYPE_TABLE);
		}

		if(omega_save_table(ps.result,set,pars,NULL,qclist,ps.proname,RECIPE,product_frame,
				           NULL,isfirst) == -1){
			cpl_msg_error(cpl_func, "Cannot save the product");
			freeplist(qclist);
			omega_qcheck_tidy();
			return -1 ;
		}

/*
		if (omega_qcheck_save(set,pars) == -1) {
			cpl_msg_error(_id, "Cannot save the product");
			omega_qcheck_tidy();
			return -1 ;
		}
*/

		freetable(ps.result);
		freeplist(qclist);
//		freefits(ps.domefits1);
//		ps.domefits1 = NULL;

	}     /* go back and do next extension */


	/*Clean up */
	omega_qcheck_tidy();

	return 0;

}

/*----------------------------------------------------------------------------*/
/**
@brief    This recipe is used to generate a quick check on the detector
responsivity for one particular chip (Calfile 547).
@param    set    frame set
@param	params parameters list
@return   0 if everything is ok

The recipe takes as input exactly one raw domeflat, and a master bias.
Optionally, a previous QuickCheckFrame can be specified for comparison,
and an overscan correction mode can be set as a parameter.

*/
/*----------------------------------------------------------------------------*/

static int omega_qcheck(cpl_frameset *set, cpl_parameterlist *pars)
{

	int j,jst,jfn,oscan1,isfirst;
	cpl_size nlab;
	cpl_frame 		*mbias_frame;
	const cpl_frame 	*frame1;
	cpl_image 		*mbias_image;
	cpl_image		*trim_raw1;
	cpl_frame *product_frame;
	cpl_parameter 	*par;
	cpl_propertylist *qclist,*alist;


	/*Start the recipe */

	if (!pars) {
		cpl_msg_error (cpl_func, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (cpl_func, "Frameset not found");
		return -1;
	}

	/* Initialise a few things */
	omega_qcheck_init();

	/* Retrieve input parameters */
	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.ExtensionNumber") ;
	omega_qcheck_config.extnum = cpl_parameter_get_int(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.OverscanMethod") ;
	omega_qcheck_config.oc = cpl_parameter_get_int(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.NumberIter") ;
	omega_qcheck_config.niter = cpl_parameter_get_int(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.rej_threshold") ;
	omega_qcheck_config.rejt = cpl_parameter_get_double(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_qcheck.PAF") ;
	omega_qcheck_config.paf = cpl_parameter_get_bool(par) ;

	/* Identify the RAW and CALIB frames in the input frameset */
	if (oc_dfs_set_groups(set)) {
		cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames") ;
		omega_qcheck_tidy();
		return -1 ;
	}


	/* Verify the frameset contents. */
	/*Make sure there is exactly 1 lifetest frames */
	if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
			&nlab)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot labelise the input frameset");
		omega_qcheck_tidy();
		return -1;
	}
	if ((ps.domelist = omega_frameset_subgroup(set,ps.labels,nlab,
			LIFETEST_RAW)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot find dome frame in input frameset");
		omega_qcheck_tidy();
		return -1;
	}
	if (cpl_frameset_get_size(ps.domelist) < 1) {
		cpl_msg_error(cpl_func,"Need exactly 1 (%s) frame to run this recipe",LIFETEST_RAW);
		omega_qcheck_tidy();
		return -1;
	}

	/* Set up some convenience variables */
	frame1 = cpl_frameset_get_frame_const(ps.domelist,0);
	cpl_msg_info (cpl_func,"Using %s frame: %s",LIFETEST_RAW,cpl_frame_get_filename(frame1));

	mbias_frame = cpl_frameset_find(set, OMEGA_CALIB_BIAS);
	if (mbias_frame == NULL) {
		cpl_msg_error (cpl_func,"No Master Bias is present in frame set");
		omega_qcheck_tidy();
		return -1;
	}

	cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(mbias_frame));

	/* Loop for each of the image extensions */
	omega_extensions(frame1,omega_qcheck_config.extnum,&jst,&jfn);
	if(omega_qcheck_config.extnum == 0){
		cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_qcheck_config.extnum);
		omega_qcheck_tidy();
		return -1;
	}

	for (j = jst; j <= jfn; j++) {
		cpl_msg_info(cpl_func,"Working on extension %d",j);
		isfirst = (j == jst);
		omega_qcheck_config.mean = 0.0;
		omega_qcheck_config.median = 0.0;
		omega_qcheck_config.stdev = 0.0;


		/* Check that this detector is live*/
		/*
    	plist = cpl_propertylist_load(cpl_frame_get_filename(frame1),j);
    	omega_pfits_get_detlive(plist,&live);
    	if (! live) {
    	    cpl_msg_warning(cpl_func,"First dome image detector not live");
             save dummy extension
            freeplist(plist);
            continue;
         }
        freeplist(plist);
		 */

		/* Check overscan correction method consistency */
		oscan1 = omega_pfits_get_overscan(mbias_frame, j);
		if(oscan1 != omega_qcheck_config.oc) {
			cpl_msg_warning (cpl_func, "Overscan correction mode for Master Bias (oc = %d) differs from "
					"the one used here (oc = %d)", oscan1, omega_qcheck_config.oc);
		}

		/* Load master bias*/
		mbias_image = cpl_image_load(cpl_frame_get_filename(mbias_frame), CPL_TYPE_FLOAT, 0, j);
		if (mbias_image == NULL) {
			cpl_msg_error(cpl_func,"Cannot load MASTER BIAS");
			omega_qcheck_tidy();
			return -1;
		}

		ps.domefits1 = omega_fits_load(frame1,CPL_TYPE_FLOAT,j);

		trim_raw1 = omega_trim_oscan_correct(ps.domefits1, omega_qcheck_config.oc);
		if(trim_raw1 == NULL){
			cpl_msg_error(cpl_func,"Cannot trim input image");
			freeimage(mbias_image);
			omega_qcheck_tidy();
			return -1;
		}

		cpl_image_subtract(trim_raw1, mbias_image);
		freeimage(mbias_image);

		/*Calculate statistics iteratively*/
		ps.stats = omega_iter_stat_opts(trim_raw1,NULL,omega_qcheck_config.rejt,omega_qcheck_config.niter);
		if(ps.stats != NULL){
			omega_qcheck_config.mean = cpl_stats_get_mean(ps.stats);
			omega_qcheck_config.median = cpl_stats_get_median(ps.stats);
			omega_qcheck_config.stdev = cpl_stats_get_stdev(ps.stats);
		}
		else{
			cpl_msg_warning(cpl_func,"Cannot calculate statistics iteratively");
		}

		freestats(ps.stats);
		freeimage(trim_raw1);

		/* FIXME: include comparison with REFERENCE frame */
		/* Save the product */
		ps.result = cpl_table_new(1);
		cpl_table_new_column(ps.result, "MEAN", CPL_TYPE_DOUBLE);
		cpl_table_new_column(ps.result, "MEDIAN", CPL_TYPE_DOUBLE);
		cpl_table_new_column(ps.result, "STDEV", CPL_TYPE_DOUBLE);
		cpl_table_set_double(ps.result, "MEAN", 0, omega_qcheck_config.mean);
		cpl_table_set_double(ps.result, "MEDIAN", 0, omega_qcheck_config.median);
		cpl_table_set_double(ps.result, "STDEV", 0, omega_qcheck_config.stdev);

		/* Create QC list */
		qclist = cpl_propertylist_new();
		cpl_propertylist_append_double(qclist, "ESO QC QUICK CHECK MEAN",
				omega_qcheck_config.mean) ;
		cpl_propertylist_set_comment(qclist,"ESO QC QUICK CHECK MEAN","Mean of difference");

		cpl_propertylist_append_double(qclist, "ESO QC QUICK CHECK MEDIAN",
				omega_qcheck_config.median) ;
		cpl_propertylist_set_comment(qclist,"ESO QC QUICK CHECK MEDIAN","Median of difference");

		cpl_propertylist_append_double(qclist, "ESO QC QUICK CHECK STDEV",
				omega_qcheck_config.stdev) ;
		cpl_propertylist_set_comment(qclist,"ESO QC QUICK CHECK STDEV","Standard deviation of difference");

		if(isfirst){
			ps.proname = cpl_sprintf("%s_%s.fits", INSTRUME,LTEST_PROCATG);
			product_frame = omega_product_frame(ps.proname, LTEST_PROCATG, CPL_FRAME_TYPE_TABLE);
		}


		alist=cpl_propertylist_load_regexp(cpl_frame_get_filename(frame1),j,"EXTNAME",0);
		if(omega_save_table(ps.result,set,pars,alist,qclist,ps.proname,RECIPE,product_frame,
				           NULL,isfirst) == -1){
			cpl_msg_error(cpl_func, "Cannot save the product");
			cpl_propertylist_delete(alist);
			freeplist(qclist);
			omega_qcheck_tidy();
			return -1 ;
		}
		cpl_propertylist_delete(alist);
/*
		if (omega_qcheck_save(set,pars) == -1) {
			cpl_msg_error(cpl_func, "Cannot save the product");
			omega_qcheck_tidy();
			return -1 ;
		}
*/

		freetable(ps.result);
		freeplist(qclist);
		freefits(ps.domefits1);
		ps.domefits1 = NULL;
	}     /* go back and do next extension */


	/*Clean up */
	omega_qcheck_tidy();

	return 0;

}


/* Initialize the pointers */
static void omega_qcheck_init(void) {
	ps.labels = NULL;
	ps.domelist = NULL;
	ps.domefits1 = NULL;
	ps.stats = NULL;
	ps.eh = NULL;
	ps.proname = NULL;
	ps.result = NULL;
}

/* Free any allocated memory */
static void omega_qcheck_tidy(void) {
	freespace(ps.labels);
	freeframeset(ps.domelist);
	freefits(ps.domefits1);
	freestats(ps.stats);
	freespace(ps.proname);
	freetable(ps.result);
}

/**@}*/







