/*
 * omega_shutter.c
 *
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  Created on: Jan 4, 2011
 *      Author: agabasch
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include "omega_recipe.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup omega_shutter  Shutter Test
 *
 * @par Synopsis: This recipe tests illumination variations due to \n
 * inhomogeneities in the shutter blade speed. The recipe always \n
 * takes as input 4 raw dome flat frames and a master bias frame The \n
 * recipe expects the following tags in the Set of Frames:
 *
 * @par Input frames:
 * DOME_SHUTTERTEST\n
 * MASTER_BIAS\n
 *
 * @par Output frames:
 * SHUTTER_TEST\n
 *
 * @code
 *      #include omega_recipe.h
 * @endcode
 */
/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_shutter_create(cpl_plugin *);
static int omega_shutter_exec(cpl_plugin *);
static int omega_shutter_destroy(cpl_plugin *);
static int omega_shutter(cpl_frameset *, cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/

int omega_shutter_combine(cpl_parameterlist *pars, int xn);
int omega_shutter_biassubtract(cpl_parameterlist *pars, int xn);
static void omega_shutter_init(void);
static void omega_shutter_tidy(void);
static cpl_error_code  omega_shutter_basicsteps(cpl_frameset * frameset,
		cpl_image * mbias, cpl_imagelist * ilist_shortexp_up,
		float exptime_lower, float exptime_upper,
		const char * direction, int extension);
static cpl_matrix * omega_fill_ccdlayout(void);
static cpl_error_code  omega_average_ccd_table(cpl_imagelist * ilist_in,
		cpl_table * table_out );
static cpl_error_code omega_table_insert_xcoordinates(cpl_table * table);
static float  omega_get_exptime_limit(cpl_frameset * frameset);

/*-----------------------------------------------------------------------------
                            Static structures
 -----------------------------------------------------------------------------*/

static struct {
	/* Inputs. Parameters */
	int         oc;
}omega_shutter_config;

/* Input and Output */
static struct {

	cpl_size         *labels;
	const cpl_frame  *mbframe;
	cpl_frameset     *domelist;
}ps;

/*-----------------------------------------------------------------------------
                            Static global variables
 -----------------------------------------------------------------------------*/

/* Static variables */
#define RECIPE "omega_shutter"

static char omega_shutter_man[] =

		"This recipe tests illumination variations due to inhomogeneities "
		"in the \nshutter blade speed.\n"
		"\n"
		"Input files:\n\n"
		"  DO category:               Type:      Explanation:      Required: \n"
		"  DOME_SHUTTERTEST           Raw        Domeflat             Y      \n"
		"  and                                                               \n"
		"  MASTER_BIAS                Raw        Master bias          Y    \n\n"
		"Output files:\n\n"
		"  DO category:               Data type: Explanation:                \n"
		"  SHUTTER_DOWN               Products:  Collapsed CCD array         \n"
		"  and                                                               \n"
		"  SHUTTER_UP                 Products:  Collapsed CCD array      \n\n";

#define omega_shutter_exit(message)                 \
		{                                          \
	if (message) cpl_msg_error(cpl_func, message); \
	cpl_imagelist_delete(ilist_shortexp_up);       \
	cpl_imagelist_delete(ilist_longexp_up);        \
	cpl_imagelist_delete(ilist_shortexp_down);     \
	cpl_imagelist_delete(ilist_longexp_down);      \
	cpl_imagelist_delete(ilist_up);                \
	cpl_imagelist_delete(ilist_down);              \
	cpl_image_delete(mbias);                       \
	cpl_table_delete(table_up);                    \
	cpl_table_delete(table_down);                  \
	cpl_propertylist_delete(plistHeader);          \
	cpl_msg_indent_less();                         \
	return -1;                                     \
		}

#define omega_average_ccd_table_exit(message)       \
		{                                          \
	if (message) cpl_msg_error(cpl_func, message); \
	cpl_matrix_delete(ccd_mapping);                \
	cpl_imagelist_delete(ilist_tmp);               \
	cpl_image_delete(image_tmp);                   \
	cpl_image_delete(image_collapsed);             \
	cpl_table_delete(table_local);                 \
	cpl_msg_indent_less();                         \
	return cpl_error_set(cpl_func, cpl_error_get_code()); \
		}

#define omega_shutter_basicsteps_exit(message)      \
		{                                          \
	if (message) cpl_msg_error(cpl_func, message); \
	cpl_propertylist_delete(plist);                \
	cpl_image_delete(image_local);                 \
	cpl_image_delete(image_local_collapsed);       \
	cpl_free(direction_local);                     \
    cpl_msg_indent_less();                         \
	return cpl_error_set(cpl_func, cpl_error_get_code()); \
		}

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
	cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
	cpl_plugin  *   plugin = &recipe->interface ;

	if (cpl_plugin_init(plugin,
			CPL_PLUGIN_API,
			OMEGA_BINARY_VERSION,
			CPL_PLUGIN_TYPE_RECIPE,
			"omega_shutter",
			"OMEGA - Test the shutter homogeneity",
			omega_shutter_man,
			"Armin Gabasch",
			PACKAGE_BUGREPORT,
			omega_get_license(),
			omega_shutter_create,
			omega_shutter_exec,
			omega_shutter_destroy)) {
		cpl_msg_error(cpl_func, "Plugin initialization failed");
		(void)cpl_error_set_where(cpl_func);
		return 1;
	}

	if (cpl_pluginlist_append(list, plugin)) {
		cpl_msg_error(cpl_func, "Error adding plugin to list");
		(void)cpl_error_set_where(cpl_func);
		return 1;
	}

	return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
static int omega_shutter_create(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	cpl_parameter * p ;

	/* Do not create the recipe if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Create the parameters list in the cpl_recipe object */
	recipe->parameters = cpl_parameterlist_new() ;

	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Parameter list allocation failed");
		cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
	}

	/* Fill the parameters list */

	/* General Parameters */


	p = cpl_parameter_new_range("omega.omega_shutter.OverscanMethod",
			CPL_TYPE_INT,
			"Overscan Correction Method",
			"omega_shutter",
			0, 0, 6);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_shutter.ExtensionNumber",
			CPL_TYPE_INT,
			"DUMMY parameter to ensure Paranal DFS compatibility!!",
			"omega_shutter",
			-1) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
	cpl_parameterlist_append(recipe->parameters, p) ;



	return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_shutter_exec(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	int recipe_status;
	cpl_errorstate initial_errorstate = cpl_errorstate_get();

	/* Return immediately if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Verify parameter and frame lists */
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}
	if (recipe->frames == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Invoke the recipe */
	recipe_status = omega_shutter(recipe->frames, recipe->parameters);

	/* Ensure DFS-compliance of the products */
	if (cpl_dfs_update_product_header(recipe->frames)) {
		if (!recipe_status) recipe_status = (int)cpl_error_get_code();
	}

	if (!cpl_errorstate_is_equal(initial_errorstate)) {
		/* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
		cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
	}

	return recipe_status;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_shutter_destroy(cpl_plugin * plugin)
{
	cpl_recipe  *recipe;

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	cpl_parameterlist_delete(recipe->parameters);

	return 0 ;
}

static int omega_shutter(cpl_frameset *set, cpl_parameterlist *pars)
{

	int j=0,jst=0,jfn=0;
	cpl_size nlab=0;
	int nflats = 0;
	int oscan1 = 0;
	cpl_parameter 	*par   =NULL;
	cpl_imagelist * ilist_shortexp_up   =NULL;
	cpl_imagelist * ilist_longexp_up    =NULL;
	cpl_imagelist * ilist_shortexp_down =NULL;
	cpl_imagelist * ilist_longexp_down  =NULL;
	cpl_imagelist * ilist_up    =NULL;
	cpl_imagelist * ilist_down =NULL;
	cpl_image     * mbias=NULL;
	cpl_table     * table_up=NULL;
	cpl_table     * table_down=NULL;
	cpl_propertylist  * plistHeader=NULL;
	float           shutter_explimit=0.;
	//cpl_errorstate     prestate=cpl_errorstate_get();

	/*Start the recipe */

	if (!pars) {
		cpl_msg_error (cpl_func, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (cpl_func, "Frameset not found");
		return -1;
	}

	/* Retrieve input parameters */
	par = cpl_parameterlist_find(pars, "omega.omega_shutter.OverscanMethod") ;
	omega_shutter_config.oc = cpl_parameter_get_int(par) ;

	/* Identify the RAW and CALIB frames in the input frameset */
	if (oc_dfs_set_groups(set)) {
		cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames") ;
		return -1 ;
	}

	/*Initialized things*/
	omega_shutter_init();

	/* Verify the frameset contents. */
	if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
			&nlab)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot labelise the input frameset");
		omega_shutter_tidy();
		return -1;
	}
	if ((ps.domelist = omega_frameset_subgroup(set,ps.labels,nlab,
			SHUTTER_RAW)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot find dome frames in input frameset");
		omega_shutter_tidy();
		return -1;
	}

	nflats = cpl_frameset_count_tags(ps.domelist, SHUTTER_RAW);
	if (nflats < 4) {
		cpl_msg_error (cpl_func, "Need at least 4 (%s) frames to run "
				"this recipe", SHUTTER_RAW);
		omega_shutter_tidy();
		return -1;
	}

	cpl_msg_info (cpl_func,"There are %d %s in frame set",nflats, SHUTTER_RAW);

	/* Check for calibration frames */
	/* Master Bias */
	ps.mbframe = cpl_frameset_find_const(set, OMEGA_CALIB_BIAS);
	if (ps.mbframe == NULL) {
		cpl_msg_error (cpl_func, "Need at least 1 (%s) frames to run "
				"this recipe", OMEGA_CALIB_BIAS);
		omega_shutter_tidy();
		return -1;
	}
	else{
		cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BIAS,
				cpl_frame_get_filename(ps.mbframe));
	}


	/* Loop for each of the image extensions */
	jst=1;   /*First extension*/
	jfn=32;  /*Last extension*/


	ilist_shortexp_up   =cpl_imagelist_new();
	ilist_longexp_up    =cpl_imagelist_new();
	ilist_shortexp_down =cpl_imagelist_new();
	ilist_longexp_down  =cpl_imagelist_new();

	shutter_explimit=omega_get_exptime_limit(ps.domelist);
	cpl_msg_info(cpl_func, "Limit to discriminate long and short exposures:"
			" %f second",shutter_explimit);

	for (j = jst; j <= jfn; j++) {
		cpl_msg_info(cpl_func,"Beginning work on extension %02d",j);
		cpl_msg_indent_more();

		/* Check overscan correction method consistency */
		if(ps.mbframe != NULL){
			oscan1 = omega_pfits_get_overscan(ps.mbframe, j);
			if(oscan1 != omega_shutter_config.oc) {
				cpl_msg_warning (cpl_func, "Overscan correction mode for Master"
						" Bias (oc = %d) differs from the one used here "
						"(oc = %d)", oscan1, omega_shutter_config.oc);
			}
		}

		/*Load master bias*/
		mbias = cpl_image_load(cpl_frame_get_filename(ps.mbframe),
				CPL_TYPE_FLOAT,0,j);
		if(mbias==NULL){
			cpl_msg_error (cpl_func, "Unable to load extension %d of the master"
					" bias file %s", j, cpl_frame_get_filename(ps.mbframe));
			omega_shutter_tidy();
			omega_shutter_exit(NULL);
		}

		/*Overscan correction, bias subtraction, exposure time normalisation*/

		if(omega_shutter_basicsteps(ps.domelist, mbias, ilist_shortexp_up  ,
				FLT_MIN       , shutter_explimit, "up"  ,j) != CPL_ERROR_NONE) {
			cpl_msg_error (cpl_func, "Unable to perform basic image calibration"
					" on extension %d ", j);
			omega_shutter_tidy();
			omega_shutter_exit(NULL);
		}

		if(omega_shutter_basicsteps(ps.domelist, mbias, ilist_longexp_up   ,
				shutter_explimit, FLT_MAX        , "up"  ,j) != CPL_ERROR_NONE){
			cpl_msg_error (cpl_func, "Unable to perform basic image calibration"
					" on extension %d ", j);
			omega_shutter_tidy();
			omega_shutter_exit(NULL);
		}

		if(omega_shutter_basicsteps(ps.domelist, mbias, ilist_shortexp_down,
				FLT_MIN         , shutter_explimit, "down",j)!= CPL_ERROR_NONE){
			cpl_msg_error (cpl_func, "Unable to perform basic image calibration"
					" on extension %d ", j);
			omega_shutter_tidy();
			omega_shutter_exit(NULL);
		}

		if(omega_shutter_basicsteps(ps.domelist, mbias, ilist_longexp_down ,
				shutter_explimit, FLT_MAX         , "down",j)!= CPL_ERROR_NONE){
			cpl_msg_error (cpl_func, "Unable to perform basic image calibration"
					" on extension %d ", j);
			omega_shutter_tidy();
			omega_shutter_exit(NULL);
		}

		cpl_image_delete(mbias); mbias=NULL;
		cpl_msg_indent_less();
	}

	/*Divide the short exposure by the long exposure*/

	cpl_msg_info(cpl_func,"Dividing long exposures through short exposures");

	if(cpl_imagelist_divide(ilist_shortexp_up,ilist_longexp_up)!=
			CPL_ERROR_NONE){
		omega_shutter_tidy();
		omega_shutter_exit("Can not divide long exposure through short exposure");
	}

	if(cpl_imagelist_divide(ilist_shortexp_down,ilist_longexp_down)!=
			CPL_ERROR_NONE){
		omega_shutter_tidy();
		omega_shutter_exit("Can not divide long exposure through short exposure");
	}


	ilist_up=cpl_imagelist_new();
	ilist_down=cpl_imagelist_new();
	table_up=cpl_table_new(0);
	table_down=cpl_table_new(0);

	cpl_table_new_column(table_up,"RowAveraged",CPL_TYPE_FLOAT);
	cpl_table_new_column(table_down,"RowAveraged",CPL_TYPE_FLOAT);


	cpl_msg_info(cpl_func,"Final average-collapsing the full CCD array in "
			"y-direction");

	if(omega_average_ccd_table(ilist_shortexp_up, table_up)!= CPL_ERROR_NONE){
		omega_shutter_tidy();
		omega_shutter_exit("Can not collapse CCD array in y-direction");
	}

	if(omega_average_ccd_table(ilist_shortexp_down, table_down)!=CPL_ERROR_NONE){
		omega_shutter_tidy();
		omega_shutter_exit("Can not collapse CCD array in y-direction");
	}

	if(omega_table_insert_xcoordinates(table_up)!= CPL_ERROR_NONE ||
			omega_table_insert_xcoordinates(table_down)!= CPL_ERROR_NONE){
		omega_shutter_tidy();
		omega_shutter_exit("Can not insert x-coordinates in the product table");
	}


	plistHeader=cpl_propertylist_new();
	cpl_propertylist_append_string(plistHeader, CPL_DFS_PRO_CATG, "SHUTTER_UP");

	if(cpl_dfs_save_table(set, NULL, pars, set, NULL, table_up, NULL,
			"omega_shutter", plistHeader, NULL, PACKAGE "/" PACKAGE_VERSION ,
			"omega_shutter_up.fits") != CPL_ERROR_NONE){
		omega_shutter_tidy();
		omega_shutter_exit("Can not save the product SHUTTER_UP");
	}

	cpl_propertylist_update_string(plistHeader, CPL_DFS_PRO_CATG, "SHUTTER_DOWN");
	if (cpl_dfs_save_table(set, NULL, pars, set, NULL, table_down, NULL,
			"omega_shutter", plistHeader, NULL, PACKAGE "/" PACKAGE_VERSION ,
			"omega_shutter_down.fits")!=CPL_ERROR_NONE){
		omega_shutter_tidy();
		omega_shutter_exit("Can not save the product SHUTTER_DOWN");
	}
	cpl_propertylist_delete(plistHeader); plistHeader=NULL;

	/*Clean up */
	cpl_imagelist_delete(ilist_shortexp_up);     ilist_shortexp_up=NULL;
	cpl_imagelist_delete(ilist_longexp_up);      ilist_longexp_up=NULL;
	cpl_imagelist_delete(ilist_shortexp_down);   ilist_shortexp_down=NULL;
	cpl_imagelist_delete(ilist_longexp_down);    ilist_longexp_down=NULL;
	cpl_imagelist_delete(ilist_up);              ilist_up=NULL;
	cpl_imagelist_delete(ilist_down);            ilist_down=NULL;
	cpl_table_delete(table_up);                  table_up=NULL;
	cpl_table_delete(table_down);                table_down=NULL;
	omega_shutter_tidy();

	return 0;
}

/**
 * @brief
 *  Averages the imagelist according to the omega ccd layout
 * @param    ilist       the input images list
 * @param    table_out   the table where to insert the result
 *
 *
 * @return cpl error code
 */
static cpl_error_code  omega_average_ccd_table(cpl_imagelist * ilist_in,
		cpl_table * table_out ){

	cpl_matrix    * ccd_mapping=NULL;
	cpl_imagelist * ilist_tmp=NULL;
	cpl_image     * image_tmp=NULL;
	cpl_image      * image_collapsed=NULL;
	cpl_table     * table_local=NULL;;
	int nrow=0, ncol=0, row=0, col=0;
	int ccd_index=0;
	cpl_errorstate    prestate=cpl_errorstate_get();

	ccd_mapping=omega_fill_ccdlayout();
	//Test: cpl_error_set(cpl_func,CPL_ERROR_ACCESS_OUT_OF_RANGE);
	if (!cpl_errorstate_is_equal(prestate)){
		omega_average_ccd_table_exit(NULL);
	}

	nrow=cpl_matrix_get_nrow(ccd_mapping); /*4*/
	ncol=cpl_matrix_get_ncol(ccd_mapping); /*8*/
	if (!cpl_errorstate_is_equal(prestate)){
		omega_average_ccd_table_exit(NULL);
	}

	for (col=0; col<ncol;col++){
		cpl_msg_debug(cpl_func,"working on column: %d", col);
		cpl_msg_indent_more();
		ilist_tmp=cpl_imagelist_new();
		for (row=0; row<nrow;row++){
			cpl_msg_debug(cpl_func,"working on row: %d", row);
			ccd_index=cpl_matrix_get(ccd_mapping,row,col)-1;
			cpl_msg_debug(cpl_func,"fits file extension: %d", ccd_index+1);
			image_tmp=cpl_image_duplicate(cpl_imagelist_get(ilist_in,ccd_index));
			cpl_imagelist_set(ilist_tmp,image_tmp,row);
		}
		if (!cpl_errorstate_is_equal(prestate)){
			omega_average_ccd_table_exit(NULL);
		}

		image_collapsed=cpl_imagelist_collapse_create(ilist_tmp);
		if (!cpl_errorstate_is_equal(prestate)){
			omega_average_ccd_table_exit(NULL);
		}
		table_local=cpl_table_new(cpl_image_get_size_x(image_collapsed));
		cpl_table_new_column(table_local,"RowAveraged",CPL_TYPE_FLOAT);
		cpl_table_copy_data_float(table_local,"RowAveraged",
				cpl_image_get_data_float(image_collapsed));
		if (!cpl_errorstate_is_equal(prestate)){
			omega_average_ccd_table_exit(NULL);
		}

		cpl_table_insert(table_out, table_local, cpl_table_get_nrow(table_out));
		cpl_table_delete(table_local); table_local=NULL;
		cpl_image_delete(image_collapsed); image_collapsed=NULL;
		cpl_imagelist_delete(ilist_tmp); ilist_tmp=NULL;
		cpl_msg_indent_less();
	}

	cpl_matrix_delete(ccd_mapping); ccd_mapping=NULL;

	/* Propagate error, if any */
	return cpl_error_set(cpl_func, cpl_error_get_code());

}


/**
 * @brief
 *   Fill a matrix with the OmegaCAM fits file extension numbers
 *
 *
 * @return Pointer to the new matriX
 */
static cpl_matrix * omega_fill_ccdlayout(void){
	cpl_matrix * ccd=NULL;


	ccd=cpl_matrix_new(4,8);

	cpl_matrix_set(ccd,0,0 ,1 );
	cpl_matrix_set(ccd,0,1 ,2 );
	cpl_matrix_set(ccd,0,2 ,3 );
	cpl_matrix_set(ccd,0,3 ,4 );
	cpl_matrix_set(ccd,0,4 ,17);
	cpl_matrix_set(ccd,0,5 ,18);
	cpl_matrix_set(ccd,0,6 ,19);
	cpl_matrix_set(ccd,0,7 ,20);

	cpl_matrix_set(ccd,1,0 ,5 );
	cpl_matrix_set(ccd,1,1 ,6 );
	cpl_matrix_set(ccd,1,2 ,7 );
	cpl_matrix_set(ccd,1,3 ,8 );
	cpl_matrix_set(ccd,1,4 ,21);
	cpl_matrix_set(ccd,1,5 ,22);
	cpl_matrix_set(ccd,1,6 ,23);
	cpl_matrix_set(ccd,1,7 ,24);

	cpl_matrix_set(ccd,2,0 ,9);
	cpl_matrix_set(ccd,2,1 ,10);
	cpl_matrix_set(ccd,2,2 ,11);
	cpl_matrix_set(ccd,2,3 ,12);
	cpl_matrix_set(ccd,2,4 ,25);
	cpl_matrix_set(ccd,2,5 ,26);
	cpl_matrix_set(ccd,2,6 ,27);
	cpl_matrix_set(ccd,2,7 ,28);


	cpl_matrix_set(ccd,3,0 ,13);
	cpl_matrix_set(ccd,3,1 ,14);
	cpl_matrix_set(ccd,3,2 ,15);
	cpl_matrix_set(ccd,3,3 ,16);
	cpl_matrix_set(ccd,3,4 ,29);
	cpl_matrix_set(ccd,3,5 ,30);
	cpl_matrix_set(ccd,3,6 ,31);
	cpl_matrix_set(ccd,3,7 ,32);

	return ccd;
}

/**
 * @brief
 *  Performs basic image operations on a single extension: Triming and  Overscan
 *  correction, bias subtraction, median collapsing in y and exposure time
 *  normalisation
 * @param    frameset      Set of frame
 * @param    mbias         Master bias
 * @param    ilist         imagelist to be filled
 * @param    exptime_lower lower exposure time limit
 * @param    exptime_upper upper exposure time limit
 * @param    direction     shutter direction
 * @param    extension     extension to work on
 *
 *
 * @return cpl error code
 */
static cpl_error_code  omega_shutter_basicsteps(cpl_frameset * frameset,
		cpl_image * mbias, cpl_imagelist * ilist, float exptime_lower,
		float exptime_upper, const char * direction, int extension)
{
	char             * direction_local=NULL;
	double             local_TMOPEN=0.;
	float              exptime_local=0;
	cpl_propertylist * plist=NULL;
	cpl_image        * image_local=NULL;
	cpl_image        * image_local_collapsed=NULL;
	cpl_frame        * cur_frame=NULL;
	cpl_errorstate     prestate = cpl_errorstate_get();

	cur_frame=cpl_frameset_get_first(frameset);

	/*Looping over frames*/
	while(cur_frame)
	{
		cpl_free(direction_local); direction_local=NULL;
		plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame),0);
		local_TMOPEN=cpl_propertylist_get_double(plist,
				"ESO DET SHUT TMOPEN");
		if(cpl_error_get_code() != CPL_ERROR_NONE)
		{
			omega_shutter_basicsteps_exit("Missing keyword "
					"ESO DET SHUT TMOPEN in scientific header");
		}
		cpl_msg_debug(cpl_func, "ESO DET SHUT TMOPEN: %g", local_TMOPEN);
		if(local_TMOPEN>0.){
			direction_local= cpl_sprintf("up");
		}
		else {
			direction_local= cpl_sprintf("down");
		}

		exptime_local=cpl_propertylist_get_float(plist, "EXPTIME");

		if ((cpl_error_get_code() != CPL_ERROR_NONE) || (exptime_local <=0)){
			cpl_error_set(cpl_func,CPL_ERROR_ILLEGAL_INPUT);
			omega_shutter_basicsteps_exit("Missing keyword "
					"EXPTIME in scientific header or EXPTIME <= ZERO");
		}

		if(strcmp(direction_local,direction)==0 && exptime_local >exptime_lower
				&& exptime_local <=exptime_upper)
		{
			cpl_msg_debug(cpl_func, "Loop: ESO DET SHUT TMOPEN: %g", local_TMOPEN);
			cpl_msg_debug(cpl_func, "Loop: direction_local: %s", direction_local);
			cpl_msg_debug(cpl_func, "Loop: direction: %s", direction);

			cpl_msg_info(cpl_func,"Trimming and overscan correcting");

			image_local=TrimOscanCorrect(cur_frame, omega_shutter_config.oc,
					extension);

			if(image_local==NULL){
				omega_shutter_basicsteps_exit("Can not trim and correct the "
						"overscan ");
			}

			//Subtract Bias !!!
			cpl_msg_info(cpl_func,"Subtracting the bias");
			cpl_image_subtract(image_local,mbias);
			if (!cpl_errorstate_is_equal(prestate)){
				omega_shutter_basicsteps_exit("Can not subtract the bias");
			}

			/*Collapse image in y direction by using the median*/
			cpl_msg_info(cpl_func,"Median collapsing in y-direction");
			image_local_collapsed=cpl_image_collapse_median_create(
					image_local, 0, 0, 0);
			if (!cpl_errorstate_is_equal(prestate)){
				omega_shutter_basicsteps_exit("Can not collapse the image in "
						"y-direction");
			}

			/*Devide image through the exposure time*/
			cpl_msg_info(cpl_func,"Normalizing the exposure time to 1 sec");
			cpl_image_divide_scalar(image_local_collapsed,exptime_local);
			if (!cpl_errorstate_is_equal(prestate)){
				omega_shutter_basicsteps_exit("can not normalize the exposure "
						"time to unity");
			}

			/*Insert image in the imagelist*/
			cpl_imagelist_set(ilist,image_local_collapsed,extension-1);

			/*Free memory*/
			cpl_image_delete(image_local); image_local=NULL;

		}
		cpl_propertylist_delete(plist); plist=NULL;
		cpl_free(direction_local); direction_local=NULL;
		cur_frame=cpl_frameset_get_next(frameset);
	}


	if (cpl_imagelist_get_size(ilist)<1){
		cpl_error_set_message(cpl_func,CPL_ERROR_ILLEGAL_INPUT,"Imagelist is "
				"empty! Please check the input raw frames");
	}

	/* Propagate error, if any */
	return cpl_error_set(cpl_func, cpl_error_get_code());

}
/**
 * @brief
 *  Inserts a new column to a table and fills it with the x-coordinates
 * @param    intable        Table where to insert the new column
 *
 * @return cpl error code
 */
static cpl_error_code omega_table_insert_xcoordinates(cpl_table * intable)
{
	/*Fill the column x_coordinate to better plot the results*/
	int            *  ptablecol=NULL;
	int               i = 0;
	cpl_errorstate    prestate = cpl_errorstate_get();

	cpl_table_new_column(intable, "x_coordinate", CPL_TYPE_INT);
	cpl_table_fill_column_window_int(intable, "x_coordinate", 0,
			cpl_table_get_nrow(intable), 0);
	ptablecol = cpl_table_get_data_int(intable, "x_coordinate");

	if (!cpl_errorstate_is_equal(prestate)){
		return cpl_error_set(cpl_func, cpl_error_get_code());
	}
	else {

		for(i = 0; i < cpl_table_get_nrow(intable); i++){
			ptablecol[i] = i + 1;
		}
	}
	/* Propagate error, if any */
	return cpl_error_set(cpl_func, cpl_error_get_code());
}

/* Initialize the pointers */
static void omega_shutter_init(void) {
	ps.labels = NULL;
	ps.mbframe = NULL;
	ps.domelist = NULL;

}

/* Free any allocated memory */
static void omega_shutter_tidy(void) {
	freespace(ps.labels);
	freeframeset(ps.domelist);

}

/**
 * @brief
 * Gets the exposure time limit to discriminate between long and short exposure
 * @param  frameset   The frameset where the exposure time is read
 *
 * @return (exptime_min+exptime_max)/2
 */

static float  omega_get_exptime_limit(cpl_frameset * frameset){

	cpl_vector       * exptime_vector=NULL;
	cpl_propertylist * plist=NULL;
	cpl_frame        * cur_frame=NULL;
	int                i=0;
	float              exptime_limit=0;

	exptime_vector=cpl_vector_new(cpl_frameset_get_size(frameset));

	cur_frame=cpl_frameset_get_first(frameset);

	i=0;
	/*Looping over frames*/
	while(cur_frame)
	{
		plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame),0);
		if(!cpl_propertylist_has(plist, "EXPTIME"))
		{
			cpl_msg_warning(cpl_func,"Can not find EXPTIME in file %s",
					cpl_frame_get_filename(cur_frame));
			cpl_propertylist_delete(plist);
			cpl_vector_delete(exptime_vector);
			return 0.;
		}

		cpl_vector_set(exptime_vector, i,
				cpl_propertylist_get_float(plist, "EXPTIME"));
		cpl_propertylist_delete(plist);
		i++;
		cur_frame=cpl_frameset_get_next(frameset);
	}

	exptime_limit=(cpl_vector_get_min(exptime_vector)+
			cpl_vector_get_max(exptime_vector))/2.;
	cpl_vector_delete(exptime_vector);

	return exptime_limit;

}

/**@}*/








