/* $Id: omega_standard_star.c,v 1.37 2012/08/22 09:51:58 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012/08/22 09:51:58 $
 * $Revision: 1.37 $
 * $Name: omega-1_0_2 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/
#include "omega_recipe.h"
#include "omega_science.h"

#include "omega_background.h"
#include "omega_cosmic.h"
#include "omega_photometry.h"
#include "omega_satellites.h"
#include "omega_wcscor.h"
#include "omega_utils.h"
/**
 * @defgroup omega_standard_star  Standard Stars
 *
 * @par Synopsis:
 * This recipe is used to reduced a series of standard star fields. The frames will be trimmed, corrected for \n
 * the overscan (if requested), corrected for the bias and the flat. A weighted average of all standard frames \n
 * will be taken. This image will be matched against the USNOA2 catalogues to derive a WCS correction. This\n
 * image is saved as the REDUCED_STD product. A photometric catalogue is created from this product and saved.\n
 * The final step is to derive the ZEROPOINTS table, which matches the image with the reference star catalogue.\n
 * The recipe expects the following tags in the Set of Frames:
 *
 * @par Input frames:\n
 * STD_ZEROPOINT\n
 * MASTER_BIAS\n
 * MASTER_FLAT\n
 * BAD_PIXELS_MAP\n
 * REFERENCE_STAR_CATALOG\n
 * MONITOR_REPORT\n
 * EXTINCTION_CURVE\n
 * MASTER_USNOA2_CATALOGUE
 *
 * @par Output frames:\n
 * REDUCED_STD\n
 * PHOTOMETRIC_TABLE\n
 * ZEROPOINTS
 *
 *
 * @code
 *      #include omega_recipe.h
 *      #include omega_background.h
 *      #include omega_cosmic.h
 *      #include omega_photometry.h
 *      #include omega_satellites.h
 *      #include omega_wcscor.h
 * @endcode
 */
/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_standard_star_create(cpl_plugin *) ;
static int omega_standard_star_exec(cpl_plugin *) ;
static int omega_standard_star_destroy(cpl_plugin *) ;
static int omega_standard_star(cpl_frameset *,cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/
omega_fits *omega_std_process(cpl_frameset *stdlist,cpl_parameterlist *pars, int ext);
static int omega_std_retrieve_input_param(const cpl_parameterlist  *parlist);
static int omega_std_load_calib(int ext);
static void omega_std_init(void);
static void omega_std_tidy(int level);


/*-----------------------------------------------------------------------------
                            Static structures
 -----------------------------------------------------------------------------*/
static struct {
	/* Inputs. Parameters */
	int extnum;
	int oc;
	int bckg;
	int paf;
	double MaxWcsShift;
	/* Outputs. QC parameters */
	int nmatches;
	int nstars;
	double mratio;
	double ZP;
	double ZPerr;
	double Extinction;
	double Exterr;
	double Seeing;
	double Ellipticity;

}omega_std_config;

/* Input and Output */
static struct {

	/* Calib frames */
	const cpl_frame *bpmfr;
	const cpl_frame *extfr;
	const cpl_frame *illumfr;
	const cpl_frame *mbiasfr;
	const cpl_frame *mflatfr;
	const cpl_frame *monitfr;
	const cpl_frame *reffr;
	const cpl_frame *stdfr;
	const cpl_frame *usnoa2;
	const cpl_frame *colfr;

	cpl_size         *labels;
	cpl_frameset     *stdlist;
	cpl_image        *mbias;
	cpl_image        *bpm;
	cpl_image        *mflat;
	cpl_image        *illum;

	/* Products */
	char *sname;
	char *pname;
	char *zname;
	omega_fits        *stdfits;
	cpl_table        *zptbl;
	cpl_table        *photbl;
	cpl_table        *extratbl;

}ps;


/* Static variables */

/* Static variables */
#define RECIPE "omega_standard_star"


/*----------------------------------------------------------------------------*/
/**
@brief    Build the list of available plugins, for this module.
@param    list    the plugin list
@return   0 if everything is ok

This function is exported.
*/
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
	cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
	cpl_plugin  *   plugin = &recipe->interface ;

	cpl_plugin_init(plugin,
			CPL_PLUGIN_API,
			OMEGA_BINARY_VERSION,
			CPL_PLUGIN_TYPE_RECIPE,
			"omega_standard_star",
			"OMEGA - This recipe is used to reduce standard star observations.",
			"The recipe assumes that the data is for a single CCD. \n"
			"The following operations are performed: \n\n"
			" 	 1. Trim and overscan correct; \n"
			"    2. Subtract the bias and divide by the flat field; \n"
			"    3. Correct for fringing, illumination and the background, if requested; \n"
			" 	 4. Extract the zero point from the data and calculate the extinction \n"
			"	    value and save both in the header and in one table. \n\n"
			" ------------------------------------------------------------------------ \n\n"
			" Mandatory inputs : \n\n"
			"	: a raw standard star frame \n"
			"	: a master bias frame (Calfile 541) \n"
			"	: a master flat field (Calfile 546) \n\n"
			"   : a master USNOA2 table \n"
			"   : a standard stars catalog table \n\n"
			" Optional inputs : \n\n"
			"	: fringe map (Calfile 545) \n"
			"	: bad pixels map (Calfile 522+535) \n"
			"	: illumination correction frame (Calfile 548) \n\n"
			" The recipe will save the extinction and zero point values in one table.",
			"Sandra Castro",
			"scastro@eso.org",
			omega_get_license(),
			omega_standard_star_create,
			omega_standard_star_exec,
			omega_standard_star_destroy) ;

	cpl_pluginlist_append(list, plugin) ;

	return 0;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Setup the recipe options
@param    plugin  the plugin
@return   0 if everything is ok

Create the recipe instance and make it available to the application using the
interface.
*/
/*----------------------------------------------------------------------------*/
static int omega_standard_star_create(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	cpl_parameter * p ;
	char *path = NULL;

	/* Do not create the recipe if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Create the parameters list in the cpl_recipe object */
	recipe->parameters = cpl_parameterlist_new() ;
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Parameter list allocation failed");
		cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
	}

	/* Fill the parameters list */
	/*FIXME: include range in all parameters */

	/* --------------- General parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_standard_star.ExtensionNumber",
			CPL_TYPE_INT,
			"FITS extension number to load (1 to 32). (-1 == all)",
			"omega_standard_star",
			-1) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_range("omega.omega_standard_star.OverscanMethod",
			CPL_TYPE_INT,
			"Overscan Correction Method",
			"overscan",
			0, 0, 6);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.PAF",
			CPL_TYPE_BOOL,
			"Boolean value to create PAF files. 1(TRUE), 0(FALSE)",
			"omega_standard_star",
			1) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "paf") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	/* --------------- Saturated pixels parameters ---------------------------*/

	p = cpl_parameter_new_value("omega.omega_standard_star.LowThreSatuPixel",
			CPL_TYPE_DOUBLE,
			"Low threshold for calculating the saturated pixels map.",
			"omega_standard_star",
			-500.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "lt-satu") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.HighThreSatuPixel",
			CPL_TYPE_DOUBLE,
			"High threshold for calculating the saturated pixels map.",
			"omega_standard_star",
			60000.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "ht-satu") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	/* --------------- Satellite detection parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_standard_star.DetectionThreSatellite",
			CPL_TYPE_DOUBLE,
			"Minimum SNR for pixels to contribute to Hough map.",
			"omega.Hough",
			4.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "det-sate") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.HoughThreshold",
			CPL_TYPE_DOUBLE,
			"Threshold for satellite tracks in Hough image.",
			"omega.Hough",
			400.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "hough-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* --------------- Fringing  parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_standard_star.SigmaFringeScaling",
			CPL_TYPE_DOUBLE,
			"Sigma threshold in image data for scaling estimate (fringes only)",
			"omega_standard_star",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sig-fr-sc") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.LowThreFringe",
			CPL_TYPE_DOUBLE,
			"Lower bound of fringes to include in scaling (fringes only)",
			"omega_standard_star",
			1.5) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "lt-fr") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.HighThreFringe",
			CPL_TYPE_DOUBLE,
			"Higher bound of fringes to include in scaling (fringes only)",
			"omega_standard_star",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "ht-fr") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* ----------------- Background  parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_standard_star.BackRestore",
			CPL_TYPE_BOOL,
			"Boolean value to restore background after illumination correction. 1(TRUE), 0(FALSE)",
			"omega_standard_star",
			1) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "backg-rest") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.BackThreshold",
			CPL_TYPE_DOUBLE,
			"Detection threshold for background in image (Sextractor DETECT_THRESH)",
			"omega_standard_star",
			1000.) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "backthre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* FIXME: need to subtract the background from the STD star?? */


	/* ------------------ PATH to external executable programs --------------*/
	path = cpl_sprintf("%s", OMEGA_BIN_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.BinPath",
			CPL_TYPE_STRING,
			"Path to any external executable program.",
			"omega.BinPath",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "bin-path");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	/* ------------------ Sextractor parameters ------------------------- */
	path = cpl_sprintf("%s/omega.sex", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.SexConfig",
			CPL_TYPE_STRING,
			"Path to Sextractor config file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-config");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);


	path = cpl_sprintf("%s/omega.conv", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.SexConv",
			CPL_TYPE_STRING,
			"Path to Sextractor convolution mask file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-conv");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);


	path = cpl_sprintf("%s/omega.param", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.SexParam",
			CPL_TYPE_STRING,
			"Path to Sextractor parameters file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-param");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	path = cpl_sprintf("%s/omega.nnw", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.SexNnw",
			CPL_TYPE_STRING,
			"Path to Sextractor neural network config file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-nnw");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	p = cpl_parameter_new_value("omega.omega_standard_star.SexDetThreshold",
			CPL_TYPE_DOUBLE,
			"Detection threshold for Sextractor sources.",
			"omega.sextractor",
			1.5);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-dthre");
	cpl_parameterlist_append(recipe->parameters, p);


	/* ------- Cosmic rays detection parameters for Sextractor ------------*/
	path = cpl_sprintf("%s/omega.cosmic.sex", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.SexCosmic",
			CPL_TYPE_STRING,
			"Path to Sextractor cosmic detection mode config file.",
			"omega.Sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-cosmic");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	path = cpl_sprintf("%s/omega.cosmic.param", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.SexCosmicParam",
			CPL_TYPE_STRING,
			"Path to Sextractor cosmic parameters file.",
			"omega.Sextractor",
			path);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"sex-cosmic-param");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	path = cpl_sprintf("%s/cosmic.ret", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_standard_star.SexCosmicFilt",
			CPL_TYPE_STRING,
			"Path to Sextractor filter mask for cosmic detection mode.",
			"omega.Sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-cosfilt");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	p = cpl_parameter_new_value("omega.omega_standard_star.SexCosmicDet",
			CPL_TYPE_DOUBLE,
			"Detection threshold for detecting cosmic rays in Sextractor",
			"omega.Sextractor",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-cosmic-det") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* ------------------ Astrometric parameters ------------------------- */
	p = cpl_parameter_new_value("omega.omega_standard_star.AstromDetThre",
			CPL_TYPE_DOUBLE,
			"Sextractor threshold to detect stars for astrometric correction.",
			"omega.Astrom",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "astrom-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_enum("omega.omega_standard_star.PlateSolution",
			CPL_TYPE_INT,
			"Number of constant plate fits to do for the astrometry solution.",
			"omega.Astrom",
			4,2,4,6);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "platesol");
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_standard_star.NumIterations",
			CPL_TYPE_INT,
			"Number of iterations for the plate solution fitting.",
			"omega.Astrom",
			3);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "niter");
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_range("omega.omega_standard_star.SextractorFlagLimit",
			CPL_TYPE_INT,
			"Sextractor FLAG upper limit to preselect good sources for the "
			"astrometric correction computation.", "omega.Astrom",
			7, 0, 256);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"SextractorFlagLimit") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_range("omega.omega_standard_star.SextractorClassStar",
			CPL_TYPE_DOUBLE,
			"Sextractor CLASS_STAR lower limit to preselect starlike objects "
			"for the astrometric correction computation.", "omega.Astrom",
			0.0, 0.0, 1.0);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"SextractorClassStar") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.MaxWcsShift",
			CPL_TYPE_DOUBLE,
			"Maximum allowed x and y pixel shift of the image when refining "
			"the WCS solution.",
			"omega.Astrom",
			200.);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "MaxWcsShift");
	cpl_parameterlist_append(recipe->parameters, p);



	/* ------------------ Photometric parameters ------------------------- */
	p = cpl_parameter_new_value("omega.omega_standard_star.PhotomDetThre",
			CPL_TYPE_DOUBLE,
			"Sextractor threshold to detect stars for photometric correction.",
			"omega.Photom",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "photom-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.MatchPhotomRadius",
			CPL_TYPE_DOUBLE,
			"Radius within which to match photometric stds and stars",
			"omega.Photom",
			10.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "radius") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_standard_star.PhotomAperture",
			CPL_TYPE_DOUBLE,
			"Aperture in pixels for photometric measurements in Sextractor",
			"omega.Photom",
			30.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "aperture") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	/* ------------------ Pattern Matching parameters ------------------------- */
	p = cpl_parameter_new_value("omega.omega_standard_star.Nstars",
			CPL_TYPE_INT,
			"Number of stars in image to use in "
			"pattern matching",
			"omega.Pattern",
			25);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "nstars");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);


	p = cpl_parameter_new_value("omega.omega_standard_star.Npattern",
			CPL_TYPE_INT,
			"Number of catalogue sources to use in "
			"pattern matching",
			"omega.Pattern",
			15);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "npattern");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);


	p = cpl_parameter_new_value("omega.omega_standard_star.Tolerance",
			CPL_TYPE_DOUBLE,
			"Max relative difference of angles and scales from their "
			"median value for match acceptance (tolerance)",
			"omega.Pattern",
			0.1);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "tol");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_standard_star.SearchPatternRadius",
			CPL_TYPE_DOUBLE,
			"Search radius for full PPM (pixels)",
			"omega.Pattern",
			10.0);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "pradius");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_standard_star.MaxPatternRadius",
			CPL_TYPE_DOUBLE,
			"Maximum search radius for full PPM (pixels)",
			"omega.Pattern",
			40.0);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "maxpradius");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	/* Return */
	return 0;
}

/*----------------------------------------------------------------------------*/
/**
@brief    Execute the plugin instance given by the interface
@param    plugin  the plugin
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_standard_star_exec(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	int recipe_status;

	/* Return immediately if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Verify parameter and frame lists */
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}
	if (recipe->frames == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Invoke the recipe */
	recipe_status = omega_standard_star(recipe->frames, recipe->parameters);

	/* Ensure DFS-compliance of the products */
	if (cpl_dfs_update_product_header(recipe->frames)) {
		if (!recipe_status) recipe_status = (int)cpl_error_get_code();
	}

	return recipe_status;

}

/*----------------------------------------------------------------------------*/
/**
@brief    Destroy what has been created by the 'create' function
@param    plugin  the plugin
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_standard_star_destroy(cpl_plugin * plugin)
{
	cpl_recipe  *   recipe;

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	cpl_parameterlist_delete(recipe->parameters) ;

	return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
@brief    This recipe searches for a single STANDARD star frame in frameset
and does a quick reduction on this image. It then derives the zeropoint for
it and the extinction table.

@param    parlist     the parameters list
@param    framelist   the frames list
@return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int omega_standard_star(cpl_frameset *set, cpl_parameterlist *pars)
{
	int j,jst,jfn,isfirst;
	cpl_size nlab;
	int nraw = 0;
	int oscan1 = 0;
	double exptime = 0.0;
	double gain = 1.0;

	char *sname = NULL;
	char *pname = NULL;
	char *zname = NULL;
	const char *stdname = "omega_std.fits";
	const char * maskname= NULL;
	omega_fits * stdfits_mask=NULL;
	cpl_image *image_simple_mask=NULL;
	cpl_frame *product_frame = NULL;
	cpl_frame *product_frame_mask=NULL ;
	cpl_frame *pframe = NULL;
	cpl_frame *zpframe = NULL;
	cpl_image *image = NULL;
	cpl_propertylist *plist,*xlist,*qclist,*alist;
	cpl_errorstate prestate = cpl_errorstate_get();

	/*Start the recipe */

	if (pars == NULL) {
		cpl_msg_error (cpl_func, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (cpl_func, "Frameset not found");
		return -1;
	}

	/* Retrieve input parameters and initiates specific parameters lists*/
	if(omega_std_retrieve_input_param(pars) != 0){
		cpl_msg_error(cpl_func,"Cannot retrieve input parameters. %s", cpl_error_get_message());
		return -1;
	}

	/* Identify the RAW and CALIB frames in the input frameset */
	if (oc_dfs_set_groups(set)) {
		cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames") ;
		return -1 ;
	}

	/*Initialized things*/
	omega_std_init();

	/* Verify the frameset contents. */
	if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
			&nlab)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot labelise the input frameset");
		omega_std_tidy(0);
		return -1;
	}

	if ((ps.stdlist = omega_frameset_subgroup(set,ps.labels,nlab,
			STD_RAW)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot find standard stars in input frameset");
		omega_std_tidy(0);
		return -1;
	}
	nraw = cpl_frameset_count_tags (ps.stdlist, STD_RAW);
	cpl_msg_info (cpl_func,"There are %d %s frames in frame set",nraw, STD_RAW);

	/* Check for calibration frames */
	/* Master Bias */
	ps.mbiasfr = cpl_frameset_find_const(set, OMEGA_CALIB_BIAS);
	if (ps.mbiasfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set", OMEGA_CALIB_BIAS);
		omega_std_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(ps.mbiasfr));

	/* Master Flat */
	ps.mflatfr = cpl_frameset_find_const(set, OMEGA_CALIB_FLAT);
	if(ps.mflatfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set", OMEGA_CALIB_FLAT);
		omega_std_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s", OMEGA_CALIB_FLAT, cpl_frame_get_filename(ps.mflatfr));


	/* Reference standard star list */
	ps.reffr = cpl_frameset_find (set, OMEGA_CALIB_REFSTAR);
	if (ps.reffr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set",OMEGA_CALIB_REFSTAR);
		omega_std_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s", OMEGA_CALIB_REFSTAR, cpl_frame_get_filename(ps.reffr));


	/* Monitoring report */
	ps.monitfr = cpl_frameset_find(set, OMEGA_CALIB_MONIT);
	if(ps.monitfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s in frame set",OMEGA_CALIB_MONIT );
		omega_std_tidy(0);
		return -1;
	}

	/* Extinction curve */
	ps.extfr = cpl_frameset_find (set,OMEGA_CALIB_EXTCUR);
	if(ps.extfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s in frame set", OMEGA_CALIB_EXTCUR);
		omega_std_tidy(0);
		return -1;
	}

	/* USNOA2 master table */
	ps.usnoa2 = cpl_frameset_find (set, OMEGA_USNOA2);
	if (ps.usnoa2 == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set",OMEGA_USNOA2);
		omega_std_tidy(0);
		return -1;
	}
	else{
		cpl_msg_info(cpl_func,"Using %s %s", OMEGA_USNOA2, cpl_frame_get_filename(ps.usnoa2));
	}

	/* Bad Pixels Map */
	ps.bpmfr = cpl_frameset_find (set,OMEGA_CALIB_BPM);
	if(ps.bpmfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s in frame set", OMEGA_CALIB_BPM);
		omega_std_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BPM,cpl_frame_get_filename(ps.bpmfr));

	/* Illumination frame */
	ps.illumfr = cpl_frameset_find(set, OMEGA_CALIB_ILLUM);
	if(ps.illumfr != NULL)
		cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_ILLUM,cpl_frame_get_filename(ps.illumfr));

	/* Colour terms */
	ps.colfr = cpl_frameset_find(set,OMEGA_CALIB_COLTERMS);
	if(ps.colfr != NULL)
		cpl_msg_info(cpl_func,"Using %s %s", OMEGA_CALIB_COLTERMS, cpl_frame_get_filename(ps.colfr));

	/* Get first frame from frame set */
	ps.stdfr = cpl_frameset_get_first_const(ps.stdlist);

	/* Check which image extension(s) to load */
	omega_extensions(ps.stdfr,omega_std_config.extnum,&jst,&jfn);
	if(omega_std_config.extnum == 0){
		cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_std_config.extnum);
		omega_std_tidy(0);
		return -1;
	}

	/* Loop over extensions */
	for (j = jst; j <= jfn; j++) {
		isfirst = (j == jst);
		cpl_msg_indent_more();
		cpl_msg_info(cpl_func,".....Working on extension %d.....",j);
		cpl_msg_indent_less();

		omega_std_config.nmatches = 0;
		omega_std_config.nstars = 0;
		omega_std_config.mratio = 0.0;
		omega_std_config.ZP = 0.0;
		omega_std_config.ZPerr = 0.0;
		omega_std_config.Extinction = 0.0;
		omega_std_config.Exterr = 0.0;
		omega_std_config.Seeing = 0.0;
		omega_std_config.Ellipticity = 0.0;

		/* Check overscan correction method consistency */
		if(ps.mbiasfr != NULL){
			oscan1 = omega_pfits_get_overscan(ps.mbiasfr, j);
			if(oscan1 != omega_std_config.oc) {
				cpl_msg_warning (cpl_func, "Overscan correction mode for Master Bias (oc = %d) differs from "
						"the one used here (oc = %d)", oscan1, omega_std_config.oc);
			}
		}

		/* Load calibration frames */
		if(omega_std_load_calib(j) != 0){
			cpl_msg_error(cpl_func,"Cannot load calibration frame(s)");
			omega_std_tidy(0);
			return -1;
		}

		/* Call the processing routine */
		if((ps.stdfits = omega_std_process(ps.stdlist,pars,j)) == NULL){
			cpl_msg_error(cpl_func, "Cannot reduce image extension");
			omega_std_tidy(0);
			return -1;
		}

		/* Free some space */
		omega_std_tidy(1);

		/* Save reduced stack */
		if(isfirst){
			ps.sname = cpl_sprintf("%s_%s.fits", INSTRUME,STD_PROCATG);
			product_frame = omega_product_frame(ps.sname, STD_PROCATG, CPL_FRAME_TYPE_IMAGE);
			maskname=cpl_sprintf("%s_%s.fits", INSTRUME,"REDUCED_STD_MASK");
			product_frame_mask = omega_product_frame(maskname, "REDUCED_STD_MASK", CPL_FRAME_TYPE_IMAGE);

		}
		cpl_msg_info(cpl_func,"Saving STD image");
		if((omega_save_fits(ps.stdfits,set,pars,NULL,CPL_BPP_IEEE_FLOAT,ps.sname,RECIPE,
				product_frame,NULL,isfirst)) == -1){
			cpl_msg_error(cpl_func,"Cannot save product");
			cpl_free(maskname);
			omega_std_tidy(0);
			return -1;
		}



		stdfits_mask=omega_fits_duplicate(ps.stdfits);
		/*Extracting the mask file from the image*/
		image_simple_mask=cpl_image_new_from_mask(cpl_image_get_bpm(stdfits_mask->image));
		cpl_image_delete(stdfits_mask->image); stdfits_mask->image=NULL;
		stdfits_mask->image=image_simple_mask;

		cpl_msg_info(cpl_func,"Saving STD MASK image");
		if((omega_save_fits(stdfits_mask,set,pars,NULL,CPL_BPP_8_UNSIGNED,maskname,RECIPE,
				product_frame_mask,NULL,isfirst)) == -1){
			cpl_msg_error(cpl_func,"Cannot save product");
			freefits(stdfits_mask);
			cpl_free(maskname);
			omega_std_tidy(0);
			return -1;
		}
		freefits(stdfits_mask);

		/* Correct image to unit time (flux) for photometry.
		 * If keyword is missing or set to 0.0, catch it and
		 * try the primary header. */
		plist = omega_fits_get_phu(ps.stdfits);
		xlist = omega_fits_get_ehu(ps.stdfits);

		alist = cpl_propertylist_new();
		cpl_propertylist_append_string(alist, "EXTNAME",
				cpl_propertylist_get_string(xlist, "EXTNAME"));
		cpl_propertylist_set_comment(alist,"EXTNAME", "Extension name");


		if(cpl_propertylist_has(xlist,"ESO QC STD PPM FAILURE")){
			cpl_propertylist_append_int(alist, "ESO QC STD PPM FAILURE",
					cpl_propertylist_get_int(xlist, "ESO QC STD PPM FAILURE"));
			cpl_propertylist_set_comment(alist, "ESO QC STD PPM FAILURE", "PPM failure if ZERO");
		}
		/*Get gain*/
		omega_pfits_get_conad(xlist,&gain);

		/* Add DRS keywords */
		cpl_propertylist_update_int(alist, "ESO DRS OVERSCAN METHOD", omega_std_config.oc );
		cpl_propertylist_set_comment(alist, "ESO DRS OVERSCAN METHOD", "overscan correction method");

		exptime = omega_pfits_get_exptime(xlist);
		if(exptime <= 0.0)
			exptime = omega_pfits_get_exptime(plist);

		image = cpl_image_duplicate(omega_fits_get_image(ps.stdfits));

		if(exptime > 0.0)
			cpl_image_divide_scalar(image, exptime);

		cpl_propertylist_save(plist, stdname, CPL_IO_CREATE);
		cpl_image_save(image,stdname,CPL_BPP_IEEE_DOUBLE,xlist,CPL_IO_EXTEND);

		freeimage(image);
		freefits(ps.stdfits);
		ps.stdfits = NULL;

		/* Create photometric table */
		if((ps.photbl = omega_photom_tbl(stdname, gain, ps.reffr,
				pars, &ps.extratbl)) == NULL){
			cpl_msg_warning(cpl_func,"Cannot create photometric table");
			cpl_msg_warning(cpl_func,"Creating an empty photometric table "
					"for this extension");
			ps.photbl = cpl_table_new(0);
			//omega_std_tidy(0);
			//return -1;
		}

		/* Save photometric table */
		if(isfirst){
			ps.pname = cpl_sprintf("%s_%s.fits", INSTRUME,PHOT_PROCATG);
			pframe = omega_product_frame(ps.pname, PHOT_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}
		cpl_msg_info(cpl_func,"Saving Photometric table");
		if(omega_save_table(ps.photbl,set,pars,alist,NULL,ps.pname,RECIPE,
				           pframe,NULL,isfirst) == -1){
			cpl_msg_error(cpl_func,"Cannot save photometric table");
			freeplist(alist);
			cpl_free(maskname); maskname=NULL;
			omega_std_tidy(0);
			return -1;
		}

		prestate = cpl_errorstate_get();
		/* Create Zeropoints table */
		if((ps.zptbl = omega_zeropoints(ps.photbl, ps.extratbl, ps.monitfr, ps.extfr,
				ps.colfr)) == NULL ){
			cpl_msg_warning(cpl_func,"Cannot create ZP table");
			cpl_msg_warning(cpl_func,"Creating an empty ZP table for this "
					"extension");
			ps.zptbl = cpl_table_new(0);
			//freeplist(alist);
			//omega_std_tidy(0);
			//return -1;
		}
		cpl_errorstate_set(prestate);

		freetable(ps.extratbl);
		freetable(ps.photbl);

		/* Get QC parameters from table */
		omega_std_config.ZP = cpl_table_get_double(ps.zptbl, "ZEROPOINT", 0, NULL);
		omega_std_config.ZPerr = cpl_table_get_double(ps.zptbl, "ZEROPOINT_ERR", 0, NULL);
		omega_std_config.Extinction = cpl_table_get_double(ps.zptbl, "EXTINCTION", 0, NULL);
		omega_std_config.Exterr = cpl_table_get_double(ps.zptbl, "EXTINCTION_ERR", 0, NULL);
		omega_std_config.nstars = cpl_table_get_int(ps.zptbl, "NSTARS", 0, NULL);
		omega_std_config.Seeing = cpl_table_get_double(ps.zptbl, "SEEING", 0, NULL);
		omega_std_config.Ellipticity = cpl_table_get_double(ps.zptbl, "ELLIPTICITY", 0, NULL);

		qclist = cpl_propertylist_new();
		cpl_propertylist_update_double(qclist, "ESO QC ZEROPOINT", omega_std_config.ZP) ;
		cpl_propertylist_set_comment(qclist, "ESO QC ZEROPOINT", "zp with flux in ADU/sec");
		cpl_propertylist_update_double(qclist, "ESO QC ZEROPOINT ERR", omega_std_config.ZPerr) ;
		cpl_propertylist_set_comment(qclist, "ESO QC ZEROPOINT ERR", "zp error");
		/*QC parameters for zeropoints calculated for electrons and not ADUs*/
		cpl_propertylist_update_double(qclist, "ESO QC ZEROPOINT_ELECTRON", omega_std_config.ZP+(2.5*log10(gain))) ;
		cpl_propertylist_set_comment(qclist, "ESO QC ZEROPOINT_ELECTRON", "zp with flux in e-/sec");
		cpl_propertylist_update_double(qclist, "ESO QC ZEROPOINT_ELECTRON ERR", omega_std_config.ZPerr) ;
		cpl_propertylist_set_comment(qclist, "ESO QC ZEROPOINT_ELECTRON ERR", "zp error");

		cpl_propertylist_update_double(qclist, "ESO QC EXTINCTION", omega_std_config.Extinction) ;
		cpl_propertylist_set_comment(qclist, "ESO QC EXTINCTION", "calculated extinction");
		cpl_propertylist_update_double(qclist, "ESO QC EXTINCTION ERR", omega_std_config.Exterr) ;
		cpl_propertylist_set_comment(qclist, "ESO QC EXTINCTION ERR", "calculated extinction error");
		cpl_propertylist_update_int(qclist, "ESO QC STD MATCHES", omega_std_config.nmatches);
		cpl_propertylist_set_comment(qclist, "ESO QC STD MATCHES", "number matches between USNOA2 cat and stars");
		cpl_propertylist_update_double(qclist, "ESO QC STD RATIO MATCHES", omega_std_config.mratio);
		cpl_propertylist_set_comment(qclist, "ESO QC STD RATIO MATCHES", "ratio USNOA2 matches/detected stars");
		cpl_propertylist_update_double(qclist, "ESO QC STD SEEING", omega_std_config.Seeing);
		cpl_propertylist_set_comment(qclist, "ESO QC STD SEEING", "calculated seeing of standard field");
		cpl_propertylist_update_double(qclist, "ESO QC STD ELLIPTICITY", omega_std_config.Ellipticity);
		cpl_propertylist_set_comment(qclist, "ESO QC STD ELLIPTICITY", "calculated ellipticity of standard field");
		cpl_propertylist_update_int(qclist, "ESO QC STD NUMBER STARS", omega_std_config.nstars);
		cpl_propertylist_set_comment(qclist, "ESO QC STD NUMBER STARS", "number of std stars used for ZP");

		/* Save ZP table */
		if(isfirst){
			ps.zname = cpl_sprintf("%s_%s.fits", INSTRUME,ZP_PROCATG);
			zpframe = omega_product_frame(ps.zname, ZP_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}
		cpl_msg_info(cpl_func,"Saving ZP table");
		if((omega_save_table(ps.zptbl,set,pars,alist,qclist,ps.zname,RECIPE,
				            zpframe,NULL,isfirst)) == -1){
			cpl_msg_error(cpl_func,"Cannot save photometric table");
			freeplist(qclist);
			freeplist(alist);
			cpl_free(maskname); maskname=NULL;
			omega_std_tidy(0);
			return -1;
		}

		freetable(ps.zptbl);
		freeplist(qclist);
		freeplist(alist);
	}      /* go and work on next extension */

	/*Clean up */
	cpl_free(maskname); maskname=NULL;
	omega_std_tidy(0);

	return 0;

}

/**
 * @brief
 *   Process the standard star frames
 * @param stdlist  Frame set containing only RAW standard frames
 * @param pars     Parameters list
 * @param ext      FITS extension to load
 *
 * @return
 *   A omega_fits structure containing the reduced image with
 *   updated headers
 */
omega_fits *omega_std_process(cpl_frameset *stdlist,cpl_parameterlist *pars, int ext)
{

	int i = 0;
	int nraw = 0;
	const char *smoothed = "omega_smoothed_std.fits";
	const char *smoothed_weight = "omega_smoothed_std_weight.fits";
	double xshift=0.,yshift=0.;
	cpl_image *trim=NULL, *mask_img=NULL, *image=NULL, *combmap=NULL;
	cpl_image *back = NULL;
	cpl_image *reduced = NULL;
	cpl_image *sumwght = NULL;
	cpl_mask *satu_map=NULL,*bpm_map=NULL;
	cpl_mask *satellite= NULL;
	cpl_mask *cosmic=NULL;
	cpl_propertylist *plist=NULL, *astrolist=NULL, *wcslist=NULL;
	cpl_frame *frame=NULL;
	omega_fits *stdfits = NULL;
	omega_fits **sfits = NULL;


	if(cpl_image_get_mean(ps.mflat) < 0.1){
		cpl_msg_error(cpl_func,"Master Flat image is NULL");
		return NULL;
	}

	if(cpl_image_get_min(ps.mflat) <= DBL_EPSILON){
		cpl_msg_warning(cpl_func,"Some pixel values of the Master Flat are "
				"smaller or equal to %g.", DBL_EPSILON);
		cpl_msg_warning(cpl_func,"Thresholding image to a minimum of %g ",
				DBL_EPSILON);
		cpl_image_threshold(ps.mflat, DBL_EPSILON, DBL_MAX,
				DBL_EPSILON, DBL_MAX);
	}

	/* Load omega_fits structure */
	nraw = cpl_frameset_get_size(stdlist);
	sfits = omega_fits_load_list(stdlist,CPL_TYPE_FLOAT,ext);

	for(i = 0; i < nraw; i++){
		/* Trim and overscan correct */
		if((trim = omega_trim_oscan_correct(sfits[i], omega_std_config.oc)) == NULL){
			cpl_msg_error(cpl_func,"Unable to trim image %d",i);
			freelfits(sfits,nraw);
			return NULL;
		}

		/* Subtract the bias */
		omega_biascor(trim, ps.mbias);

		/* Divide by the flat */
		if(omega_flatcor(trim, ps.mflat, NULL) != 0){
			cpl_msg_error(cpl_func,"Error in flat correction");
			freeimage(trim);
			freelfits(sfits,nraw);
			return NULL;
		}

		/* Create BPM+SATURATION mask */
		satu_map = omega_saturated_map(trim, omega_std_config.oc, pars);
		bpm_map = cpl_mask_threshold_image_create(ps.bpm, 0.5, 1.5) ;
		if(satu_map != NULL){
//			cpl_mask_and(bpm_map, satu_map);
			cpl_mask_or(bpm_map, satu_map);
			freemask(satu_map);
		}
		mask_img = cpl_image_new_from_mask(bpm_map);
		cpl_image_save(mask_img,"bpmsatu.fits",CPL_BPP_16_SIGNED,NULL,CPL_IO_CREATE);
		freemask(bpm_map);

		/* Get background */
		back = omega_background(trim, pars);

		if(back == NULL)
			cpl_msg_debug(cpl_func,"Unable to create background image");

		/* Detect cosmic rays */
		image = cpl_image_subtract_create(trim, back);
		cosmic = omega_cosmic_rays(image,ps.mflat,ps.illum,mask_img,pars);

		/* Subtract background from std */
		if(omega_std_config.bckg == 1){
			freeimage(trim);
			trim = cpl_image_duplicate(image);
		}
		freeimage(image);
		freeimage(back);


		/* Detect satellite tracks in background subtracted image */
		satellite = detsat(trim, pars);


		/* Create combination map (BPM,saturated,cosmic,satellite) */
		combmap = create_weightframe(ps.mflat,mask_img,cosmic,satellite);
		freemask(satellite);
		freemask(cosmic);
		freeimage(mask_img);

		if((ps.illum != NULL) && (combmap != NULL))
			cpl_image_divide(combmap, ps.illum);

		/* Multiply stds by their weights */
		cpl_image_multiply(trim, combmap);

		if(i == 0){
			reduced = cpl_image_duplicate(trim);
			sumwght = cpl_image_duplicate(combmap);
		}
		else{
			cpl_image_add(reduced, trim);
			cpl_image_add(sumwght, combmap);
		}
		freeimage(trim);
		freeimage(combmap);

		/* Go to next image */

	}

	freelfits(sfits,nraw);

	/* Weighted average of STDs */
	if(cpl_image_divide(reduced, sumwght) != CPL_ERROR_NONE){
		cpl_msg_error(cpl_func,"Cannot create weighted average of images. %s", cpl_error_get_message());
		freeimage(sumwght);
		return NULL;
	}


	/* Correct coordinates after trimming */
	frame = cpl_frameset_get_first(stdlist);
	plist = cpl_propertylist_load(cpl_frame_get_filename(frame),0);
	astrolist = cpl_propertylist_load(cpl_frame_get_filename(frame),ext);
	omega_shift_refpix(frame, ext, astrolist);

	/* Use pattern matching to detect stars */
	/* Smooth image before creating catalogue */
	image = omega_smooth_image(reduced, 1);

	/* Save star as intermediate product */
	cpl_propertylist_save(plist, smoothed, CPL_IO_CREATE);
	cpl_image_save(image, smoothed, BITPIX, astrolist, CPL_IO_EXTEND);

	/* Save star weight as intermediate product */
	cpl_propertylist_save(plist, smoothed_weight, CPL_IO_CREATE);
	cpl_image_save(sumwght, smoothed_weight, BITPIX, astrolist, CPL_IO_EXTEND);

	freeimage(image);
	freeimage(sumwght);

	/* Calculate WCS distortion */
	wcslist = omega_match_points(smoothed, smoothed_weight, ps.usnoa2, pars, ext,
			&omega_std_config.nmatches, &omega_std_config.mratio);

	omega_get_wcsshift(astrolist, wcslist, &xshift, &yshift);


	if(wcslist == NULL || fabs(xshift)> omega_std_config.MaxWcsShift ||
			fabs(yshift)>omega_std_config.MaxWcsShift){
		cpl_msg_warning(cpl_func,"Unable to refine WCS for image");
		cpl_msg_warning(cpl_func,"Taking original WCS coordinates");
		wcslist = cpl_propertylist_duplicate(astrolist);
		cpl_propertylist_update_int (wcslist, "ESO QC STD PPM FAILURE", 0);
		cpl_propertylist_set_comment(wcslist, "ESO QC STD PPM FAILURE", "PPM failure if ZERO");
		xshift=0.;
		yshift=0.;
	}
	else{
		cpl_propertylist_update_int (wcslist, "ESO QC STD PPM FAILURE", 1);
		cpl_propertylist_set_comment(wcslist, "ESO QC STD PPM FAILURE", "PPM failure if ZERO");
	}

	cpl_propertylist_update_double (astrolist, "ESO QC DELTAX SHIFT", xshift);
	cpl_propertylist_set_comment(astrolist, "ESO QC DELTAX SHIFT", "WCS shift in x");
	cpl_propertylist_update_double (astrolist, "ESO QC DELTAY SHIFT", yshift);
	cpl_propertylist_set_comment(astrolist, "ESO QC DELTAY SHIFT", "WCS shift in y");

	/* Append WCS header to extension header */
	if(omega_pfits_update_header(astrolist, wcslist) != 0)
		cpl_msg_debug(cpl_func,"Cannot append WCS header to extension header");


	/* Wrap product into omega fits structure using 1st frame headers */
	stdfits = omega_fits_wrap(reduced,NULL,plist,astrolist);

	freeplist(plist);
	freeplist(astrolist);
	freeplist(wcslist);

	return stdfits;
}


int omega_std_retrieve_input_param(const cpl_parameterlist  *pars)
{

	const cpl_parameter *par = NULL;
	cpl_errorstate prestate = cpl_errorstate_get();

	par = cpl_parameterlist_find_const(pars, "omega.omega_standard_star.ExtensionNumber") ;
	omega_std_config.extnum = cpl_parameter_get_int(par) ;
	par = cpl_parameterlist_find_const(pars, "omega.omega_standard_star.OverscanMethod") ;
	omega_std_config.oc = cpl_parameter_get_int(par) ;
	par = cpl_parameterlist_find_const(pars, "omega.omega_standard_star.BackRestore");
	omega_std_config.bckg = cpl_parameter_get_bool(par) ;
	par = cpl_parameterlist_find_const(pars, "omega.omega_standard_star.PAF");
	omega_std_config.paf = cpl_parameter_get_bool(par) ;
	par = cpl_parameterlist_find_const(pars, "omega.omega_standard_star.MaxWcsShift") ;
	omega_std_config.MaxWcsShift = cpl_parameter_get_double(par) ;

	if(!cpl_errorstate_is_equal(prestate)){
		cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
		return -1;
	}

	return 0;
}

static int omega_std_load_calib(int ext)
{

	/* Load mandatory calibration frames */
	/* master bias */
	ps.mbias = cpl_image_load(cpl_frame_get_filename(ps.mbiasfr), CPL_TYPE_FLOAT, 0, ext);
	if(ps.mbias == NULL){
		cpl_msg_error(cpl_func,"Cannot load Master Bias. %s", cpl_error_get_message());
		return -1;
	}
	/* master flat */
	ps.mflat = cpl_image_load(cpl_frame_get_filename(ps.mflatfr),CPL_TYPE_FLOAT,0,ext);
	if (ps.mflat == NULL){
		cpl_msg_error(cpl_func,"Cannot load Master Flat. %s",cpl_error_get_message());
		return -1;
	}

	/* bad pixels map */
	ps.bpm = cpl_image_load(cpl_frame_get_filename(ps.bpmfr), CPL_TYPE_INT, 0, ext);
	if(ps.bpm == NULL){
		cpl_msg_error(cpl_func, "Cannot load BPM. %s", cpl_error_get_message());
		return -1;
	}

	/* illumination map */
	if(ps.illumfr != NULL)
		ps.illum = cpl_image_load(cpl_frame_get_filename(ps.illumfr),CPL_TYPE_FLOAT,0,ext);

	return 0;
}

/* Initialize the pointers */
static void omega_std_init(void) {
	ps.stdlist = NULL;
	ps.labels = NULL;
	ps.bpmfr = NULL;
	ps.extfr = NULL;
	ps.monitfr = NULL;
	ps.stdfr = NULL;
	ps.usnoa2 = NULL;
	ps.reffr = NULL;
	ps.illumfr = NULL;
	ps.mbiasfr = NULL;
	ps.mflatfr = NULL;
	ps.mbias = NULL;
	ps.mflat = NULL;
	ps.colfr = NULL;
	ps.stdfits = NULL;
	ps.zptbl = NULL;
	ps.photbl = NULL;
	ps.sname = NULL;
	ps.pname = NULL;
	ps.zname = NULL;
}

/* Free any allocated memory */
static void omega_std_tidy(int level) {
	freeimage(ps.mbias);
	freeimage(ps.bpm);
	freeimage(ps.illum);
	freeimage(ps.mflat);
	if(level == 1)
		return;

	freeframeset(ps.stdlist);
	freefits(ps.stdfits);
	freetable(ps.extratbl);
	freetable(ps.zptbl);
	freetable(ps.photbl);
	freespace(ps.labels);
	freespace(ps.sname);
	freespace(ps.pname);
	freespace(ps.zname);
}

/**@}*/

